# Slowly heating a sphere of atoms

import sys
# Unix
sys.path.append('../../src/MD/SWIGBuild/i686/obj3D')
# Windows
sys.path.append('../../src/MD/SWIGBuild/')

from MD3D import *
from Numeric import *

# Sphere of atoms
atoms = DynamicListOfAtoms()

# Lennard Jones potential, cut off smoothly at 2.7 units
potential = CutLennardJonesPotential()

# Sphere of diameter = 2*radius, in box of side 5*radius
# fcc has unit cell = sqrt(2) times interatomic spacing
latticeConstant = potential.GetLengthScale()*sqrt(2.0)
lattice = FCCLattice(latticeConstant)
radius = 2.5 * latticeConstant

# Put atoms in periodic boundary conditions ([0,L]x[0,L]x[0,L])
bc = SmartPeriodicBoundaryConditions()
bc.SetLength(array([1, 1, 1.])*radius*5)
atoms.SetBoundaryConditions(bc)

# Build a spherical fcc cluster of the right spacing
sphericalInitializer = SphericalClusterInitializer(radius, lattice)
# Set center to center of periodic boundary conditions box
# sphericalInitializer.SetCenter(array([1, 1, 1.]) * radius * 1.5)
sphericalInitializer.Create(atoms)


# neighborLocator will give atoms up to cutoff (default noNeighborLocator)
cutoff = potential.GetCutoffDistance()
# Simplest, best for small numbers of atoms
simpleNeighborLocator = SimpleNeighborLocator(cutoff)
# Jakob Schiotz's old version: WARNING -- doesn't seem to give exactly the same answers
#neighborList = NeighborList(cutoff, 0.05*cutoff)
# Current state-of-the-art for low temperatures
cellNeighborList = CellNeighborList(cutoff, 0.05*cutoff)
# More efficient for very sparse systems, less storage
cellNeighborLocator = CellNeighborLocator(cutoff, 0.05*cutoff)

atoms.SetNeighborLocator(cellNeighborList)

# Warning: our version of Verlet isn't standard!
timeStep = 0.002
nStepsBetweenDraws = 20
initialTemperature = 0.
verlet = VerletAtomsMover(potential, timeStep, nStepsBetweenDraws)
verletThermalize = VerletThermalizeAtomsMover(potential, initialTemperature, timeStep, nStepsBetweenDraws)
# Langevin doesn't seem to be working right
friction = 0.1
langevin = LangevinAtomsMover(potential, timeStep, nStepsBetweenDraws, friction, initialTemperature)
# Warning: Hoover is tricky to get working right
tau = 1.0
hoover = HooverAtomsMover(potential, initialTemperature, timeStep, tau, nStepsBetweenDraws)
thermalizer = verletThermalize

# Add graphics observer
sys.path.append('Observers')
from RasMolObserver import RasMolObserver3D
ras = RasMolObserver3D(radius=150)
atoms.AddObserver(ras)

# Add energy observer
energy = EnergyObserver(potential)
# Update observer by hand (so graphics doesn't consume too much time)
# atoms.AddObserver(energy)

print 'number of atoms is', atoms.GetNumber()

# Change temperature in steps: then measure at constant energy using Verlet
print "T, mean temperature, mean energy, c_diff, c_fluc"
for i in range(1,15):
    thermalizer.SetTemperature(i/10.)
    # thermalize for 50 steps, then measure for 100
    # EnergyObserver's specific heat measurement assumes microcanonical ensemble:
    # must use constant energy simulation (no thermalization)
    for j in range(50):
        thermalizer.Move(atoms)
    energy.Reset()
    for j in range(100):
        verlet.Move(atoms)
        energy.Update(atoms)
    atoms.Notify()
    PE = potential.CalculateEnergy(atoms)
    KE = atoms.GetKineticEnergy()
    meanEnergy = energy.GetMeanEnergy()
    meanTemperature = energy.GetKineticTemperature()
    if (i==1):
        print thermalizer.GetTemperature(), meanTemperature, meanEnergy
    else:
        print thermalizer.GetTemperature(), meanTemperature, meanEnergy, \
              (meanEnergy-oldEnergy)/((meanTemperature-oldTemperature) \
                                     *atoms.GetNumber()), \
              energy.GetKineticSpecificHeat()
    oldEnergy = meanEnergy
    oldTemperature = meanTemperature

# Cool more slowly...
for i in range(15,0,-1):
    thermalizer.SetTemperature(i/10.)
    for j in range(200):
        thermalizer.Move(atoms)
    energy.Reset()
    for j in range(400):
        verlet.Move(atoms)
        energy.Update(atoms)
    atoms.Notify()
    PE = potential.CalculateEnergy(atoms)
    KE = atoms.GetKineticEnergy()
    meanEnergy = energy.GetMeanEnergy()
    meanTemperature = energy.GetKineticTemperature()
    print thermalizer.GetTemperature(), meanTemperature, meanEnergy, \
          (meanEnergy-oldEnergy)/((meanTemperature-oldTemperature) \
                                  *atoms.GetNumber()), \
                                  energy.GetKineticSpecificHeat()
    oldEnergy = meanEnergy
    oldTemperature = meanTemperature

# Minimize energy from this state
min = QuickMinAtomsMover(potential, 0.01)
min.Move(atoms)
atoms.Notify()
