import tempfile
from sys import path
path.append('../../src/MD/SWIGBuild')
path.append('../../src/MD/SWIGBuild/i686/obj2D')
path.append('../../src/MD/SWIGBuild/i686/obj3D')

# I had the problem that the RasMol window wasn't deleted when we exit python.
# The reason was the callback mechanism with C++: the shadow class keeps a reference
# to the python Update method of the RasMolObserver, which itself owns the shadow
# object. Not increasing the reference count of the Update function from C++ didn't
# help. So my solution is to have a container that keeps references to both the
# shadow objects and the RasMolObserver instances. This is a global object (called
# shadowContainer, which will be deleted when python exits and will kill all the
# rasmol processes. The drawback of this approach is that some other modules seem to
# be deleted earlier (like RasMolInterface), so we have to keep also references to
# them, so they are not deleted before we need them.  Thierry Cretegny

class _ShadowInstances:
    def __init__(self):
        self._list = []
        self._obslist = []
    def Register(self, obs, shadow):
        self._list.append(shadow)
        self._obslist.append(obs)
    def __del__(self):
        del self._list
        del self._obslist

shadowContainer = _ShadowInstances()
from RasMolInterface import *

class _RasMolObserver(RasMolInterface):
    """ This is our RasMol Observer, designed as a ListOfAtomsObserver. It can be attached
    to ListOfAtoms in the usual way (i.e. atoms.AddObserver(observer) ). The radius of the
    sphere can be specified, as well as a function used to determine the color of the atoms,
    like potential.CalculateAtomicEnergy. This function should handle calls like func(atoms, i).
    """
    def __init__(self,_shadow_,atoms=None,options="", radius=300, colorMethod=None):
        self._start = 1
        shadowContainer.Register(self,_shadow_)
        self._shadow_ = len(shadowContainer._list)-1  #_shadow_
        self._atoms_ = atoms
        from copy import deepcopy
        self.this = deepcopy(_shadow_.this)    # deep copy to avoid ref increase. (Necessary ?)
        self._tmppdbfilename_=tempfile.mktemp()    # the tmp file where we write RasMol data
        self._sphereRadius = radius                # the radius of the atoms
        self._colorMethod = colorMethod
        self._writefile(atoms)
        
        RasMolInterface.__init__(self,filename=self._tmppdbfilename_)
        self._delete_interface = RasMolInterface.__del__ # we keep a reference to the delete method

        self("spacefill " + str(self._sphereRadius))
        if colorMethod: self("color temperature")
        self("wireframe off")

    def _writefile(self, subject):
        fil=open(self._tmppdbfilename_,"w")
        if subject:
            pos=subject.GetCartesianPositions()[:]
            n = len(subject)
            for i in xrange(n):
                p = pos[:][i]
                if(len(p) == 2): p = [p[0], p[1], 0.]
                if self._colorMethod is None: color = 0.
                else: color = self._colorMethod(subject,i)
                fil.write("ATOM  %5d %-4s              %8.3f%8.3f%8.3f  0.00  %8.3f\n" % (i%100000, "NN", p[0], p[1], p[2], color))
        fil.close()

    def Update(self, atoms=None):
        if atoms is None:
            atoms = shadowContainer._list[self._shadow_].GetLastLOA()
        self._writefile(atoms)
        if self._start :
            self.Load(self._tmppdbfilename_)
            self("spacefill " + str(self._sphereRadius))
            if self._colorMethod: self("color temperature")
            self("wireframe off")
            self._start = 0
        else:
           self("spacefill off")
           self.Reload()
           self("spacefill " + str(self._sphereRadius))
            
    def __del__(self):
        self._delete_interface(self)
        try:
            self._unlink(self._tmppdbfilename_) # cf comment in RasMolInterface.__init__
        except:
            pass   # The file might not exist.



class RasMolObserver2D(_RasMolObserver):
    def __init__(self,atoms=None,options="", radius=300, colorMethod=None, notifyLevel = 0):
        import MD2D
        shadow = MD2D.PythonLOAObserver(self.Update, notifyLevel) # the C++ component
        _RasMolObserver.__init__(self,shadow, atoms, options, radius, colorMethod)


class RasMolObserver3D(_RasMolObserver):
    def __init__(self,atoms=None,options="", radius=300, colorMethod=None, notifyLevel = 0):
        import MD3D
        shadow = MD3D.PythonLOAObserver(self.Update, notifyLevel) # the C++ component
        _RasMolObserver.__init__(self,shadow, atoms, options, radius, colorMethod)


       
