// ConjugateGradientMinimizer.h: interface for the ConjugateGradientMinimizer class.
//
//////////////////////////////////////////////////////////////////////

#ifndef ConjugateGradientMinimizer_H
#define ConjugateGradientMinimizer_H

#include <math.h>
#include <string>
#include <iostream>
#include "DigitalMaterial.h"
#include "Minimizer.h"


//! Does conjugate gradient minimization

/**
 *
 * Other things to clean up: numerical recipes macros have been turned into
 * functions. They still have names like SHFT though, which we should change.
 * Also the error codes are now just integers, not good.
 *
 *
 * \author nbailey 
 * \date 6/11/2000
 * \ingroup moleculardynamics
 */




class ConjugateGradientMinimizer: public Minimizer  
{
public:
	ConjugateGradientMinimizer(double tolerance, int maximumIterations);
	virtual ~ConjugateGradientMinimizer();
	double Minimize(double *currentPoint, Minimizable *minimizable);
	
	double GetTolerance() const {return tolerance;}
	void SetTolerance(double tol) {tolerance = tol;}
	int GetNIterations() {return nIterations;} // returns number of iterations taken during the last call to minimize
	int GetMaxIterations() const {return maximumIterations;}
	void SetMaxIterations(int maxIt) {maximumIterations = maxIt;}
	
	double GetConvergence() const { return gradSq; }

	void SetStepsBetweenUpdate(int steps){ stepsBetweenUpdate = steps; }
	int GetStepsBetweenUpdate(){ return stepsBetweenUpdate; }


	void SetLineMinimizer(std::string line_min)
		{
			if ( line_min == "BRENT" )
				this->LINE_MIN = BRENT;
			else if( line_min == "DRENT" )
				this->LINE_MIN = DBRENT;
			else if( line_min == "NEWTON" )
				this->LINE_MIN = NEWTON;
			else throw DMException("ConjugateGradientMinimizer","SetLineMinimizer","unrecognized line minimizer option");
		}
	void SetFiniteDifferenceParameter(double fdp) {finiteDifferenceParameter = fdp;}
	double GetFiniteDifferenceParameter() const {return finiteDifferenceParameter;}

	void SetOutputStream(std::ostream *stream){this->outputStream = stream;}
	void Verbose() {verbose = true;}
	void Quiet() {verbose = false;}

	//! Use the objective function for testing when done (default)
	void UseFunction() {useFunction = true; }
	//! Ignore the objective function for the purpose of testing when done
	void IgnoreFunction() {useFunction = false; }


#ifndef SWIG	
protected:
	enum {BRENT, DBRENT, NEWTON} LINE_MIN;// determines which method of line minimzation is being used
	int maximumIterations;
	double tolerance;
	double gradSq;
	int nIterations;	
	int stepsBetweenUpdate;
	Minimizable *minimizable;

	bool verbose;
	std::ostream *outputStream;
	static const int DISPLAYPRECISION;
	bool useFunction;

	// These are the Numerical Recipes routines
	double Derivative1D(double x, const double *currentGradient = 0);
	double Function1D(double x);
	double DerivBrentLineMin(double ax,double bx,double cx,double tol,double *xmin);
	double BrentLineMin(double ax,double bx,double cx,double tol,double *xmin);
	// Have LineMinimize call either BrentLineMin or DerivBrentLineMin depending on whether derivative information is to be used in the line minimization
	double NewtonLineMin(double ax, double bx, double cx, double tol, double *xmin, const double *gradient = 0);
	double NewtonDerivLineMin(double ax, double bx, double cx, double tol, double *xmin);
	
	double LineMinimize(double *direction, double currentValue, const double *currentGradient = 0);
	void Bracket(double *ax,double *bx,double *cx,double *fa,double *fb,double *fc);
	
	
	// These were NR macros (defined in "nrutil.h"). I made them inline member functions
	double FMAX(double a,double b) {return (a> b ?  a:b);}
	double SIGN(double a,double b) {return ( b >= 0.0  ? fabs(a) : -fabs(a));}
	
	const double GOLD;  // used by mnbrak
	const double CGOLD; // used by brent
	
	void MOV3(double &a,double &b,double &c,double d,double e,double f)
		{a=d;b=e;c=f;}	
	
	
	int		ITMAX_DerivBrentLineMin; 
	double	ZEPS_DerivBrentLineMin; 
	int		ITMAX_BrentLineMin; 
	double	ZEPS_BrentLineMin; 
	double	EPS_Minimizer;
	double	TOL_LineMinimizer;
	int		GLIMIT_Bracket;
	double	TINY_Bracket;
	double  finiteDifferenceParameter;
    
	double *currentPoint;
	double *currentDirection;
	double *tempPoint;
#endif // SWIG
  
};



#endif // !defined(CONJUGATEGRADIENTMINIMIZER_H)
