#ifndef _OBSERVER_LIST_H_
#define _OBSERVER_LIST_H_

#include <vector>
#include <algorithm>
#include "DigitalMaterial.h"
#include "DMException.h"

//! Responsible for keeping track of observers of a subject.
/** This class helps implement the Subject-Observer pattern from 
  * Gamma et al. by taking care of the evil standard template library
  * functions the subject has to use to keep track of its observers.
  * The subject class owns an ObserverList as a member, where the
  * template parameter is the class of the observer.
  * Then the subject class forwards requests to this class.
  *
  * Here is an example of how to instantiate the class
  * \code
  * class SweepLogic {
  *   ObserverList<AvalancheObserver> observers;
  * \endcode
  * Then you forward calls to this class
  * \code
  * void SweepLogic::AddObserver( AvalancheObserver* observer ) {
  *   observers->AddObserver( observer );
  * }
  * \endcode
  *
  * You can write a Notify() function in simple terms:
  * \code
  * void SweepLogic::Notify( double dVal ) {
  *   int i;
  *   for (i=0; i<observers.size(); i++) {
  *     observers.at(i)->NotificationFunction( dVal );
  *   }
  * \endcode
  *
  *
  * \author ajd27
  * \date 18 Aug 2000
  */
template<class Observer>
class ObserverList
{
	typedef std::vector<Observer*> observer_vector;
	observer_vector observers;
public:

	//! Add an observer to the list of those to notify.
	/**
	  *If the observer is already in the list, the request is ignored.
	  *
	  * @param observer Observer* Pointer to the observer to add.
	  * @return bool Whether we added it.  False indicates we already
	  * had a pointer to this observer.
	  */
	bool AddObserver( Observer *observer );

	//! Remove an observer from the list of those to notify.
	/**
	  * If the observer is not on our list, the request is ignored.
	  * @param observer Observer* Pointer to observer to remove.
	  * @return bool Whether we removed it.  False indicates we didn't
	  * have a pointer to this observer in the first place.
	  */
	bool RemoveObserver( Observer *observer );

	//! The size of the list of observers.
	/*! This is for iterating through the list.
		*/
	size_t size() const { return observers.size(); }

	//! Return the nth observer.
	/*! For iterating through the observers.  It is traditional
		to use at() to represent checked access to a list.

		This function can throw the STL out of bounds exception
		which vector.at() throws.

		@param n size_t The index of the observer to return.
		@return Observer* Pointer to the nth observer.
		*/
	Observer* at( size_t n ) {
		ASSERT((n >= 0) && (n < observers.size()));
		return observers[n]; 
	}

	//! Notify all observers with the given function.
	/** This executes a for_each from the standard library.
	  * The argument is the same sort of function/functor you
	  * would give to for_each.  This function is the STL
	  * alternative to using size() and at().
	  */
	template<class Notification>
	void Notify( Notification notify ) {
		int i, iMax = observers.size();
		Observer* listen;
		for (i=0; i<iMax; i++) {
			listen = observers[i];
			notify(listen);
		}
		// for_each commented out.  VC++ troubles.
//		std::for_each( observers.begin(), observers.end(), notify );
	}
};

template<class Observer>
inline bool ObserverList<Observer>::AddObserver( Observer *observer )
{
	ASSERT( 0 != observer );
	if ( 0 == observer) throw
		DMException( "ObserverList","AddObserver",
		"Passed null pointer.");

	//listen_vector::iterator
	if (std::find( observers.begin(), observers.end(), observer ) ==
		observers.end())
		observers.push_back( observer );
	else return false;
	return true;
}

template<class Observer>
inline bool ObserverList<Observer>::RemoveObserver( Observer *observer )
{
	ASSERT( 0 != observer );
	if ( 0 == observer) throw
		DMException( "ObserverList","AddObserver",
		"Passed null pointer.");

	typename observer_vector::iterator tlist;
	tlist = std::find( observers.begin(), observers.end(), observer);
	if (tlist != observers.end())
		observers.erase( tlist );
	else return false;
	return true;
}

#endif //_OBSERVER_LIST_H_
