// MPIProtocol.cpp: implementation of the MPIProtocol class.
//
//////////////////////////////////////////////////////////////////////
#include "DMFirstHeader.h"
#include "DigitalMaterial.h"
#ifdef DMPARALLEL
#include "MPIProtocol.h"
#include "NotEnoughMemory.h"
#include "SprintString.h"
#include "DMLastHeader.h"


using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

MPIProtocol::MPIProtocol()
{
    int result = MPI_Comm_dup(MPI_COMM_WORLD, &myComm);
    if(result != MPI_SUCCESS) 
        throw ProtocolError("MPIProtocol", "MPIProtocol", 
        "Error duplicating MPI_COMM_WORLD");

    // get the number of processes in myComm;
    result = MPI_Comm_size(myComm, &nProcs);
    if(result != MPI_SUCCESS) 
        throw ProtocolError("MPIProtocol", "GetNumberOfProcessors", 
        "Error in MPI_Comm_size");
    
    // get my process id in myComm
    result = MPI_Comm_rank(myComm, &myProc);
    if(result != MPI_SUCCESS) 
        throw ProtocolError("MPIProtocol", "GetProcessorNumber", 
        "Error in MPI_Comm_rank");
    
    // set the translation between our Operations and MPI's .
    Op_to_MPI_Op[SUM] = MPI_SUM;
    Op_to_MPI_Op[MAX] = MPI_MAX;
    Op_to_MPI_Op[MIN] = MPI_MIN;

}


void MPIProtocol::Broadcast(int *values, int size, int master)
{
    int result = MPI_Bcast(values, size, MPI_INT, master, myComm);	
    if(result != MPI_SUCCESS) 
        throw ProtocolError("MPIProtocol", "Broadcast", 
        "Error in broadcast");
}

void MPIProtocol::Broadcast(double *values, int size, int master)
{
    int result = MPI_Bcast(values, size, MPI_DOUBLE, master, myComm);	
    if(result != MPI_SUCCESS) 
        throw ProtocolError("MPIProtocol", "Broadcast", 
        "Error in broadcast");
}

void MPIProtocol::Broadcast(string &str, int master)
{
    const int maxSize = 1024;
    int result;
    if(myProc == master)
    {
        int thisSize  = str.size();
        char *thisStr = const_cast<char *>(str.c_str());  // the broadcast won't change the string.
        if(thisSize > maxSize)
        {
            string empty = "";
            Broadcast(empty, master);   // tells to allocate more space
            Broadcast(&thisSize, 1, master);
        }
        result = MPI_Bcast(thisStr, thisSize, MPI_CHAR, master, myComm);
    }
    else
    {
        char *buff = new char[maxSize];
        result = MPI_Bcast(buff, maxSize, MPI_CHAR, master, myComm);
        if(buff[0] == 0)
        {
            int newSize;
            Broadcast(&newSize, 1, master);
            delete [] buff;
            buff = new char[newSize];
            result = MPI_Bcast(buff, newSize, MPI_CHAR, master, myComm);
        }
        str = buff;
        delete [] buff;
    }
    if(result != MPI_SUCCESS) 
        throw ProtocolError("MPIProtocol", "Broadcast", 
        "Error in broadcast");
}


void MPIProtocol::Barrier()
{
    int result = MPI_Barrier(myComm);
    if(result != MPI_SUCCESS) 
        throw ProtocolError("MPIProtocol", "Barrier", 
        "Error in barrier");
}


std::vector<int> MPIProtocol::GatherValue(int value, int root)
{
    int *buffer = new int[nProcs];
    int result = MPI_Gather(&value, 1, MPI_INT, buffer, 1, MPI_INT, root, myComm);
    if(result != MPI_SUCCESS) 
        throw ProtocolError("MPIProtocol", "GatherValue", 
        "Error in Gather");
    vector<int> ret;
    if(myProc == root) ret.insert(ret.begin(), buffer, buffer+nProcs);
    delete [] buffer;
    return ret;
}


void MPIProtocol::AllReduce(const double *in, double *out, int size, Operation op)
{
    int result = MPI_Allreduce(const_cast<double *>(in), 
                               out, size, MPI_DOUBLE, Op_to_MPI_Op[op], myComm);
    if(result != MPI_SUCCESS) 
        throw ProtocolError("MPIProtocol", "AllReduce<double>", 
        "Error in MPI_AllReduce");
}

void MPIProtocol::AllReduce(const int *in, int *out, int size, Operation op)
{
    int result = MPI_Allreduce(const_cast<int *>(in), 
                               out, size, MPI_INT, Op_to_MPI_Op[op], myComm);
    if(result != MPI_SUCCESS) 
        throw ProtocolError("MPIProtocol", "AllReduce<int>", 
        "Error in MPI_AllReduce");
}




MPIProtocol::MPIBuffers::MPIBuffers(MPIProtocol *protocol):
                                    myProc(protocol->GetProcessorNumber()), 
                                    nProcs(protocol->GetNumberOfProcessors()),
                                    myComm(protocol->GetCommunicator())
{
    this->sendBuffers  = new MPISendBuffer[nProcs];
    this->recvBuffers  = new MPIRecvBuffer[nProcs];
    this->otherRecvBuffersSize = new int[nProcs];
    for(int proc=0; proc<nProcs; ++proc)
    {
        recvBuffers[proc].SetComm(myComm);
        sendBuffers[proc].SetComm(myComm);
        sendBuffers[proc].EmptyBuffer();
        sendBuffers[proc].PutInt(recvBuffers[proc].GetBufferSize(),"");
        otherRecvBuffersSize[proc] = recvBuffers[proc].GetBufferSize();
    }
    mersenne = 1;
    while(mersenne < nProcs) mersenne = mersenne*2;
    mersenne = mersenne-1;

}

MPIProtocol::MPIBuffers::~MPIBuffers()
{
    delete [] this->otherRecvBuffersSize;
    delete [] this->sendBuffers;
    delete [] this->recvBuffers;
}



void MPIProtocol::MPIBuffers::AllToAll()
{
    // use an alltoallv based on hypercube algorithm
    // (http://www.sdsc.edu/~tkaiser/my_mpi.html)
    int proc, ierr;
    MPI_Status status;
    int i,my_tag;
    
    // Make the receive buffers ready for the communication
    for(proc=0; proc<nProcs; ++proc) recvBuffers[proc].EmptyBuffer();
    
    for(i=1;i<=mersenne;i++) {
        my_tag=4925;
        /*! do xor to find the processor xchng */
        int xchng= i ^ myProc;
        if(xchng >= nProcs) continue;
        if(myProc < xchng)
        {
            int sendSize = sendBuffers[xchng].GetSize();
            if(sendSize > otherRecvBuffersSize[xchng])
            {
                SendAnticipatedOverflow(xchng, sendSize, my_tag++);
                otherRecvBuffersSize[xchng] = sendSize;
            }
            MPI_Send(sendBuffers[xchng].GetBuffer(),
                sendSize,MPI_PACKED, xchng,my_tag++,myComm);
            
            ierr =  MPI_Recv(recvBuffers[xchng].GetBuffer(),
                recvBuffers[xchng].GetBufferSize(),MPI_PACKED,
                xchng,my_tag++,myComm,&status);
            ASSERT(ierr == MPI_SUCCESS);
            int tmpSize = recvBuffers[xchng].GetInt("");
            if(tmpSize < 0)
            {   // we're told our receive buffer is too small
                recvBuffers[xchng].Allocate(-tmpSize);
                ierr = MPI_Recv(recvBuffers[xchng].GetBuffer(),
                    recvBuffers[xchng].GetBufferSize(),MPI_PACKED,
                    xchng,my_tag++,myComm,&status);
                ASSERT(ierr == MPI_SUCCESS);
                tmpSize = recvBuffers[xchng].GetInt("");
            }
            if(sendSize < tmpSize) otherRecvBuffersSize[xchng] = tmpSize;
            // otherwise there has been a reallocation
        }
        else  // the opposite of the previous block
        {
            ierr = MPI_Recv(recvBuffers[xchng].GetBuffer(),
                recvBuffers[xchng].GetBufferSize(),MPI_PACKED,
                xchng,my_tag++,myComm,&status);
            int tmpSize = recvBuffers[xchng].GetInt("");
            if(tmpSize < 0) 
            {   // our recv buffer is too small
                recvBuffers[xchng].Allocate(-tmpSize);
                // now we are ready to receive
                ierr = MPI_Recv(recvBuffers[xchng].GetBuffer(),
                    recvBuffers[xchng].GetBufferSize(),MPI_PACKED,
                    xchng,my_tag++,myComm,&status);
                ASSERT(ierr == MPI_SUCCESS);
                tmpSize = recvBuffers[xchng].GetInt("");
                ASSERT(tmpSize > 0);
            }
            otherRecvBuffersSize[xchng] = tmpSize;
            int sendSize = sendBuffers[xchng].GetSize();
            if(sendSize > otherRecvBuffersSize[xchng])
            {
                SendAnticipatedOverflow(xchng, sendSize, my_tag++);
                otherRecvBuffersSize[xchng] = sendSize;
            }
            MPI_Send(sendBuffers[xchng].GetBuffer(),
                sendSize,MPI_PACKED, xchng,my_tag++,myComm);
        }
        if(ierr != MPI_SUCCESS) throw ProtocolError("MPIProtocol", "AllToAll", 
            SprintString("MPI error with code %i", ierr));
    }
    /* ! exchange data with myself */
    if(recvBuffers[myProc].GetBufferSize() < sendBuffers[myProc].GetSize())
        recvBuffers[myProc].Allocate(sendBuffers[myProc].GetSize());
    memcpy(recvBuffers[myProc].GetBuffer(), sendBuffers[myProc].GetBuffer(), 
        sendBuffers[myProc].GetSize());
    
    otherRecvBuffersSize[myProc] = recvBuffers[myProc].GetInt("");
    
    for(proc=0; proc<nProcs; ++proc) 
    {
        sendBuffers[proc].EmptyBuffer();
        sendBuffers[proc].PutInt(recvBuffers[proc].GetBufferSize(),"");
    }
}



void MPIProtocol::MPIBuffers::SendAnticipatedOverflow(int proc,
                                                      int neededBufferSize, 
                                                      int tag)
{
    char tmpBuff[20];
    int tmp = -neededBufferSize;
    int currentSize = 0;
    MPI_Pack(&tmp, 1, MPI_INT, tmpBuff, 20, &currentSize, myComm);
    MPI_Send(tmpBuff, currentSize, MPI_PACKED, proc , tag, myComm);
}

#endif //DMPARALLEL
