// MPIProtocol.h: interface for the MPIProtocol class.
//
//////////////////////////////////////////////////////////////////////

#ifndef MPIPROTOCOL_H
#define MPIPROTOCOL_H

#include "DMFirstHeader.h"
#include <map>
#include "DigitalMaterial.h"

#include "DMProtocol.h"
#include "mpi.h"
#include "MPISendBuffer.h"
#include "MPIRecvBuffer.h"

#include "DMException.h"

class MPIProtocol : public DMProtocol
{
public:
    MPIProtocol();
    virtual ~MPIProtocol() {}
    
    int      GetProcessorNumber()    const {return myProc;}
    int      GetNumberOfProcessors() const {return nProcs;}
    MPI_Comm GetCommunicator()       const {return myComm;}

    void Broadcast(double *values, int size, int master);
    void Broadcast(int *values, int size, int master);
    void Broadcast(std::string &string, int master);
    
    void Barrier();
    
    std::vector<int> GatherValue(int number, int root);

    void AllReduce(const double *in, double *out, int size, Operation op);
    void AllReduce(const int *in   , int    *out, int size, Operation op);


    DMProtocol::Buffers *NewBuffers() 
        {return new MPIBuffers(this);}

    class MPIBuffers : public DMProtocol::Buffers
    {
    public:
        MPIBuffers(MPIProtocol *protocol);
        ~MPIBuffers();
        //! an alltoall communication using buffers.
        /*!Useful for mixed-type data and/or variable length messages
        */
        void AllToAll();
        DMWriter *GetSendBuffer(int proc) {return &sendBuffers[proc];}
        DMReader *GetRecvBuffer(int proc) {return &recvBuffers[proc];}
    private:
        const int myProc;
        const int nProcs;
        const MPI_Comm myComm;
        //! the smallest Mersenne number bigger than nProcs
        int mersenne;
        //! The size of the receive buffer on the destination processor
        /*! we use that in order to predict whether a processor will have enough 
        space to receive what we send. If it is too small we first ask to resize
        the array by calling \a SendAnticipatedOverflow */
        int *otherRecvBuffersSize;
        MPISendBuffer *sendBuffers;
        MPIRecvBuffer *recvBuffers;
    
        //! Notify a processor that its recv buffer is too small to get our message
        void SendAnticipatedOverflow(int proc, int neededBufferSize, int tag);
    
    };
    
    class ProtocolError: public DMException
    {
        std::string sFunction, sFile, sNCMessage;	
    public:
        ProtocolError( const std::string& sFunction,
            const std::string& sFile,
            const std::string& sNCMessage ): 
        DMException(sFunction, sFile, sNCMessage){}
        ~ProtocolError(){}
    };
private:    
    MPI_Comm myComm;
    int nProcs;
    int myProc;
    
    std::map<Operation, MPI_Op> Op_to_MPI_Op;
};



#endif // !defined(MPIPROTOCOL_H)
