// SubjectObserver.h: interface for the Subject and Observer classes
//
//////////////////////////////////////////////////////////////////////

#ifndef OBSERVER_H
#define OBSERVER_H

#include <list>
#include "DigitalMaterial.h"

class Subject;
 //! Observers follow receive messages from generic Subjects.
/** 
  * \sa Subject
  * \ingroup dmsystem
  */
class Observer
{
public:
	virtual ~Observer();
	virtual void Update(Subject *theChangedSubject) = 0;

protected:
	Observer();
	
};


//! A Subject sends message to attached Observers.
/** Adapted from the Subject-Observer section of 
 * "Design Patterns, Elements of Reusable Object-Oriented Software"
 * by E. Gamma, R. Helm, R. Johnson, and J. Vlissides, 
 * Addison Wesley, Massachusetts, 1995
 *ISBN 0-201-63361-2
 *
 * The book suggests a type-specific subject-observer pattern which we have
 * abstracted here at the cost of using RTTI and dynamic casts in the 
 * concrete Subjects and Observers.
 * \sa Observer
 * \ingroup dmsystem
 */
class Subject  
{
public:
	virtual ~Subject(); 
	
	virtual void  AttachObserver(Observer* o) 
	{
		if (o)
			_observers->push_back(o);
	}
	virtual void Detach(Observer* o) 
	{
		if (o)
			_observers->remove(o);
	}
	virtual void Notify() 
	{
		std::list<Observer*>::iterator i = _observers->begin();
		while (i != _observers->end()) 
		{
			(*i)->Update(this);
			i++;
		}
	}

protected:
	Subject();
private:
#ifndef SWIG // XXX SWIG doesn't support standard template libraries
	std::list<Observer*> *_observers;
#endif // SWIG
};

#endif // !defined(OBSERVER_H)
