#include <math.h>
#include <algorithm>
#include <iostream>
#include "DiscreteAvalanche.h"
#include "SpinLattice.h"
#include "Rand.h"

Rand RandomGen;

void Allocation( SpinLattice* spin, int SideLength )
{
	int DimCnt = 2;
	spin->SetDimension( 2 );
	SpinCoord* dims = new SpinCoord[DimCnt];

	int DimIdx;
	for (DimIdx=0; DimIdx<DimCnt; DimIdx++) {
		dims[DimIdx] = SideLength;
	}

	spin->SetLengths( dims );

	delete[] dims;
}


void RandomSampleForSpin( const SpinLattice*const lattice,
						 Rand* const rand )
{
	int nup = 131313;
	bool bFound = false;
	bool bQuit = false;
	int N = lattice->GetLength();
	int timeToSlice = lattice->GetLength();
	int time = 0;
	SpinIndex loc;
	
	// Do a completely random search if there are still a lot of up spins.
	
		time = 0;
		while (!bFound && (time<timeToSlice) ) {
			loc = rand->discrete(0, N-1);
			if ( (-1 == lattice->Get(loc)) &&
				(nup == lattice->NeighborsUp(loc)) )
				bFound = true;
			++time;
		}
	if (bQuit) {
		TRACE("ProbabilisticStart::FindRandomSpin Quitting because batchTimer said to.\n");
	}
}

void LinearSearchForSpin( const SpinLattice*const lattice )
{
	int nup = 121212;
	bool bFound = false;
	bool bQuit = false;
	int N = lattice->GetLength();
	int timeToSlice = N;
	int time = 0;
	Rand* rand = &RandomGen;
	SpinIndex loc;
	
	TRACE("ProbabilisticStart::FindRandomSpin linear search.\n");
	// Do a linear search if there are only a few up spins.
	int i = 0;
	loc = 0;
	// First find all candidates in the lattice.
		time = 0;
		while ( (loc<N) && (time<timeToSlice) ) {
			if ( (-1 == lattice->Get(loc)) &&
				(nup == lattice->NeighborsUp(loc)) ) {
				//candidates[i] = loc;
				i++;
			}
			++loc;
			++time;
		}
		// Assert that we aren't looping forever because
		// the psNUp array is wrong.
	
	// Then choose one candidate at random.
	int choice = rand->discrete(0,N-1);
	
	if (bQuit) {
		TRACE("ProbabilisticStart::FindRandomSpin Quitting because batchTimer said to.\n");
	}
}



/*! This routine may spend too much time generating random numbers.
	*/
int RandomReads( SpinLattice* spins, int DeltaCount, int ReadCount )
{
	int ComputedQuantity;
	int ReadIndex;
	int DeltaIndex;
	int SpinCount;

	SpinCount = spins->GetLength();
	ComputedQuantity = 0;

	int NextSpinPick = 0;
	for ( DeltaIndex=0; DeltaIndex<DeltaCount; DeltaIndex++ ) {
		int RandomDelta = RandomGen.discrete(0,SpinCount-1);
		
		for ( ReadIndex=0; ReadIndex<ReadCount; ReadIndex++ ) {
			NextSpinPick = (NextSpinPick+RandomDelta) % SpinCount;
			ComputedQuantity += spins->Get( NextSpinPick );
		}
	}
	return ComputedQuantity;
}

/*! This routine may spend too much time generating random numbers.
	*/
void RandomWrites( SpinLattice* spins, int DeltaCount, int ReadCount )
{
	int ReadIndex;
	int DeltaIndex;
	int SpinCount;

	SpinCount = spins->GetLength();

	int NextSpinPick = 0;
	for ( DeltaIndex=0; DeltaIndex<DeltaCount; DeltaIndex++ ) {
		int RandomDelta = RandomGen.discrete(0,SpinCount-1);
		
		for ( ReadIndex=0; ReadIndex<ReadCount; ReadIndex++ ) {
			NextSpinPick = (NextSpinPick+RandomDelta) % SpinCount;
			spins->flip( NextSpinPick, 1 );
		}
	}
}

int RandomNeighbors( SpinLattice* spins, int DeltaCount, int ReadCount )
{
	int ComputedQuantity;
	int ReadIndex;
	int DeltaIndex;
	int SpinCount;

	SpinCount = spins->GetLength();
	ComputedQuantity = 0;

	int NextSpinPick = 0;
	for ( DeltaIndex=0; DeltaIndex<DeltaCount; DeltaIndex++ ) {
		int RandomDelta = RandomGen.discrete(0,SpinCount-1);
		
		for ( ReadIndex=0; ReadIndex<ReadCount; ReadIndex++ ) {
			NextSpinPick = (NextSpinPick+RandomDelta) % SpinCount;
			ComputedQuantity += spins->NeighborsUp( NextSpinPick );
		}
	}
	return ComputedQuantity;
}

int RandomWhichNeighbors( SpinLattice* spins, int DeltaCount, int ReadCount )
{
	int ComputedQuantity;
	int ReadIndex;
	int DeltaIndex;
	int SpinCount;
	SpinIndex Neighbors[12];

	SpinCount = spins->GetLength();
	ComputedQuantity = 0;

	int NextSpinPick = 0;
	for ( DeltaIndex=0; DeltaIndex<DeltaCount; DeltaIndex++ ) {
		int RandomDelta = RandomGen.discrete(0,SpinCount-1);
		
		for ( ReadIndex=0; ReadIndex<ReadCount; ReadIndex++ ) {
			NextSpinPick = (NextSpinPick+RandomDelta) % SpinCount;

			spins->GetNeighbors( NextSpinPick, Neighbors );

			ComputedQuantity += Neighbors[0];
			ComputedQuantity += Neighbors[1];
			ComputedQuantity += Neighbors[2];
			ComputedQuantity += Neighbors[3];
		}
	}
	return ComputedQuantity;
}

void GetNeighbors
	(
	const SpinIndex iPos,
	SpinIndex* const neighbors,
	const SpinIndex* const Strides,
	const SpinCoord* const Dimensions,
	int DimensionCnt
	)
{
	SpinIndex DecreasingIndex = iPos;

	int DimIdx; // still not right.  Compare with below.
	for ( DimIdx=0; DimIdx < DimensionCnt; DimIdx++ ) {

		// Find the coordinate along this side.
		SpinCoord SideLength = Dimensions[DimIdx];
		SpinCoord coord = DecreasingIndex % SideLength;
		DecreasingIndex /= SideLength;

		// Find the upper and lower neighbors of that coordinate.
		if ( coord == SideLength-1 ) {
			neighbors[2*DimIdx] = iPos +
				(Strides[DimIdx] - Strides[DimIdx+1] );
			neighbors[2*DimIdx+1] = iPos - Strides[DimIdx];

		} else if ( coord == 0 ) {
			neighbors[2*DimIdx] = iPos + Strides[DimIdx];
			neighbors[2*DimIdx+1] = iPos +
				Strides[DimIdx+1] - Strides[DimIdx];

		} else {
			neighbors[2*DimIdx] = iPos + Strides[DimIdx];
			neighbors[2*DimIdx+1] = iPos - Strides[DimIdx];
		}
	}
}



void CheckWhichNeighbors( SpinLattice* spins )
{
	const SpinCoord* Dimensions	= spins->GetSideLengths();
	int DimensionCnt		= spins->GetDimension();

	SpinIndex* GeneralNeighbors = new SpinIndex[DimensionCnt];
	SpinIndex* TheseNeighbors = new SpinIndex[DimensionCnt];

	SpinIndex* Strides = new SpinIndex[DimensionCnt+1];
	Strides[0] = 1;
	int DimIdx;
	for ( DimIdx=0; DimIdx<DimensionCnt; DimIdx++ ) {
		Strides[DimIdx+1] =
			Dimensions[DimIdx] * Strides[DimIdx];
	}

	int XIdx;
	for (XIdx = 0; XIdx<Dimensions[0]; XIdx++) {
		int YIdx;
		for (YIdx = 0; YIdx<Dimensions[1]; YIdx++) {
			SpinIndex iPos = YIdx*Dimensions[0]+XIdx;

			GetNeighbors
				(
				iPos,
				GeneralNeighbors,
				Strides,
				Dimensions,
				DimensionCnt
				);
			std::sort( GeneralNeighbors, GeneralNeighbors+DimensionCnt );

			spins->GetNeighbors( iPos, TheseNeighbors );
			std::sort( TheseNeighbors, TheseNeighbors+DimensionCnt );

			bool bNeighborsAreCorrect = true;
			int CoordinationIdx;
			for (
				CoordinationIdx=0;
				CoordinationIdx<spins->GetCoordinationNumber();
				CoordinationIdx++
				)
			{
				if ( GeneralNeighbors[ CoordinationIdx ] !=
					TheseNeighbors[ CoordinationIdx ] )
				{
					bNeighborsAreCorrect = false;
				}	ASSERT( bNeighborsAreCorrect );
			}

			if ( !bNeighborsAreCorrect ) {
				std::cout << "The neighbors do not match." << std::endl;
			}
		}
	}
}

void SecondCheckWhichNeighbors( SpinLattice* spins )
{
	const SpinCoord* Dimensions	= spins->GetSideLengths();
	int DimCnt					= spins->GetDimension();
	SpinIndex length			= spins->GetLength();
	int CoordinationCnt			= spins->GetCoordinationNumber();

	SpinIndex* Neighbors = new SpinIndex[CoordinationCnt];
	SpinCoord* CenterCoords = new SpinCoord[DimCnt];
	SpinCoord* NeighborCoords = new SpinCoord[DimCnt];

	// Loop over all lattice locations.
	SpinIndex scIdx;
	for ( scIdx = 0; scIdx < length; scIdx++ ) {
		spins->GetNeighbors( scIdx, Neighbors );
		spins->GetCoords( scIdx, CenterCoords );

		// Check that no two coordinates are the same.
		std::sort( Neighbors, Neighbors + CoordinationCnt );
		int* location = std::adjacent_find( Neighbors, Neighbors + CoordinationCnt );
		if ( location != Neighbors + CoordinationCnt ) {
			std::cout << "GetNeighbors returned two identical neighbors." << std::endl;
		}

		// Check the distance between spins for each lattice location.
		int CoordinationIdx;
		for (
			CoordinationIdx = 0;
			CoordinationIdx < CoordinationCnt;
			CoordinationIdx++
			)
		{
			spins->GetCoords( Neighbors[ CoordinationIdx ], NeighborCoords );

			// Check the distance for each dimension.
			int TotalDistance = 0;
			int DimIdx;
			for ( DimIdx = 0; DimIdx < DimCnt; DimIdx++ ) {
				int Distance = CenterCoords[DimIdx] - NeighborCoords[DimIdx];
				Distance = abs( Distance );
				if ( Distance == Dimensions[DimIdx]-1 ) Distance = 1;
				TotalDistance += Distance*Distance;
			}
			if ( TotalDistance != 1 ) {
				std::cout << "There is an error in neighbors of " << scIdx << std::endl;
				std::cout << "TotalDistanceSq is " << TotalDistance << std::endl;
				ASSERT( TotalDistance == 1 );
			}
		}
	}

	delete[] Neighbors;
	delete[] CenterCoords;
	delete[] NeighborCoords;
}

// Time the various functions in SpinLattice
int main( int argc, char *argv[] )
{
	int DeltaCount = 1000;
	int ReadCount = 1000;
	SpinLattice spins;

	Allocation( &spins, 1024 );

	// Check whether the WhichNeighbors routine works
	CheckWhichNeighbors( &spins );

	// Check WhichNeighbors relying on different kind of check.
	SecondCheckWhichNeighbors( &spins );

	// Search through whole lattice for a spin.
	LinearSearchForSpin( &spins );

	// Sample randomly for a spin.
	RandomSampleForSpin( &spins, &RandomGen );

	// Time getting random spins.
	RandomReads( &spins, DeltaCount, ReadCount );

	// Time setting the value of random spins.
	RandomWrites( &spins, DeltaCount, ReadCount );

	// Time getting the total number of spins up.
	int nUp = spins.GetTotalNumberUp();

	// Time getting the neighbors of a spin.
	RandomNeighbors( &spins, DeltaCount, ReadCount );
	RandomWhichNeighbors( &spins, DeltaCount, ReadCount );

	// Time getting edge neighbors of a spin.
	// Time getting bulk neighbors of a spin.

	return 0;
}