#ifndef _AVALANCHE_ALGORITHM_H_
#define _AVALANCHE_ALGORITHM_H_

#include "DMFirstHeader.h"
#include <queue>
#include "DiscreteAvalanche.h"
#include "SpinIndex.h"
#include "Rand.h"
#include "SpinLattice.h"
#include "FlipListener.h"
#include "ObserverList.h"

class BadArgument;
class GaussianInitializer;

//! Avalanche algorithm interface.
/** This avalanche algorithm stores the value of the random field internally.
  * It also saves the number of neighbors each spin has.
  * It evaluates avalanches using a spinFlipQueue, one shell at a time.
  *
  */
class AvalancheAlgorithm
{
public:
	enum Status {
		INVALID,
		SUCCESS,
		ALLFLIPPED,
		TIMEDOUT
	};

	//! Tell this object what SpinLattice it should use.
	virtual void SetLattice( SpinLattice *const lattice ) PUREVIRTUAL;
	//! Tell it how to initialize the random field.
	virtual void SetInitializer( GaussianInitializer *initializer ) PUREVIRTUAL;
	//! This call permits the class to set up cached internal variables.
	virtual void Initialize() PUREVIRTUAL;
	//! Ask the class whether it is all ready to go.
	virtual void Verify( void ) const PUREVIRTUAL;
	//! Ask what lattice this class uses.
	virtual const SpinLattice* GetLattice() const PUREVIRTUAL;
	//! Finds the start of the next avalanche.
	virtual Status FindNextStart( double* H, SpinIndex *iFirstSpin ) PUREVIRTUAL;
	//! Run through a single avalanche.
	virtual Status Cascade( double H, SpinIndex iFirstSpin ) PUREVIRTUAL;
	//! Return the magnetization
	virtual int GetM() const PUREVIRTUAL;

	//! Add an observer to the avalanche
	bool AddObserver( FlipListener* observer );
	//! Remove an observer from the avalanche
	bool RemoveObserver( FlipListener* observer );
	//! Notify the listeners
	void Notify( SpinLattice* lattice, const int* coords, int dDimension );
private:
	ObserverList<FlipListener> observers;

	//! This class helps us tell the observers what spin flipped.
	class FlipNotification
	{
		SpinLattice* lattice;
		const int* coords;
		int dDimension;
	public:
		FlipNotification( SpinLattice* lattice, const int* coords, int dDimension ) : 
			lattice(lattice), coords(coords), dDimension(dDimension)  {}
		void operator()( FlipListener* observer ) {
			observer->FlipCompleted( lattice, coords, dDimension );
		}
	};

};


#endif // _AVALANCHE_ALGORITHM_H_
