#include <algorithm>
#include "AvalancheStart.h"
#include "AvalancheAlgorithm.h"
#include "SpinLattice.h"
#include "FailedVerify.h"
#include "OrderByFields.h"

AvalancheStart::AvalancheStart()
{
	lattice = 0;
	piSortedSpins = 0;
	piNextSpin = 0;
	pbSpinDone = 0;

	pdRandomField = 0;
	pcNUp = 0;

	iVerify = 0;
}

AvalancheStart::~AvalancheStart()
{
	DeleteArrays();
}

/** You need to give it an AvalancheAlgorithm that has already
  * done its internal initializations, because this function will
  * steal const copies of some of the internal pointers.
  */
void AvalancheStart::SetAlgorithm( AvalancheAlgorithm* avalanche )
{
	this->lattice = avalanche->GetLattice();
	this->pdRandomField = avalanche->GetRandomField();
	this->pcNUp = avalanche->GetNUp();
	this->Z = lattice->GetCoordinationNumber();
	int L = lattice->GetLength();
	this->N = L*L;
	if (pdRandomField && pcNUp) iVerify |= ALGORITHM;
}

void AvalancheStart::Initialize()
{
	if ( (iVerify & ALGORITHM) == 0)
		throw FailedVerify("Can\'t call initialize until algorithm is set.");
	AllocateArrays();
	InitializeSortedSpins();
	InitializeSortedList();
	iVerify |= INITIALIZE;
}

void AvalancheStart::Verify()
	if (iVerify != 3) throw FailedVerify( "AvalancheStart failed to verify" );
}

/**
  * @param H returns the magnetic field where the next avalanche starts.
  * @param iFirstSpin returns the spin index where the avalanche starts.
  * @ret false if you can't have any more avalanches.  true if you can.
  */
bool AvalancheStart::FindNextStart( double &H, SpinIndex &iFirstSpin )
{
	int i;
	SpinIndex iMinSpin = -1, iMinUp = -1;
	double dTestH, dMinH = 1.0e30;
	bool bFound = false;

	while (!bFound) {

		bool bQuit = true;
		for (i=0; i<=Z; i++) {
			if (!pbSpinDone[i]) bQuit = false;
		}
		if (bQuit) return false;
		
		dMinH = 1.0e30;

		for (i=0; i<=Z; i++) {
			if (!pbSpinDone[i]) {
				dTestH = -(pdRandomField[ piSortedSpins[ piNextSpin[i] ] ] + 2*i - Z);
				if (dTestH < dMinH) {
					iMinSpin = piSortedSpins[ piNextSpin[i] ];
					dMinH = dTestH;
					iMinUp = i;
				}
			}
		}
		
		piNextSpin[iMinUp] += 1;

		if ( piNextSpin[iMinUp] >=N ) pbSpinDone[i] = true;

		if ( (lattice->Get( iMinSpin ) == -1) && (iMinUp == pcNUp[iMinSpin])) bFound = true;
	}

	H = dMinH;
	iFirstSpin = iMinSpin;
	return true;
}


// ******* Member functions below here are private. *************/
void AvalancheStart::AllocateArrays()
{
	try {
		piSortedSpins = DMAllocate<SpinIndex>(N);
		piNextSpin = DMAllocate<SpinIndex>(Z+1);
		pbSpinDone = DMAllocate<bool>(Z+1);
	} catch (std::bad_alloc e) {
		if (piSortedSpins!=0) {
			delete[] piSortedSpins;
			piSortedSpins = 0;
		}
		else throw FailedVerify( "Couldn\'t allocate enough memory in AvalancheAlgorithm.");
		if (piNextSpin!=0) {
			delete[] piNextSpin;
			piNextSpin = 0;
		}
		else throw FailedVerify( "Couldn\'t allocate enough memory in AvalancheAlgorithm.");
		if (pbSpinDone!=0) {
			delete[] pbSpinDone;
			pbSpinDone = 0;
		}
		else throw FailedVerify( "Couldn\'t allocate enough memory in AvalancheAlgorithm.");
	}
}

void AvalancheStart::DeleteArrays()
{
	if (piNextSpin) delete[] piNextSpin;
	if (pbSpinDone) delete[] pbSpinDone;
	if (piSortedSpins) delete[] piSortedSpins;

	piNextSpin = 0;
	pbSpinDone = 0;
	piSortedSpins = 0;
	// don't delete pdRandomField or pcNUp b/c we don't own them.
}

void AvalancheStart::InitializeSortedSpins()
{
	int i;
	for (i=0; i<N; i++) piSortedSpins[i] = i;

	OrderByFields order;
	order.randomField = pdRandomField;
	std::sort( piSortedSpins, piSortedSpins+N, order );
}

void AvalancheStart::InitializeSortedList()
{
	int i, tNS;
	double dt;
	double H = -pdRandomField[piSortedSpins[0]]+Z;

	for (i=0; i<Z; i++) {
		tNS = 0;
		pbSpinDone[i] = false;
		dt = -(2*i+H-Z);
		while ( (tNS<N) && ( pdRandomField[piSortedSpins[tNS]] > dt) )
			tNS++;
		piNextSpin[i] = tNS;
		if (piNextSpin[i]==N) pbSpinDone[i] = true;
	}
}