#include "DMFirstHeader.h"
#include <windows.h> // for GetMessage
#include <iostream>
#include "BatchTimer.h"
#include "DMException.h"
#include "Checkpointer.h"
#include "Tracer.h"
#include "ProcessFactory.h"
#include "DMLastHeader.h"



BatchTimer::BatchTimer()
{
	pCheckpointer = 0;
	pTracer = 0;
	EndTime = 0;
	SoonestStopTime = 0;
	bShouldQuit = false;
	bFirstCalled = true;
	bCheckTimes = true; // Go into loop first time.
	WallTimer.Start();
	eggTimer = ProcessFactory::Instance()->GetEggTimer();
	ASSERT( 0 != eggTimer );
	if ( 0 == eggTimer ) throw DMException("BatchTimer",
		"BatchTimer","Cannot get an egg timer from the ProcessFactory.");

	eggTimer->SetDueTime( 1 ); // start signaled.
	SetTimeCalculationFunction();
}

BatchTimer::~BatchTimer()
{
}

/*! Set when the simulation must end.  The simulation may end shortly
	after this time, shortly being definitely within the minute.
	
	@param uiNumberOfSeconds Seconds until the batch must quit.
	*/
void BatchTimer::SetEndTime( int NumberOfSeconds )
{
	ASSERT( NumberOfSeconds > 0 );
	if (NumberOfSeconds <= 0) throw DMException("BatchTimer",
		"SetEndTime","The number of seconds should be greater than zero.");

	this->EndTime = NumberOfSeconds;
	SetTimeCalculationFunction();
}

/*! The Checkpointer could be an observer, but there is only one for now.
	You shouldn't have to call this function, but you do for now.

	@param pCheckpointer Pointer to an object implementing the Checkpointer interface.
	*/
void BatchTimer::SetCheckpointer( Checkpointer *pCheckpointer )
{
	ASSERT( 0 != pCheckpointer );
	if (0 == pCheckpointer) throw DMException("BatchTimer",
		"SetCheckpointer","Passed a null pointer.");

	this->pCheckpointer = pCheckpointer;
	SetTimeCalculationFunction();
}

/*! The Tracer could be an observer, but there is only one for now.
	You shouldn't have to call this function, but you do for now.

	@param pTracer Pointer to an object implementing the Tracer interface.
	*/
void BatchTimer::SetTracer( Tracer *pTracer )
{
	ASSERT( 0 != pTracer );
	if ( 0 == pTracer ) throw DMException("BatchTimer",
		"SetTracer","Passed a null pointer");

	this->pTracer = pTracer;
	SetTimeCalculationFunction();
}

/*! Call this to make sure the class was properly constructed.
	It will throw a descriptive exception if it wasn't.
	*/
void BatchTimer::Verify( void )
{
	bool bSomeFunctionWasSet =
		( (0 != EndTime) || (0 != pTracer) || (0 != pCheckpointer) );
	ASSERT( bSomeFunctionWasSet );
	if (!bSomeFunctionWasSet) throw DMException("BatchTimer","Verify",
		"You must set one of a Tracer, a Checkpointer, or an EndTime.");
}


/*! When you call this, it doesn't always do recalculations.  It tries
	to do calculations only about every fifteen seconds so it won't interfere
	with program execution.
	*/
bool BatchTimer::TimeSlice()
{
	// Improve responsiveness more than a little on win32 systems.
#ifdef _DEBUG
	MSG msg;
	::PeekMessage( &msg, NULL, 0, 0, PM_NOREMOVE );
#endif
	Sleep(0);
	if ( !eggTimer->DoneYet() ) return false;

	CheckTimes();
	return bShouldQuit;
}

// *************** Private below this line ******************* //

void BatchTimer::SetTimeCalculationFunction( void )
{
	if (0 == pCheckpointer) {
		if (0 == pTracer) {
			CalculateStopAndQuit = SoonestStopFromEndTime;
		} else {
			CalculateStopAndQuit = SoonestStopFromTrace;
		}
	} else {
		if (0 == pTracer) {
			CalculateStopAndQuit = SoonestStopFromCheckpoint;
		} else {
			CalculateStopAndQuit = SoonestStopFromCheckpointOrTrace;
		}
	}

	this->bFirstCalled = true;
}


////////////////// Find the next time to stop ///////////////////////

BatchTimer::TimeType
BatchTimer::CheckCheckpoint( TimeType CurrentTime )
{
	// Set next checkpoint on first run through
	if (bFirstCalled) {
		NextCheckpoint =
			pCheckpointer->GetNextCheckpoint( 0, EndTime );
	}

	if ( CurrentTime >= NextCheckpoint ) {
		pCheckpointer->Checkpoint();
		NextCheckpoint =
			pCheckpointer->GetNextCheckpoint( CurrentTime, EndTime );
		if ( 0 > NextCheckpoint ) {
			bShouldQuit = true;
		}
	}
	return NextCheckpoint;
}

BatchTimer::TimeType
BatchTimer::CheckTrace( TimeType CurrentTime )
{
	// Set next trace on first run through
	if (bFirstCalled) {
		NextTrace =
			pTracer->GetNextTrace( 0, EndTime );
	}

	if ( CurrentTime >= NextTrace ) {
		pTracer->Trace( CurrentTime );
		NextTrace =
			pTracer->GetNextTrace( CurrentTime, EndTime );
	}
	return NextTrace;
}


/////////////// Logic of this particular combination of choices ///////


BatchTimer::TimeType
BatchTimer::SoonestStopFromCheckpoint( TimeType CurrentTime )
{
	CheckCheckpoint( CurrentTime );
	return NextCheckpoint;
}

BatchTimer::TimeType
BatchTimer::SoonestStopFromTrace( TimeType CurrentTime )
{
	CheckTrace( CurrentTime );
	if ( (this->EndTime != 0 ) && (CurrentTime > this->EndTime) ) {
		bShouldQuit = true;
	}
	return NextTrace;
}

BatchTimer::TimeType
BatchTimer::SoonestStopFromCheckpointOrTrace( TimeType CurrentTime )
{
	CheckCheckpoint( CurrentTime );
	CheckTrace( CurrentTime );

	SoonestStopTime =
		(NextCheckpoint<NextTrace) ? NextCheckpoint : NextTrace;

	return SoonestStopTime;
}

BatchTimer::TimeType
BatchTimer::SoonestStopFromEndTime( TimeType CurrentTime )
{
	ASSERT( this->EndTime > 0 );

	if ( CurrentTime > this->EndTime ) {
		bShouldQuit = true;
	}
	return EndTime;
}


////////////// The main routine called each time ///////////////////


void BatchTimer::CheckTimes()
{
	TimeType CurrentTime;

	CurrentTime = WallTimer.GetTime();

	SoonestStopTime = (this->*CalculateStopAndQuit)( CurrentTime );

	int iDelta = SoonestStopTime - CurrentTime;

	if (iDelta > 0) {
		TRACE( "BatchTimer::CheckTimes New timer %i.\n",iDelta);
		eggTimer->SetDueTime( iDelta );
	} // else let the timer keep stopping us.

	bFirstCalled = false;

	TRACE( "BatchTimer::CheckTimes Now %i Check %i Trace %i Soonest %i\n",
		CurrentTime, NextCheckpoint, NextTrace, SoonestStopTime );
}
