#ifndef _CHECK_SPANNING_H_
#define _CHECK_SPANNING_H_

#include "DMFirstHeader.h"
#include "DiscreteAvalanche.h"
#include "SpinIndex.h"
#include "SpinLattice.h"
#include "FlipListener.h"
#include "AvalancheListener.h"
#include "FlipListener.h"
#include "Serializable.h"
#include "SerialBuilder.h"

//! Check whether the current avalanche is spanning.
/*! This class is an observer of avalanches, so it gets thrown in
	with the other observers, but a lot of observers depend on the
	state of this one, so it should usually be first in the list
	of observers.

	\author Matt Kuntz
	\author ajd27 Added serialization, Initialize()
	*/
class CheckSpanning : public FlipListener, public AvalancheListener,
	public Serializable
{
	bool isSpanning; //!< Whether this avalanche is spanning.
	int numSpanning; //!< How many avalanches were spanning.
	const SpinLattice* lattice;
	int* before;  //!< One row before the avalanche (in each dimension)
	int* after; //!< One row after the avalanche (in each dimension)
	bool* spanDir; //!< Whether this dimension is spanning.
	int* extent;  //!< The size of this avalanche (in each dimension)
	bool firstSpin;  //! Whether this is the first spin of this avalanche.
public:
	//! Constructor.
	CheckSpanning();

	//! Destructor.
	~CheckSpanning();

	//! Tell it which lattice this algorithm uses.
	void SetLattice( const SpinLattice* lattice );

	//! Perform internal initializations prior to use.
	void Initialize();

	//! Call before starting a sweep.
	void BeginSweep( bool bKeep = false );

	//! This is called after every spin flip.
	void FlipCompleted( SpinLattice* lattice, const int* coords,
		int dDimension );

	//! This is called after every avalanche.
	void AvalancheCompleted( AvalancheAlgorithm* avalanche, double dField );

	//! Whether the avalanche is spanning
	bool IsSpanning() const { return isSpanning; }

	//! How many spanning avalanches have completed.
	int NumSpanning() const { return numSpanning; }

	//! The size of the current avalanche in the given dimension.
	int SpatialExtent(int dim) const {
		return extent[dim];
	}

	//! The row before the current avalanche in given dimension.
	int EmptyRow(int dim) const {
		return before[dim];
	}

	/*! Serializable interface */
	//@{
	void Save( DMWriter *writer ) const;
	void Load( DMReader *reader );
	const std::string GetType() const;
	//@}
private:
	/*! Internal pointer cleanup. */
	//@{
	void ZeroArrays();
	void AllocateArrays();
	void DeleteArrays();
	//@}

	/*! These are both cached variables. */
	//@{
	int D; //!< Dimension of the lattice.
	SpinCoord* sideLength; //!< Length of a lattice side.
	//@}

	enum { VERIFY_LATTICE=0x1 };
	int iVerify;
	bool bSerialized; //!< Whether we are reloading from being serialized.
};

#endif //_CHECK_SPANNING_H_
