#include "DMFirstHeader.h"
#include <math.h> // for pow()
#include <iostream>
#include <stdlib.h>
#include "DigitalMaterial.h"
#include "ColoringObserver.h"
#include "SpinLattice.h"
#include "DMException.h"
#include "NotEnoughMemory.h"
#include "FailedVerify.h"
#include "DMAllocator.h"
#include "SerialBuilder.h"
#include "DMReader.h"
#include "DMWriter.h"
#include "DMLastHeader.h"

const std::string ColoringObserver::sType("ColoringObserver");

/*! This static object permits the DMReader to create an empty ColoringObserver. */
const SerialBuilder<ColoringObserver> *_creator = new
	SerialBuilder<ColoringObserver>(ColoringObserver::sType);


/*! The constructor just makes everything zero except the default decimation
	of one.
	*/
ColoringObserver::ColoringObserver( )
{
	this->iDecimation = 1;		// default value.
	this->iMinDecimation = 0;
	this->iBoundsDimension = 0;
	this->maxMemory = 0;
	this->lattice = 0;

	this->D = 0;

	this->iCurrentColor = 1; // Start with 1 so zero is untouched region.
	ZeroArrays();

	this->iVerify = 0;
	this->bSerialized = false;
}

/*! Deletes allocated arrays including the color array, so
	make a copy before you kill this.
	*/
ColoringObserver::~ColoringObserver( )
{
	DeleteArrays();
}

/*! This sets a constraint on the maximum allowed amount of memory.
	The class will set the decimation to meet this constraint.

	If you do not set a maximum amount of memory, none will be
	enforced.

	In the interests of paranoia, what do we do if the maxmem is so small
	that no amount of decimation can make our lattice fit?  Maybe if
	someone thought the maximum amount of memory was set in megabytes
	and chose maxmem=1.

	\param maxmem This is in bytes.
	*/
void ColoringObserver::SetMaxMemory( size_t maxmem )
{
	ASSERT( 0 < maxmem );

	if ( 0 >= maxmem ) throw DMException("ColoringObserver",
		"SetMaxMemory", "Allowed memory must be greater than zero.");

	this->maxMemory = maxmem;
	this->iVerify |= MAXMEM;
}

/*! This sets the maximum size of the array which this class
	will save.  This class will make a copy of this array.

	\param piBounds The array with the size of each dimension.
			Entries must be greater than zero.
	\param dDimension The number of elements in the array, which
			must be greater than or equal to the dimension of the
			lattice.
	*/
void ColoringObserver::SetMaxBounds( const int *piBounds, int dDimension )
{
	// Check input arguments.
	ASSERT( 0 != piBounds );
	if ( 0 == piBounds ) throw DMException("ColoringObserver",
		"SetMaxBounds","Passed null pointer to bounds.");

	ASSERT( 0 < dDimension );
	if ( 0 >= dDimension ) throw DMException("ColoringObserver",
		"SetMaxBounds", "Dimension must be greater than zero.");

	int i;
	for (i=0; i<dDimension; i++)
		ASSERT( piBounds[i] > 0 );
	for (i=0; i<dDimension; i++)
		if (piBounds[i] <= 0) throw DMException("ColoringObserver",
			"SetMaxBounds", "Bounds should be greater than zero.");

	// Store user's choices.
	this->iBoundsDimension = dDimension;

	if (this->piMaxBounds) delete[] this->piMaxBounds;
	try {
		this->piMaxBounds = DMAllocate<int>(dDimension);
	} catch ( std::bad_alloc ) {
		if (0 == this->piMaxBounds) throw NotEnoughMemory("ColoringObserver",
			"SetMaxBounds", "Allocating small array of dimension ints.");
	}

	for (i=0; i<dDimension; i++) {
		this->piMaxBounds[i] = piBounds[i];
	}

	// Note that this was called.
	this->iVerify |= MAXBOUNDS;
}

/*! The decimation factor says that the coloring matrix will
	store the value of every Nth spin.

	The decimation is the variable used internally to decide how
	to cut down on the number of points stored.  If the decimation
	you set determines a grid size outside the maximum bounds or
	outside the maximum memory limit, then this will throw an
	exception.
	*/
void ColoringObserver::SetDecimation( int iDecimation )
{
	ASSERT( 0 < iDecimation );
	if ( 0 >= iDecimation ) throw DMException("ColoringObserver",
		"SetDecimation", "Decimation must be greater than zero.");

	this->iMinDecimation = iDecimation;

	this->iVerify |= DECIMATE;
}

void ColoringObserver::SetLattice( SpinLattice* lattice )
{
	ASSERT( 0 != lattice );
	if (0 == lattice) throw DMException("ColoringObserver",
		"SetLattice","Passed a null pointer.");

	this->lattice = lattice;

	this->iVerify |= LATTICE;
}

/*! The only precondition for calling Initialize is SetLattice, and
	you can call Initialize as often as you want.  It will clear out the
	current set of avalanche information each time.  Any settings
	you made with SetMaxMemory, SetMaxBounds, or SetDecimation will
	remain set through an initialization.
	*/
void ColoringObserver::Initialize()
{
	// Check parameters.
	ASSERT( 0 != this->lattice );
	if (0 == this->lattice) throw DMException("ColoringObserver",
		"Initialize","You must call SetLattice first.");

	// Set variables cached from the lattice.
	this->D = this->lattice->GetDimension();
	SUPERASSERT( D > 0 ); // wildly paranoid asserts get SUPERASSERT

	// Check that if someone set boundaries they have enough dimensions.
	if ( (this->iVerify&MAXBOUNDS) != 0 ) {
		ASSERT( this->D <= this->iBoundsDimension );
		if ( this->D > this->iBoundsDimension ) throw 
			DMException("ColoringObserver","Initialize",
			"The lattice has more dimensions than the bounds from SetMaxBounds.");
	}

	// This is temporary.  SpinLattice should provide this.
	const SpinCoord* piLatticeBounds = this->lattice->GetSideLengths();

	// Collect estimates of the decimation from our requirements.
	int iBoundsDecimate = this->BoundsToDecimation(piLatticeBounds,D);
	ASSERT( iBoundsDecimate > 0 );
	int iMemoryDecimate = this->MemoryToDecimation(piLatticeBounds,D);
	ASSERT( iMemoryDecimate > 0 );

	if (0 != this->iMinDecimation) this->iDecimation = this->iMinDecimation;
	else this->iDecimation = 1;

	// Choose the largest required decimation.
	if (iBoundsDecimate > this->iDecimation)
		this->iDecimation = iBoundsDecimate;
	if (iMemoryDecimate > this->iDecimation)
		this->iDecimation = iMemoryDecimate;

	this->SetBoundsFromDecimation(piLatticeBounds,D);
	this->AllocateColors();
	this->ClearColorArray();

	this->iVerify = INITIALIZED;
}

void ColoringObserver::Verify()
{
	ASSERT( 0 != lattice );
	if (0 == lattice) throw DMException("ColoringObserver",
		"Verify","You need to initialize the lattice.");

	ASSERT( INITIALIZED == iVerify );
	if ( INITIALIZED != iVerify ) throw FailedVerify("ColoringObserver",
		"Have not yet called Initialize.");
}

void ColoringObserver::BeginSweep( bool bKeepData )
{
	this->Verify();
	this->ClearColorArray();
}

/*! This should be very fast, but that requires a tight algorithm
	for this simple loop.  I remember Matt's code being nice here.
	It is worth looking up.
	*/
void ColoringObserver::FlipCompleted( SpinLattice* lattice,
											const int* coords,
											int dDimension)
{
	ASSERT( 0 != lattice );
	ASSERT( this->lattice == lattice );
	ASSERT( 0 != coords );
	ASSERT( 0 < dDimension );
	ASSERT( this->D == dDimension );

	size_t pos = CoordToPosition( coords );
	pColors[pos] = iCurrentColor;
}


/*! At the completion of an avalanche, we increment the counter which
	tells us which avalanches the spin flips are in.  This value is 
	stored in an unsigned int, so it can wrap around if you have more
	than 2^32 avalanches.

	Maybe we should store how many times it wraps around?  Or we could
	let you set the number at which it wraps around so it would match
	your color table.
	*/
void ColoringObserver::AvalancheCompleted( AvalancheAlgorithm* avalanche,
												 double dField )
{
	++iCurrentColor;
	// if we went all the way around, keep 0 as unflipped.
	if ( 0 == iCurrentColor ) ++iCurrentColor;
}


// *************** PRIVATE BELOW THIS LINE ********************* //


int ColoringObserver::BoundsToDecimation( const int* piLatticeBounds,
										 int dDimension )
{
	int i;
	int iDecimation = 1;

	if ( (iVerify&MAXBOUNDS) == 0) return iDecimation;

	for (i=0; i<dDimension; i++) {
		int iTrialDecimation = 1+(piLatticeBounds[i]-1)/piMaxBounds[i];
		iDecimation = (iTrialDecimation > iDecimation) ? iTrialDecimation : iDecimation;
	}
	return iDecimation;
}

/*! This figures out what decimation would make our total memory
	usage less than the value set with SetMaxMemory or, if that
	wasn't called, the default value of decimation of 1.

	The actual memory usage isn't just the product of (L+1) for
	each dimension divided by the decimation, because that would
	overestimate memory usage.  It's really, [(x1+1)/d][(x2+1)/d]
	and so on, where the brackets indicate that you take the floor
	of the value.  Therefore we start with an overestimate of the 
	decimation and decrement it until it just fits in memory.
	*/
int ColoringObserver::MemoryToDecimation( const int* piLatticeBounds,
										 int dDimension )
{
	int i;
	size_t iMem;

	// Set a default value of the max memory if it wasn't set.
	if ( (this->iVerify&MAXMEM) == 0 ) return 1;

	// Use straight division to overshoot the correct decimation.
	int iTrialDecimation = sizeof(COLOR);
	for (i=0; i<dDimension; i++) iTrialDecimation *= (piLatticeBounds[i]+1);
	iTrialDecimation = (int) pow(iTrialDecimation/this->maxMemory,0.3333333);

	// Decrement the decimation until we overshoot the memory mark.
	// We know we overshot, so we can start by decrementing it.
	do {
		--iTrialDecimation;
		iMem = sizeof(COLOR);
		for (i=0; i<dDimension; i++)
			iMem *= piLatticeBounds[i]/iTrialDecimation;
	} while ( iMem < this->maxMemory );
	
	// +1 because the loop above stops when the decimation is one too large.
	return iTrialDecimation+1;
}

void ColoringObserver::SetBoundsFromDecimation( const int* piLatticeBounds,
											   int dDimension )
{
	if (this->piBounds) delete[] this->piBounds;

	try {
		this->piBounds = DMAllocate<int>(dDimension);
	} catch ( std::bad_alloc ) {
		throw NotEnoughMemory("ColoringObserver",
			"SetBoundsFromDecimation","Allocating dimension ints.");
	}


	int maybeBadPadding = 1;
	if ( 1 == iDecimation ) maybeBadPadding = 0;

	int i;
	for (i=0; i<dDimension; i++)
		this->piBounds[i] = (piLatticeBounds[i]+maybeBadPadding)/iDecimation;
}

/*! Does not allocate piMaxBounds because it is a user
	setting.  Handle it separately.
	*/
void ColoringObserver::AllocateColors()
{
	if (this->pColors) delete[] this->pColors;
	this->pColors = 0;
	iColorCnt = 0;

	// Don't allocate bounds, because it was done above.
	size_t iMax = 1;
	int i;
	for (i=0; i<this->D; i++)
		iMax *= this->piBounds[i];

	try {
		this->pColors = DMAllocate<COLOR>(iMax);
	} catch ( std::bad_alloc ) {
		throw NotEnoughMemory( "ColoringObserver","AllocateColors",
			"Allocating large array of avalanche ordering." );
	}
	iColorCnt = iMax;
}

void ColoringObserver::ClearColorArray()
{
	::memset( this->pColors, 0, sizeof(COLOR)*iColorCnt );
}

void ColoringObserver::ZeroArrays()
{
	this->pColors		= 0;
	this->piBounds		= 0;
	this->piMaxBounds	= 0;
}


void ColoringObserver::DeleteArrays()
{
	if (this->pColors)		delete[] this->pColors;
	if (this->piBounds)		delete[] this->piBounds;
	if (this->piMaxBounds)	delete[] this->piMaxBounds;

	ZeroArrays();
}


void ColoringObserver::Save( DMWriter* writer ) const
{
	// User settings
	writer->PutInt( iMinDecimation, "minDecimation" );
	writer->PutInt( maxMemory,		"maxMemory" );

	// Class choices
	writer->PutInt( iBoundsDimension, "boundsDimension" );
	writer->PutInt( iDecimation,	  "decimation" );
	writer->PutInt( iColorCnt,		  "colorCnt" );

	// Current color state
	writer->PutInt( iCurrentColor, "currentColor" );

	// Verification
	writer->PutInt( D, "dimension" );
	writer->PutInt( iVerify, "verify" );

	std::vector<int> colorShape(1);
	colorShape[0] = iColorCnt;
	writer->PutIntArray( reinterpret_cast<const int*>(pColors),
		colorShape, "colors" );

	std::vector<int> boundsShape(1);
	boundsShape[0] = D;
	writer->PutIntArray( piBounds, boundsShape, "bounds" );

	std::vector<int> maxBoundsShape(1);
	maxBoundsShape[0] = D;
	writer->PutIntArray( piMaxBounds, maxBoundsShape, "maxBounds" );
}


void ColoringObserver::Load( DMReader *reader )
{
	// User settings
	iMinDecimation  = reader->GetInt( "minDecimation" );
	maxMemory		= reader->GetInt( "maxMemory" );

	// Class choices
	iBoundsDimension = reader->GetInt( "boundsDimension" );
	iDecimation		 = reader->GetInt( "decimation" );
	iColorCnt		 = reader->GetInt( "colorCnt" );

	// Current color state
	iCurrentColor = reader->GetInt( "currentColor" );

	D		= reader->GetInt( "dimension" );
	iVerify = reader->GetInt( "verify" );

	// Turn off the sign we were initialized.
	iVerify ^= INITIALIZED;

	std::vector<int> colorShape;
	int* piColors = reader->GetIntArray( colorShape, "colors" );
	pColors = reinterpret_cast<unsigned int*>( piColors );
	ASSERT( colorShape.size() == 1 );
	ASSERT( colorShape[0] == iColorCnt );
	if ( colorShape[0] != iColorCnt ) {
		throw DMException("ColoringObserver","Load",
			"The color array is the wrong length.");
	}

	std::vector<int> boundsShape;
	piBounds = reader->GetIntArray( boundsShape, "bounds" );
	ASSERT( boundsShape.size() == 1 );
	ASSERT( boundsShape[0] == D );
	if ( boundsShape[0] != iColorCnt ) {
		throw DMException("ColoringObserver","Load",
			"The bounds array is the wrong length.");
	}

	std::vector<int> maxBoundsShape;
	piMaxBounds = reader->GetIntArray( maxBoundsShape, "maxBounds" );
	ASSERT( maxBoundsShape.size() == 1 );
	ASSERT( maxBoundsShape[0] == D );
	if ( maxBoundsShape[0] != iColorCnt ) {
		throw DMException("ColoringObserver","Load",
			"The max bounds array is the wrong length.");
	}
}


std::ostream& ColoringObserver::WriteVTKFile( std::ostream& s ) const
{
	int* coords = new int[this->D];
	int dimIdx;
  
  s << "# vtk DataFile Version 2.0" << std::endl;
  s << "Coloring for a spin dataset" << std::endl;
  s << "ASCII" << std::endl;
	s << "DATASET STRUCTURED_POINTS" << std::endl;
	s << "DIMENSIONS " << this->piBounds[0];
	for ( dimIdx = 1; dimIdx < this->D; dimIdx++ )
		s << " " << this->piBounds[ dimIdx ];
	s << std::endl;

	s << "SPACING 1";
	for ( dimIdx = 1; dimIdx < this->D; dimIdx++ )
		s << " 1";
	s << std::endl;

  s << "ORIGIN 0";
	for ( dimIdx = 1; dimIdx < this->D; dimIdx++ )
		s << " 0";
	s << std::endl;

  s << "POINT_DATA " << this->iColorCnt << std::endl;

	s << "SCALARS AvalancheIndex unsigned_int 1" << std::endl;
	s << "LOOKUP_TABLE default" << std::endl;

	for ( int pos = 0; pos < this->iColorCnt; pos++ ) {
			s << this->pColors[ pos ] << std::endl;
	}

	delete[] coords;
	return s;
}

std::ostream& operator<<( std::ostream& s, const ColoringObserver& r )
{
  return r.WriteVTKFile( s );
}

std::ostream& ColoringObserver::WriteSimpleFile( std::ostream& s ) const
{
  int* coords = new int[ this->D ];
	s << "DIMENSIONS " << this->piBounds[0];
	for ( int dimIdx = 1; dimIdx < this->D; dimIdx++ )
		s << " " << this->piBounds[ dimIdx ];
	s << std::endl;

  for ( int pos = 0; pos < this->iColorCnt; pos++ ) {

		if ( 0 != this->pColors[pos] ) {
			this->GetCoords( pos, coords );
			s << coords[0];
			for ( int dimIdx = 1; dimIdx < this->D; dimIdx++ ) {
				s << " " << coords[ dimIdx ];
			}
			s << " ";
			s << this->pColors[ pos ] << std::endl;
		}

  }
  delete[] coords;
  return s;
}


#ifndef NDEBUG
void ColoringObserver::TestMethod()
{
	int i;
	int iMemDecimation, iBoundsDecimation;

	int *piSaveMaxBounds;
	// Copy current values to saved ones.
	int iSaveVerify = iVerify;
	if (iVerify&MAXBOUNDS) {
		piSaveMaxBounds = new int[D];
		for (i=0; i<D; i++) piSaveMaxBounds[i] = piMaxBounds[i];
	} else {
		piMaxBounds = new int[6];
	}
	size_t iSaveMaxMemory = maxMemory;

	// Set up a piLatticeBounds;
	int piLatticeBounds[6] = { 5, 6, 7, 5, 5, 5 };

	// Test with no settings.
	iVerify = 0;
	iMemDecimation = MemoryToDecimation( piLatticeBounds, 1 );
	ASSERT( iMemDecimation == 1 );
	iBoundsDecimation = BoundsToDecimation( piLatticeBounds, 1 );
	ASSERT( iBoundsDecimation == 1 );

	// Test with settings
	// On 1D lattice
	iVerify &= MAXMEM;
	iVerify &= MAXBOUNDS;
	maxMemory = 6*sizeof(COLOR);
	piMaxBounds[0] = 6;
	iMemDecimation = MemoryToDecimation( piLatticeBounds, 1 );
	ASSERT( iMemDecimation == 1 );
	iBoundsDecimation = BoundsToDecimation( piLatticeBounds, 1 );
	ASSERT( iBoundsDecimation == 1 );

	maxMemory = 5*sizeof(COLOR);
	piMaxBounds[0] = 5;
	iMemDecimation = MemoryToDecimation( piLatticeBounds, 1 );
	ASSERT( iMemDecimation == 1 );
	iBoundsDecimation = BoundsToDecimation( piLatticeBounds, 1 );
	ASSERT( iBoundsDecimation == 1 );

	maxMemory = 4*sizeof(COLOR);
	piMaxBounds[0] = 4;
	iMemDecimation = MemoryToDecimation( piLatticeBounds, 1 );
	ASSERT( iMemDecimation == 2 );
	iBoundsDecimation = BoundsToDecimation( piLatticeBounds, 1 );
	ASSERT( iBoundsDecimation == 2 );

	// On 3D lattice
	maxMemory = (5*6*7+1)*sizeof(COLOR);
	piMaxBounds[0] = 6;
	piMaxBounds[1] = 7;
	piMaxBounds[2] = 8;
	iMemDecimation = MemoryToDecimation( piLatticeBounds, 3 );
	ASSERT( iMemDecimation == 1 );
	iBoundsDecimation = BoundsToDecimation( piLatticeBounds, 3 );
	ASSERT( iBoundsDecimation == 1 );

	maxMemory = (5*6*7)*sizeof(COLOR);
	piMaxBounds[0] = 5;
	piMaxBounds[1] = 6;
	piMaxBounds[2] = 8;
	iMemDecimation = MemoryToDecimation( piLatticeBounds, 3 );
	ASSERT( iMemDecimation == 1 );
	iBoundsDecimation = BoundsToDecimation( piLatticeBounds, 3 );
	ASSERT( iBoundsDecimation == 1 );

	maxMemory = (5*6*7-1)*sizeof(COLOR);
	piMaxBounds[0] = 5;
	piMaxBounds[1] = 6;
	piMaxBounds[2] = 6;
	iMemDecimation = MemoryToDecimation( piLatticeBounds, 3 );
	ASSERT( iMemDecimation == 2 );
	iBoundsDecimation = BoundsToDecimation( piLatticeBounds, 3 );
	ASSERT( iBoundsDecimation == 2 );

	// Restore internal variables.
	iVerify = iSaveVerify;
	if (iVerify&MAXBOUNDS) {
		for (i=0; i<D; i++) piSaveMaxBounds[i] = piMaxBounds[i];
		delete[] piSaveMaxBounds;
	} else {
		delete[] piMaxBounds;
		piMaxBounds = 0;
	}
	maxMemory = iSaveMaxMemory;
}

#endif //NDEBUG