#ifndef _COLORING_OBSERVER_H_
#define _COLORING_OBSERVER_H_

#include "DMFirstHeader.h"
#include <stdlib.h>
#include "DigitalMaterial.h"
#include "SpinLattice.h"
#include "FlipListener.h"
#include "AvalancheListener.h"

class ColoringObserver;
class AvalancheAlgorithm;

//! Stores grid showing what avalanche each spin flip belongs to.
/*! This class tells us the index of the avalanche for which a 
	spin flipped.  Because memory requirements will restrict how much
	data we can store, we may only be able to save some fraction of the
	lattice's spins or only the index modulo a small number (so we 
	can store with bytes instead of integers).

	It is called a ColoringObserver because the index of the avalanche
	is what we use to determine the avalanche color on the screen.

	There should be two separate classes for measuring the coloring,
	one for showing it on the screen and one for storing it for calculations.
	This one seems better suited right now for storing calculations.
	It works in N dimensions.
	If you want to make a coloring observer which, for instance, 
	doesn't decimate under low memory conditions but instead combines
	colors in some way, then start by creating a new class.

	\author ajd27
	\date 31 Oct 2000 (boo!)
	*/
class ColoringObserver : public FlipListener, public AvalancheListener,
	public Serializable
{
public:
	typedef unsigned int COLOR;	//!< We could store colors as bytes.
private:
	// Internal arrays with current configuration.
	COLOR *pColors;		//!< Array of colors.
	int iColorCnt;		//!< Size of color array.
	int *piBounds;		//!< Bounds of color array.
	int iDecimation;	//!< Decimation factor for coloring array.

	// User settings through public interface.
	int *piMaxBounds;		//!< Maximum size of color array.
	int iBoundsDimension;	//!< Dimension of the user-supplied bounds.
	int iMinDecimation;		//!< Set by SetDecimation.
	size_t maxMemory;		//!< Max amount of memory we can use.

	int D;					//!< Dimension of the lattice
	COLOR iCurrentColor;	//!< Color of this avalanche.

	SpinLattice* lattice;	//!< Pointer to the single lattice we track.
public:
	//! Constructor.
	ColoringObserver( );
	//! Destructor.
	~ColoringObserver( );
	//! Limit the amount of memory we can use.
	void SetMaxMemory( size_t maxmem );
	//! Limit the size of the grid to save.
	void SetMaxBounds( const int *piBounds, int dDimension );
	//! Set the decimation factor.
	void SetDecimation( int iDecimation );
	//! Give us a SpinLattice to initialize from.
	void SetLattice( SpinLattice* lattice );
	//! Initializes internal data.
	void Initialize();
	//! Verifies that everything is ready to run.
	void Verify();
	//! Call at the start of a sweep.
	void BeginSweep( bool bKeepData );
	//! To be a FlipListener.
	void FlipCompleted( SpinLattice* lattice, const int* coords,
		int dDimension );
	//! To be an AvalancheListener.
	void AvalancheCompleted( AvalancheAlgorithm* avalanche,
		double dField );
	//! Get the data.
	const COLOR* GetColoring( void ) const { return pColors; }
	//! Find the dimensions of this array.
	const int* GetBounds( void ) const { return piBounds; }
	//! May as well give the dimension.
	int GetDimension( void ) const { return D; }
	//! Returns the color at a given position.
	int GetColor( const int* coords ) const;

	//! Serializable interface
	//@{
	void Save( DMWriter *writer ) const;
	void Load( DMReader *reader );
	const std::string GetType() const { return sType; }
	//@}
	static const std::string sType;

  std::ostream& WriteSimpleFile( std::ostream& s ) const;
  std::ostream& WriteVTKFile( std::ostream& s ) const;

	friend std::ostream& operator<<( std::ostream& s,
		const ColoringObserver& r );
private:
	int CoordToPosition( const int* coords ) const;
	void GetCoords( int iPos, int *const coords ) const;	
	//! Compute the decimation from the maximum lattice bounds.
	int BoundsToDecimation(const int *piLatticeBounds, int dDimension);
	//! Compute the decimation from the maximum memory.
	int MemoryToDecimation(const int *piLatticeBounds, int dDimension);
	//! Now that we know the decimation, set the actual color array bounds.
	void SetBoundsFromDecimation(const int *piLatticeBounds, int dDimension);

	void ZeroArrays();
	void DeleteArrays();
	//! Allocate internal arrays after dimensions are set.
	void AllocateColors();
	void ClearColorArray();

	// The iVerify flags tell us whether one of the settings was changed.
	// The first time the class is instantiated, we can tell whether a setting
	// was changed by whether it is nonzero, but the second time through,
	// we need a flag to figure out whether someone reset a setting.
	enum {
		DECIMATE = 0x1,		//!< The user requested a decimation.
		MAXMEM = 0x2,		//!< The user set a max memory.
		MAXBOUNDS = 0x4,	//!< The user set max boundaries to the lattice.
		LATTICE = 0x8,		//!< Someone set the lattice.
		INITIALIZED = 16
	};
	int iVerify;			//!< Stores the bit flags defined in the enum.
	bool bSerialized;		//!< Whether we were just read from disk.
#ifndef NDEBUG
	void TestMethod();
#endif // NDEBUG
};


/*! Convert the lattice location (held in _coords_) into an index
	in the object's coloring array.

	XXX ajd27 - Why is this optimized when it wasn't tested?
	*/
inline int ColoringObserver::CoordToPosition( const int* coords ) const
{
	int DimLim = D-1;

	const int* pc = coords + DimLim;
	const int* pb = piBounds + DimLim;

	size_t pos = *pc/iDecimation;
	while ( pc > coords ) {
		--pc;
		--pb;
		pos *= *pb;
		pos += (*pc)/iDecimation;
	}
	return pos;
}

inline int ColoringObserver::GetColor( const int* coords ) const
{
	size_t pos = CoordToPosition( coords );
	return 	pColors[pos];
}

inline void
ColoringObserver::GetCoords( int iPos, int *const coords ) const
{
	int DimIdx;
	for( DimIdx=0; DimIdx < this->D; DimIdx++ ) {

		int SideLength = this->piBounds[DimIdx];
		coords[DimIdx] = iPos % SideLength;
		iPos /= SideLength;

	}
}


#endif //_COLORING_OBSERVER_H_
