#include "DMFirstHeader.h"
#include "DigitalMaterial.h"
#include "ExceptionHandler.h"
#include "DMLastHeader.h"

// Tell the compiler to load this static
// instance early.  Turn off the warning because, of course,
// we know what we're doing, right?
#pragma warning( disable : 4073 )
#pragma init_seg( lib )

ExceptionHandler* _exceptionHandler = new ExceptionHandler();
#pragma warning( default : 4073 )

/*!	Write to the console without using ostream services.
	Avoid printf because it can use a library which competes
	with ostream.  Additionally, by the time we get to 
	this handler, our application may not have much stack
	left, so it's safest to use someone else's.
	*/
void ReportError( const char * message )
{
	ASSERT( message != 0 );
	if ( message == 0 ) return;
	HANDLE console = GetStdHandle( STD_OUTPUT_HANDLE );
	ASSERT( console != INVALID_HANDLE_VALUE );
	if (console == INVALID_HANDLE_VALUE) {
		return;
	}
	
	int iMessageLen = lstrlen( message );
	DWORD dwNumWritten = 0;
	BOOL iRet = WriteConsole( console,
		message,
		iMessageLen,
		&dwNumWritten,
		NULL );
	ASSERT( iRet != 0 );
	ASSERT( dwNumWritten > 0 );
	
	iRet = WriteConsole( console,
		TEXT("\n"),
		1,
		&dwNumWritten,
		NULL );
	ASSERT( iRet != 0 );
	ASSERT( dwNumWritten > 0 );
}

/*! This is called whenever there is an unhandled exception.
	It prints the results and lets you die in peace.
	*/
LONG __stdcall ExcepCallBack ( EXCEPTION_POINTERS * pExPtrs )
{
#ifdef HAVE_BUGSLAYER
	DWORD dwOpts = GSTSO_PARAMS | GSTSO_MODULE | GSTSO_SYMBOL |
		GSTSO_SRCLINE;
	ReportError( GetFaultReason( pExPtrs ) );
	
	ReportError( GetRegisterString( pExPtrs ) );
	const char * szBuff = GetFirstStackTraceString ( dwOpts  ,
		pExPtrs  ) ;
	do
	{
		ReportError( szBuff );
		szBuff = GetNextStackTraceString ( dwOpts , pExPtrs ) ;
	}
	while ( NULL != szBuff ) ;
#endif	
	return ( EXCEPTION_CONTINUE_SEARCH ) ;
}

/*! The constructor does two things:  It sets crash
	handler to catch exceptions, and it tells the 
	operating system to ignore our exceptions (no
	dialog boxes).

	It also sets the thread priority below normal so you can still
	kill it on a user machine.  Is it counter-intuitive to set the 
	thread priority low?  Processes spend most of their time waiting 
	to be scheduled by the system.  In our case, we are either running
	in batch or running while we test the code.  If we are in batch,
	then our program has no competition for processor time, and it makes
	no difference that it has low priority.  If we are in on the machine,
	then we need it to be responsive, and you will find that the simulation
	still gets 99% of the processor time.
	*/
ExceptionHandler::ExceptionHandler()
{
#ifdef HAVE_BUGSLAYER
	if ( TRUE != SetCrashHandlerFilter ( ExcepCallBack ) )
    {
		ReportError( "Failed to set crash handler." );
	}
	// Turn off use of msgbox in release builds.
#ifndef _DEBUG
	SetDiagAssertOptions( DA_SHOWODS | DA_SHOWSTACKTRACE );
#endif //_DEBUG
#endif
	/* These print debug strings to stdout.
	_CrtSetReportMode( _CRT_WARN, _CRTDBG_MODE_FILE );
	_CrtSetReportFile( _CRT_WARN, _CRTDBG_FILE_STDOUT );
	_CrtSetReportMode( _CRT_ERROR, _CRTDBG_MODE_FILE );
	_CrtSetReportFile( _CRT_ERROR, _CRTDBG_FILE_STDOUT );
	_CrtSetReportMode( _CRT_ASSERT, _CRTDBG_MODE_FILE );
	_CrtSetReportFile( _CRT_ASSERT, _CRTDBG_FILE_STDOUT );
	*/

	uiErrorMode = SetErrorMode( SEM_FAILCRITICALERRORS |
		SEM_NOGPFAULTERRORBOX |
		SEM_NOOPENFILEERRORBOX |
		SEM_NOALIGNMENTFAULTEXCEPT );

#ifdef _DEBUG
	// Get the current state of the flag
	// and store it in a temporary variable
	int tmpFlag = _CrtSetDbgFlag( _CRTDBG_REPORT_FLAG );
	
	// Turn On (OR) - Keep freed memory blocks in the
	// heaps linked list and mark them as freed
//	tmpFlag |= _CRTDBG_DELAY_FREE_MEM_DF;
	tmpFlag |= _CRTDBG_ALLOC_MEM_DF; // use checking at all
	tmpFlag |= _CRTDBG_LEAK_CHECK_DF; // end of program
//	tmpFlag |= _CRTDBG_CHECK_ALWAYS_DF; // check at every allocation

	// Turn Off (AND) - prevent _CrtCheckMemory from
	// being called at every allocation request
	tmpFlag &= ~_CRTDBG_CHECK_ALWAYS_DF;
	tmpFlag &= ~_CRTDBG_DELAY_FREE_MEM_DF;
	
	// Set the new state for the flag
	_CrtSetDbgFlag( tmpFlag );
#endif //_DEBUG
}

/*! On exit, return the error mode to its original state.
	This just seemed good form.
	*/
ExceptionHandler::~ExceptionHandler()
{
	// Return error mode to previous state.
	SetErrorMode( uiErrorMode );
}