#ifndef _FIXED_FIELD_AVALANCHE_H_
#define _FIXED_FIELD_AVALANCHE_H_

#include "DMFirstHeader.h"
#include <queue>
#include "DiscreteAvalanche.h"
#include "SpinIndex.h"
#include "SpinLattice.h"
#include "Rand.h"
#include "AvalancheAlgorithm.h"

class BadArgument;
class GaussianInitializer;
class FixedFieldAvalancheStart;

//! A standard avalanche algorithm
/** This avalanche algorithm stores the value of the random field internally.
  * It also saves the number of neighbors each spin has.
  * It evaluates avalanches using a spinFlipQueue, one shell at a time.
  *
  */
class FixedFieldAvalanche : public AvalancheAlgorithm
{
	FixedFieldAvalancheStart *starter;
	SpinLattice* lattice;
	GaussianInitializer* initializer;

	double *pdRandomField;
	char *pcNUp;
	int M;
public:
	//! Constructor.
	FixedFieldAvalanche();
	//! Destructor.
	~FixedFieldAvalanche();
	//! Tell this object what SpinLattice it should use.
	void SetLattice( SpinLattice *lattice ) throw (BadArgument);
	//! How do we decide where the next avalanche begins?
	void SetStarter( FixedFieldAvalancheStart *starter );
	//! Tell it how to initialize the random field.
	void SetInitializer( GaussianInitializer *initializer );
	//! This call permits the class to set up cached internal variables.
	void Initialize();
	//! Ask the class whether it is all ready to go.
	void Verify( void ) const;
	//! Ask what lattice this class uses.
	const SpinLattice* GetLattice() const;
	//! Get a const copy of the internal random field.
	const double* GetRandomField() const;
	//! Get a const copy of the up neighbors.
	const char* GetNUp() const;
	//! Return magnetization.
	int GetM() const { return M; }
	//! Finds the start of the next avalanche.
	Status FindNextStart( double* H, SpinIndex *iFirstSpin );
	//! Run through a single avalanche.
	Status Cascade( double H, SpinIndex iFirstSpin );

private:
	void InitializeRandomField();
	void InitializeNUp();
	void AllocateArrays();
	void DeleteArrays();
	void ZeroArrays();
	void SetCachedVariables();

	//! The spinFlipQueue is the main tool of this class but not part of its state.
	std::queue<SpinIndex> spinFlipQueue;

	//! length of the sides of the lattice. Cached from SpinLattice.
	SpinIndex iLength;
	//! Number of spins. Cached from SpinLattice.
	SpinIndex N;
	//! coordination number of lattice. Cached from SpinLattice.
	int Z;
	//! These constants are quick notes about what initializations have been performed.
	enum {
		INITIALIZED = 0x1
	};
	//! Keep track of which initializations have been performed.
	int iVerify;
};


#endif //_FIXED_FIELD_AVALANCHE_H_
