#include "DMFirstHeader.h"
#include <windows.h>
#include <string>
#include <algorithm>
#include <sstream>
#include "GeneralSpinLattice.h"
#include "DMReader.h"
#include "DMWriter.h"
#include "SerialBuilder.h"
#include "DMException.h"
#include "NotEnoughMemory.h"
#include "DMAllocator.h"
#include "DMLastHeader.h"

using std::stringstream;

const std::string sType =
	std::string("GeneralSpinLattice");

/*! This static object permits the DMReader to create an empty GeneralSpinLattice. */
const SerialBuilder<GeneralSpinLattice> *_creator = new
	SerialBuilder<GeneralSpinLattice>(sType);

GeneralSpinLattice::GeneralSpinLattice()
{
	DimensionCnt = 2;
	ZeroArrays();
	AllocateArrays();
	ZeroSpinArray();
}

GeneralSpinLattice::~GeneralSpinLattice()
{
	DeleteArrays();
}

bool GeneralSpinLattice::CheckDimensionArray( const SpinIndex* Dimensions )
{
	int DimIdx;
	for ( DimIdx=0;	DimIdx<this->DimensionCnt; DimIdx++ ) {
		if ( Dimensions[DimIdx] <= 0 ) return false;
	}
	return true;
}

void GeneralSpinLattice::SetDimension( int NumberOfDimensions )
{
	ASSERT( NumberOfDimensions>0 );
	if ( NumberOfDimensions <= 0 ) throw DMException("GeneralSpinLattice",
		"SetDimension","Dimension must greater than zero.");

	this->DimensionCnt = NumberOfDimensions;

	AllocateArrays();
	ZeroSpinArray();
}

void GeneralSpinLattice::ZeroSpinArray( )
{
	int DimIdx;
	for ( DimIdx=0; DimIdx<this->DimensionCnt; DimIdx++ ) {
		this->Dimensions[DimIdx] = 0;
	}
	if ( !_SetLengths( this->Dimensions ) ) throw
		DMException("GeneralSpinLattice","ZeroSpinArray",
		"Couldn't resize the spin array to zero length.");
}


bool GeneralSpinLattice::_SetLengths( const SpinIndex* Dimensions )
{
	// Find total number of spins.
	int DimIdx;
	int NewSpinCnt = 1;
	for ( DimIdx=0; DimIdx<this->DimensionCnt; DimIdx++ ) {
		NewSpinCnt *= Dimensions[DimIdx];
	}

	try {
		this->resize( NewSpinCnt );
	} catch (std::bad_alloc e) {
		// failed to resize;
		if (NewSpinCnt > 0 )
			ZeroSpinArray();
		return false;
	}

	this->SpinCnt = NewSpinCnt;

	for ( DimIdx=0; DimIdx<this->DimensionCnt; DimIdx++ ) {
		this->Dimensions[DimIdx] = Dimensions[DimIdx];
	}

	this->SetStrides();

	return true;
}


void GeneralSpinLattice::SetStrides( )
{
	int DimIdx;
	this->Strides[0] = 1;
	for ( DimIdx=0; DimIdx<this->DimensionCnt; DimIdx++ ) {
		this->Strides[DimIdx+1] =
			this->Dimensions[DimIdx] * this->Strides[DimIdx];
	}
}


bool GeneralSpinLattice::SetLengths( const SpinIndex* Dimensions )
{
	// Check arguments
	bool bPointerInvalid = (0 == Dimensions);
	ASSERT( !bPointerInvalid );
	if ( bPointerInvalid ) throw DMException( "GeneralSpinLattice",
		"SetDimensions", "Passed a null pointer." );

	bool bDimensionArrayMalformed = (!CheckDimensionArray(Dimensions));
	ASSERT( !bDimensionArrayMalformed );
	if (bDimensionArrayMalformed) throw DMException( "GeneralSpinLattice",
		"SetDimensions", "Every dimension must be greater than zero.");

	return _SetLengths( Dimensions );
}


void GeneralSpinLattice::SetAllDown()
{
	std::fill( this->begin(), this->end(), false );
}


/*! Because each spin is bit-shifted into a dword,
	using the std:: algorithm and iterators is 30% faster than
	using a regular for loop.
	*/
SpinIndex GeneralSpinLattice::GetTotalNumberUp() const
{
	SpinIndex total = std::count( this->begin(), this->end(), true );
	
	return total;
}


#define CHAR_BITS 8
//typedef unsigned int DWORD;
const int VBITS = CHAR_BITS* sizeof(DWORD);

void GeneralSpinLattice::Save( DMWriter *writer ) const
{
	static const int nBlocks = 10;
	SpinIndex N = this->size();
	writer->PutInt( DimensionCnt, "dimensionCnt" );
	writer->PutInt( SpinCnt, "spinCnt" );
	std::vector<int> dimShape(1);
	dimShape[0] = DimensionCnt;
	writer->PutIntArray( Dimensions, dimShape, "dimensions" );
	
	int dwLen = (N+VBITS-1)/VBITS;
	int dwBlockLen = dwLen/nBlocks + 1;
	int iBlockLen = dwBlockLen*VBITS;
	int iBlocks = (N+iBlockLen-1)/iBlockLen;
	writer->PutInt( iBlocks, "nBlocks" );

	std::vector<int> shape(1);
	SpinIndex end;
	int i;
	for (i=0; i<iBlocks; i++) {
		end = (i+1)*iBlockLen;
		if (end>N) end = N;
		int dwLen;
		int* data = WriteSlice( i*iBlockLen, end, dwLen );
		stringstream sCharBuffer;
		sCharBuffer << "block" << i;
		shape[0] = dwLen;
		writer->PutIntArray( (const int*) data, shape, sCharBuffer.str() );
		delete[] data;
	}
}


void GeneralSpinLattice::Load( DMReader *reader )
{
	DimensionCnt = reader->GetInt( "dimensionCnt" );
	SpinCnt = reader->GetInt( "spinCnt" );
	std::vector<int> dimShape;
	Dimensions = reader->GetIntArray( dimShape, "dimensions" );

	int N = SpinCnt;
	this->resize( SpinCnt );
	this->SetStrides();

	int iBlocks = reader->GetInt( "nBlocks" );
	std::vector<int> shape(1);
	SpinIndex start, finish;
	start = finish = 0;
	int i;
	for (i=0; i< iBlocks; i++) {
		stringstream sCharBuffer;
		int *data;
		sCharBuffer << "block" << i;
		data = reader->GetIntArray( shape, sCharBuffer.str() );
		ASSERT( shape[0] > 0 );
		finish = start+shape[0]*VBITS;
		if (finish>N) finish = N;
		ReadSlice( start, finish, data );
		start = finish;
		delete[] data;
	}
	
}

/*
	int iLen = (N+VBITS-1)/VBITS;
	int i, j;
	DWORD *data = new DWORD[iLen];
	for (i=0; i<iLen; i++) {
		DWORD it = 0;
		for (j=0; j<VBITS; j++) {
			DWORD tr = ( (*this)[i*VBITS+j] ) ? 1 : 0;
			it <<= 1;
			it |= tr;
		}
		data[i] = it;
	}
	std::vector<int> shape;
	shape.push_back( iLen );
	writer->PutIntArray( (const int*) data, shape, "lattice" );
	delete[] data;
*/
int* GeneralSpinLattice::WriteSlice( SpinIndex start, SpinIndex finish, int& dwLen ) const
{
	
	SpinIndex len = finish-start;
	dwLen = (len+VBITS-1)/VBITS;
	int i,j;
	DWORD *data = new DWORD[dwLen];
	for (i=0; i< dwLen; i++) {
		DWORD it = 0;
		for (j=VBITS-1; j>=0; --j) {
			DWORD tr = ( (*this)[start+i*VBITS+j] ) ? 1 : 0;
			it <<= 1;
			it |= tr;
		}
		data[i] = it;
	}
	return (int*)data;
	
	return 0;
}

void GeneralSpinLattice::ReadSlice( SpinIndex start, SpinIndex finish, int*data )
{
	
	SpinIndex len = finish-start;
	int dwLen = (len+VBITS-1)/VBITS;

	ASSERT( 0 == IsBadReadPtr( (CONST VOID *) data, sizeof(int)*dwLen) );

	DWORD* dwData = (DWORD*) data;
	DWORD tr;
	int i,j;
	for (i=0; i<dwLen; i++) {
		DWORD it = dwData[i];
		for (j=0; j<VBITS; ++j) {
			tr = ( it & 0x1 );
			(*this)[start+i*VBITS+j] = ( tr == 1 );
			it >>= 1;
		}
	}
	
}

/*
	std::vector<int> shape;
	this->resize( iX*iY );
	int *data = (int*) reader->GetIntArray( shape, "lattice" );
	int iLen = shape[0];
	int i, j;
	DWORD tr;
	for (i=0; i<iLen; i++) {
		DWORD it = ((DWORD*)data)[i];
		for (j=0; j<VBITS; j++) {
			tr = ( it & 0x1 );
			(*this)[i*VBITS+VBITS-j-1] = ( tr == 1 );
			it >>= 1;
		}
	}
	delete[] data;

*/

/** The signature of this method is an error.  It should return a reference.
  */
const std::string GeneralSpinLattice::GetType() const
{
	return sType;
}


void GeneralSpinLattice::AllocateArrays( )
{
	DeleteArrays();

	try {
		Dimensions = DMAllocate<int>(DimensionCnt);
		Strides = DMAllocate<SpinIndex>( DimensionCnt+1 );
		_neighbors = DMAllocate<SpinIndex>( GetCoordinationNumber() );

	} catch ( std::bad_alloc ) {
		DeleteArrays();
		throw NotEnoughMemory( "GeneralSpinLattice","AllocateArrays",
			"Allocating three tiny arrays." );
	}
}

void GeneralSpinLattice::DeleteArrays()
{
	if (Dimensions) delete[] Dimensions;
	if (Strides) delete[] Strides;
	if (_neighbors) delete[] _neighbors;

	ZeroArrays();
}

void GeneralSpinLattice::ZeroArrays()
{
	Dimensions = 0;
	Strides = 0;
	_neighbors = 0;
}
