#ifndef _WHOLEGRIDBASE_
#define _WHOLEGRIDBASE_

#include "DigitalMaterial.h"
#include <windows.h>

//! A helper class for classes which display 2D grids.
/*! 
	\author ajd27
	*/
class WholeGridBase
{
public:
	//! Constructor.
	WholeGridBase();
	//! Destructor.
	~WholeGridBase();
	//! Return a palette
	/** This will return whatever palette you set.
	  * If you did not set a palette, it will return a palette
	  * with 216 entries taken from Petzold's sample code.
	  */
	RGBQUAD* GetPalette();
	//! Number of entries in the palette.
	int GetPaletteSize();
	//! Set the colors.
	int SetPalette(RGBQUAD* colors, int iColors);

	//! This function helps fit the display into a rectangle.
	/**
	  * \param cWidth Width of display
	  * \param cHeight Height of display
	  * \param xDim Width of grid
	  * \param yDim Height of grid
	  * \return If the return value is positive n, it means you can 
	  * display every nth point.  If the return value is negative n,
	  * you each point should be n pixels wide.
	  * A return value of 0 indicates an error.
	  */
	int CalculateDecimation(int cWidth, int cHeight, int xDim, int yDim);

	//! Find preferred size smaller than bounds
	/** Returns a rect larger than bounds if bounds cannot fit
	  * whole grid as single pixels.
	  */
	int GetPreferredSize(RECT& bounds, RECT& preferred);

	//! Give this object a pointer to the grid data.
	/** This command does not copy the data.  It just
	  * gives it a pointer so we don't have to pass it later
	  * in the Draw() command.
	  */
	int SetData(int *data, int xDim, int yDim, int iMaxVal);

	//! Draw the grid to a device context.
	/** The point of this routine is to draw to either
	  * an enhanced metafile or to print.
	  * This will fail if you try to use the default palette for
	  * more than 216 colors.
	  * \bug Does not account for printing to plotters.
	  * It currently includes calls to FillRect instead of
	  * Rectangle.
	  */
	int Draw(HDC hDC, RECT& bounds);
	//! Create a packed Device-independent bitmap.
	HGLOBAL CreatePackedDIB(RECT& bounds);
	//! Tell you how many bits are needed to represent a number.
	static int NumToBits(int n);
private:
	void PixelFill(HDC hDC, RECT& rect);
	void RectFill(HDC hDC, RECT& bounds, int iDecimate);
	void BrushRectFill(HDC hDC, RECT& bounds, int iDecimate,
								  HBRUSH *brushes);
	void PlainRectFill(HDC hDC, RECT& bounds, int iDecimate);
	HGLOBAL CreateExactSizeDIB(RECT& bounds, int iDecimate);
	void DrawToVoid(RGBQUAD* buf, int iDecimate, int pitch);
	//! Make an array of brushes which correspond to the palette.
	/** If this succeeds, you need to call ::DeleteObject() on
	  * every last brush.  You may need to call GetPaletteSize()
	  * to find out how many there are.
	  * \return Pointer to array of HBRUSH or 0 on failure.
	  */
	HBRUSH* BrushesFromPalette();
	RGBQUAD* colors; // palette
	int iColors; // size of palette
	int *data; // pointer to the data (not a copy)
	int xDim, yDim; // dimensions of data array int *data.
	int iMaxVal; // max possible value of array element.
};


#endif // _WHOLEGRIDBASE_
