#ifndef _WHOLEGRIDDISPLAY_
#define _WHOLEGRIDDISPLAY_

#include "DigitalMaterial.h"
#include <windows.h>

class WholeGridSource;

//! Displays a 2D grid of whole numbers.
/**
 * This is meant to display things like the Ising Model.  It is a purely abstract
 * base class.
 *
 * \author Drew Dolgert
 * \date 6 April 2000
 * \bug The calling requirements are actually less stringent than shown here.
 *		If you try to show it before the data thread has set the data, you just
 *		get either nothing or a blank screen.
 */
class WholeGridDisplay
{
public:
	//! Possible return values
	enum {
			SUCCESS,
			FAILURE, //! General failure you can't do anything about.
			NOMEM,	//! Not enough memory.
			BADPARAM, //! Method argument is incorrect.
			NOTINITIALIZED, //! You should call Initialize() first.
			DIMNOTSET	//! You must first call SetGridDimensions().
	};
	//! Initialize the display by associating a window.
	/*! This is a main initialization for this object.
		Be sure to test whether it succeeded.
		This must be called before showing the data.
		\return BADPARAM if hwnd = 0.
	*/
	virtual int Initialize(HWND hwnd) = 0;
	//! Set the size of the 2D data grid of whole numbers.
	/*! This must be called before calling SetData().
		\param iX Width of grid.
		\param iY Height of grid.
		\param iMaxVal Max numintegers in grid (2 for Ising).
		\return NOMEM
		\return BADPARAM if (iX <=0 || iY <= 0)
		\return FAIL - GDI Won't return this.
	*/
	virtual int SetGridDimensions(int iX, int iY, int iMaxVal) = 0;
	//! Tell the object what its display rectangle is in client screen coords.
	/*! If the grid gets the whole client rectangle, this should
		be the result of GetClientRect(&rect). 
		This must be called before calling Show().
		Any failure to return here happens because we could
		not allocate the bitmap.
		\return BADPARAM if rectangle isn't oriented right-side up.
		\return NOMEM if doesn't have enough display memory.
		\return NOTINITIALIZED if you didn't first call Initialize()
	*/
	virtual int SetDisplayRect(RECT &dispRect ) = 0;

	//! Find largest rect smaller than bounds preferred by this grid
	/** Will return grid larger than bounds if bounds is smaller than
	  * a single-pixel grid.
	  * \return DIMNOTSET if you haven't called SetGridDimension yet.
	  * \return SUCCESS if all is well.
	  */
	virtual int GetPreferredSize(RECT& bounds, RECT& preferred) = 0;

	//! Tell the grid to make a bitmap from the data.
	/*! This does not show the data.  It must be called before
		you call Show().  The implementation promises to convert
		the data to a bitmap during this call.  There are two
		effects of this:  First, memory usage will be only be the
		bitmap.  Second, if you resize the screen, this will have
		only the old size bitmap to show until the next update.
		\return BADPARAM if the data is NULL
		\return DIMNOTSET if you didn't first call SetGridDimensions()
	*/
	virtual int SetData(WholeGridSource *data) = 0;

	//! Draw the grid in the given rectangle.
	/** This routine is for drawing to Metafiles or printers.
	  * \bug It cannot draw to a plotter because it uses FillRect.
	  * It should check for a plotter and switch to Rectangle.
	  * \bug Rectangle uses a slightly different definition
	  * under Win98 from WinNT when you make a metafile.  This
	  * doesn't check for Win98.  Actually, I think the NT one
	  * won't work correctly.
	  */
	virtual int DrawToDC(HDC hDC, RECT& rect) = 0;

	//! Create a Packed Device-independent bitmap.
	/** This bitmap is 32 bpp in depth.
	  * \return 0 on failure.
	  */
	virtual HGLOBAL CreatePackedDIB(RECT& bounds) = 0;

	//! Display the data in the given rect in the given window.
	/*! If you call Show() while another thread calls SetData(),
		there could be a problem.  Just define a mutex across them
		if this could happen. */
	virtual void Show() = 0;
};

#endif // _WHOLEGRIDDISPLAY_