#include "DigitalMaterial.h"
#include "WholeGridGDI.h"
#include "WholeGridSource.h"
#include "WholeGridBase.h"
#include "DMLastHeader.h"

WholeGridGDI::WholeGridGDI()
{
	// create a device context compatible with the display
	// This can return NULL if the display device isn't a raster device.
	// Maybe a plotter or plasma display isn't raster.
	// I'd be delighted to test it on a plasma display.
	gridhelp = new WholeGridBase();
	if (gridhelp==0) return;
	hBitmap = 0;
	hwnd = 0;
	iMaxVal = 0;
	xDim = yDim = 0;
	storage = 0;
	bBitmapCurrent = true;
}

WholeGridGDI::~WholeGridGDI()
{
	if (hBitmap) {
		::DeleteObject(hBitmap);
	}
	if (storage) delete [] storage;
	delete gridhelp;
}

int WholeGridGDI::SetGridDimensions(int iX, int iY, int iMaxVal)
{
	if (iX<0 || iY<0 || iMaxVal<0) return BADPARAM;
	if (this->xDim==iX && this->yDim==iY) return SUCCESS;
	this->xDim = iX;
	this->yDim = iY;
	this->iMaxVal = iMaxVal;
	this->storage = new int[iX*iY];
	if (storage == 0) return NOMEM;
	bBitmapCurrent = false;
	iDecimate = gridhelp->CalculateDecimation(cWidth,cHeight,xDim,yDim);
	return SUCCESS;
}


int WholeGridGDI::Initialize(HWND hwnd)
{
	if (hwnd == 0) return BADPARAM;
	if (gridhelp==0) return FAILURE;
	this->hwnd = hwnd;
	return SUCCESS;
}

int WholeGridGDI::GetPreferredSize(RECT& bounds, RECT& preferred)
{
	if (xDim==0 || yDim==0) return DIMNOTSET;
	if (gridhelp->GetPreferredSize(bounds, preferred) == 0)
			return FAILURE;
	return SUCCESS;
}


int WholeGridGDI::SetDisplayRect(RECT &dispRect)
{
	if (hwnd==0) return NOTINITIALIZED;
	if (dispRect.top-dispRect.bottom>0) return BADPARAM;
	if (dispRect.right-dispRect.left<0) return BADPARAM;
	::CopyRect(&displayRect,&dispRect);
	int width = displayRect.right-displayRect.left;
	int height = displayRect.bottom-displayRect.top;
	// Only re-initialize if we need to.
	if (this->cWidth != width || this->cHeight != height) {
		this->cWidth = width;
		this->cHeight = height;
	} else {
		return SUCCESS;
	}
	bBitmapCurrent = false;
	if (hBitmap) ::DeleteObject(hBitmap);

	hBitmap = this->CreateMemoryBitmap(hwnd, cWidth, cHeight);
	if (hBitmap == NULL) return NOMEM;

	// Now we can draw.
	// Turn it all white to start.
	HDC bufferDC = this->MemoryDCWithBitmap(hBitmap);
	::PatBlt( bufferDC, 0, 0, cWidth, cHeight, BLACKNESS);
	::DeleteDC(bufferDC);
	iDecimate = gridhelp->CalculateDecimation(cWidth,cHeight,xDim,yDim);
	return SUCCESS;
}

HBITMAP WholeGridGDI::CreateMemoryBitmap(HWND hwnd, int cWidth, int cHeight)
{
	HDC hDC = GetDC(hwnd);
	hBitmap = ::CreateCompatibleBitmap(hDC, cWidth, cHeight);
	if (hBitmap == NULL) return NULL;
	::ReleaseDC(hwnd,hDC);
	return hBitmap;
}

HDC WholeGridGDI::MemoryDCWithBitmap(HBITMAP hBitmap)
{
	HDC bufferDC = ::CreateCompatibleDC(NULL);
	// Select the bitmap into the buffer.
	hOldBitmap = ::SelectObject(bufferDC,hBitmap);
	return bufferDC;
}

// Maybe a GDIFlush could fix this?
int WholeGridGDI::SetData(WholeGridSource *data)
{
	// We should decimate with an average color of some sort.
	// For now, lose extra pixels.
	// The plus one is for data ranges from -1 to 254.

	// We can't copy the pixels until someone set our display rect
	// and we allocate a bitmap to draw into.
	if (data == 0) return BADPARAM;
	if (storage == 0) return DIMNOTSET;
	int i,j;
	for (j=0; j<yDim; j++) {
		for (i=0; i<xDim; i++) {
			storage[i+j*xDim] = data->GetGridPoint(i,j);
		}
	}
	// Invalidate our part of the rectangle and 
	// ask that the background not be cleared for us.
//	if (::InvalidateRect(hwnd, &displayRect, false)==0) {
//		i=3; // say what?
//	}
	bBitmapCurrent = false;
	return SUCCESS;
}


void WholeGridGDI::DrawToBitmap()
{
	if (storage==0) return;
	HDC bufferDC = this->MemoryDCWithBitmap(hBitmap);

	this->DrawToDC(bufferDC, displayRect);

	::DeleteDC(bufferDC);
	bBitmapCurrent = true;
}

int WholeGridGDI::DrawToDC(HDC hDC, RECT& rect)
{
	gridhelp->SetData(storage, xDim, yDim, iMaxVal);
	gridhelp->Draw(hDC, displayRect);
	return SUCCESS;
}

HGLOBAL WholeGridGDI::CreatePackedDIB(RECT& bounds)
{
	gridhelp->SetData(storage, xDim, yDim, iMaxVal);
	return gridhelp->CreatePackedDIB(bounds);
}


void WholeGridGDI::Show()
{
	// Can't display if I don't have a bitmap.
	if (hBitmap==NULL) return;
	// I have nothing against showing a white screen if no one has yet
	// set the data.
	if (!bBitmapCurrent) this->DrawToBitmap();
	HDC bufferDC = this->MemoryDCWithBitmap(hBitmap);

	HDC hDC = GetDC(hwnd);
	int bRes =
		::BitBlt( hDC, 0, 0, cWidth, cHeight, bufferDC, 0, 0, SRCCOPY);
	if (bRes ==0) {
		bRes = 1; // huh?
	}
	::DeleteDC(bufferDC);
	::ReleaseDC(hwnd,hDC);
	// We should really only validate our rect.
	// This is an XXX.
//	::ValidateRect(hwnd, NULL);
}
