#pragma warning(disable:4786) // turns off spurious Windows STL warnings
#include "DMFirstHeader.h"
#include <map>
#include <fstream>
#include <string>
#include <sstream>
#include <iterator>

#include "HysteresisSimulation.h"
#include "FixedFieldAvalanche.h"
#include "SortedListAvalancheStart.h"
#include "ProbabilisticAvalanche.h"
#include "ProbabilisticStart.h"
#include "GaussianInitializer.h"
#include "Pow2SpinLattice.h"
#include "SweepLogic.h"
#include "Rand.h"
#include "CheckSpanning.h"
#include "OnSpanning.h"
#include "SizeDistribution.h"
#include "CorrelationFunctionObserver.h"
#include "MagnetizationObserver.h"
#include "ColoringObserver.h"
#include "TimeSeries.h"
#include "DMWriter.h"
#include "DMReader.h"
#include "DMException.h"
#include "NotEnoughMemory.h"
#include "NetCDFWriter.h"
#include "SerialBuilder.h"
#include "BatchTimer.h"
#include "DMAllocator.h"
#include "SprintString.h"
#include "SimpleTracer.h"
#include "MagnetizationTracer.h"
#include "NetCDFCheckpointer.h"
#include "DMLastHeader.h"

using std::stringstream;
using std::string;

// ******************** Serialization object ***********

const std::string HysteresisSimulation::sType = 
	std::string("HysteresisSimulation");
const SerialBuilder<HysteresisSimulation> *_creator = new
SerialBuilder<HysteresisSimulation>( HysteresisSimulation::sType );


// ******************* Construction / Destruction ******

HysteresisSimulation::HysteresisSimulation()
{
	lNumberOfMinutes = 0;
	this->ZeroSubjectPointers();
	this->ZeroObserverPointers();

	dimCnt = 2;
	iLength = 1024;
	iSeed = 1;
	dDisorder = 0.95;

	bAverageSizeDistributionOn	  = false;
	bAverageCorrelationFunctionOn = false;

	bSerialized = false;

	bSubjectInitialized = false;
	bObserverInitialized = false;

	bOwnBatchTimer = false;
	bFirstAvalanche = true;
}

HysteresisSimulation::~HysteresisSimulation()
{
	DeleteSubjectComponents();
	DeleteObserverComponents();
}


// ******************* Get / Set ******************

void HysteresisSimulation::SetDimension( int dimCnt )
{
	this->dimCnt = dimCnt;

	this->bSubjectInitialized = false;
}

void HysteresisSimulation::SetSideLength( SpinCoord length )
{
	this->iLength = length;

	this->bSubjectInitialized = false;
}

void HysteresisSimulation::SetEachSideLength( int dimIdx, SpinCoord length )
{
	if ( dimIdx > sideLength.size() ) {
		sideLength.resize( dimIdx, 0 );
	}
	sideLength[dimIdx] = length;

	this->bSubjectInitialized = false;
}

void HysteresisSimulation::SetSeed( int iSeed )
{
	this->iSeed = iSeed;
}

void HysteresisSimulation::IncrementSeed( )
{
	this->iSeed++;
}

void HysteresisSimulation::SetDisorder( double dDisorder )
{
	this->dDisorder = dDisorder;
}


void HysteresisSimulation::EnableSpanning()
{
	if ( 0 == spanning ) {
		spanning = DMAllocate<CheckSpanning>();
	}

	this->bObserverInitialized = false;
}


void HysteresisSimulation::DisableSpanning()
{
	if ( 0 == correlationFunction &&
		 0 == sizeDistribution &&
		 0 == onSpanning ) {

		if ( 0 != spanning ) {
			RemoveFromFlipObservers( spanning );
			RemoveFromAvalancheObservers( spanning );
			delete spanning;
		}
		spanning = 0;

		this->bObserverInitialized = false;

	}
}


void HysteresisSimulation::EnableSpanningWriter()
{
	EnableSpanning();

	if ( 0 == onSpanning ) {
		onSpanning = DMAllocate<OnSpanning>();
	}

	this->bObserverInitialized = true;
}


void HysteresisSimulation::DisableSpanningWriter()
{
	if ( 0 != onSpanning ) {
		RemoveFromAvalancheObservers( onSpanning );
		delete onSpanning;
	}
	onSpanning = 0;
	
	this->bObserverInitialized = false;

	DisableSpanning();
}


void HysteresisSimulation::EnableColoringObserver( )
{
	if ( 0 == coloringObserver ) {
		coloringObserver = DMAllocate<ColoringObserver>();
	}

	this->bObserverInitialized = false;
}


void HysteresisSimulation::DisableColoringObserver( )
{
	if ( 0 != coloringObserver ) {
		RemoveFromFlipObservers( coloringObserver );
		RemoveFromAvalancheObservers( coloringObserver );
		delete coloringObserver;
	}
	coloringObserver = 0;

	this->bObserverInitialized = false;
}


void HysteresisSimulation::EnableCorrelationFunction( bool bCumulative )
{
	EnableSpanning();

	if ( 0 == correlationFunction ) {
		correlationFunction = DMAllocate<CorrelationFunctionObserver>();
	}

	this->bAverageCorrelationFunctionOn = bCumulative;
	this->bObserverInitialized = false;
}


void HysteresisSimulation::DisableCorrelationFunction( )
{
	if ( 0 != correlationFunction ) {
		RemoveFromFlipObservers( correlationFunction );
		RemoveFromAvalancheObservers( coloringObserver );
		delete correlationFunction;
	}
	correlationFunction = 0;

	this->bObserverInitialized = false;

	DisableSpanning();
}


void HysteresisSimulation::EnableSizeDistribution( bool bCumulative )
{
	EnableSpanning();

	if ( 0 == sizeDistribution ) {
		sizeDistribution = DMAllocate<SizeDistribution>();
	}

	this->bAverageSizeDistributionOn = bCumulative;
	this->bObserverInitialized = false;
}


void HysteresisSimulation::DisableSizeDistribution( )
{
	if ( 0 != sizeDistribution ) {
		RemoveFromFlipObservers( sizeDistribution );
		RemoveFromAvalancheObservers( sizeDistribution );
		delete sizeDistribution;
	}
	sizeDistribution = 0;

	this->bObserverInitialized = false;

	DisableSpanning();
}


void HysteresisSimulation::EnableMagnetizationObserver( )
{
	if ( 0 == magnetization ) {
		magnetization = DMAllocate<MagnetizationObserver>();
	}

	this->bObserverInitialized = false;
}


void HysteresisSimulation::DisableMagnetizationObserver( )
{
	if ( 0 != magnetization ) {
		RemoveFromAvalancheObservers( magnetization );
		delete magnetization;
	}
	magnetization = 0;

	this->bObserverInitialized = false;
}


void HysteresisSimulation::EnableTimeSeries( )
{
	if ( 0 == timeSeries ) {
		timeSeries = DMAllocate<TimeSeries>();
	}

	this->bObserverInitialized = false;
}


void HysteresisSimulation::DisableTimeSeries( )
{
	if ( 0 != timeSeries ) {
		RemoveFromAvalancheObservers( timeSeries );
		delete timeSeries;
	}
	timeSeries = 0;

	this->bObserverInitialized = false;
}


void HysteresisSimulation::EnableBatchTimer()
{
	// Create a new timer if we don't own an internal one.
	if ( 0 == batchTimer ) {
		try {
			batchTimer = DMAllocate<BatchTimer>();
		} catch (std::bad_alloc) {
			throw DMException("HysteresisSimulation","EnableBatchTimer",
				"Could not allocate a new BatchTimer.");
		}
		this->bSubjectInitialized = false;
	}
}


void HysteresisSimulation::DisableBatchTimer()
{
	if ( 0 != batchTimer ) {
		delete batchTimer;
		this->bSubjectInitialized = false;
	}
	batchTimer = 0;
}

void HysteresisSimulation::EnableCheckpointer()
{
	// Create a new timer if we don't own an internal one.
	if ( 0 == checkpointer ) {
		try {
			checkpointer = DMAllocate<NetCDFCheckpointer>();
		} catch (std::bad_alloc) {
			throw DMException("HysteresisSimulation","EnableCheckpointer",
				"Could not allocate a new NetCDFCheckpointer.");
		}
		this->bSubjectInitialized = false;
	}

	this->checkpointer->SetSerializable( this );
}


void HysteresisSimulation::DisableCheckpointer()
{
	if ( 0 != checkpointer ) {
		delete checkpointer;
		this->bSubjectInitialized = false;
	}
	checkpointer = 0;
}


void HysteresisSimulation::EnableSimpleTracer()
{
	DisableTracer();

	// Create a new timer if we don't own an internal one.
	if ( 0 == tracer ) {
		try {
			tracer = DMAllocate<SimpleTracer>();
		} catch (std::bad_alloc) {
			throw DMException("HysteresisSimulation","EnableSimpleTracer",
				"Could not allocate a new SimpleTracer.");
		}
		this->bSubjectInitialized = false;
	}
}


void HysteresisSimulation::EnableMagnetizationTracer()
{
	MagnetizationTracer* magTracer;

	DisableTracer();
	// Create a new timer if we don't own an internal one.
	try {
		magTracer = DMAllocate<MagnetizationTracer>();
	} catch (std::bad_alloc) {
		throw DMException("HysteresisSimulation",
			"EnableMagnetiztionTracer",
			"Could not allocate a new SimpleTracer.");
	}

	EnableMagnetizationObserver();
	magTracer->SetMagnetizationObserver( magnetization );
	this->tracer = magTracer;

	this->bSubjectInitialized = false;
}


void HysteresisSimulation::DisableTracer()
{
	// Don't disable if the tracer or checkpointer need us.
	if ( (0 != tracer) || (0 != checkpointer) ) return;

	if ( 0 != tracer ) {
		delete tracer;
		this->bSubjectInitialized = false;
	}
	tracer = 0;
}



/*!	You don't have to call this if there is no ending time.
	It creates a timer to make the class serialize.

	You normally want to call this before you call Initialize.
	*/
void HysteresisSimulation::SetEndTime( long lNumberOfMinutes )
{
	ASSERT( lNumberOfMinutes > 0 );
	if (lNumberOfMinutes <= 0) {
		throw DMException( "HysteresisSimulation","SetEndTime",
			"The end time should be a positive number of minutes.");
	}

	this->lNumberOfMinutes = lNumberOfMinutes;

	this->EnableBatchTimer();

	batchTimer->SetEndTimeMinutes( lNumberOfMinutes );
}

void HysteresisSimulation::SetTraceFrequency( long traceSecondCnt )
{
	ASSERT( traceSecondCnt >= 0 );
	if ( traceSecondCnt < 0 ) {
		throw DMException( "HysteresisSimulation","SetTraceFrequency",
			"The trace time should be a positive number of seconds.");
	}

	this->traceSecondCnt = traceSecondCnt;

	if ( traceSecondCnt > 0 ) {
		EnableBatchTimer();
		EnableMagnetizationTracer();

		this->tracer->SetTraceFrequency( this->traceSecondCnt );
	} else {
		DisableTracer();
		DisableBatchTimer();
	}
}


void HysteresisSimulation::SetCheckpointFrequency( long checkpointMinuteCnt )
{
	ASSERT( checkpointMinuteCnt >= 0 );
	if ( checkpointMinuteCnt < 0 ) {
		throw DMException( "HysteresisSimulation", "SetCheckpointFrequency",
			SprintString("Checkpoint frequency of %i is not positive.",
			checkpointMinuteCnt ) );
	}

	this->checkpointMinuteCnt = checkpointMinuteCnt;

	if ( checkpointMinuteCnt > 0 ) {
		EnableBatchTimer();
		EnableCheckpointer();
		
		this->checkpointer->SetFrequencyMinutes( checkpointMinuteCnt );
	} else {
		DisableCheckpointer();
		DisableBatchTimer();
	}
}


// ********************* Initialization ****************

void HysteresisSimulation::Initialize()
{
	if ( !bSerialized ) {
		CreateSubjectComponents();

		SetSubjectResetParameters();
		SetSubjectNoResetParameters();
	}
	MakeSubjectConnections();
	bSubjectInitialized = true;

	InitializeObservers();

	bSerialized = false;
}


void HysteresisSimulation::InitializeObservers()
{
	if ( !bSerialized ) {
		SetObserverResetParameters();
		SetObserverNoResetParameters();
	}
	MakeObserverConnections();
	AttachObservers();

	bObserverInitialized = true;
}


void HysteresisSimulation::CreateSubjectComponents()
{
	this->DeleteSubjectComponents();

	try {
		//	avalancheAlgorithm = new FixedFieldAvalanche();
		//	ASSERT( 0 != avalancheAlgorithm );
		//if (0 == avalancheAlgorithm ) throw std::bad_alloc();
		//	avalancheStarter = new SortedListAvalancheStart();
		//	ASSERT( 0 != avalancheStarter );
		//if (0 == avalancheAlgorithm ) throw std::bad_alloc();
		pAvalanche = DMAllocate<ProbabilisticAvalanche>();
		initializer = DMAllocate<GaussianInitializer>();
		spinLattice = DMAllocate<SpinLattice>();
		sweepLogic = DMAllocate<SweepLogic>();
		rand = DMAllocate<Rand>();

	} catch (std::bad_alloc e) {
		DeleteSubjectComponents();
		throw NotEnoughMemory("HysteresisSimulation",
			"CreateComponents","Making all components.");
	}
}

void HysteresisSimulation::SetSubjectResetParameters()
{
	if (spinLattice) {
		spinLattice->SetDimension( this->dimCnt );
		SpinCoord* dims;
		try {
			dims = DMAllocate<SpinCoord>(dimCnt);
		} catch (std::bad_alloc) {
			throw NotEnoughMemory("HysteresisSimulation",
				"SetParameterValues","Allocating DimCnt coordinates.");
		}

		if ( sideLength.size() > 0 ) {
			int dimIdx;
			for (dimIdx=0; dimIdx<dimCnt; dimIdx++) {
				dims[dimIdx] = sideLength[dimIdx];
			}
		} else if (0 != iLength ) {

			int dimIdx;
			for (dimIdx=0; dimIdx<dimCnt; dimIdx++) {
				dims[dimIdx] = this->iLength;
			}
		} else {
				throw DMException("HysteresisSimulation","SetValues",
					"No dimension is defined.");
		}

		spinLattice->SetLengths( dims );
		delete[] dims;
	}
}


void HysteresisSimulation::SetSubjectNoResetParameters()
{
	if (rand)
		rand->SetSeed( iSeed );
	if (initializer)
		initializer->SetDisorder( dDisorder );

	this->bFirstAvalanche = true;
}


void HysteresisSimulation::SetObserverResetParameters()
{
	if (sizeDistribution) {
		; // nothing to do.
	}
	if (correlationFunction) {
		correlationFunction->SetMaxCorrelationLength();
	}
	if (magnetization) {
		magnetization->SetNumMeasurements();
	}

	if (coloringObserver) {
		coloringObserver->SetDecimation( 1 );
	}
	if (checkpointer) checkpointer->SetFrequency( 60*this->checkpointMinuteCnt );
}


void HysteresisSimulation::SetObserverNoResetParameters()
{
}


void HysteresisSimulation::MakeSubjectConnections()
{
	if (batchTimer) {
		if (tracer) {
			batchTimer->SetTracer( tracer );
		}
		if (checkpointer) {
			batchTimer->SetCheckpointer( checkpointer );
		}
	}

	if (initializer) {
		initializer->SetRand( rand );
	}
	if (avalancheAlgorithm != 0) {
		avalanche = avalancheAlgorithm;
	} else if (pAvalanche != 0) {
		pAvalanche->SetRand( rand );
		if (batchTimer) {
			pAvalanche->SetBatchTimer( batchTimer );
		}
		avalanche = pAvalanche;
	} else {
		throw DMException("HysteresisSimulation","MakeConnections",
			"You need to make an AvalancheAlgorithm.");
	}

	avalanche->SetLattice( spinLattice );
	avalanche->SetInitializer( initializer );
	avalanche->Initialize();
	avalanche->Verify();

	sweepLogic->SetAvalanche( avalanche );
}


void HysteresisSimulation::MakeObserverConnections()
{
	if (spanning) {
		spanning->SetLattice( spinLattice );
		spanning->Initialize();
	}

	if (onSpanning) {
		onSpanning->SetSpanning( spanning );
	}

	if (sizeDistribution) {
		sizeDistribution->SetSpanning( spanning );
		sizeDistribution->Initialize();
	}
	if (correlationFunction) {
		correlationFunction->SetSpanning( spanning );
		correlationFunction->SetLattice( spinLattice );
		correlationFunction->Initialize();
		correlationFunction->Verify();
	}
	if (magnetization) {
		magnetization->SetLattice( spinLattice );
		magnetization->SetAlgorithm( avalanche );
		magnetization->Initialize();
	}
	if (coloringObserver) {
		coloringObserver->SetLattice( spinLattice );
		coloringObserver->Initialize();
	}
	if (timeSeries) {
		timeSeries->SetAlgorithm( avalanche );
		timeSeries->Initialize();
	}

}

void HysteresisSimulation::AttachObservers()
{
	bool bSuccess;
	if (spanning) {
		bSuccess = sweepLogic->AddListener( spanning );
		bSuccess = avalanche->AddObserver( spanning );
	}
	if ( onSpanning ) {
		bSuccess = sweepLogic->AddListener( onSpanning );
	}
	if ( sizeDistribution ) {
		bSuccess = sweepLogic->AddListener( sizeDistribution );
		bSuccess = avalanche->AddObserver( sizeDistribution );
	}
	if ( correlationFunction ) {
		bSuccess = sweepLogic->AddListener( correlationFunction );
		bSuccess = avalanche->AddObserver( correlationFunction );
	}
	if ( magnetization ) {
		bSuccess = sweepLogic->AddListener( magnetization );
	}
	if ( coloringObserver ) {
		bSuccess = sweepLogic->AddListener( coloringObserver );
		bSuccess = avalanche->AddObserver( coloringObserver );
	}
	if ( timeSeries ) {
		bSuccess = sweepLogic->AddListener( timeSeries );
	}
}


bool HysteresisSimulation::Verify() //throw (MissingSimulationElement)
{
	if ( pAvalanche ) pAvalanche->Verify();
	// SweepLogic has no verify.
	// SpinLattice has no verify.
	// GaussianInitializer has no verify.
	if ( batchTimer ) batchTimer->Verify();

	// CheckSpanning has no verify.
	if ( coloringObserver )		coloringObserver->Verify();
	if ( correlationFunction )	correlationFunction->Verify();
	// MagnetizationObserver has no verify.
	if ( sizeDistribution )		sizeDistribution->Verify();
	// TimeSeries has no verify.


	return true;
}

/*!
	\param Number of avalanches to do.  A RunCnt of 0 means we
	should run until the end of the avalanche.
	\return true if avalanches are done, false if they are not.
	*/
AvalancheAlgorithm::Status
HysteresisSimulation::Run( int RunCnt )
{
	ASSERT( RunCnt >= 0 );
	if ( RunCnt < 0 ) throw DMException("HysteresisSimulation",
		"Run","You should specify a positive number of runs to do.");

	AvalancheAlgorithm::Status AvalancheStatus;

	if ( bFirstAvalanche ) {
		if ( !bSubjectInitialized ) {
			Initialize();
		} else if ( !bObserverInitialized ) {
			InitializeObservers();
		} else {
			this->SetSubjectNoResetParameters();
			this->SetObserverNoResetParameters();
		}
		this->BeginSweep();
		this->bFirstAvalanche = false;
	}

	if ( RunCnt > 0 ) {
		AvalancheStatus = sweepLogic->Run( RunCnt );
		if (AvalancheStatus != AvalancheAlgorithm::SUCCESS) {
			this->bFirstAvalanche = true;
		}

	} else {

		RunCnt = 10000;
		do {
			AvalancheStatus = sweepLogic->Run( RunCnt );
		} while (AvalancheStatus == AvalancheAlgorithm::SUCCESS);

		this->bFirstAvalanche = true;
	}

	return AvalancheStatus;
}

std::ostream& HysteresisSimulation::WriteState( std::ostream& s )
{
	s << "HysteresisSimulation" << std::endl;
	s << "\tdim\t" << this->dimCnt << std::endl;
	s << "\tlength\t";
	if ( sideLength.size() > 0 ) {
		int dimIdx;
		for ( dimIdx = 0; dimIdx < sideLength.size(); dimIdx++ ) {
			s << " " << sideLength[ dimIdx ];
		}
		s << std::endl;
	} else if ( iLength > 0 ) {
		s << iLength << std::endl;
	}
	s << "\tseed\t" << iSeed << std::endl;
	s << "\tdisorder\t" << dDisorder << std::endl;
	
	s << "\tObservers:";
	if (spanning) s << " spanning";
	if (onSpanning) s << " spanWriter";
	if (sizeDistribution) s << " size";
	if (bAverageSizeDistributionOn) s << " avgSize";
	if (correlationFunction) s << " correl";
	if (bAverageCorrelationFunctionOn) s << " avgCorrel";
	if (magnetization) s << " magnetization";
	if (coloringObserver) s << " coloring";
	if (timeSeries) s << " timeSeries";
	s << std::endl;

	if (batchTimer) {
		if (this->bOwnBatchTimer) {
			std::cout << "Using own batch timer." << std::endl;
		} else {
			std::cout << "Using external batch timer." << std::endl;
		}
	}

	return s;
}


/*! I tried to work out a scheme where you can save and 
	load any combination of components.  As it is done
	here, you save all the possible component names
	and whether or not they are included.  This method 
	works fine for now, but it would be hard to tell if
	there were some component you had neglected to retrieve.

	I think a better method might be to save a single
	list of all the sub-components.  You can read that list
	and search for ones you know how to load.  That way,
	you would know if there are some you don't know how to load.

	*/
void HysteresisSimulation::Save( DMWriter* writer ) const
{
	int iFirstAvalanche = (bFirstAvalanche) ? 1 : 0;
	writer->PutInt( iFirstAvalanche, "isFirstAvalanche" );

	int iAvgSizeDistribution = (bAverageSizeDistributionOn) ? 1 : 0;
	writer->PutInt( iAvgSizeDistribution, "isAverageSizeDistribution" );

	int iAvgCorrelDist = (bAverageCorrelationFunctionOn) ? 1 : 0;
	writer->PutInt( iAvgCorrelDist, "isAverageCorrelationFunction" );

	writer->PutInt( iSeed, "randSeed" );
	writer->PutDouble( dDisorder, "disorder" );
	writer->PutInt( dimCnt, "dimCnt" );
	writer->PutInt( iLength, "length" );
	writer->PutInt( sideLength.size(), "sideLengthCnt" );
	// Only store sidelengths if they were set.
	if ( sideLength.size() > 0 ) {
		std::vector<int> sideShape(1);
		sideShape[0] = sideLength.size();
		writer->PutIntArray( &sideLength[0], sideShape, "sideLength" );
	}


	int iBool;

	iBool = (avalancheAlgorithm) ? 1 : 0;
	writer->PutInt( iBool, "isAvalancheAlgorithm");
	iBool = (avalancheStarter) ? 1 : 0;
	writer->PutInt( iBool, "isAvalancheStarter");
	iBool = (pAvalanche) ? 1 : 0;
	writer->PutInt( iBool, "isPAvalanche");
	iBool = (pStart) ? 1 : 0;
	writer->PutInt( iBool, "isPStart");
	iBool = (initializer) ? 1 : 0;
	writer->PutInt( iBool, "isInitializer");
	iBool = (spinLattice) ? 1 : 0;
	writer->PutInt( iBool, "isSpinLattice");
	iBool = (sweepLogic) ? 1 : 0;
	writer->PutInt( iBool, "isSweepLogic");
	iBool = (rand) ? 1 : 0;
	writer->PutInt( iBool, "isRand");
	iBool = (spanning) ? 1 : 0;
	writer->PutInt( iBool, "isSpanning");
	iBool = (onSpanning) ? 1 : 0;
	writer->PutInt( iBool, "isOnSpanning");
	iBool = (sizeDistribution) ? 1 : 0;
	writer->PutInt( iBool, "isSizeDistribution");
	iBool = (correlationFunction) ? 1 : 0;
	writer->PutInt( iBool, "isCorrelationFunction");
	iBool = (magnetization) ? 1 : 0;
	writer->PutInt( iBool, "isMagnetization");

	iBool = (coloringObserver) ? 1 : 0;
	writer->PutInt( iBool, "isColoringObserver" );

	iBool = (timeSeries) ? 1 : 0;
	writer->PutInt( iBool, "isTimeSeries");

// These aren't implemented.
//	if (avalancheAlgorithm)
//		writer->Put( avalancheAlgorithm, "avalancheAlgorithm" );
//	if (avalancheStart)
//		writer->Put( avalancheStart, "avalancheStart" );
	if (pAvalanche)
		writer->Put( pAvalanche, "probAvalanche" );
	if (pStart)
		writer->Put( pStart, "probStart" );
	if (initializer)
		writer->Put( initializer, "gaussianInitializer" );
	if (spinLattice)
		writer->Put( spinLattice, "spinLattice" );
	if (sweepLogic)
		writer->Put( sweepLogic, "sweepLogic" );
	if (rand)
		writer->Put( rand, "rand" );
	if (spanning)
		writer->Put( spanning, "checkSpanning" );
	if (onSpanning)
		writer->Put( onSpanning, "onSpanning" );
	if (sizeDistribution)
		writer->Put( sizeDistribution, "sizeDistribution" );
	if (correlationFunction)
		writer->Put( correlationFunction, "correlationFunction" );
	if (magnetization)
		writer->Put( magnetization, "magnetization" );
//	if (coloringObserver)
//		writer->Put( coloringObserver, "coloringObserver" );
	if (timeSeries)
		writer->Put( timeSeries, "timeSeries" );
}

void HysteresisSimulation::Load( DMReader *reader )
{
	DeleteSubjectComponents();
	DeleteObserverComponents();
	bSerialized = true;
	bSubjectInitialized = false;
	bObserverInitialized = false;

	int iFirstAvalanche = reader->GetInt( "isFirstAvalanche" );
	bFirstAvalanche = ( 1 == iFirstAvalanche );

	int iAvgSizeDist = reader->GetInt( "isAverageSizeDistribution" );
	bAverageSizeDistributionOn = ( 1 == iAvgSizeDist );

	int iAvgCorrelDist = reader->GetInt( "isAverageCorrelationFunction" );
	this->bAverageCorrelationFunctionOn = ( 1 == iAvgCorrelDist );

	this->iSeed = reader->GetInt( "randSeed" );
	this->dDisorder = reader->GetDouble( "disorder" );
	this->dimCnt = reader->GetInt( "dimCnt" );
	this->iLength = reader->GetInt( "length" );

	this->sideLength.clear();
	int sideCnt = reader->GetInt( "sideLengthCnt" );
	if ( sideCnt > 0 ) {
		std::vector<int> sideShape;
		int* piSide = reader->GetIntArray( sideShape, "sideLength" );
		ASSERT( sideShape.size() > 0 );
		ASSERT( sideShape[0] == sideCnt );
		if ( sideCnt != sideShape[0] ) {
			throw DMException( "HysteresisSimulation","Load",
			SprintString("Expected side length array of size %i, got size %i.",
			sideCnt, sideShape[0] ) );
		}
		std::copy( piSide, piSide+sideCnt, std::back_inserter(sideLength) );
		delete[] piSide;
	}

	std::map<const std::string,bool> uses;
	uses["avalancheAlgorithm"] =
		(reader->GetInt( "isAvalancheAlgorithm" ) == 1);
	uses["avalancheStart"] =
		(reader->GetInt( "isAvalancheStarter" ) == 1);
	uses["probAvalanche"] =
		(reader->GetInt( "isPAvalanche" ) == 1);
	uses["probStart"] =
		(reader->GetInt( "isPStart" ) == 1);
	uses["gaussianInitializer"] =
		(reader->GetInt( "isInitializer" ) == 1);
	uses["spinLattice"] =
		(reader->GetInt( "isSpinLattice" ) == 1);
	uses["sweepLogic"] =
		(reader->GetInt( "isSweepLogic" ) == 1);
	uses["rand"] =
		(reader->GetInt( "isRand" ) == 1);
	uses["checkSpanning"] =
		(reader->GetInt( "isSpanning" ) == 1);
	uses["onSpanning"] =
		(reader->GetInt( "isOnSpanning" ) == 1);
	uses["sizeDistribution"] =
		(reader->GetInt( "isSizeDistribution" ) == 1);
	uses["correlationFunction"] =
		(reader->GetInt( "isCorrelationFunction" ) == 1);
	uses["magnetization"] =
		(reader->GetInt( "isMagnetization" ) == 1);
	uses["coloringObserver"] = 
		(reader->GetInt( "isColoringObserver" ) == 1);
	uses["timeSeries"] =
		(reader->GetInt( "isTimeSeries" ) == 1);

	if (uses["avalancheAlgorithm"])
		avalancheAlgorithm = dynamic_cast<FixedFieldAvalanche*>(
		reader->Get( "avalancheAlgorithm" ) );
	if (uses["avalancheStart"])
		avalancheStarter = dynamic_cast<FixedFieldAvalancheStart*>(
		reader->Get( "avalancheStart" ) );
	if (uses["probAvalanche"])
		pAvalanche = dynamic_cast<ProbabilisticAvalanche*>(
		reader->Get( "probAvalanche" ) );
	if (uses["probStart"])
		pStart = dynamic_cast<ProbabilisticStart*>(
		reader->Get( "probStart" ) );
	if (uses["gaussianInitializer"])
		initializer = dynamic_cast<GaussianInitializer*>(
		reader->Get( "gaussianInitializer" ) );
	if (uses["spinLattice"])
		spinLattice = dynamic_cast<SpinLattice*>(
		reader->Get( "spinLattice" ) );
	if (uses["sweepLogic"])
		sweepLogic = dynamic_cast<SweepLogic*>(
		reader->Get( "sweepLogic" ) );
	if (uses["rand"])
		rand = dynamic_cast<Rand*>(
		reader->Get( "rand" ) );
	if (uses["checkSpanning"])
		spanning = dynamic_cast<CheckSpanning*>(
		reader->Get( "checkSpanning" ) );
	if (uses["onSpanning"])
		onSpanning = dynamic_cast<OnSpanning*>(
		reader->Get( "onSpanning" ) );
	if (uses["sizeDistribution"])
		sizeDistribution = dynamic_cast<SizeDistribution*>(
		reader->Get( "sizeDistribution" ) );
	if (uses["correlationFunction"])
		correlationFunction = dynamic_cast<CorrelationFunctionObserver*>(
		reader->Get( "correlationFunction" ) );
	if (uses["magnetization"])
		magnetization = dynamic_cast<MagnetizationObserver*>(
		reader->Get( "magnetization" ) );
//	if (uses["coloringObserver"])
//		coloringObserver = dynamic_cast<ColoringObserver*>(
//		reader->Get( "coloringObserver" ) );
	if (uses["timeSeries"])
		timeSeries = dynamic_cast<TimeSeries*>(
		reader->Get( "timeSeries" ) );
}

const std::string HysteresisSimulation::GetType() const
{
	return sType;
}

void HysteresisSimulation::SaveResults() const
{
	stringstream sFilename;
	sFilename << "L" << iLength << "_R" << dDisorder << ".S" << iSeed;
	if (sizeDistribution) {
		string sTemp = "SizeDistribution";
		sTemp.append( sFilename.str() );
		std::ofstream sdf( sTemp.c_str() );
		sdf << *sizeDistribution;
	}
	if (correlationFunction) {
		string sTemp = "CorrelationFunction";
		sTemp.append( sFilename.str() );
		std::ofstream cff(sTemp.c_str());
		cff << *correlationFunction;
	}
	if (magnetization) {
		string sTemp = "Magnetization";
		sTemp.append( sFilename.str() );
		std::ofstream mf(sTemp.c_str());
		mf << *magnetization;
	}
	if (timeSeries) {
		string sTemp = "TimeSeries";
		sTemp.append( sFilename.str() );
		std::ofstream tsf( sTemp.c_str() );
		tsf << *timeSeries;
	}
	if (coloringObserver) {
		string sTemp = "Coloring";
		sTemp.append( sFilename.str() );
		std::ofstream tsf( sTemp.c_str() );
		tsf << *coloringObserver;
	}
}


void HysteresisSimulation::SaveAverageResults() const
{
	stringstream sFilename;
	sFilename << "L" << iLength << "_R" << dDisorder << ".S" << iSeed;

	if ( sizeDistribution && bAverageSizeDistributionOn ) {
		string sTemp = "AverageSizeDistribution";
		sTemp.append( sFilename.str() );
		std::ofstream sdf( sTemp.c_str() );
		sizeDistribution->SaveAverageData( sdf );
	}

	if ( correlationFunction && bAverageCorrelationFunctionOn ) {
		string sTemp = "AverageCorrelationFunction";
		sTemp.append( sFilename.str() );
		std::ofstream cff(sTemp.c_str());
		correlationFunction->SaveAverageData( cff );
	}
}


void HysteresisSimulation::DeleteSubjectComponents()
{
	if (avalancheAlgorithm) delete avalancheAlgorithm;
	if (avalancheStarter) delete avalancheStarter;
	if (pAvalanche) delete pAvalanche;
	if (pStart) delete pStart;
	if (initializer) delete initializer;
	if (spinLattice) delete spinLattice;
	if (sweepLogic) delete sweepLogic;
	if (rand) delete rand;

	this->ZeroSubjectPointers();
}


void HysteresisSimulation::DeleteObserverComponents()
{
	if (spanning) delete spanning;
	if (onSpanning) delete onSpanning;
	if (sizeDistribution) delete sizeDistribution;
	if (correlationFunction) delete correlationFunction;
	if (magnetization) delete magnetization;
	if (coloringObserver) delete coloringObserver;
	if (timeSeries) delete timeSeries;


	if (bOwnBatchTimer) {
		if (batchTimer) delete batchTimer;
	}
	if (tracer) delete tracer;
	if (checkpointer) delete checkpointer;

	this->ZeroObserverPointers();
}

void HysteresisSimulation::ZeroSubjectPointers()
{
	avalancheAlgorithm = 0;
	avalanche = 0;
	start = 0;
	initializer = 0;
	avalancheStarter = 0;
	spinLattice = 0;
	sweepLogic = 0;
	pAvalanche = 0;
	pStart = 0;
	initializer = 0;
	spinLattice = 0;
	sweepLogic = 0;
	rand = 0;
}


void HysteresisSimulation::ZeroObserverPointers()
{
	spanning = 0;
	onSpanning = 0;
	sizeDistribution = 0;
	correlationFunction = 0;
	magnetization = 0;
	coloringObserver = 0;
	timeSeries = 0;

	// These are zeroed with observers at beginning
	// and end of HysteresisSimulation's life.
	batchTimer = 0;
	tracer = 0;
	checkpointer = 0;
}

void HysteresisSimulation::RemoveFromAvalancheObservers
	(
	AvalancheListener* observer
	)
{
	if ( 0 != sweepLogic) {
		sweepLogic->RemoveListener( observer );
	}
}

void HysteresisSimulation::RemoveFromFlipObservers
	(
	FlipListener* observer
	)
{
	if ( 0 != avalanche ) {
		avalanche->RemoveObserver( observer );
	}
}


void HysteresisSimulation::BeginSweep( bool bKeepData )
{
	if (spinLattice)
		spinLattice->SetAllDown();

	if (pAvalanche)
		pAvalanche->BeginSweep( bKeepData );

	this->sweepLogic->BeginSweep( bKeepData );

	if ( spanning )				spanning->BeginSweep( bKeepData );

	if ( onSpanning ) onSpanning->BeginSweep( bKeepData );

	if ( sizeDistribution )
		sizeDistribution->BeginSweep( bKeepData | bAverageSizeDistributionOn );
	if ( correlationFunction )
		correlationFunction->BeginSweep( bKeepData | bAverageCorrelationFunctionOn );

	if ( magnetization )		magnetization->BeginSweep( bKeepData );
	if ( coloringObserver )		coloringObserver->BeginSweep( bKeepData );
	if ( timeSeries )			timeSeries->BeginSweep( bKeepData );
}
