#include "DMFirstHeader.h"
#include "MagnetizationObserver.h"
#include "DMException.h"
#include "NotEnoughMemory.h"
#include "DMWriter.h"
#include "DMReader.h"
#include "SpinLattice.h"
#include "SerialBuilder.h"
#include "DMAllocator.h"
#include "DMLastHeader.h"

const std::string sType = std::string("MagnetizationObserver");
const SerialBuilder<MagnetizationObserver> *_creator = new
	SerialBuilder<MagnetizationObserver>(sType);

MagnetizationObserver::MagnetizationObserver()
{
	numPoints = 0;
	lattice = 0;
	ZeroArrays();
	bSerialized = false;
	iVerify = 0;
}

MagnetizationObserver::~MagnetizationObserver()
{
	DeleteArrays();
}

void MagnetizationObserver::SetNumMeasurements( int numPoints )
{
	ASSERT( numPoints > 0 );
	if (numPoints <= 0) throw
		DMException("MagnetizationObserver","SetNumMeasurements",
		"The number of measurements should be greater than zero.");

	this->numPoints = numPoints;

	iVerify |= VERIFY_NUMPOINTS;
}

void MagnetizationObserver::SetLattice( SpinLattice* lattice )
{
	ASSERT( 0 != lattice );
	if (lattice == 0) throw
		DMException("MagnetizationObserver","SetLattice",
		"Argument pointer should be nonzero.");
	
	this->lattice = lattice;

	iVerify |= VERIFY_LATTICE;
}

void MagnetizationObserver::SetAlgorithm( AvalancheAlgorithm* avalanche )
{
	ASSERT( 0 != avalanche );
	if (avalanche == 0) throw
		DMException("MagnetizationObserver","SetAlgorithm",
		"Argument pointer should be nonzero.");

	this->avalanche = avalanche;

	iVerify |= VERIFY_AVALANCHE;
}

void MagnetizationObserver::Initialize()
{
	ASSERT( numPoints > 0 );
	if (numPoints <= 0) throw
		DMException("MagnetizationObserver","Initialize",
		"Call SetNumMeasurements to set the number of points.");

	ASSERT( 0 != lattice );
	if (lattice == 0) throw
		DMException("MagnetizationObserver","Initialize",
		"Call SetLattice to give us a SpinLattice.");

	ASSERT( 0 != avalanche );
	if (avalanche== 0) throw
		DMException("MagnetizationObserver","Initialize",
		"Call SetAvalanche to give us an AvalancheAlgorithm.");

	N = lattice->GetLength();
	
	if (!bSerialized) {
		DeleteArrays();
		AllocateArrays();
	}

	bSerialized = false;
	iVerify = 0;
}

void MagnetizationObserver::BeginSweep( bool bKeep )
{
	ASSERT( 0 == iVerify );
	if ( 0 != iVerify ) {
		throw DMException( "MagnetizationObserver","BeginSweep",
			"You have to call Initialize again if you set parameters." );
	}

	curPoint = 0;
	MagStepSize = 2.0/numPoints;
	LastMagnetization = -1.0;
	OutputMagStep = -1.0;
	Hint = 0.0;
}

void MagnetizationObserver::Save( DMWriter *writer ) const
{
	writer->PutInt( numPoints, "numPoints" );
	writer->PutInt( curPoint, "curPoint" );
	writer->PutDouble( MagStepSize, "MagStepSize" );
	writer->PutDouble( Hint, "Hint" );
	writer->PutDouble( LastMagnetization, "LastMagnetization" );
	writer->PutDouble( OutputMagStep, "OutputMagStep" );

	std::vector<int> shape;
	shape.push_back(numPoints+1);
	writer->PutDoubleArray( H, shape, "H" );
	writer->PutDoubleArray( M, shape, "M" );
}

void MagnetizationObserver::Load( DMReader *reader )
{
	bSerialized = true;
	std::vector<int> shape;

	numPoints = reader->GetInt( "numPoints" );
	curPoint = reader->GetInt( "curPoint" );
	MagStepSize = reader->GetDouble( "MagStepSize" );
	Hint = reader->GetDouble( "Hint" );
	LastMagnetization = reader->GetDouble( "LastMagnetization" );
	OutputMagStep = reader->GetDouble( "OutputMagStep" );

	DeleteArrays();
	H = reader->GetDoubleArray( shape, "H" );
	M = reader->GetDoubleArray( shape, "M" );
}

const std::string MagnetizationObserver::GetType() const
{
	return sType;
}

void MagnetizationObserver::DeleteArrays( void )
{
	if (H) delete[] H;
	if (M) delete[] M;

	ZeroArrays();
}

void MagnetizationObserver::AllocateArrays( void )
{
	try {
		H = DMAllocate<double>(numPoints+1);
		M = DMAllocate<double>(numPoints+1);
	} catch (std::bad_alloc) {
		DeleteArrays();
		throw NotEnoughMemory("MagnetizationObserver",
			"AllocateArrays","H and M");
	}
}

void MagnetizationObserver::ZeroArrays( void )
{
	H = 0;
	M = 0;
}

std::ostream& operator<<( std::ostream& s, const MagnetizationObserver& r )
{
	s << "Magnetization" << std::endl;
	s << "NumberOfEntries:\t" << r.numPoints << std::endl;
	s << "Field : Magnetization " << std::endl;
	int i;
	for (i=0; i< r.numPoints; i++)
		s << r.H[i] << "\t" << r.M[i] << std::endl;
	return s;
}