#include "DMFirstHeader.h"
#include <windows.h>
#include <string>
#include <algorithm>
#include <sstream>
#include "Pow2SpinLattice.h"
#include "DMReader.h"
#include "DMWriter.h"
#include "SerialBuilder.h"
#include "DMException.h"
#include "NotEnoughMemory.h"
#include "DMAllocator.h"
#include "DMLastHeader.h"

using std::stringstream;

const std::string sType =
	std::string("Pow2SpinLattice");

/*! This static object permits the DMReader to create an empty Pow2SpinLattice. */
const SerialBuilder<Pow2SpinLattice> *_creator = new
	SerialBuilder<Pow2SpinLattice>(sType);

Pow2SpinLattice::Pow2SpinLattice()
{
	DimensionCnt = 2;
	ZeroArrays();
	AllocateArrays();
	ZeroSpinArray();
}

Pow2SpinLattice::~Pow2SpinLattice()
{
	DeleteArrays();
}

bool Pow2SpinLattice::CheckDimensionArray( const SpinIndex* Dimensions )
{
	int DimIdx;
	for ( DimIdx=0;	DimIdx<this->DimensionCnt; DimIdx++ ) {
		if ( Dimensions[DimIdx] <= 0 ) return false;
	}
	return true;
}

void Pow2SpinLattice::SetDimension( int NumberOfDimensions )
{
	ASSERT( NumberOfDimensions>0 );
	if ( NumberOfDimensions <= 0 ) throw DMException("Pow2SpinLattice",
		"SetDimension","Dimension must greater than zero.");

	this->DimensionCnt = NumberOfDimensions;

	AllocateArrays();
	ZeroSpinArray();
}

void Pow2SpinLattice::ZeroSpinArray( )
{
	int DimIdx;
	for ( DimIdx=0; DimIdx<this->DimensionCnt; DimIdx++ ) {
		this->Dimensions[DimIdx] = 0;
	}
	if ( !_SetLengths( this->Dimensions ) ) throw
		DMException("Pow2SpinLattice","ZeroSpinArray",
		"Couldn't resize the spin array to zero length.");
}

bool Pow2SpinLattice::_SetLengths( const SpinIndex* Dimensions )
{
	// Find total number of spins.
	int DimIdx;
	int NewSpinCnt = 1;
	for ( DimIdx=0; DimIdx<this->DimensionCnt; DimIdx++ ) {
		NewSpinCnt *= Dimensions[DimIdx];
	}

	try {
		this->resize( NewSpinCnt );
	} catch (std::bad_alloc e) {
		// failed to resize;
		if (NewSpinCnt > 0 )
			ZeroSpinArray();
		return false;
	}

	this->SpinCnt = NewSpinCnt;

	for ( DimIdx=0; DimIdx<this->DimensionCnt; DimIdx++ ) {
		this->Dimensions[DimIdx] = Dimensions[DimIdx];
	}

	if ( 0 != this->SpinCnt ) SetStrides();

	
	return true;
}


void Pow2SpinLattice::SetStrides()
{
	// Find out how many bits wide each dimension is.
	int* BitWidth;
	try {
		BitWidth = DMAllocate<int>(this->DimensionCnt);
	} catch ( std::bad_alloc ) {
		throw NotEnoughMemory("Pow2SpinLattice","_SetLengths",
			"Allocating dimension integers." );
	}
	int DimIdx;
	for ( DimIdx = 0; DimIdx<this->DimensionCnt; DimIdx++ ) {

		int width = 0;
		SpinCoord sideLength = this->Dimensions[DimIdx];
		while ( 0 == (sideLength & 0x1) ) {
			sideLength >>= 1;
			width++;
		}

		BitWidth[DimIdx] = width;
	}

	// Find the number of bits to shift each dimension.
	int CumulativeShift = 0;
	for ( DimIdx=0; DimIdx < this->DimensionCnt; DimIdx++ ) {
		this->BitShift[DimIdx] = CumulativeShift;
		CumulativeShift += BitWidth[DimIdx];
	}

	// Use the number of bits to shift to find the lowest bit.
	for ( DimIdx=0; DimIdx < this->DimensionCnt; DimIdx++ ) {
		this->LowestBit[DimIdx] = ( 0x1 << BitShift[DimIdx] );
	}
	this->LowestBit[this->DimensionCnt] = ( 0x1 << CumulativeShift );

	// Now make the mask.
	for ( DimIdx=0; DimIdx < this->DimensionCnt; DimIdx++ ) {
		SpinCoord AllOnes = this->Dimensions[DimIdx]-1;
		this->BitMask[DimIdx] = (AllOnes << BitShift[DimIdx]);
	}

	delete[] BitWidth;
}


bool Pow2SpinLattice::SetLengths( const SpinCoord* Dimensions )
{
	// Check arguments
	bool bPointerInvalid = (0 == Dimensions);
	ASSERT( !bPointerInvalid );
	if ( bPointerInvalid ) throw DMException( "Pow2SpinLattice",
		"SetDimensions", "Passed a null pointer." );

	bool bDimensionArrayMalformed = (!CheckDimensionArray(Dimensions));
	ASSERT( !bDimensionArrayMalformed );
	if (bDimensionArrayMalformed) throw DMException( "Pow2SpinLattice",
		"SetDimensions", "Every dimension must be greater than zero.");

	// Check that the dimensions are all powers of two.
	int DimIdx;
	for ( DimIdx=0; DimIdx < this->DimensionCnt; DimIdx++ ) {

		SpinCoord DimCoord = Dimensions[DimIdx];
		int BitIdx;
		int BitTotal = 0;
		for ( BitIdx=0; BitIdx < 8*sizeof(SpinCoord); BitIdx++ ) {
			BitTotal += ( DimCoord & 0x1 );
			DimCoord >>= 1;
		}

		bool bDimensionsArePowersOfTwo = ( 1 == BitTotal );
		ASSERT( bDimensionsArePowersOfTwo );
		if ( !bDimensionsArePowersOfTwo ) {
			throw DMException("Pow2SpinLattice","SetLengths",
				"Each dimension must be a power of 2 for this SpinLattice.");
		}
	}

	return _SetLengths( Dimensions );
}


void Pow2SpinLattice::SetAllDown()
{
	std::fill( this->begin(), this->end(), false );
}


/*! Because each spin is bit-shifted into a dword,
	using the std:: algorithm and iterators is 30% faster than
	using a regular for loop.
	*/
SpinIndex Pow2SpinLattice::GetTotalNumberUp() const
{
	SpinIndex total = std::count( this->begin(), this->end(), true );
	
	return total;
}


#define CHAR_BITS 8
//typedef unsigned int DWORD;
const int VBITS = CHAR_BITS* sizeof(DWORD);

void Pow2SpinLattice::Save( DMWriter *writer ) const
{
	static const int nBlocks = 10;
	SpinIndex N = this->size();
	writer->PutInt( DimensionCnt, "dimensionCnt" );
	writer->PutInt( SpinCnt, "spinCnt" );
	std::vector<int> dimShape(1);
	dimShape[0] = DimensionCnt;
	writer->PutIntArray( Dimensions, dimShape, "dimensions" );
	
	
	int dwLen = (N+VBITS-1)/VBITS;
	int dwBlockLen = dwLen/nBlocks + 1;
	int iBlockLen = dwBlockLen*VBITS;
	int iBlocks = (N+iBlockLen-1)/iBlockLen;
	writer->PutInt( iBlocks, "nBlocks" );

	std::vector<int> shape(1);
	SpinIndex end;
	int i;
	for (i=0; i<iBlocks; i++) {
		end = (i+1)*iBlockLen;
		if (end>N) end = N;
		int dwLen;
		int* data = WriteSlice( i*iBlockLen, end, dwLen );
		stringstream sCharBuffer;
		sCharBuffer << "block" << i;
		shape[0] = dwLen;
		writer->PutIntArray( (const int*) data, shape, sCharBuffer.str() );
		delete[] data;
	}
}


void Pow2SpinLattice::Load( DMReader *reader )
{
	DimensionCnt = reader->GetInt( "dimensionCnt" );
	SpinCnt = reader->GetInt( "spinCnt" );
	std::vector<int> dimShape;
	Dimensions = reader->GetIntArray( dimShape, "dimensions" );

	int N = SpinCnt;
	this->resize( SpinCnt );
	this->SetStrides();

	int iBlocks = reader->GetInt( "nBlocks" );
	std::vector<int> shape(1);
	SpinIndex start, finish;
	start = finish = 0;
	int i;
	for (i=0; i< iBlocks; i++) {
		stringstream sCharBuffer;
		int *data;
		sCharBuffer << "block" << i;
		data = reader->GetIntArray( shape, sCharBuffer.str() );
		ASSERT( shape[0] > 0 );
		finish = start+shape[0]*VBITS;
		if (finish>N) finish = N;
		ReadSlice( start, finish, data );
		start = finish;
		delete[] data;
	}
}

/*
	int iLen = (N+VBITS-1)/VBITS;
	int i, j;
	DWORD *data = new DWORD[iLen];
	for (i=0; i<iLen; i++) {
		DWORD it = 0;
		for (j=0; j<VBITS; j++) {
			DWORD tr = ( (*this)[i*VBITS+j] ) ? 1 : 0;
			it <<= 1;
			it |= tr;
		}
		data[i] = it;
	}
	std::vector<int> shape;
	shape.push_back( iLen );
	writer->PutIntArray( (const int*) data, shape, "lattice" );
	delete[] data;
*/
int* Pow2SpinLattice::WriteSlice( SpinIndex start, SpinIndex finish, int& dwLen ) const
{
	SpinIndex len = finish-start;
	dwLen = (len+VBITS-1)/VBITS;
	int i,j;
	DWORD *data = new DWORD[dwLen];
	for (i=0; i< dwLen; i++) {
		DWORD it = 0;
		for (j=VBITS-1; j>=0; --j) {
			DWORD tr = ( (*this)[start+i*VBITS+j] ) ? 1 : 0;
			it <<= 1;
			it |= tr;
		}
		data[i] = it;
	}
	return (int*)data;
	return 0;
}

void Pow2SpinLattice::ReadSlice( SpinIndex start, SpinIndex finish, int*data )
{
	SpinIndex len = finish-start;
	int dwLen = (len+VBITS-1)/VBITS;

	ASSERT( 0 == IsBadReadPtr( (CONST VOID *) data, sizeof(int)*dwLen) );

	DWORD* dwData = (DWORD*) data;
	DWORD tr;
	int i,j;
	for (i=0; i<dwLen; i++) {
		DWORD it = dwData[i];
		for (j=0; j<VBITS; ++j) {
			tr = ( it & 0x1 );
			(*this)[start+i*VBITS+j] = ( tr == 1 );
			it >>= 1;
		}
	}
}

/*
	std::vector<int> shape;
	this->resize( iX*iY );
	int *data = (int*) reader->GetIntArray( shape, "lattice" );
	int iLen = shape[0];
	int i, j;
	DWORD tr;
	for (i=0; i<iLen; i++) {
		DWORD it = ((DWORD*)data)[i];
		for (j=0; j<VBITS; j++) {
			tr = ( it & 0x1 );
			(*this)[i*VBITS+VBITS-j-1] = ( tr == 1 );
			it >>= 1;
		}
	}
	delete[] data;

*/

/** The signature of this method is an error.  It should return a reference.
  */
const std::string Pow2SpinLattice::GetType() const
{
	return sType;
}


void Pow2SpinLattice::AllocateArrays( )
{
	DeleteArrays();

	try {
		Dimensions = DMAllocate<SpinCoord>( DimensionCnt );
		_neighbors = DMAllocate<SpinIndex>( GetCoordinationNumber() );
		LowestBit = DMAllocate<SpinIndex>( DimensionCnt + 1 );
		BitMask   = DMAllocate<SpinIndex>( DimensionCnt );
		BitShift  = DMAllocate<int>( DimensionCnt );
	} catch ( std::bad_alloc ) {
		DeleteArrays();
		throw NotEnoughMemory("Pow2SpinLattice","AllocateArrays",
			"Allocating 5*dimension SpinIndices.");
	}
}

void Pow2SpinLattice::DeleteArrays()
{
	if (Dimensions) delete[] Dimensions;

	if (_neighbors) delete[] _neighbors;

	if (LowestBit) delete[] LowestBit;
	if (BitMask)   delete[] BitMask;
	if (BitShift)  delete[] BitShift;


	ZeroArrays();
}

void Pow2SpinLattice::ZeroArrays()
{
	Dimensions = 0;

	_neighbors = 0;

	LowestBit = 0;
	BitMask   = 0;
	BitShift  = 0;
}
