
#include "DMFirstHeader.h"
#include <stdexcept>
#include <iostream>
#include "ProbabilisticStart.h"
#include "SpinLattice.h"
#include "GaussianInitializer.h"
#include "ProbabilisticAvalanche.h"
#include "FailedVerify.h"
#include "DMException.h"
#include "SerialBuilder.h"
#include "BatchTimer.h"
#include "DMLastHeader.h"

const std::string sType =
	std::string("ProbabilisticStart");

/*! This static object permits the DMReader to create an empty ProbabilisticStart. */
const SerialBuilder<ProbabilisticStart> *_creator = new
	SerialBuilder<ProbabilisticStart>(sType);


ProbabilisticStart::ProbabilisticStart()
{
	lattice = 0;
	initializer = 0;
	rand = 0;
	batchTimer = 0;
	iVerify = 0;
#ifndef NDEBUG
	startIdx = 0;
#endif
}

ProbabilisticStart::~ProbabilisticStart()
{
}

void ProbabilisticStart::SetRand( Rand *rand )
{
	ASSERT( 0 != rand );
	if (rand == 0) throw
		DMException("ProbabilisticStart","SetRand","Passed a null pointer." );

	this->rand = rand;
}

void ProbabilisticStart::SetInitializer( GaussianInitializer *initializer )
{
	ASSERT( 0 != initializer );
	if (initializer == 0) throw
		DMException("ProbabilisticStart","SetInitializer","Passed a null pointer." );

	this->initializer = initializer;
}

void ProbabilisticStart::SetLattice( const SpinLattice *lattice )
{
	ASSERT( 0 != lattice );
	if (0 == lattice) throw DMException("ProbabilisticStart",
		"SetLattice","Passed a null pointer.");

	this->lattice = lattice;
}

void ProbabilisticStart::SetDownSpins( SpinIndex *psNUp )
{
	ASSERT( 0 != psNUp );
	if (0 == psNUp) throw DMException("ProbabilisticStart",
		"SetDownSpins","Passed a null pointer.");

	this->psNUp = psNUp;
}

/*! This method may not be called if we aren't running under batch.
	This routine needs a batch timer because finding the next spin can
	take a very long time.
	*/
void ProbabilisticStart::SetBatchTimer( BatchTimer* batchTimer )
{
	ASSERT( 0 != batchTimer );
	if ( 0 == batchTimer ) throw DMException("ProbabilisticStart",
		"SetBatchTimer","Passed a null pointer.");

	this->batchTimer = batchTimer;
}

/*! The interface for AvalancheStart classes requires you call this before
	asking for the first avalanche location and that you call this after
	calling SetInitializer(), SetAlgorithm() and SetRand().
	*/
void ProbabilisticStart::Initialize()
{
	ASSERT( 0 != rand );
	ASSERT( 0 != initializer );
	ASSERT( 0 != lattice );
	ASSERT( 0 != psNUp );
	if ( 0 == psNUp )
		throw DMException("ProbabilisticStart","Initialize",
		"Can\'t call initialize until you call SetDownSpins.");
	if ( 0 == rand )
		throw DMException("ProbabilisticStart","Initialize",
		"Can\'t call initialize until you call SetRand.");
	if ( 0 == initializer )
		throw DMException("ProbabilisticStart","Initialize",
		"Can\'t call initialize until you call SetInitializer.");
	if (0 == lattice) throw DMException("ProbabilisticStart",
		"Initialize","You need to call SetLattice before Initialize.");
	if (0 == lattice) throw DMException("ProbabilisticStart",
		"Initialize","You need to call SetLattice before Initialize.");

	SetCachedVariables();

	iVerify |= INITIALIZE;
}

/*! This BeginSweep() doesn't have anything to check because there
	is no internal data.
	*/
void ProbabilisticStart::BeginSweep( bool bKeepData )
{
	this->Verify();
#ifdef _DEBUG
	startIdx = 0;
#endif
}

void ProbabilisticStart::Verify() const
{
	ASSERT( (iVerify&INITIALIZE) != 0 );
	if ( (iVerify & INITIALIZE) == 0) throw FailedVerify( "ProbabilisticStart","Need to call Initialize." );
}

/*! There isn't anything to save. */
void ProbabilisticStart::Save( DMWriter *writer ) const
{
}

/*! And there is nothing to load. */
void ProbabilisticStart::Load( DMReader *reader )
{
}

const std::string ProbabilisticStart::GetType() const
{
	return sType;
}

/*! This routine will return false if a) you can't have any more avalanches
	or b) it is time to quit.
  \param H returns the magnetic field where the next avalanche starts.
  \param iFirstSpin returns the spin index where the avalanche starts.
  \return false if you can't have any more avalanches.  true if you can.
  */
AvalancheAlgorithm::Status
ProbabilisticStart::FindNextStart( double *H, SpinIndex *iFirstSpin )
{
	ASSERT( 0 != H );
	if ( 0 == H ) throw DMException("ProbabilisticStart", "FindNextStart",
		"Passed null pointer for magnetic field." );

	ASSERT( 0 != iFirstSpin );
	if ( 0 == iFirstSpin ) throw DMException("ProbabilisticStart", "FindNextStart",
		"Passed null pointer for first spin." );

#ifndef NDEBUG
	startIdx++;
	//TRACE( "ProbabilisticStart::FindNextStart %i Enter\n", startIdx );
#endif

	int i;
	SpinIndex iTotalUp = 0;
	for (i=0; i<=Z; i++) {
		iTotalUp += psNUp[i];
	}

	if (iTotalUp==0) {
		TRACE("ProbabilisticStart::FindNextStart No more up spins.\n");
		return AvalancheAlgorithm::ALLFLIPPED;
	}

	//	TRACE("ProbabilisticStart::FindNextStart Finding field.\n");
	double dField = this->FindNextField( *H );
	//	TRACE("ProbabilisticStart::FindNextStart Finding position.\n");
	SpinIndex loc;
	AvalancheAlgorithm::Status status = this->FindNextPosition( dField, loc );
	if ( AvalancheAlgorithm::SUCCESS != status ) return status;

	*H = dField;
	*iFirstSpin = loc;
	//	TRACE("ProbabilisticStart::FindNextStart Leave\n");
	return AvalancheAlgorithm::SUCCESS;
}

// **************** functions private below this line **************** //

double ProbabilisticStart::FindNextField( double dField )
{
	// This has the important side-effect of setting dInvProbFlipped
	// for pNoneError.
	double rate = this->SpinFlipRate( dField );
	// Set the rate to a reasonable value if zero.
	if ( rate == 0 ) rate = 0.05;

		// Make linear and quadratic guesses of dh which will give
		// pNone(dh)==dPH

	double dh1,dh2,dh1err,dh2err;

	double dPH = rand->uniform();
	dh1 = -log(dPH)/rate;
	if(dh1>20)
	{
		dh1=0;
		dh1err = pNoneError(dField,dh1,dPH);
		dh2=20;
		dh2err = pNoneError(dField,dh2,dPH);
	}
	else
	{
		dh1err = pNoneError(dField,dh1,dPH);
		dh2 = -log(dPH-dh1err)/rate;
		dh2err = pNoneError(dField,dh2,dPH);
	}

		// Use guesses to bracket root
	BracketRoot(dField,dPH,&dh1,&dh1err,&dh2,&dh2err);

		// Find the root
	double dh = FindRoot(dField,dPH,dh1,dh1err,dh2,dh2err);
	return dField + dh;
}

/*!
	\param dField The known field at which the next avalanche will start.
	\return The spin to start the next avalanche or -1 if we should quit.
	*/
AvalancheAlgorithm::Status
ProbabilisticStart::FindNextPosition( double dField, SpinIndex& foundSpin )
{
	double rate = SpinFlipRate( dField );
	int nup = FindNeighborsUp( rate );
	return FindRandomSpin( nup, foundSpin );
}

int ProbabilisticStart::FindNeighborsUp( double dRate )
{
	int nup;
	double total = 0;
	double dProbNUp = rand->uniform()*dRate;

	// The code below has a subtle bug when the dProbNUp is 0, I think.
	// It was crashing the program when it returned nup=-1.
	// It has been replaced with Matt's old code until my brain can handle
	// thinking about this. - ajd27
	//	nup = -1;
	//	while ( (nup<Z) && (total<dProbNUp) )
	//		total += dRates[++nup];
	//	ASSERT( -1 != nup );

	for (nup = 0; nup <= Z; nup++)
	{
		total += dRates[nup];
		if (total>=dProbNUp) break;
	}
	
	return nup;
}

/*!
	This is a painful function because we have to randomly search
	the lattice for a spin, and there isn't enough memory to store where
	previous spins were.

	Just on the stupid tip, we can work out a small optimization.  The
	average number of times we need to loop in order to find the correct
	down spin is (the number of downs with correct neighbors) out of (the 
	total number of spins).  The time per iteration is a
	call to rand plus a call to NeighborsUp.  If we search linearly for
	the spins, then it will always take N times the time for a single
	NeighborsUp call.  Denote the time to call rand t_r, the 
	time to call NeighborsUp t_n, and the number of spins N_down.  Then if
	N_down is less than 1 + t_r/t_n, it is preferable to do a linear search.
	Lots of words.  Simple concept.  Timings show t_r = 2*t_n, so
	we choose the linearThreshhold = 3.

	\param nup The number of up neighbors this spin has.
	\return The spin to start the next avalanche or -1 if we should quit.
	*/
AvalancheAlgorithm::Status ProbabilisticStart::FindRandomSpin
	(
	int nup,
	SpinIndex& foundSpin
	) const
{
	static const int linearThreshhold = 3;
	SpinIndex candidates[linearThreshhold];
	ASSERT( nup >=0 );
	ASSERT( nup <= Z );
	AvalancheAlgorithm::Status status;

	//TRACE("ProbabilisticStart::FindRandomSpin Finding %i up neighbors.\n", nup);
	// Do a completely random search if there are still a lot of up spins.
	if (psNUp[nup] >= linearThreshhold) {
		status = RandomSampleForSpin( nup, lattice, foundSpin );

	} else {
		status = LinearSearchForSpin( nup, candidates,
			linearThreshhold, lattice, foundSpin );
	}

	return status;
}

/*	We pass the pointer to the lattice in so that it will be const in the 
	time-consuming inner loop.  This allows the compiler more optimizations
	and decreases running time by thirty percent.
	*/
AvalancheAlgorithm::Status ProbabilisticStart::RandomSampleForSpin
	(
	int nup,
	const SpinLattice* const spinLattice,
	SpinIndex& foundSpin
	) const
{
	bool bFound = false;
	bool bQuit = false;
	static const int timeToSlice = 1000000;
	int time = 0;
	SpinIndex loc;
	AvalancheAlgorithm::Status status = AvalancheAlgorithm::SUCCESS;
	int latticeDimCnt = spinLattice->GetDimension();
	const SpinCoord const* latticeDim = spinLattice->GetSideLengths();
	SpinCoord* randomLoc = new SpinCoord[latticeDimCnt];
	
	// Do a completely random search if there are still a lot of up spins.
	int totalLoops = 0;
	while (!bFound && !bQuit) {

#ifndef NDEBUG
		if ( totalLoops > 0 ) TRACE( "  Loop %i.", totalLoops );
#endif

		// Rand::discrete find integers from start to finish _inclusive_.
		time = 0;
		while (!bFound && (time<timeToSlice) ) {

			for ( int dimIdx = 0; dimIdx < latticeDimCnt; dimIdx++ )
				randomLoc[ dimIdx ] = rand->discrete( 0, latticeDim[ dimIdx ]-1 );
			loc = spinLattice->GetIndex( randomLoc );

			if ( (-1 == spinLattice->Get(loc)) &&
				(nup == spinLattice->NeighborsUp(loc)) )
				bFound = true;
			++time;

		}

#ifndef NDEBUG
		if ( 0 == totalLoops )
			ASSERT( bFound || CheckSpinNeighborhoodExists( nup ) );
#endif

		// Don't check the batchtimer if you found a spin because it will
		// change the spin found if you return from a serialized state.
		if (!bFound && batchTimer) {
			bQuit = batchTimer->TimeSlice();
		}
		totalLoops++;
	}
	if (bQuit) {
		TRACE("ProbabilisticStart::FindRandomSpin Quitting because batchTimer said to.\n");
		status = AvalancheAlgorithm::TIMEDOUT;
	}
	foundSpin = loc;

	delete[] randomLoc;
	return status;
}


/*	We pass the pointer to the lattice in so that it will be const in the 
	time-consuming inner loop.  This allows the compiler more optimizations
	and decreases running time by thirty percent.
	*/
AvalancheAlgorithm::Status ProbabilisticStart::LinearSearchForSpin
	(
	int nup,
	SpinIndex* candidates,
	int linearThreshhold,
	const SpinLattice* const spinLattice,
	SpinIndex& foundSpin
	) const
{
	bool bFound = false;
	bool bQuit = false;
	static const int timeToSlice = 2000000;
	AvalancheAlgorithm::Status status = AvalancheAlgorithm::SUCCESS;
	int time = 0;
	SpinIndex loc;
	
	TRACE("ProbabilisticStart::FindRandomSpin linear search.\n");
	// Do a linear search if there are only a few up spins.
	int i = 0;
	loc = 0;
	// First find all candidates in the lattice.
	while ( (i<psNUp[nup]) && !bQuit) {
		time = 0;
		while ( (loc < N) && (time<timeToSlice) ) {
			if ( (-1 == spinLattice->Get(loc)) &&
				(nup == spinLattice->NeighborsUp(loc)) ) {
				candidates[i] = loc;
				i++;
			}
			++loc;
			++time;
		}

		if ( batchTimer ) bQuit = batchTimer->TimeSlice();

		// Assert that we aren't looping forever because
		// the psNUp array is wrong.
		ASSERT( (time==timeToSlice) || (i == psNUp[nup]) );
	}
	ASSERT( i == psNUp[nup] );
	
	// Then choose one candidate at random.
	int choice = rand->discrete(0,i-1);
	loc = candidates[choice];
	
	if (bQuit) {
		TRACE("ProbabilisticStart::FindRandomSpin Quitting because batchTimer said to.\n");
		status = AvalancheAlgorithm::TIMEDOUT;
	}

	foundSpin = loc;
	return status;
}


bool ProbabilisticStart::CheckSpinNeighborhoodExists( int nup ) const
{
	bool bFound = false;
	SpinIndex loc = 0;
	
	while (!bFound && loc < N) {
		if ( (-1 == lattice->Get(loc)) &&
			(nup == lattice->NeighborsUp(loc)) )
			bFound = true;
		++loc;
	}
	if (bFound)
		TRACE("ProbabilisticStart::CheckValid Found it at %i.  All is okay.\n",loc-1);
	else {
		TRACE("ProbabilisticStart::CheckValid Couldn't find it.  All is lost.\n");
	}
	return bFound;
}


/** Should we be caching 1/intrhoc(dLocalField) from here?
  * Must first call ProbSpinDown to initialize dInvProbNotFlipped.
  *
  * Sets the variable dInvProbNotFlipped with the inverse
  * of the probability a spin will be down given N up neighbors.
  *
  * Sets dRates, the array of rates for each number of neighbors.
  */
double ProbabilisticStart::SpinFlipRate( double dField )
{
	int nup;
	double dLocalField, rate = 0;

	for ( nup=0; nup <= Z; ++nup ) {
		dLocalField = dField + 2*nup - Z;
		dInvProbNotFlipped[nup] = 1.0/initializer->intrhoc( dLocalField );
		dRates[nup] = psNUp[nup]*initializer->rho(dLocalField)*
			dInvProbNotFlipped[nup];
		rate += dRates[nup];
	}
	return rate;
}

/** In finding the change in H, dh,  needed to start the next
  * avalanche, we try to pick dh so that the probability no spins have flipped
  * in the interval dh is equal to the random number dPH.  This function returns
  * the difference between the actuall probability given dh and the desired
  * probability dPH.
  *
  * Requires dInvProbNotFlipped, psNUp, Z, current field, and initializer.
  */
double ProbabilisticStart::pNoneError(double dField, double dh, double dPH)
{
	double probNoneFlipped = 1.0;
	double hLoc;

	// Calculate the probability that no spins will be flipped in the
	// interval dh
	for(int nup=0;nup<=Z;nup++)
	{
		hLoc = dField+dh+2*nup-Z;
		probNoneFlipped *= pow(initializer->intrhoc(hLoc)*
			dInvProbNotFlipped[nup],psNUp[nup]);
	}
		// return the error
	return probNoneFlipped-dPH;
}

/*! Given two guesses for dh, and a desired probability dPH, make sure
	that the guesses bracket the correct solution.
  */
double ProbabilisticStart::BracketRoot(double dField, double dPH,
									   double* dh1, double* dh1error,
									   double* dh2, double* dh2error)
{
	static const double FACTOR=1.6;
	static const int N_TRY=50;

	// The two guesses must be different
	if (*dh1 == *dh2)
	{
	    if(*dh1error == 0.) return 1;
		else throw std::invalid_argument("BracketRoot: dh1=dh2!");
	}
	
	// Take up to N_TRY steps to bracket the root
	for (int j=1;j<=N_TRY;j++) {
		// If the errors have opposite sign, then the root is bracketed
		if (*dh1error*(*dh2error) < 0.0) return 1;
		
		// Otherwise, move the smaller error away from
		// the larger error by the amount
		// FACTOR * (distance between guesses)
		if (fabs(*dh1error) < fabs(*dh2error))
			*dh1error=pNoneError(dField,*dh1 += FACTOR*(*dh1-*dh2),dPH);

		else if (fabs(*dh1error) > fabs(*dh2error))
			*dh2error=pNoneError(dField,*dh2 += FACTOR*(*dh2-*dh1),dPH);

		// If the errors are equal, then we randomly choose one of
		// our guesses to move
		else
		{
			// if j is even, move the first guess, if j is odd, move the second guess
			if ((j&0x1)==0)
				*dh1error=pNoneError(dField,*dh1 += FACTOR*(*dh1-*dh2),dPH);
			else
			    *dh2error=pNoneError(dField,*dh2 += FACTOR*(*dh2-*dh1),dPH);
		}
	}

	// We failed to bracket the root
	return 0;
}

/** Given two guesses for dh that bracket the desired probability dPH,
  * find the value of dh which gives dPH.  Uses the Brent algorithm.  The brent
  * algorithm keeps track of three guesses: the best guess, the last guess, and
  * an earlier guess which has an error in the opposite direction from the error
  * in the best guess.  Each step, it tries to do a quadratic interpolation,
  * or a linear interpolation if it only has two distinct points, but if this
  * produces a bad guess, it uses bisection instead.
  */
inline double
ProbabilisticStart::FindRoot(double dField, double dPH, double dh1, double dh1error,
										   double dh2, double dh2error)
{
	static const double EPSILON=2.22045e-16;	// The minimum error
	double tol=1e-16;							// The tolerance

	// dh1 is the previous best guess, dh2 is the best guess, and dh3 is an earlier guess
	// such that its error is in the opposite direction from the error of dh2.
	// dh1error, dh2error, and dh3error are the distance from the root for each of these
	// guesses.
	double dh3 = dh1;
	double dh3error = dh1error;

	// Loop until we find the root
	for(;;)
	{
		// How big was the previous step?
		double prevStep = dh2-dh1;

		// Make sure that dh2 has the smallest error
		if( fabs(dh3error) < fabs(dh2error) )
		{
			dh1 = dh2;
			dh2 = dh3;
			dh3 = dh1;

			dh1error = dh2error;
			dh2error = dh3error;
			dh3error = dh1error;
		}

		// Calculate the tolerance in our guess
		double actualTolerance = 2.0*EPSILON*fabs(dh2) + 0.5*tol;

		// This is the step we will take if we use bisection
		double newStep = 0.5*(dh3-dh2);

		// If the bisection step is smaller than the tolerance, then we
		// have found the root.  Return it.
		if( fabs(newStep) <= actualTolerance || dh2error == 0)
			return dh2;

		// If the previous step was big enough, and in the right direction,
		// then try doing an interpolation
		if( fabs(prevStep) >= actualTolerance 
			&& fabs(dh1error) > fabs(dh2error) )
		{
			double p,q;

			double diff = dh3-dh2;

			// If two of our points are equal, then we can only
			// try a linear interpolation
			if( dh1==dh3 )
			{
				double t1 = dh2error / dh1error;
				p = diff*t1;
				q = 1.0 - t1;
			}
			// Otherwise, try a quadratic inverse interpolation
			else
			{
				double t1 = dh2error / dh3error;
				double t2 = dh2error / dh1error;

				q = dh1error / dh3error;
				p = t2 * ( diff*q*(q-t1) - (dh2-dh1)*(t1-1.0) );
				q = (q-1.0) * (t1-1.0) * (t2-1.0);
			}

			// Correct the sign of the new step
			if( p>0 )
				q = -q;
			else
				p = -p;

			// If our step falls inside the old range, and isn't too big,
			// use the interpolation step instead of the bisection step
			if( 2*p < (1.5*diff*q-fabs(actualTolerance*q))
				&& 2*p < fabs(prevStep*q) )
				newStep = p/q;
		}

		// Don't make a step smaller than the tolerance
		if( fabs(newStep) < actualTolerance )
			newStep = newStep > 0 ? actualTolerance : -actualTolerance;

		// Take the step
		dh1 = dh2;
		dh1error = dh2error;
		dh2 += newStep;
		dh2error = pNoneError(dField,dh2,dPH);

		// If the errors of dh2 and dh3 have the same sign, set dh3 to dh1
		// so its error will have opposite sign.
		if( (dh2error > 0 && dh3error > 0) || (dh2error < 0 && dh3error < 0) )
		{
			dh3 = dh1;
			dh3error = dh1error;
		}
	}
}

void ProbabilisticStart::SetCachedVariables()
{
	SUPERASSERT( 0 != lattice );
	if (0 == lattice) throw DMException("ProbabilisticStart",
		"SetCachedVariables","The lattice is a null pointer.");

	this->Z = lattice->GetCoordinationNumber();
	this->N = lattice->GetLength();
}