#ifndef _PROBABILISTIC_START_H_
#define _PROBABILISTIC_START_H_

#include "DMFirstHeader.h"
#include "DiscreteAvalanche.h"
#include "SpinIndex.h"
#include "SpinLattice.h"
#include "AvalancheStart.h"
#include "Serializable.h"

class GaussianInitializer;
class ProbabilisticAvalanche;
class Rand;
class BatchTimer;

//! Finds the next field and starting spin for a probabilistic avalanche
/*! This uses only the information stored in a ProbabilisticAvalanche in order
	to find which spin flips next at which magnetic field.  It does some
	intensive root finding to get the next magnetic field and determine how
	many neighboring up spins the next avalanche starter will have.  Then it
	samples the lattice at random looking for a spin with that number of neighbors.

	\author Matt Kuntz
	\author ajd27 Refactored out of Matt's BitsHysteresisSimulation
*/
class ProbabilisticStart : public AvalancheStart, public Serializable
{
	GaussianInitializer *initializer; //!< supplies probability distribution of field.
	Rand *rand;
	const int *psNUp; //!< Number of down spins with N up neighbors.
	double dRates[10]; //!< Flip rates by number of up neighbors.
	//! 1/probability a spin with n up neighbors is down.
	double dInvProbNotFlipped[10];
	const SpinLattice* lattice; //!< the lattice of spins we work with.
	BatchTimer *batchTimer; //!< Checkpoints and traces runs under batch.
public:
	//! Constructor
	ProbabilisticStart();
	//! Destructor
	~ProbabilisticStart();
	//! Give us an initializer for the probability distribution.
	void SetInitializer( GaussianInitializer* initializer );
	//! Give us a lattice to point to.
	void SetLattice( const SpinLattice *lattice );
	//! Give us a random number generator
	virtual void SetRand( Rand* rand );
	//! Give us a pointer to the number of neighbors up
	virtual void SetDownSpins( SpinIndex *psNUp );
	//! Give us a batch timer.
	virtual void SetBatchTimer( BatchTimer *batchTimer );
	//! You must call Initialize after setting the algorithm.
	virtual void Initialize();
	//! Call before starting a sweep.
	virtual void BeginSweep( bool bKeepData = false );
	//! Ask whether everything is ready to rumble.
	virtual void Verify() const; // throw (FailedVerify);
	//! Finds the start of the next avalanche.
	virtual AvalancheAlgorithm::Status
		FindNextStart( double* H, SpinIndex *iFirstSpin );

	/*! Serializable interface */
	//@{
	virtual void Save( DMWriter *writer ) const;
	virtual void Load( DMReader *reader );
	virtual const std::string GetType() const;
	//@}

private:
	//! Find the field at which the next avalanche begins.
	double FindNextField( double dField );

	//! Find the spin where the next avalanche will start.
	AvalancheAlgorithm::Status FindNextPosition( double dField, SpinIndex& foundSpin );

	//! How many up neighbors will the first spin of the next avalanche have?
	int FindNeighborsUp( double dRate );

	//! Find a random spin with nup up neighbors.
	AvalancheAlgorithm::Status
		FindRandomSpin( int nup, SpinIndex& foundSpin ) const;

	//! Find a random spin by randomly sampling the lattice.
	AvalancheAlgorithm::Status RandomSampleForSpin( int nup, const SpinLattice* const, 
		SpinIndex& foundSpin ) const;

	//! Find a random spin by storing every nup spin in lattice.
	AvalancheAlgorithm::Status LinearSearchForSpin( int nup, SpinIndex* candidates,
		int linearThreshhold, const SpinLattice* const, SpinIndex& foundSpin ) const;

	//! Calculate the rate for spins of various up neighbors to flip.
	double SpinFlipRate( double dField );
	//! Probability none will flip.
	double pNoneError(double dField, double dh, double dPH);
	//! Find bounds on the possible new values of the field.
	double BracketRoot(double dField, double dPH,double* dh1, double* hd1error,
		double* dh2, double* dh2error);
	//! Determine the root of the equation to find the next magnetic field.
	double FindRoot(double dField, double dPH, double dh1, double dh1error,
		double dh2, double dh2error);
	//! Set internally cached variables
	void SetCachedVariables();
	//! Finds whether there is a spin with this # of up neighbors.
	bool CheckSpinNeighborhoodExists( int nup ) const;

	int Z;
	SpinIndex L, N;

	enum {
		INITIALIZE = 0x1,
	};
	int iVerify;
#ifdef _DEBUG
	int startIdx;
#endif
};

#endif //_PROBABILISTIC_START_H_
