#ifndef _SPIN_LATTICE_2D_H_
#define _SPIN_LATTICE_2D_H_

#include "DMFirstHeader.h"
#include <vector>
#include "DiscreteAvalanche.h"
//#include "stl_bvector.h"
#include "SpinIndex.h"
#include "Serializable.h"

// std::_Bvector
// stl_bvector
typedef std::_Bvector vector_bool;
// typedef bit_vector vector_bool;

//! Rectangular lattice in 2D of up or down spins.

/*!	Matt Kuntz's code defined an enum representing UP or DOWN, but it
	was clear that the integer values +1 and -1 were very often useful
	in his code, so we're taking that experience and making this class
	specify that you treat the values of the
	spins as +1 or -1.

	You can switch this class to handle more than 2^32-1 spins by 
	changing the typedef of vector_bool to the stl_vector version.
	I tool the Gnu STL vector<bool> and hacked it so it will
	address as many spins as it can fit in 2^32 bytes.  That's eight
	times as many, and we can use them.

	It's only a useful option if you also switch the SpinIndex
	to a 64 bit integer (ULONGLONG in MSVC++).

	\author ajd27
	\date 15 Aug 2000
*/
class SpinLattice2D : private vector_bool, public Serializable
{
	SpinIndex iLength;
	SpinIndex N;
public:
//	typedef SpinIndex* iterator;
	//! Constructor
	SpinLattice2D();
	//! Destructor.
	~SpinLattice2D();
	//! Set the length of a side.  The number of spins is length squared.
	bool SetLength( const SpinIndex iLength );
	//! Identical to Get.  This returns the value of the given spin.
	int operator()( const SpinIndex iPos ) const;
	//! Return the value of a given spin.
	int Get( const SpinIndex iPos ) const;
	//! Flip a spin up or down.
	void flip( const SpinIndex iPos, int orientation );
	//! How many up neighbors does this spin have?
	int NeighborsUp( const SpinIndex iPos ) const;
	//! Who are the neighbors of this spin?
	void GetNeighbors( const SpinIndex iPos, SpinIndex* const neighbors ) const;
	//! Get the coordinates of this index.
	void GetCoords( SpinIndex iPos, SpinIndex* coords ) const;
	//! Return the length of the sides.
	SpinIndex GetLength() const;
	//! What is the coordination number of this lattice?
	SpinIndex GetCoordinationNumber() const;
	//! Return the spatial dimension of the lattice.
	int GetDimension() const { return 2; }
	//! How many spins are up?
	SpinIndex GetTotalNumberUp() const;

	//! This saves the SpinLattice2D.
	virtual void Save(DMWriter *writer) const;
	//! Read in new values for this SpinLattice2D.
	virtual void Load(DMReader *reader);
	//! Return the signature type of this spinlattice.
	virtual const std::string GetType() const;
	//! Consider this private.  Use GetType instead.
private:
	int* WriteSlice( SpinIndex start, SpinIndex finish, int& dwLen ) const;
	void ReadSlice( SpinIndex start, SpinIndex finish, int*data );
};

/*! The fastest way to turn a bool into an int.
  */
inline int SpinLattice2D::operator()( const SpinIndex iPos ) const
{
	return ((*this)[iPos]) ? 1 : -1;
}

inline int SpinLattice2D::Get( const SpinIndex iPos ) const
{
	return ((*this)[iPos]) ? 1 : -1;
}

inline void SpinLattice2D::flip( const SpinIndex iPos, int orientation )
{
	(*this)[iPos] = ( orientation == 1 );
}

inline int SpinLattice2D::NeighborsUp( const SpinIndex iPos ) const
{
	int nup = 0;
	SpinIndex iX = iPos%iLength, iY = iPos/iLength;
	if (iX==0) {
		nup += ((*this)[iPos+iLength-1]) ? 1 : 0;
		nup += ((*this)[iPos+1]) ? 1 : 0;
	} else if (iX == iLength-1) {
		nup += ((*this)[iPos-1]) ? 1 : 0;
		nup += ((*this)[iPos+1-iLength]) ? 1 : 0;
	} else {
		nup += ((*this)[iPos-1]) ? 1 : 0;
		nup += ((*this)[iPos+1]) ? 1 : 0;
	}

	if (iY==0) {
		nup += ((*this)[iPos+iLength*(iLength-1)]) ? 1 : 0;
		nup += ((*this)[iPos+iLength]) ? 1 : 0;
	} else if (iY == iLength-1) {
		nup += ((*this)[iPos-iLength]) ? 1 : 0;
		nup += ((*this)[iX]) ? 1 : 0;
	} else {
		nup += ((*this)[iPos-iLength]) ? 1 : 0;
		nup += ((*this)[iPos+iLength]) ? 1 : 0;
	}
	return nup;
}

/*! 
	@param iPos position of the spin in the lattice
	@param neighbors Array in which the neigbors will be returned.
	Must be able to accept Coordination Number elements.
  */
inline void SpinLattice2D::GetNeighbors( const SpinIndex iPos, SpinIndex *const neighbors ) const
{
	SpinIndex iX = iPos%iLength, iY = iPos/iLength;
	if (iX == 0) {
		neighbors[0] = iPos+iLength-1;
		neighbors[1] = iPos+1;
	} else if (iX == iLength-1) {
		neighbors[0] = iPos-1;
		neighbors[1] = iPos+1-iLength;
	} else {
		neighbors[0] = iPos-1;
		neighbors[1] = iPos+1;
	}

	if (iY == 0) {
		neighbors[2] = iPos+iLength*(iLength-1);
		neighbors[3] = iPos+iLength;
	} else if (iY == iLength-1) {
		neighbors[2] = iPos-iLength;
		neighbors[3] = iX;
	} else {
		neighbors[2] = iPos-iLength;
		neighbors[3] = iPos+iLength;
	}
}

inline void SpinLattice2D::GetCoords( SpinIndex iPos, SpinIndex* coords ) const
{
	coords[0] = iPos%iLength;
	coords[1] = iPos/iLength;
}


#endif //_SPIN_LATTICE_2D_H_
