#include "DMFirstHeader.h"
#include "TimeSeries.h"
#include "DMException.h"
#include "DMWriter.h"
#include "DMReader.h"
#include "SerialBuilder.h"
#include "DMLastHeader.h"

const std::string sType = std::string("TimeSeries");
const SerialBuilder<TimeSeries> *_creator = new
	SerialBuilder<TimeSeries>(sType);

TimeSeries::TimeSeries()
{
	avalanche = 0;
	bSerialized = false;
	iVerify = 0;
}

void TimeSeries::SetAlgorithm( AvalancheAlgorithm* avalanche )
{
	ASSERT( 0 != avalanche );
	if (avalanche == 0) throw
		DMException("TimeSeries","SetAlgorithm",
		"You gave this method a null pointer.");

	this->avalanche = avalanche;

	iVerify |= VERIFY_ALGORITHM;
}

void TimeSeries::Initialize()
{
	ASSERT( 0 != avalanche );
	if (avalanche == 0) throw
		DMException("TimeSeries","Initialize",
		"You need to call SetAlgorithm before initializing.");

	bSerialized = false;
	iVerify = 0;
}

/*! This is called by Initialize for the first sweep, but you
	can call it again without harm.

	There is no way to keep data for this observer.
	*/
void TimeSeries::BeginSweep( bool bKeepData )
{
	ASSERT( 0 == iVerify );
	if ( 0 != iVerify ) {
		throw DMException( "TimeSeries","BeginSweep",
			"If you change the algorithm, you need to Initialize again." );
	}

	lastM = avalanche->GetM();
	this->DeleteArrays();
}

void TimeSeries::Save( DMWriter *writer ) const
{
	std::vector<int> shape;
	double *dData;
	int *iData;

	writer->PutInt( lastM, "lastM" );
	writer->PutInt( H.size(), "length");
	if ( H.size() > 0 ) {
		shape.push_back( H.size() );
		iData = (int*) &size[0];
		writer->PutIntArray( iData, shape, "size" );
	
		dData = (double*) &H[0];
		writer->PutDoubleArray( dData, shape, "H" );
	}

}

/*! If you call Load, you have to call Initialize.
	*/
void TimeSeries::Load( DMReader *reader )
{
	std::vector<int> shape;
	double* dData;
	int* iData;

	this->DeleteArrays();

	lastM = reader->GetInt( "lastM" );
	int length = reader->GetInt( "length" );

	if (length > 0) {
		iData = reader->GetIntArray( shape, "size" );
		std::back_insert_iterator< std::vector<int> > ii(size);
		std::copy( iData, iData+shape[0], ii);
		delete[] iData;

		dData = reader->GetDoubleArray( shape, "H" );
		std::back_insert_iterator< std::vector<double> > ih(H);
		std::copy( dData, dData+shape[0], ih );
		delete[] dData;
	}
	bSerialized = true;
}

const std::string TimeSeries::GetType() const
{
	return sType;
}

/*! Doesn't actually delete arrays, since they are STL containers.
	Just sets them to zero.
	*/
void TimeSeries::DeleteArrays()
{
	H.erase(H.begin(),H.end());
	size.erase(size.begin(),size.end());
}



std::ostream& operator<<( std::ostream& s, const TimeSeries& r )
{
	size_t i;
	for (i=0; i<r.H.size(); i++)
		s << r.H[i] << "\t" << r.size[i] << std::endl;
	return s;
}

