#ifndef _XY_AVERAGE_H_
#define _XY_AVERAGE_H_

#include "DMFirstHeader.h"
#include <math.h>
#include <vector>
#include <iostream>
#include "DigitalMaterial.h"
#include "XYData.h"
#include "Serializable.h"

//! Keeps running average and standard deviation of XY data.
/*! Stores standard deviation as well.

	This class has two modes, represented by bAbsentEqualsZero.
	In one case, every time data accumulates data sets of different
	length, missing data at the end is treated as a measurement of 
	zero.  Here, the SampleCnt for each data point is the same
	as the SetCnt of datasets.  In the other case, absent data is
	treated as though there were no measurement.  SampleCnt is incremented
	only when the nonzero data is accumulated by the class, and this
	value of SampleCnt is used to calculate standard deviations.

	\author Matt Kuntz
	\author Drew Dolgert
	*/
class XYAverage : public XYData, public Serializable
{
	struct Point {
		double X;
		double Y; // Cumulative sum.
		double YSq; // For cumulative sum of squares.
		int SampleCnt; // Number of times this point nonzero.
	};

	int SetCnt; //!< The number of data sets added.
	std::vector<Point> points;
	bool bAbsentEqualsZero;

public:
	XYAverage( bool bAbsentEqualsZero = true );
	~XYAverage();

	int GetPointCnt( ) const { return points.size(); }
	double GetX( int PointIdx ) const { return points[PointIdx].X; }
	double GetY( int PointIdx ) const { return points[PointIdx].Y; }
	double GetStandardDeviation( int PointIdx ) const;
	int GetSampleCnt( int PointIdx ) const { return points[PointIdx].SampleCnt; }

	int GetSetCnt( ) const { return SetCnt; }

	void AddDataSet( const XYData& data );

	void Clear();

	//! Serializable interface
	//@{
	void Save( DMWriter *writer ) const;
	void Load( DMReader *reader );
	const std::string GetType() const { return sType; }
	//@}
	friend std::ostream& operator<<( std::ostream& out, const XYAverage& data );
	static const std::string sType;
};

std::ostream& operator<<( std::ostream& out, const XYAverage& data );

/*! This returns a standard deviation divided by an extra factor of
	sqrt( SampleCnt).
	*/
inline double XYAverage::GetStandardDeviation( int PointIdx ) const
{
	const Point& pt = points[PointIdx];
	if ( pt.SampleCnt > 1 ) {

		double AverageOfSquares = pt.YSq / pt.SampleCnt;
		double Average = pt.Y/pt.SampleCnt;
		double SquareOfAverages = Average*Average;
		return sqrt( ( AverageOfSquares - SquareOfAverages ) /
			(pt.SampleCnt - 1) );

	}
	return pt.Y / pt.SampleCnt;
}

#endif // _XY_AVERAGE_H_
