#include "wx/statline.h" // for new wxStaticLine
#include "HysteresisFrame.h"
#include "ColoringObserver.h"
#include "HysteresisCanvas.h"
#include "EventID.h"

BEGIN_EVENT_TABLE(HysteresisFrame, wxFrame)

	EVT_MENU	(ID_EXIT,	HysteresisFrame::OnMenu)
	EVT_CLOSE	(			HysteresisFrame::OnClose)

	EVT_MENU	(ID_ZOOMIN,		HysteresisFrame::OnZoom)
	EVT_MENU	(ID_ZOOMOUT,	HysteresisFrame::OnZoom)

	EVT_TIMER	(ID_TIMER, HysteresisFrame::OnTimer)
	EVT_MENU	(ID_START, HysteresisFrame::OnMenu)
	EVT_MENU	(ID_STOP,  HysteresisFrame::OnMenu)
	EVT_MENU	(ID_STEP,  HysteresisFrame::OnMenu)

END_EVENT_TABLE()


HysteresisFrame::HysteresisFrame()
	: wxFrame((wxFrame*) 0, -1, _("Hysteresis"), wxPoint(200,200))
{
	// Create all the menus
	// File menu
	wxMenu* menuFile = new wxMenu("", wxMENU_TEAROFF);
	menuFile->Append(ID_EXIT, _("E&xit\tAlt-X"), _("Quit this program"));

	// Control of running
	wxMenu* menuRun = new wxMenu("", wxMENU_TEAROFF);
	menuRun->Append(ID_START, _("&Start\tCtrl-S"),_("Start sim") );
	menuRun->Append(ID_STOP, _("Sto&p\tCtrl-P"),_("Stop sim") );
	menuRun->Append(ID_STEP, _("&Next\tCtrl-N"),_("Next step") );

	// Changing simulation options
	wxMenu* menuSim = new wxMenu("", wxMENU_TEAROFF);
	menuSim->Append(ID_PARAMETERS, _("Para&meters...\tCtrl-M"),
		_("Set sim parameters") );
	menuSim->AppendSeparator();
	// Make a series of checkable menus.
	menuSim->Append(ID_TOGGLE_COLORING, _("Coloring"),
		_("Toggle coloring observer"), true );
	menuSim->Append(ID_TOGGLE_MAGNETIZATION, _("Magnetization"),
		_("Toggle magnetization observer"), true );
	menuSim->Append(ID_TOGGLE_TIME_SERIES, _("Time series"),
		_("Toggle time series observer"), true );
	menuSim->Append(ID_TOGGLE_CORRELATION_FUNCTION, _("Correlation Function"),
		_("Toggle correlation function observer"), true );

	// Menu options for viewing coloring.
	wxMenu* menuView = new wxMenu("", wxMENU_TEAROFF);
	menuView->Append(ID_ZOOMIN,_("Zoom &in\tCtrl-i"),_("Zoom in") );
	menuView->Append(ID_ZOOMOUT,_("Zoom &out\tCtrol-o"),_("Zoom out") );

	wxMenuBar* menuBar = new wxMenuBar();
	menuBar->Append(menuFile, _("&File"));
	menuBar->Append(menuRun, _("&Run"));
	menuBar->Append(menuSim, _("&Simulation"));
	menuBar->Append(menuView, _("&View"));
	SetMenuBar( menuBar );

	m_simulation = new HysteresisSimulation();
	m_simulation->SetSideLength( 64 );
	m_simulation->SetDisorder( 0.95 );
	m_simulation->EnableColoringObserver();
	m_simulation->Initialize();
	m_simulation->Verify();

	m_timer = new wxTimer( this, ID_TIMER );
	m_running = false;
	m_interval = 250;
	m_tics = 0;
	

	const ColoringObserver* coloring = m_simulation->GetColoringObserver();

	// pass a pointer to the simulation to the canvas.
	m_canvas = new HysteresisCanvas( coloring, this, -1, wxPoint(0,0), wxSize(100,100) );
	m_canvas->SetCellSize( 2 );

	wxBoxSizer* topSizer = new wxBoxSizer( wxVERTICAL );
	topSizer->Add( m_canvas, 1, wxEXPAND );

	SetAutoLayout( TRUE );
	SetSizer( topSizer );
}


HysteresisFrame::~HysteresisFrame()
{
	if (m_timer) delete m_timer;
	if (m_simulation) delete m_simulation;
}


void HysteresisFrame::OnMenu( wxCommandEvent& event )
{
	switch (event.GetId())
	{
	case ID_EXIT:
		{
			// Force the frame to close.
			Close( TRUE );
			break;
		}
	case ID_START:
		{
			OnStart();
			break;
		}
	case ID_STOP:
		{
			OnStop();
			break;
		}
	case ID_STEP:
		{
			OnStep();
			break;
		}
	default: ; // do nothing.
	}
}


void HysteresisFrame::OnClose( wxCloseEvent& event )
{
	Destroy();
}

void HysteresisFrame::OnTimer( wxTimerEvent& WXUNUSED(event))
{
	if (m_running)
		OnStep();
}

void HysteresisFrame::OnStart()
{
	if (!m_running)
	{
		m_timer->Start(m_interval);
		m_running = true;
	}
}

void HysteresisFrame::OnStop()
{
	if (m_running) {
		m_timer->Stop();
		m_running = false;
	}
}


void HysteresisFrame::OnStep()
{
	AvalancheAlgorithm::Status status = m_simulation->Run( 10 );
	if ( status != AvalancheAlgorithm::SUCCESS ) {
		OnStop();
	}
	m_canvas->Refresh( false );
}

void HysteresisFrame::OnZoom( wxCommandEvent& event )
{
	int cellsize = m_canvas->GetCellSize();

	if ( (event.GetId() == ID_ZOOMIN) && (cellsize < 32) ) {
		m_canvas->SetCellSize( cellsize * 2 );
	} else if ( (event.GetId() == ID_ZOOMOUT) && (cellsize > 1) ) {
		m_canvas->SetCellSize( cellsize / 2 );
	} else {
		; //wxASSERT( false, "There is a Zoom event which isn\'t in or out." );
	}
}
