// MDMain.cpp : main loop for Unix, Windows Console applications

#include "DMFirstHeader.h"
#include <iostream>
#include <math.h>
#include <set>
#include "DigitalMaterial.h"

#include "MDMain.h"
#include "DMArray.h"
#ifdef NETCDF_DIR
#include "NetCDFWriter.h"
#include "NetCDFReader.h"
#endif

#include "ASCIIWriter.h"
#include "ASCIIReader.h"
#include "NetCDFReader.h"
#include "CellNeighborList.h"
#include "DMLastHeader.h"

// Main loop: sets up and controls the simulation
// The software is designed so that this can be done in an interpreted, 
// scripting language

using namespace std;

int main(int argv, char **argc)
{
	try
	{
#ifdef WIN32
	WinProcessFactory::Create();
#else
	UnixProcessFactory::Create();
#endif

// Define a lattice. Dimension dependent.
#if DIMENSION == 2
        CutLennardJonesPotential potential;
	TriangularLattice lattice(potential.GetLengthScale());   
// potential.GetLengthScale() returns the interatomic distance in the crystal at equilibrium
#else
	EMTPotential potential;
	FCCLattice lattice(potential.GetLengthScale()*sqrt(2.));
#endif // DIMENSION == 2

// Define an Atom Initializer. 
	SphericalClusterInitializer makeSphere(4.*potential.GetLengthScale(), &lattice);  // radius and BravaisLattice
#if DIMENSION == 2
	double center[DIMENSION] = {-10.0, 17.};	// Center of atomic cluster
#else
	double center[DIMENSION] = {-10.0, 17., 0.};	// Center of atomic cluster
#endif // DIMENSION == 2

	makeSphere.SetCenter(center);

	DynamicListOfAtoms *atoms0 = new DynamicListOfAtoms(0);
// Initialize atoms positions
	makeSphere.Create(atoms0);

// print the number of atoms in the sphere
	cout << atoms0->GetNumber() << " atoms in the sphere" << endl;
// set the sphere into motion
	int i;
	for(i=0; i<atoms0->GetNumber(); i++) atoms0->SetCartesianVelocity(i,0,-0.4);



	DynamicListOfAtoms *atoms1 = new DynamicListOfAtoms(0);

	
// now create a rectangle of atoms
#if DIMENSION==2
	double rectSize[DIMENSION] = {10.0,10.};
#else
	double rectSize[DIMENSION] = {10.0,10.0,10.0};
	double rotation[DIMENSION][DIMENSION];  // rotation to get a 111 surface on the rectangle to be built
	for(i = 0; i < DIMENSION; i++) 
		rotation[i][0] = 1./sqrt(6.)*(i==0? 2.:-1.);
	for(i = 0; i < DIMENSION; i++) 
		rotation[i][1] = -1./sqrt(2.)*(i==0? 0.: 1.-2*(i%2));
	for(i = 0; i < DIMENSION; i++) 
		rotation[i][2] = 1./sqrt(3.);
#endif // DIMENSION==2
	
	RectangularClusterInitializer makeRectangle(rectSize,&lattice);

	center[0] = 20.; center[1] = 20.0;
#if DIMENSION == 3
	lattice.RotateLatticeVectors(rotation);
    center[2] = 0.;
#endif // DIMENSION == 3 

	makeRectangle.SetCenter(center);

	makeRectangle.Create(atoms1);			// Sets up cluster

	cout << atoms1->GetNumber() << " atoms in the rectangle " << endl;

//	atoms1->Merge(atoms0); // merge the lists into atoms.

	delete atoms0; // now we don't need atoms0.

	// Finally we add a Sphere of fixed atoms at the origin
	for(i=0; i<DIMENSION; ++i) center[i] = 100.;
	makeSphere.SetCenter(center);
	atoms0 = new DynamicListOfAtoms(0);
	makeSphere.Create(atoms0);
	Constraint *fixedAtoms = new FixedBody(atoms0->GetCartesianPositions());
	atoms0->SetConstraint(fixedAtoms);
	cout << atoms0->GetNumber() << " atoms in the fixed sphere " << endl;


// now we create a composite list of atoms with atoms0 and atoms1;
//	CompositeListOfAtoms *atoms = new CompositeListOfAtoms;
	DynamicListOfAtoms *atoms = new DynamicListOfAtoms;
	atoms->Attach(atoms1);
	atoms->Attach(atoms0);

	cout << atoms->GetNumber() << " atoms in the composite list before pruning and removing" << endl;

// some manipulation tests:
	atoms->RemoveBranch(atoms1);
	cout << "=? " << atoms->GetNumber() << " + " << atoms1->GetNumber() << endl;

	// in case some atoms are overlapping we apply overlap pruner to atoms1. (at this time we cannot prune constrained lists of atoms)
	OverlapPruner overlapPruner(lattice.GetLatticeSpacing()*0.9/sqrt(2.));
//	overlapPruner.Transform(atoms1);

    ListOfAtoms *atoms2 = atoms1->Copy();

	atoms->Attach(atoms1);
	cout << "Again: " << atoms->GetNumber() << " atoms in the composite list before pruning and removing" << endl;

//	we can also directly remove atoms from a list, given their index:
	std::set<int> toRemove;
	toRemove.insert(3);
	toRemove.insert(10);
//	atoms1->RemoveAtoms(toRemove);

	cout << atoms->GetNumber() << " atoms in the composite list " << endl;

	assert(atoms->GetShape() == atoms->GetCartesianPositions()->GetShape());
	
	// Here is a Write/Read test
#ifdef NETCDF_DIR
	NetCDFWriter ww("test.nc");
#else
	ASCIIWriter ww("test.nc");
#endif
	ww.Put(atoms);

#ifdef NETCDF_DIR
	NetCDFReader rr("test.nc");
#else
	ASCIIReader rr("test.nc");
#endif
	ListOfAtoms *test = dynamic_cast<ListOfAtoms *>(rr.Get());
//	DMMATRIX<double, DIMENSION> *test = dynamic_cast<DMMATRIX<double, DIMENSION>*>(rr.Get());

	std::cout << test->GetNumber() << " atoms in the file " <<endl;
	delete test;

	//	Shape shape(452);
	//	DynamicListOfAtoms *aaa = new DynamicListOfAtoms(shape);
	cerr << atoms->GetShape() << endl;
	cerr << atoms0->GetShape()<< endl;
	cerr << atoms1->GetShape()<< endl;
	//	atoms->RemoveSubList(atoms1);
	DynamicListOfAtoms *copied = dynamic_cast<DynamicListOfAtoms *>(atoms->Copy());

	cerr << "done"; 
	//	SimplePeriodicBoundaryConditions *bc = new SimplePeriodicBoundaryConditions();     // Free Boundary Condition
	// 	atoms.SetBoundaryConditions(bc);

	// AtomsMover(&potential, dt, nSteps)	
	// dt = time step, nSteps * dt = time elapsed

	// VerletAtomsMover verlet(&potential, 0.3, 100);
	VerletAtomsMover verlet(&potential, 0.01, 10);

	NeighborLocator *neighborLocator =
	 	  new CellNeighborList(potential.GetCutoffDistance(),0.05*potential.GetCutoffDistance());  // the 2nd argument is the allowed drift distance (if an atoms moved more than the drift, then we rebuild the neighborlist)
//	 	  new CellNeighborLocator(potential.GetCutoffDistance(),0.05*potential.GetCutoffDistance());  // the 2nd argument is the allowed drift distance (if an atoms moved more than the drift, then we rebuild the neighborlist)
		  //	  new CellNeighborList(potential.GetCutoffDistance(),0.05*potential.GetCutoffDistance());  // the 2nd argument is the allowed drift distance (if an atoms moved more than the drift, then we rebuild the neighborlist)
	//new NeighborList(potential.GetCutoffDistance(),1.1*potential.GetCutoffDistance());  // the 2nd argument is the allowed drift distance (if an atoms moved more than the drift, then we rebuild the neighborlist)

	  //new SimpleNeighborLocator(potential.GetCutoffDistance());


	atoms->SetNeighborLocator(neighborLocator);


	double Etot = 0;
	for(int at=0; at<atoms->GetNumber(); ++at) Etot += potential.CalculateAtomicEnergy(atoms, at);
	std::cout << Etot << std::endl << potential.CalculateEnergy(atoms) << std::endl;


#ifdef LASSPTOOLS
	// Attach observer
	PlotAtomsObserver plot(0, 30.0, 0, 30.0, 25);
	// 	atoms->AttachObserver(&plot);
#endif
	double t=0;

	for (i = 0; i < 10; i++)
	{
		verlet.Move(atoms);
		atoms->Notify();
		cout << "Energy = " << potential.CalculateEnergy(atoms) + atoms->GetKineticEnergy() << endl;
	}

	delete atoms1;
	cout << atoms->NumberOfBranches() << endl;
	cout << atoms->GetNumber() << endl;
	cout << atoms0->NumberOfBranches() << endl;
	delete atoms;
	delete neighborLocator;
	}
	catch (DMException e){
		std::cout<< e;}

	return 0;
}
