// MDMain.cpp : main loop for Unix, Windows Console applications

#include "DMFirstHeader.h"
#include <fstream>
#include <iostream>
#include "DigitalMaterial.h"
#include "MDMain.h"

#include "DMLastHeader.h"

/*
  Main function for testing a simple parallel application of the MD code for the purpose of applying message logging.

 */

using namespace std;

void main(int argc, char **argv)
{
	try
	{
#ifdef WIN32
	WinProcessFactory::Create();
#else
	UnixProcessFactory::Create();
#endif

#if DMPARALLEL == 1
	MPI_Init(&argc, &argv);

	//MPIProtocol protocol;
	//dmProtocol = &protocol;
	cout.precision(12);

	ProtocolFactory::Instance()->SetProtocol(ProtocolFactory::World);
#endif	

	int nProcs = dmProtocol->GetNumberOfProcessors();
	int myProc = dmProtocol->GetProcessorNumber();
	if(myProc == 0)	
		cout << "DMPARALLEL is " << DMPARALLEL << " Number of processors is " << nProcs << endl;

	// Choose a cpu layout (ie how many processors in each direction)
	int cpulayout[DIMENSION], tmp0, tmp1;
#if DIMENSION == 2
	tmp0 = int(sqrt(nProcs));
	cpulayout[0] = tmp0;
	cpulayout[1] = nProcs/tmp0;
#else
	tmp0 = int(pow(nProcs, (1/3.)));
	tmp1 = int(sqrt(nProcs/tmp0));
	cpulayout[0] = tmp0;
	cpulayout[1] =  tmp1;
	cpulayout[2] =  nProcs/tmp0/tmp1;
#endif

	if(0 == myProc)
	{
		cout << "CPU layout" << endl;
		for( int coord=0; coord < DIMENSION; ++coord )
			cout << cpulayout[coord] << " ";
		cout << endl;
	}

	// run parameters
	int plotting = 0;
	int nUnitCells[DIMENSION];
	nUnitCells[0] = 32;
	nUnitCells[1] = 16;
	if(DIMENSION == 3)
		nUnitCells[2] = 32;
	double temperature = 0.2;
	double friction = 0.1;
	double dt = 0.005;
	int nMinorSteps = 50;
	int nMajorSteps = 10;



// Define a lattice. Dimension dependent.
#if DIMENSION == 2
	CutLennardJonesPotential concretePotential;
	double latticeConstant = concretePotential.GetLengthScale();
	TriangularLattice lattice(latticeConstant);   

// potential.GetLengthScale() returns the interatomic distance in the crystal at equilibrium
#else
	StillingerWeberPotential concretePotential;
	//EMTPotential potential;
	double latticeConstant = concretePotential.GetLengthScale()*4./sqrt(3.);
	//FCCLattice lattice(latticeConstant);
	DiamondLattice lattice(latticeConstant);
#endif // DIMENSION == 2

	Potential *potential;
#if DMPARALLEL == 1
	potential = new ParallelPotential(&concretePotential);
#else
	potential = &concretePotential;
#endif
	double cutoff = concretePotential.GetCutoffDistance();

	// Define an Atom Initializer. 
	double rectangleSize[DIMENSION], center[DIMENSION];
	for(int coord=0;coord<DIMENSION;++coord)
	{
		double edgeLength = nUnitCells[coord] * latticeConstant;
		rectangleSize[coord] = edgeLength;
		center[coord] = 0.5 * edgeLength;
	}
	if(2 == DIMENSION)
	{
		if(nUnitCells[1]%2 != 0)
			++nUnitCells[1];
		rectangleSize[1] = nUnitCells[1] * latticeConstant * sqrt(3.)/2;
	}
	RectangularClusterInitializer makeRect(rectangleSize, &lattice);  
	makeRect.SetCenter(center);

	DynamicListOfAtoms *atoms = new DynamicListOfAtoms(0);

	// Initialize atoms positions
#if DMPARALLEL == 1
	ParallelAtomsInitializer makeRectParallel = ParallelAtomsInitializer(&makeRect);
	makeRectParallel.Create(atoms);
	HomogeneousDomainSubdivision domain(cpulayout, myProc, cutoff);
		
	DMArray<double, DIMENSION> *pos =
		const_cast<DMArray<double, DIMENSION>*>(atoms->GetCartesianPositions());


	Synchronizer synchronizer(pos);
	Migrator migrator(&domain, pos);

#else
	makeRect.Create(atoms);
#endif



	PeriodicBoundaryConditions *pbc = new SimplePeriodicBoundaryConditions();  
	pbc->SetLength(rectangleSize);
	atoms->SetBoundaryConditions(pbc);

	

	VerletAtomsMover verlet(potential, dt, nMinorSteps);
	LangevinAtomsMover langevin(potential, dt, nMinorSteps, friction, temperature);

	NeighborLocator *neighborLocator =
	 	  new CellNeighborList(concretePotential.GetCutoffDistance(),0.05*concretePotential.GetCutoffDistance());

#if DMPARALLEL == 1
	ParallelNeighborLocator neighborLocatorParallel(&migrator, neighborLocator);
	atoms->SetNeighborLocator(&neighborLocatorParallel); // this will migrate the atoms if it hasn't been done
#else
	atoms->SetNeighborLocator(neighborLocator);
#endif
	
	cout <<  "on processor " << myProc << ", we have:" << atoms->GetNumber() << " atoms" << endl;


#ifdef LASSPTOOLS
	if(plotting  && 0 == myProc)
	{
		// Attach observer
		PlotAtomsObserver plot(30.0, 30.0, 25);
		atoms->AddObserver(&plot);
	}
#endif

	//CheckPointObserver checkPointer(potential, &langevin, "siliconCube");
	//atoms->AddObserver(&checkPointer);

	ofstream *fileOut = 0;
	if( 0 == myProc )
	{
		fileOut = new ofstream("energies.dat");
		fileOut->precision(16);
	}
	cout.precision(12);
	for ( int step = 0; step < nMajorSteps; ++step )
	{
		langevin.Move(atoms);
		//verlet.Move(atoms);
		atoms->Notify();
		double potentialEnergy = potential->CalculateEnergy(atoms);
		double tmp = atoms->GetKineticEnergy();
		double kineticEnergy;
		dmProtocol->AllReduce(&tmp, &kineticEnergy,1,DMProtocol::SUM);
		
		
		if( 0 == myProc )
		{
			cout << "Potential Energy = " << potentialEnergy << endl;
			cout << "Kinetic Energy = " << kineticEnergy << endl;
			cout << "Total Energy = " << kineticEnergy + potentialEnergy << endl;
			(*fileOut) << kineticEnergy + potentialEnergy << endl;
		}
	}

	delete atoms;
	delete pbc;
	delete neighborLocator;
#if DMPARALLEL == 1
	delete potential;
#endif
	delete fileOut;
	}
	catch (DMException e){
		std::cout << e;}

#if DMPARALLEL == 1
	MPI_Finalize();
#endif
}
