// ParallelepipedClusterInitializer.cpp: implementation of the ParallelepipedClusterInitializer class.
//
//////////////////////////////////////////////////////////////////////

#include "DigitalMaterial.h"
#include "ParallelepipedClusterInitializer.h"
#include "SimpleMatrixInverter.h"
#include "BravaisLatticeWithBasis.h"

#include "DMLastHeader.h"

#if DIMENSION == 3

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ParallelepipedClusterInitializer::ParallelepipedClusterInitializer(
                                   const double boundaryVectors[DIMENSION][DIMENSION], 
                                   BravaisLatticeWithBasis *bravais) : ClusterInitializer(bravais)
{
	SetBoundaryVectors(boundaryVectors);
}


ParallelepipedClusterInitializer::~ParallelepipedClusterInitializer()
{

}

void ParallelepipedClusterInitializer::SetBoundaryVectors(
                                      const double boundaryVectors[DIMENSION][DIMENSION])
{
	int vec, coord;
	double transposedVectors[DIMENSION][DIMENSION];

    for( coord = 0; coord < DIMENSION; ++coord ) 
		for( vec = 0; vec < DIMENSION; ++vec ) 
		{
			this->boundaryVectors[coord][vec] = boundaryVectors[coord][vec];
			this->inverseBoundaryVectors[coord][vec] = boundaryVectors[coord][vec];
			transposedVectors[vec][coord] = boundaryVectors[coord][vec];
		}
	
	TinyVector<double, DIMENSION> sides[DIMENSION];
	double lengths[DIMENSION];
	for( vec = 0; vec < DIMENSION; ++vec ) 
	{
		sides[vec] = transposedVectors[vec];
		lengths[vec] = sqrt(sides[vec].norm2());
		sides[vec] *= 1. + 2 * bravais->GetLatticeSpacing()/lengths[vec];
	}

	TinyVector<double, DIMENSION> areaVector = sides[0].Cross(sides[1]);

	if( 2 == DIMENSION )
		this->volume = sqrt(areaVector.norm2());
	else if( 3 == DIMENSION )
		this->volume = fabs(areaVector * sides[2]);
	else throw DMException("ParallelepipedClusterInitializer","SetBoundaryVectors","Volume formula not available for general DIMENSION");

	SetMaxSize();

	SimpleMatrixInverter<DIMENSION> inverter;
	inverter.Invert(inverseBoundaryVectors);

}

int ParallelepipedClusterInitializer::SetMaxSize()
{

 	int coord, vec;
	double corner[DIMENSION], ind[DIMENSION], maxi[DIMENSION], mini[DIMENSION];
	for(int index = 0; index < pow(2,DIMENSION); ++index)
	{
		int itmp = index;
		for( coord = 0; coord < DIMENSION; ++coord ) 
			corner[coord] = clusterCenter[coord];
		
		for( vec = 0; vec < DIMENSION; ++vec )
		{
			for( coord = 0; coord < DIMENSION; ++coord ) 
				
				corner[coord] += boundaryVectors[coord][vec]/2.*(double) (itmp % 2 == 0 ? -1:1);
			itmp /= 2;
		}
		bravais->LatticeCoordinatesFromCartesianCoordinates(ind,corner);
		for( coord = 0; coord < DIMENSION; ++coord )
		{

			if((index == 0) || (maxi[coord] < ind[coord])) maxi[coord] = ind[coord];
			if((index == 0) || (mini[coord] > ind[coord])) mini[coord] = ind[coord];
		}
	}
	for( coord = 0; coord < DIMENSION; ++coord ) 
	{
		maxSize[coord] = (int)maxi[coord] + 1;
		minSize[coord] = (int)mini[coord] - 1;
	}

	return int(volume/bravais->GetVolumeOfCell()*bravais->GetNAtomsPerCell());

}


bool ParallelepipedClusterInitializer::Inside(double X[DIMENSION])
{
	// Return outside if near negative boundaries, inside if near positive boundaries

	double transformedPos[DIMENSION];
    int coord, vec;
 
	for(vec=0;vec<DIMENSION;++vec)
		transformedPos[vec] = 0;
	for(vec=0;vec<DIMENSION;++vec)
		for(coord=0;coord<DIMENSION;++coord)
			transformedPos[vec] += inverseBoundaryVectors[vec][coord] * (X[coord] - clusterCenter[coord]);

	bool in = true;
	for ( vec = 0; vec < DIMENSION; ++vec)
	{
		double smallFraction = 1e-14;

		if (fabs(fabs(transformedPos[vec]) - 0.5) < smallFraction)
		{
			// Near boundary
			if (transformedPos[vec] < 0) in = false;
		}
		else
		{
			in = in && (fabs(transformedPos[vec]) <= 0.5);
		}
	}
	return in;
}



#endif //DIMENSION == 3

