// ExtendedQuickMinAtomsMover.cpp: implementation of the ExtendedQuickMinAtomsMover class.
//
//////////////////////////////////////////////////////////////////////


#include "DMFirstHeader.h"
#include <iostream>
#include <fstream>
#include "DigitalMaterial.h"
#include "ExtendedQuickMinAtomsMover.h"
#include "QuickMinAtomsMover.h"
#include "Potential.h"
#include "ListOfAtoms.h"
#include "DMArray.h"
#include "lapack.h"
#include "SerialBuilder.h"
#include "Rand.h"

#include "DMLastHeader.h"


using namespace std;

static SerialBuilder<ExtendedQuickMinAtomsMover> *_extQMcreator = 
    new SerialBuilder<ExtendedQuickMinAtomsMover>("ExtendedQuickMinAtomsMover");

const string ExtendedQuickMinAtomsMover::className = string("ExtendedQuickMinAtomsMover");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////



ExtendedQuickMinAtomsMover::ExtendedQuickMinAtomsMover(AtomsExtension *atomsExtension,QuickMinAtomsMover *qm,Potential *potential,double dt,double tolerance,int nBdyIterations)
{
	maxIterations = 10;
	verbose = false;
	this->nParameters = atomsExtension->GetNParameters();
	this->atomsExtension = atomsExtension;
	this->qm = qm;
	this->potential = potential;
	this->dt = dt;
	this->tolerance = tolerance;
	this->nBdyIterations = nBdyIterations;
	nAtoms = 0;
	nIterations = 0;
	convergence = 0.;
	forces = 0;
	
	parameters = new double[nParameters];
	parameterVelocities= new double[nParameters];
	parameterForces= new double[nParameters];
	int n;
	for(n=0;n<nParameters;++n)
		parameters[n] = parameterVelocities[n] = parameterForces[n] = 0.;
	
	kineticMatrix = new double[nParameters * nParameters];
	forces = 0;
}

ExtendedQuickMinAtomsMover::~ExtendedQuickMinAtomsMover()
{
	delete parameters;
	delete parameterVelocities;
	delete parameterForces;	
	delete kineticMatrix;

	if(forces)
		delete forces;
}

void ExtendedQuickMinAtomsMover::SetInitialParameters(double *initialParameters)
{
	for(int n=0;n<nParameters;++n)
		parameters[n] = initialParameters[n];
}


void ExtendedQuickMinAtomsMover::Allocate(const DMArray<double, DIMENSION> *array)
{
	if(forces && forces->GetShape() == array->GetShape())
		return;
	else
	{
		if(forces)
			delete forces;
		
		forces = new DMArray<double, DIMENSION>(array->GetShape());
	}
}

void ExtendedQuickMinAtomsMover::Move(ListOfAtoms *atoms)
{
    int n;
    double parameterForceSq, atomicEnergy, totalEnergy;
	
	Allocate(atoms->GetCartesianPositions());
    
    atomsExtension->SetParameters(parameters,atoms);  
	int toMinimize = atomsExtension->GetQMBranch(atoms);
    nIterations=0;
	
	int qmN = qm->GetMaxIterations();
	int oldPrec = cout.precision(16);	

	// Now do main part of XQM
	for( n = 0; n < nParameters; ++n )
		parameterVelocities[n] = 0.;

	// initial calculation of forces
	atomicEnergy = atomsExtension->AtomicForcesEnergy(atoms, parameters, potential, forces);
	totalEnergy = atomsExtension->ParameterForcesEnergy(atoms,parameters, parameterForces, kineticMatrix, forces, atomicEnergy, &parameterForceSq);
	convergence = atomsExtension->GetTotalForceSq();

	while(convergence > tolerance && nIterations < maxIterations)    {
		qm->Move(atoms->GetBranch(toMinimize));
	
		
		atomicEnergy = atomsExtension->AtomicForcesEnergy(atoms, parameters, potential, forces);
		totalEnergy = atomsExtension->ParameterForcesEnergy(atoms,parameters, parameterForces, kineticMatrix, forces, atomicEnergy, &parameterForceSq);
		
		for(int bdyIter = 0; bdyIter < nBdyIterations; ++bdyIter )
		{
			atomsExtension->IncrementVelocitiesFromForces(parameterForces, kineticMatrix, parameterVelocities,atoms,dt);
			
			// zero velocities if dot product with forces is negative
			double prmDotProd = 0.;
			for(n=0;n<nParameters;++n)
				prmDotProd += parameterForces[n]*parameterVelocities[n];
			
			if(prmDotProd < 0.)
			{
				for(n=0;n<nParameters;++n)
					parameterVelocities[n] = 0.;
			}
				
			// increment the parameters according to their velocities
			for(n=0;n<nParameters;++n)
				parameters[n] += dt*parameterVelocities[n];
		
			atomsExtension->SetParameters(parameters,atoms);

			// forces on boundary parameters
			atomicEnergy = atomsExtension->AtomicForcesEnergy(atoms, parameters, potential, forces);
			totalEnergy = atomsExtension->ParameterForcesEnergy(atoms,parameters, parameterForces, kineticMatrix, forces, atomicEnergy, &parameterForceSq);
		}
		convergence = atomsExtension->GetTotalForceSq();


		if(verbose)
		{
			cout << "Iteration: " << nIterations << endl;
			cout << "Parameters "; 
			for(n=0;n<nParameters;++n)
				cout << " " << parameters[n];
			cout << endl;

			cout << "qm iterations: " << qm->GetNIterations() << " qm convergence: " << qm->GetConvergence() << " ParameterForceSq: " << parameterForceSq << endl;	
			cout << "Energy: " << totalEnergy << " Total convergence: " << convergence << endl << endl;
			atoms->Notify();
		}
		++nIterations;
    }
	
	cout.precision(oldPrec);
}


const double *ExtendedQuickMinAtomsMover::GetKineticMatrix(ListOfAtoms *atoms, const double *parameters)
{ 
	Allocate(atoms->GetCartesianPositions());
	
	double atomicEnergy = atomsExtension->AtomicForcesEnergy(atoms, parameters, potential, forces);
	double parameterForceSq;

	atomsExtension->ParameterForcesEnergy(atoms, parameters, parameterForces, kineticMatrix, forces, atomicEnergy, &parameterForceSq);
	
	return (const double*) kineticMatrix; 
}


void ExtendedQuickMinAtomsMover::EstimateModeFrequencies(ListOfAtoms *atoms)
{
	// Calculate frequencies via generalized eigenvalue problem 
	// compute hessian for the generalized forces using numerical derivatives 
	char uplo='U', jobz = 'N';
    int matrixSize = nParameters,info, lwork=nParameters*10, itype = 1, index, paramIdx, paramIdx2;
	double *work = new double[lwork];
	double *parameterHessian = new double[matrixSize*matrixSize];
	double *pForcesPlus = new double[nParameters];
	double *pForcesMinus = new double[nParameters];
	double *eValues = new double[matrixSize];
	double epsilon = 0.000001;
	double atomicEnergy, parameterForceSq;
	

	index = 0;
	for(paramIdx=0;paramIdx<nParameters;++paramIdx)
	{
		// 'i'th derivative of the forces
		parameters[paramIdx] += epsilon;
		atomsExtension->SetParameters(parameters,atoms);
		
		atomicEnergy = atomsExtension->AtomicForcesEnergy(atoms, parameters, potential, forces);
		atomsExtension->ParameterForcesEnergy(atoms, parameters, pForcesPlus, kineticMatrix, forces, atomicEnergy, &parameterForceSq);

		// now subtract 2*epsilon to get the original value minus epsilon
		parameters[paramIdx] -= 2*epsilon; 
		atomsExtension->SetParameters(parameters,atoms);
		
		atomicEnergy = atomsExtension->AtomicForcesEnergy(atoms, parameters, potential, forces);
		atomsExtension->ParameterForcesEnergy(atoms, parameters, pForcesMinus, kineticMatrix, forces, atomicEnergy, &parameterForceSq);
		
		
		for( paramIdx2 = 0; paramIdx2 < nParameters; ++paramIdx2 )
		{
			parameterHessian[index++] = -0.5 * (pForcesPlus[paramIdx2] - pForcesMinus[paramIdx2])/epsilon;
		}
		
		// Now add epsilon to get back the original value
		parameters[paramIdx] += epsilon;
		atomsExtension->SetParameters(parameters,atoms);
	}

	atomicEnergy = atomsExtension->AtomicForcesEnergy(atoms, parameters, potential, forces);
	atomsExtension->ParameterForcesEnergy(atoms, parameters, pForcesMinus, kineticMatrix, forces, atomicEnergy, &parameterForceSq);
	
	
	dsygv_(&itype, &jobz, &uplo, &matrixSize, parameterHessian, &matrixSize, kineticMatrix, &matrixSize, eValues, work, &lwork, &info );
	
	cout << "info = " << info << endl;
	
	cout << "Eigenvalues (squares of frequencies)" << endl;
	for(int a=0;a<matrixSize;++a)
		cout << eValues[a] << " ";
	cout << endl;
	
	


	delete pForcesPlus;
	delete pForcesMinus;
	delete parameterHessian;
	delete eValues;



}

/////////////////////// FOR DEBUGGING ////////////////////////////////
/*
void ExtendedQuickMinAtomsMover::CheckEnergyForce(ListOfAtoms *atoms)
{
	int idx;
	double delta;
	double maxDelta = 1.e-6;
	Rand rand;
	double *increments = new double[nParameters];
	double parameterForceSq;

	Allocate(atoms->GetCartesianPositions());

	for(idx = 0; idx < nParameters; ++idx)
		parameters[idx] += 0.0;

	atomsExtension->SetParameters(parameters,atoms);  
	double atomicEnergy = atomsExtension->AtomicForcesEnergy(atoms, parameters, potential, forces);
	double initialEnergy = atomsExtension->ParameterForcesEnergy(atoms, parameters, parameterForces, kineticMatrix, forces, atomicEnergy, &parameterForceSq);	
	cout << "initialEnergy: " << initialEnergy << endl;

	double analyticGradient = 0.;
	for(idx = 0; idx < nParameters; ++idx)
		{
			increments[idx] = rand.uniform();
			if(idx < 2)
				increments[idx] = 0.;
			analyticGradient -= increments[idx] * parameterForces[idx];
		}

	for(delta = 0.1;delta > maxDelta; delta /= 10.)
	{
		for(idx = 0; idx < nParameters; ++idx)
			parameters[idx] += delta * increments[idx];
		
		atomsExtension->SetParameters(parameters,atoms);  
		atomicEnergy = atomsExtension->AtomicForcesEnergy(atoms, parameters, potential, forces);
		double energyPlus = atomsExtension->ParameterForcesEnergy(atoms, parameters, parameterForces, kineticMatrix, forces, atomicEnergy, &parameterForceSq);

		for(idx = 0; idx < nParameters; ++idx)
			parameters[idx] -= 2 * delta * increments[idx];


		atomsExtension->SetParameters(parameters,atoms);  
		atomicEnergy = atomsExtension->AtomicForcesEnergy(atoms, parameters, potential, forces);
		double energyMinus = atomsExtension->ParameterForcesEnergy(atoms, parameters, parameterForces, kineticMatrix, forces, atomicEnergy, &parameterForceSq);

		
		for(idx = 0; idx < nParameters; ++idx)
			parameters[idx] += delta * increments[idx];
		
		double numericGradient = 0.5 * (energyPlus - energyMinus) / delta;
		cout << delta << " " << analyticGradient << " " << numericGradient << endl;
		
	}

	delete increments;
}

*/

////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////


void ExtendedQuickMinAtomsMover::Save(DMWriter *writer) const
{
	AtomsMover::Save(writer);	
	
	writer->PutDouble(dt, "dt");
	writer->PutDouble(convergence, "convergence");
	writer->PutDouble(tolerance, "tolerance");
	writer->PutInt(nIterations, "nIterations");
	writer->PutInt(maxIterations, "maxIterations");
	writer->PutInt(nBdyIterations, "nBdyIterations");
	writer->PutBool(verbose, "verbose");
}

void ExtendedQuickMinAtomsMover::Load(DMReader *reader)
{
	// NOTE: After loading, must set the pointers to the AtomsExtension, QuickMinAtomsMover 
	// and Potential

	AtomsMover::Load(reader);

	SetDt(reader->GetDouble("dt"));
	convergence = reader->GetDouble("convergence");
	SetTolerance(reader->GetDouble("tolerance"));
	nIterations = reader->GetInt("nIterations");
	SetMaxIterations(reader->GetInt("maxIterations"));
	SetBoundaryIterations(reader->GetInt("nBdyIterations"));
	verbose = reader->GetBool("verbose");
}
