// MinimizerAtomsMover.cpp: implementation of the MinimizerAtomsMover class.
//
//////////////////////////////////////////////////////////////////////

#include "DigitalMaterial.h"
#include "MinimizerAtomsMover.h"
#include "BoundaryConditions.h"
#include "NeighborLocator.h"
#include "ExternalWork.h"
#include "SerialBuilder.h"
#include "DMLastHeader.h"


static SerialBuilder<MinimizerAtomsMover> *_Minimizercreator = 
    new SerialBuilder<MinimizerAtomsMover>("MinimizerAtomsMover");

const std::string MinimizerAtomsMover::className = std::string("MinimizerAtomsMover");

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

MinimizerAtomsMover::MinimizerAtomsMover(Minimizer *minimizer, Potential *potential, ExternalWork *work)
{
  this->potential = potential;
  double tolerance = 1.0e-8;
  this->minimizer = minimizer;
  this->work = work;
  this->verbose = false;
}

MinimizerAtomsMover::~MinimizerAtomsMover()
{

}

void MinimizerAtomsMover::Move(ListOfAtoms *atoms)
{
	int nAtoms = atoms->GetNumber();
    int maxIterations = nAtoms*DIMENSION*2;
    
    //minimizer->SetMaxIterations(maxIterations);
    

    double *positions = new double[nAtoms*DIMENSION];
    atomsMinimizable = new AtomsMinimizable(atoms, potential, work, verbose);

    
    for(int i = 0;i< nAtoms;i++)
		for(int coord = 0;coord<DIMENSION;coord++)	
			positions[DIMENSION*i + coord] = atoms->GetCartesianPosition(i,coord);
    minimizer->Minimize(positions,atomsMinimizable);

/*
    try
    {
		minimizer->Minimize(positions,atomsMinimizable);
    }
    catch (ConjugateGradientMinimizer::Error error)
    {
		std::cerr << "Error in ConjugateGradient: " << error.what << std::endl;
		throw;
    }
    
    if(minimizer->GetNIterations() == minimizer->GetMaxIterations())
	std::cerr <<
	    "Reached maximum number of iterations in conjugate gradient" << std::endl;
  

*/
  
    delete atomsMinimizable;
    delete [] positions;
}


// The following functions "mediate" between the CG minimizer function and the
// actual forces and energy functions in the atoms class.


double *MinimizerAtomsMover::AtomsMinimizable::Gradient(double *positions)
{
    int index = 0,coord;
    DMArrayIterator<double, DIMENSION> posIter(atoms->GetCartesianPositions());
    for( posIter.Init(); !posIter.IsDone(); posIter.Next() )
		for( coord = 0; coord < DIMENSION; ++coord)	
		{
			posIter.Current()[coord] = positions[index++];
		}	   
    
    atoms->AdjustPositions();
    bc->EnforceBoundaryConditions(const_cast< DMArray<double,DIMENSION>* > (atoms->GetCartesianPositions()) );
    neighborLocator->UpdatePosition();
    
    potential->CalculateForces(atoms,forces);
    atoms->AdjustForces(forces);

	if(verbose)
	{
		double nrm = forces->Nrm2();
		cout << "MinimizerAtomsMover: Norm of forces = " << nrm << endl;
	}

    index = 0;
    DMArrayIterator<double, DIMENSION> forcesIter(forces);
    for(forcesIter.Init();!forcesIter.IsDone();forcesIter.Next())
		for(coord = 0;coord<DIMENSION;coord++)	
		{
			gradients[index++] = -forcesIter.Current()[coord];
		}
    return gradients;
}


double MinimizerAtomsMover::AtomsMinimizable::ObjectiveFunction(double *positions)
{	
    int index=0,coord;
    DMArrayIterator<double, DIMENSION> posIter(atoms->GetCartesianPositions());
    for(posIter.Init();!posIter.IsDone();posIter.Next())
		for(coord = 0;coord<DIMENSION;coord++)	
		{
			posIter.Current()[coord] = positions[index++];
		}	   

	
    atoms->AdjustPositions();


    bc->EnforceBoundaryConditions(const_cast< DMArray<double,DIMENSION>* > (atoms->GetCartesianPositions()));
    neighborLocator->UpdatePosition();
    
	double value = potential->CalculateEnergy(atoms) - (work ? work->CalculateWork(atoms):0);

	if(verbose)
	{
		cout << value << endl; 
		atoms->Notify();
	}

    return value;
}


////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////


void MinimizerAtomsMover::Save(DMWriter *writer) const
{
	AtomsMover::Save(writer);	
}

void MinimizerAtomsMover::Load(DMReader *reader)
{
	AtomsMover::Load(reader);
}
