// ExtendedNudgedElasticBand.cpp: implementation of the ExtendedNudgedElasticBand class.
//
//////////////////////////////////////////////////////////////////////


#include "DMFirstHeader.h"
#include <fstream>
#include <iostream>
#include "blas.h"
#include "ExtendedNudgedElasticBand.h"
#include "ListOfAtoms.h"
#include "Potential.h"
#include "BoundaryConditions.h"
#include "NeighborLocator.h"
#include "FixedBody.h"
#include "DMException.h"
#include "DMLastHeader.h"

using namespace std;


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


ExtendedNudgedElasticBand::ExtendedNudgedElasticBand(int nReplicas,double dt,double *springConstant, BoundaryConditions *bc,Potential *potential,NeighborLocator *neighborLocator,double tol,AtomsExtension *atomsExtension, bool wantAbsoluteEnergies,bool usingClimbingImage) :
	NudgedElasticBandBarrierFinder(0, dt, springConstant, bc, potential, neighborLocator, tol, 0,  wantAbsoluteEnergies,usingClimbingImage) , repParamFile("")
{
	this->atomsExtension = atomsExtension;
	
	nParameters = atomsExtension->GetNParameters();
	mainList = atomsExtension->GetMainList();
	
	// operational arrays
	parameters = 0;
	parameterVelocities = 0;
	parameterForces = 0;

	nFreeAtoms = 0;
	Allocate(0,nReplicas);

	for( int repIdx =0; repIdx < nReplicas + 1; ++repIdx )
		this->springConstant[repIdx] = springConstant[repIdx];

	gradientArray = 0;
	sizeGradientArray = 0;

	saddleParameters = new double[nParameters];
	for( int prmIdx = 0; prmIdx < nParameters; ++prmIdx )
		saddleParameters[prmIdx] = 0.;

	parameterMetric = new double[nParameters * nParameters];
	SetParameterMetric(0);
}


ExtendedNudgedElasticBand::~ExtendedNudgedElasticBand()
{
	if(nReplicas > 0)
		FreeMemoryNReplicas();

	if(sizeGradientArray)
		delete [] gradientArray;

	delete parameterMetric;
	delete saddleParameters;
}

void ExtendedNudgedElasticBand::Allocate(int arrayLength, int nReplicas)
{
    int oldNReplicas = 0; // default
	bool changeNReplicas = false;

	if(nReplicas != this->nReplicas)
	{
		changeNReplicas = true;
		// need to store this since the base class Allocate will set nReplicas
		oldNReplicas = this->nReplicas;
	}
	NudgedElasticBandBarrierFinder::Allocate(arrayLength, nReplicas);

	if(changeNReplicas)
	{
		if(oldNReplicas > 0)
			FreeMemoryNReplicas();
		if(nReplicas > 0)
		{
			double **parametersMinusOne = new double*[nReplicas + END_POINTS];
			parameters = parametersMinusOne + 1;
			parameterVelocities = new double*[nReplicas];
			parameterForces = new double*[nReplicas];
			kineticMatrix = new double*[nReplicas];

			for( int repIdx = 0; repIdx < nReplicas; ++repIdx )
			{      
				parameters[repIdx] = new double[nParameters];
				parameterVelocities[repIdx] = new double[nParameters];
				parameterForces[repIdx] = new double[nParameters];
				for( int prmIdx = 0; prmIdx < nParameters; ++prmIdx )
					parameters[repIdx][prmIdx] = parameterVelocities[repIdx][prmIdx] = parameterForces[repIdx][prmIdx] = 0.;
				kineticMatrix[repIdx] = new double[nParameters * nParameters];
			}
		}
    }  
}

void ExtendedNudgedElasticBand::FreeMemoryNReplicas()
{
	if(parameters[0])
	{
		for(int repIdx = 0; repIdx < nReplicas; ++repIdx )
		{
			delete parameters[repIdx];
			delete parameterVelocities[repIdx];
			delete parameterForces[repIdx];
			delete kineticMatrix[repIdx];
		}
    }
	delete [] (parameters-1);
	delete [] parameterVelocities;
	delete [] parameterForces;
	delete [] kineticMatrix;
}


void ExtendedNudgedElasticBand::Initialize()
{
	// NEB_ArrayLength will actually be equal to the number of free Atoms plus
	// the number of boundary parameters
	// since the arrays have width DIMENSION, DIMENSION-1 slots
	// for each parameter will be unused and therefore set to zero
	
	nFreeAtoms = atoms1->GetLeaf(mainList)->GetNumber();
	
	if(NEB_ArrayLength != nFreeAtoms + nParameters)
		Allocate(nFreeAtoms + nParameters, nReplicas);

	// create the replicas as copies of atoms1, and set in straight-line path
	// the boundary atoms will be set in literal straight line values
	// which won't be consistent with parameters but they'll be set properly
	// below
	CreateReplicas();

	// I'm putting the cast to avoid warnings---they really should be const
	// but I want pointers to them that look like the pointers to the non-const
	// replica parameters
	parameters[-1] = (double *) parameters1;
	parameters[nReplicas] = (double *) parameters2;

	// do the straight-line thing for the parameters also
	double fractionalSpacing = 1. / (nReplicas + 1);
	double *parameterDifference = new double[nParameters];
	int repIdx, prmIdx;
    for( prmIdx = 0; prmIdx < nParameters; ++prmIdx )
		parameterDifference[prmIdx] = parameters2[prmIdx] - parameters1[prmIdx];
	double *savedParams;
	int repPrmIdx;
	if(initializeFromSaved)
	{
		savedParams = new double[(nReplicas+2) * nParameters];
		ifstream paramFile(repParamFile.c_str());
		repPrmIdx = 0;
		for ( int chainIdx = 0; chainIdx < nReplicas+2; ++chainIdx )
			for( prmIdx = 0; prmIdx < nParameters; ++prmIdx )
				paramFile >> savedParams[repPrmIdx++];
	}
    
	repPrmIdx = nParameters; // skip the first set of params (for atoms1)
	for( repIdx = 0; repIdx < nReplicas; ++repIdx )
    {
		for( prmIdx = 0; prmIdx < nParameters; ++prmIdx )
		{
			if(initializeFromSaved)
				parameters[repIdx][prmIdx] = savedParams[repPrmIdx++];
			else
				parameters[repIdx][prmIdx] = parameters1[prmIdx] + (repIdx+1) * fractionalSpacing * parameterDifference[prmIdx];
			parameterVelocities[repIdx][prmIdx] = 0.;
		}
		atomsExtension->SetParameters(parameters[repIdx], replicas[repIdx]);
	}

    delete [] parameterDifference;
	if(initializeFromSaved)
		delete savedParams;
}

void ExtendedNudgedElasticBand::SetParameterMetric( const double *pMetric)
{
	int prmIdx;
	if(pMetric)
		for ( prmIdx = 0; prmIdx < nParameters * nParameters; ++prmIdx )
			parameterMetric[prmIdx] = pMetric[prmIdx];
	else
	{
		// set to identity matrix
		memset(parameterMetric, 0, nParameters * nParameters * sizeof(double));
		for ( prmIdx = 0; prmIdx < nParameters; ++prmIdx )
			parameterMetric[prmIdx * (nParameters+1)] = 1.;
	}
	
	// check for symmetry
	for ( prmIdx = 0; prmIdx < nParameters; ++prmIdx )
		for ( int prmIdx2 = 0; prmIdx2 < prmIdx; ++prmIdx2 )
		{
			double diff = fabs(parameterMetric[prmIdx * nParameters + prmIdx2] - parameterMetric[prmIdx2 * nParameters + prmIdx]);
			if( diff > 1.0e-8 )
				throw DMException("ExtendedNudgedElasticBand","SetParameterMetric",SprintString("Non-symmetric matrix passed: %d %d %20.16f",prmIdx,prmIdx2,diff));
		}
}



double ExtendedNudgedElasticBand::ReplicaForceEnergy(int chainIdx)
{
	const double *paramsLoc;
	double atomisticEnergy, energy, *parameterForcesLoc, parameterForceSq;
	ListOfAtoms *atomsLoc;
	DMArray<double, DIMENSION> *forcesLoc;

	if( chainIdx <= 0 )
	{ atomsLoc = atoms1; paramsLoc = parameters1; forcesLoc = 0; parameterForcesLoc = 0; }
	else if( chainIdx > nReplicas )
	{ atomsLoc = atoms2; paramsLoc = parameters2; forcesLoc = 0; parameterForcesLoc = 0;  }
	else
	{ atomsLoc = replicas[chainIdx-1]; paramsLoc = parameters[chainIdx-1];
	forcesLoc = forces[chainIdx-1]; ; parameterForcesLoc = parameterForces[chainIdx-1]; }
	
	// AtomicForcesEnergy and ParameterForcesEnergy can deal with a zero 
	// pointer for the forces---they'll just calculate the energy
	atomisticEnergy =  atomsExtension->AtomicForcesEnergy(atomsLoc, paramsLoc, potential, forcesLoc);
	energy = atomsExtension->ParameterForcesEnergy(atomsLoc, paramsLoc, parameterForcesLoc, kineticMatrix[chainIdx-1], forcesLoc, atomisticEnergy, &parameterForceSq);	
	// maybe only apply constraints to the region1 (mainList) atoms. Because we need the region2 forces later.
	// atomsLoc->AdjustForces(forces[repIdx]);   // Apply constraints

	if(!wantAbsoluteEnergies)
		energy -= energy1;

	return energy;
}


void ExtendedNudgedElasticBand::QuickMinStep(ListOfAtoms *repAtoms, DMArray<double, DIMENSION> *repForces, int repIdx)
{
	NudgedElasticBandBarrierFinder::QuickMinStep(repAtoms->GetLeaf(mainList), repForces->GetLeaf(mainList), repIdx);

	// calculate forces and velocities for the parameters
	// do the parameter forces within the main forces function

	atomsExtension->IncrementVelocitiesFromForces(parameterForces[repIdx], kineticMatrix[repIdx], parameterVelocities[repIdx], repAtoms,dt);
	
	int prmIdx;
	double prmDotProd = 0.;
	for(prmIdx = 0; prmIdx < nParameters; ++prmIdx )
		prmDotProd += parameterForces[repIdx][prmIdx] * parameterVelocities[repIdx][prmIdx];
	if(prmDotProd < 0.)
		for(prmIdx = 0; prmIdx < nParameters; ++prmIdx )
			parameterVelocities[repIdx][prmIdx] = 0.;

	
	for(prmIdx = 0; prmIdx < nParameters; ++prmIdx )
		parameters[repIdx][prmIdx] += dt * parameterVelocities[repIdx][prmIdx];


	atomsExtension->SetParameters(parameters[repIdx],repAtoms);
}


/*!

  Uses base class CalculateLinkVector on the "main" sublists, then adds the 
  contribution from the parameters. Also adjusts the length returned by
  the base class function to account for the latter (using the metric)

 */

void ExtendedNudgedElasticBand::CalculateLinkVector( int repIdxSecond, double(*linkVector)[DIMENSION], double &linkLength)
{
	// first do the part that comes from the core atoms, using the base class
	// function
	int atomIdx;
	DMArrayIterator<double, DIMENSION> firstIter(replicas[repIdxSecond-1]->GetLeaf(mainList)->GetCartesianPositions());
	DMArrayIterator<double, DIMENSION> secondIter(replicas[repIdxSecond]->GetLeaf(mainList)->GetCartesianPositions());

	for( atomIdx = 0, firstIter.Init(), secondIter.Init();
		!firstIter.IsDone();
		 ++atomIdx, firstIter.Next(), secondIter.Next() )
	{
		const double *firstPos = firstIter.Current();
		const double *secondPos = secondIter.Current();

		for( int coord = 0; coord < DIMENSION; ++coord )
			linkVector[atomIdx][coord] = secondPos[coord] - firstPos[coord];
	} 
	ASSERT(atomIdx == nFreeAtoms);
	bc->DifferenceBoundaryConditions(linkVector,nFreeAtoms);	

	linkLength = cblas_ddot(atomIdx*DIMENSION,(const double *)linkVector, 1, (const double *)linkVector, 1);


	// set the parameter differences. First zero that part of the array
	// since only the first component will be used
	memset((double *)(linkVector + nFreeAtoms),0,nParameters * DIMENSION * sizeof(double));
	for( int prmIdx = 0; prmIdx < nParameters; ++prmIdx)
	{
		int px = nFreeAtoms + prmIdx;
		linkVector[px][0] = parameters[repIdxSecond][prmIdx] - parameters[repIdxSecond-1][prmIdx];
		// contributions from parameters via metric: sub-diagonal terms*2
		for( int prmIdx2 = 0; prmIdx2 < prmIdx; ++prmIdx2)
			linkLength += 2 * parameterMetric[prmIdx * nParameters + prmIdx2] * linkVector[px][0] * linkVector[nFreeAtoms+prmIdx2][0];
		// diagonal terms
		linkLength += parameterMetric[prmIdx*(nParameters+1)] * linkVector[px][0] * linkVector[px][0];
	}
	linkLength = sqrt(linkLength);
	ASSERT(linkLength >= 0.); // detects nan, for example

}


void ExtendedNudgedElasticBand::GetDotProducts(int repIdx, double (*linkVectorMinus)[DIMENSION], double (*linkVectorPlus)[DIMENSION], double *tangentLength, double *f_UDotTangent, double *linkPlusDotMinus)
{
	// length of tangent vector
	*tangentLength = cblas_ddot(nFreeAtoms*DIMENSION, (const double *)tangent, 1, (const double *)tangent, 1);


	// dot product between tangent and potential forces
	*f_UDotTangent = 0.;
	DMArrayIterator<double, DIMENSION> forceIter(forces[repIdx]->GetLeaf(mainList));
	int atIdx;
	for( atIdx  = 0, forceIter.Init(); !forceIter.IsDone(); forceIter.Next(), ++atIdx )
	{
		double *currentForce = forceIter.Current(); 
		for( int coord = 0; coord < DIMENSION; ++coord )
			*f_UDotTangent += currentForce[coord] * tangent[atIdx][coord];
	}  

	// dot product between consecutive link vectors
	*linkPlusDotMinus = cblas_ddot( nFreeAtoms * DIMENSION, (double*)linkVectorPlus, 1, (double *)linkVectorMinus, 1);

	
	// contributions from parameters
	for( int prmIdx = 0; prmIdx < nParameters; ++prmIdx )
	{
		int px = nFreeAtoms + prmIdx;
		
		// no metric because contracting covariant with contravariant
		*f_UDotTangent += parameterForces[repIdx][prmIdx] * tangent[nFreeAtoms+prmIdx][0];

		for( int prmIdx2 = 0; prmIdx2 < nParameters; ++prmIdx2 )
		{ 
			int px2 = nFreeAtoms + prmIdx2;
			double g12 = parameterMetric[prmIdx*nParameters+prmIdx2];

			*tangentLength += g12 * tangent[px][0] * tangent[px2][0];		
			*linkPlusDotMinus += g12 * linkVectorMinus[px][0] * linkVectorPlus[px2][0];
		}
	}
	// square root to get the actual length
	*tangentLength = sqrt(*tangentLength);	
}


double ExtendedNudgedElasticBand::AddSpringForce(int repIdx, double (*linkVectorMinus)[DIMENSION], double (*linkVectorPlus)[DIMENSION], double linkMinusLength, double linkPlusLength, double f_UDotTangent, double cosPhi)
{
	double repForceSq = 0., g12;
	int atIdx, prmIdx, prmIdx2;
	// meaning of spring methods is as in base class
	if( springMethod == 0 )
	{
		memcpy(springForce, linkVectorPlus, nFreeAtoms*DIMENSION*sizeof(double));
		cblas_dscal(nFreeAtoms*DIMENSION, springConstant[repIdx+1], (double*)springForce, 1);
		cblas_daxpy(nFreeAtoms*DIMENSION, - springConstant[repIdx], (double *)linkVectorMinus, 1, (double *)springForce, 1);
		
		for( prmIdx = 0; prmIdx < nParameters; ++prmIdx)
		{
			int px = nFreeAtoms + prmIdx;
			for( prmIdx2 = 0; prmIdx2 < nParameters; ++prmIdx2)
			{
				int px2 = nFreeAtoms + prmIdx2;
				g12 = parameterMetric[prmIdx*nParameters+prmIdx2];
				springForce[px][0] = g12 * (springConstant[repIdx+1] * linkVectorPlus[px2][0] - springConstant[repIdx] * linkVectorMinus[px2][0]);
			}
		}
	}
	else
	{
		memcpy(springForce, tangent, NEB_ArrayLength*DIMENSION*sizeof(double));
		cblas_dscal(NEB_ArrayLength*DIMENSION, springConstant[repIdx+1] * linkPlusLength - springConstant[repIdx] * linkMinusLength, (double*)springForce, 1);
		double springFactor = springConstant[repIdx+1] * linkPlusLength - springConstant[repIdx] * linkMinusLength; 
		for( prmIdx = 0; prmIdx < nParameters; ++prmIdx)
		{
			int px = nFreeAtoms + prmIdx;
			for( prmIdx2 = 0; prmIdx2 < nParameters; ++prmIdx2)
			{
				int px2 = nFreeAtoms + prmIdx2;
				g12 = parameterMetric[prmIdx*nParameters+prmIdx2];
				springForce[px][0] = g12 * tangent[px2][0] * springFactor;
			}
		}
	}
	// dot product of spring force with tangent
	double f_SDotTangent = cblas_ddot(NEB_ArrayLength*DIMENSION, (double *)springForce, 1 , (double *)tangent,1);

	double unKink = GetUnKinkFactor(cosPhi, repIdx);
	int coord;
	// add the appropriate parts of spring force and subtract the appropriate
	// parts of the potential force
	DMArrayIterator<double, DIMENSION> forceIter(forces[repIdx]->GetLeaf(mainList));
	for( atIdx = 0, forceIter.Init(); !forceIter.IsDone(); forceIter.Next(), ++atIdx )
	{
		double *currentForce = forceIter.Current();
		if( springMethod == 0 )
			for( coord = 0; coord < DIMENSION; ++coord )      
			{
				currentForce[coord] += 
					tangent[atIdx][coord] * (f_SDotTangent - f_UDotTangent )
					+ unKink * (springForce[atIdx][coord] - tangent[atIdx][coord] * f_SDotTangent ); 
				repForceSq += currentForce[coord] * currentForce[coord];
			}
		else
			for( int coord = 0; coord < DIMENSION; ++coord )      
			{
				currentForce[coord] += springForce[atIdx][coord] - f_UDotTangent * tangent[atIdx][coord] ;
				repForceSq += currentForce[coord] * currentForce[coord];
			}
			
	}
	// repeat for parameters
	for (prmIdx = 0; prmIdx < nParameters; ++prmIdx )
	{
		int px = nFreeAtoms + prmIdx;
		for( int prmIdx2 = 0; prmIdx2 < nParameters; ++prmIdx2 )
		{ 
			int px2 = nFreeAtoms + prmIdx2;
			g12 = parameterMetric[prmIdx*nParameters+prmIdx2];

			if( springMethod == 0 )
				parameterForces[repIdx][prmIdx] += g12 * tangent[px2][0] * (f_SDotTangent - f_UDotTangent) + unKink * (springForce[px][0] - g12 * tangent[px2][0] * f_SDotTangent );
			else
				parameterForces[repIdx][prmIdx] +=  springForce[px][0] - g12 * tangent[px][0] * f_UDotTangent;
		}
		repForceSq += parameterForces[repIdx][prmIdx] * parameterForces[repIdx][prmIdx];
	}
	
	return repForceSq;
}


double ExtendedNudgedElasticBand::FlipTangentialForce(int repIdx, double f_UDotTangent)
{
	int atIdx;
	double repForceSq = 0.;
	DMArrayIterator<double, DIMENSION> forceIter(forces[repIdx]->GetLeaf(mainList));
	for( atIdx = 0, forceIter.Init();
		!forceIter.IsDone();
		forceIter.Next(), ++atIdx )
	{
		double *currentForce = forceIter.Current();
		for(int coord=0;coord<DIMENSION;coord++)      
		{
			currentForce[coord] += - 2 * f_UDotTangent * tangent[atIdx][coord];
			repForceSq += currentForce[coord] * currentForce[coord];
		}
	}

	for ( int prmIdx = 0; prmIdx < nParameters; ++prmIdx )
	{
		int px = nFreeAtoms + prmIdx;
		for( int prmIdx2 = 0; prmIdx2 < nParameters; ++prmIdx2 )
		{ 
			int px2 = nFreeAtoms + prmIdx2;
			double g12 = parameterMetric[prmIdx*nParameters+prmIdx2];
			
			parameterForces[repIdx][prmIdx] += -2 * f_UDotTangent * g12 * tangent[px][0];				
		}
		
		repForceSq += parameterForces[repIdx][prmIdx] * parameterForces[repIdx][prmIdx];

	}

	return repForceSq;
}



//////// FUNCTIONS ASSOCIATED WITH BEING A MINIMIZABLE //////////////////

/*!

  SetCurrentPoint does the reverse of SetExternalParameters; it fills the
  given array wih doubles which are compiled from the local parameter values
  and atomic positions; SetExternalParameters takes a list of doubles
  and sets the local parameters and atomic positions accordingly

 */

void ExtendedNudgedElasticBand::SetCurrentPoint(double *currentPoint)
{
	int nGlobParameters = GetNParameters(), repIdx, globParIdx = 0, locParIdx, coord;

 	// the indexing of the parameters is replica by replica
	for( repIdx = 0; repIdx < nReplicas; ++repIdx)
	{
		// for each replica, the first this->nParameters 
		// of the parameters are boundary parameters		
		
		for( locParIdx = 0; locParIdx < nParameters; ++locParIdx )
			currentPoint[globParIdx++] = this->parameters[repIdx][locParIdx];
		
		// rest are from the atoms in the main list
		DMArrayIterator<double,DIMENSION> posIter(replicas[repIdx]->GetLeaf(mainList)->GetCartesianPositions());
		for( posIter.Init(); !posIter.IsDone(); posIter.Next())
		{
			double *current = posIter.Current();
			for( coord = 0; coord < DIMENSION; ++coord)
				currentPoint[globParIdx++] = current[coord]; 
		}
	}
	ASSERT(globParIdx == nGlobParameters);
}


void ExtendedNudgedElasticBand::SetExternalParameters(double *parameters)
{
	int nGlobParameters = GetNParameters(), repIdx, globParIdx = 0, locParIdx, coord;
	DMArray<double, DIMENSION> mainListPos(atoms1->GetLeaf(mainList)->GetShape());
	DMArrayIterator<double,DIMENSION> posIter(&mainListPos);
	
	
	// use parameters to set atom positions
	// the indexing of the parameters is replica by replica
	for( repIdx = 0; repIdx < nReplicas; ++repIdx)
	{
		// for each replica, the first this->nParameters 
		// of the parameters are boundary parameters		
		
		for( locParIdx = 0; locParIdx < nParameters; ++locParIdx )
			this->parameters[repIdx][locParIdx] = parameters[globParIdx++];
	
		// Set boundary positions according to parameters
		atomsExtension->SetParameters(this->parameters[repIdx],replicas[repIdx]);	
	
		// rest is for the atoms in the main list
		for( posIter.Init(); !posIter.IsDone(); posIter.Next())
		{
			double *current = posIter.Current();
			for( coord = 0; coord < DIMENSION; ++coord)
				current[coord] = parameters[globParIdx++]; 
		}
		replicas[repIdx]->GetLeaf(mainList)->SetCartesianPositions(&mainListPos);
	}
	ASSERT(globParIdx == nGlobParameters);
}

double ExtendedNudgedElasticBand::ObjectiveFunction(double *parameters)
{

	SetExternalParameters(parameters);

	// calculate some measure of the energy?
	return 1.0;
}


double *ExtendedNudgedElasticBand::Gradient(double *parameters)
{
	int repIdx, nGlobParameters = GetNParameters();
	if(sizeGradientArray != nGlobParameters)
	{
		if(sizeGradientArray) delete [] gradientArray;
		sizeGradientArray = GetNParameters();
		gradientArray = new double[sizeGradientArray];
	}

	SetExternalParameters(parameters);


	// Calculate the gradient forces and the potential energies:
	for( repIdx = 0; repIdx < nReplicas; ++repIdx )
		barrierPoints[repIdx+1][1] = ReplicaForceEnergy(repIdx + 1);
	
	totalForceSq = CalculateNudging(); // includes adjusts to remove parallel/perpendicular components etc.

	// still need to calculate parameter forces (do in rep loop)

	
	// put forces into final array
	int globParIdx = 0, locParIdx, coord;

	for( repIdx = 0; repIdx < nReplicas; ++repIdx)
	{
		// parameter forces first

		// calculate forces on boundary parameters
		// atomsExtension->CalculateParameterForces(replicas[repIdx], this->parameters[repIdx], parameterForces[repIdx], kineticMatrix[repIdx], forces[repIdx]);	
		
		// transfer to gradientArray
		for( locParIdx = 0; locParIdx < nParameters; ++locParIdx )
			gradientArray[globParIdx++] = -parameterForces[repIdx][locParIdx];
		
		// now do same for mainList atoms
		DMArrayIterator<double,DIMENSION> forceIter(forces[repIdx]->GetLeaf(mainList));		
		for( forceIter.Init(); !forceIter.IsDone(); forceIter.Next())
		{
			double *current = forceIter.Current();
			for( coord = 0; coord < DIMENSION; ++coord)
				gradientArray[globParIdx++] = -current[coord];
		}
	}
	
	ASSERT(globParIdx == nGlobParameters);

	return gradientArray;
}

int ExtendedNudgedElasticBand::GetNParameters()
{
	return nReplicas * (nParameters + DIMENSION * atoms1->GetLeaf(mainList)->GetNumber());
}


ListOfAtoms *ExtendedNudgedElasticBand::GetSaddlePointConfiguration()
{
	ListOfAtoms *saddle = NudgedElasticBandBarrierFinder::GetSaddlePointConfiguration();


	int prmIdx, repIdx; 
	int nInterpPoints = 4; // number of points used for the interpolation
	double dy,value = 0.;
	double **values = new double*[nParameters];
	for( prmIdx = 0; prmIdx < nParameters; ++prmIdx )
		values[prmIdx] = new double[nReplicas+END_POINTS];

	double *interpValues = new double[nParameters];
	double *c = new double[nInterpPoints];
	double *d = new double[nInterpPoints];

	
	for( prmIdx = 0; prmIdx < nParameters; ++prmIdx )
	{
		values[prmIdx][0] = parameters1[prmIdx];
		for( repIdx = 0; repIdx < nReplicas; ++repIdx )
			values[prmIdx][repIdx+1] = parameters[repIdx][prmIdx];

		values[prmIdx][nReplicas+END_POINTS-1] = parameters2[prmIdx];	
	}

	int offset = 0;
	while(barrierPoints[offset][0] < saddleCoord) ++offset;
	offset -= 2; // Want to use two points on each side of saddleCoord
	if(offset < 0) offset = 0;
	if(offset > nReplicas+END_POINTS-nInterpPoints) offset = nReplicas+END_POINTS-nInterpPoints;

	// The following is adapted from Numerical Recipes routine polint:
	for( prmIdx = 0; prmIdx < nParameters; ++prmIdx )
	{
		int i,m,ns=0;
		double den,dif,dift,ho,hp,w;
		
		dif = fabs(saddleCoord-barrierPoints[offset+0][0]);
		
		for ( i = 0; i < nInterpPoints; ++i )
		{
			if ( (dift=fabs(saddleCoord-barrierPoints[offset+i][0])) < dif)
			{
				ns=i;
				dif=dift;
			}
			c[i]=values[prmIdx][i+offset];
			d[i]=values[prmIdx][i+offset];
		}
		
		interpValues[prmIdx]=values[prmIdx][offset+ns--];
		for ( m = 1; m < nInterpPoints; ++m )
		{
			for ( i = 0; i < nInterpPoints - m; ++i) 
			{
				ho=barrierPoints[offset+i][0]-saddleCoord;
				hp=barrierPoints[offset+i+m][0]-saddleCoord;
				w=c[i+1]-d[i];
				if ( (den=ho-hp) == 0.0)
				{
					throw DMException("ExtendedNudgedElasticBand",
									  "GetSaddlePointConfiguration",
									  "Error in SaddleInterpolate");
				}
					den=w/den;
					d[i]=hp*den;
					c[i]=ho*den;
			}
			interpValues[prmIdx] += (dy=(2*ns < (nInterpPoints-m) ? c[ns+1] : d[ns--]));
		}
	}

	for( prmIdx = 0; prmIdx < nParameters; ++prmIdx )
		saddleParameters[prmIdx] = interpValues[prmIdx];

	atomsExtension->SetParameters(saddleParameters, saddle);


	delete d;
	delete c;
	delete interpValues;
	for( prmIdx = 0; prmIdx < nParameters; ++prmIdx )
		delete values[prmIdx];
	delete values;
	


	return saddle;	
}
