// SmartPeriodicBoundaryConditions.cpp: implementation of the SmartPeriodicBoundaryConditions class.
//
//////////////////////////////////////////////////////////////////////
#include "DMFirstHeader.h"

#include <math.h>

#include "DigitalMaterial.h"

#include "SmartPeriodicBoundaryConditions.h"
#include "DMArray.h"
#include "SerialBuilder.h"
#include "DMLastHeader.h"

static SerialBuilder<SmartPeriodicBoundaryConditions> *_SmartPBCcreator = 
    new SerialBuilder<SmartPeriodicBoundaryConditions>
        ("SmartPeriodicBoundaryConditions");

const std::string SmartPeriodicBoundaryConditions::className = std::string("SmartPeriodicBoundaryConditions");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

SmartPeriodicBoundaryConditions::SmartPeriodicBoundaryConditions(const bool wrap[DIMENSION])
{

    for(int coord=0; coord<DIMENSION; ++coord)
    {
        this->wrap[coord] = (wrap ? wrap[coord]: true);
        center[coord] = 0.;
        length[coord] = 1.;
    }
    SetLength(length); // this is to set upper- and lower-Edge
}


SmartPeriodicBoundaryConditions::~SmartPeriodicBoundaryConditions()
{
    
}


void SmartPeriodicBoundaryConditions::SetLength(const double length[DIMENSION])
{
    for(int coord=0; coord<DIMENSION; ++coord)
    {
        this->length[coord] = length[coord];
        lowerEdge[coord] = center[coord] - 0.5*length[coord];
        upperEdge[coord] = center[coord] + 0.5*length[coord];
    }
}


void SmartPeriodicBoundaryConditions::SetBoxCenter(const double center[DIMENSION])
{
    for(int coord=0; coord<DIMENSION; ++coord)
    {
        this->center[coord] = center[coord];
        lowerEdge[coord] = center[coord] - 0.5*length[coord];
        upperEdge[coord] = center[coord] + 0.5*length[coord];
    }
}

void SmartPeriodicBoundaryConditions::SetLowerEdge(const double lowerEdge[DIMENSION])
{
    for(int coord=0; coord<DIMENSION; ++coord)
    {
        this->lowerEdge[coord] = lowerEdge[coord];
        center[coord] = 0.5*(lowerEdge[coord] + upperEdge[coord]);
        length[coord] = upperEdge[coord] - lowerEdge[coord];
    }
}

void SmartPeriodicBoundaryConditions::SetUpperEdge(const double upperEdge[DIMENSION])
{
    for(int coord=0; coord<DIMENSION; ++coord)
    {
        this->upperEdge[coord] = upperEdge[coord];
        center[coord] = 0.5*(lowerEdge[coord] + upperEdge[coord]);
        length[coord] = upperEdge[coord] - lowerEdge[coord];
    }
}


void SmartPeriodicBoundaryConditions::EnforceBoundaryConditions(
                                     DMArray<double, DIMENSION> *positions) const
{
    DMArrayIterator<double, DIMENSION> posIter(positions);
    for(posIter.Init(); !posIter.IsDone(); posIter.Next())
    {
        double *atomPos = posIter.Current();
        for(int coord=0;coord<DIMENSION;coord++)
        {
            if(wrap[coord])
            {
                if(atomPos[coord] >= upperEdge[coord])
                {
                    atomPos[coord] -= length[coord];			// Do this if fmod is more expensive than two "if"s.
                    if (atomPos[coord] >= upperEdge[coord])
                    {
                        atomPos[coord] = lowerEdge[coord]+fmod(atomPos[coord]-lowerEdge[coord], length[coord]);
                    }
                }
                if(atomPos[coord] < lowerEdge[coord])
                {
                    atomPos[coord] += length[coord];			// Do this if fmod is more expensive than two "if"s.
                    if(atomPos[coord] < lowerEdge[coord])
                    {
                        atomPos[coord] = lowerEdge[coord]+fmod(atomPos[coord]-lowerEdge[coord],length[coord])+length[coord];
                    }
                }
            }
        }
    }
}

void SmartPeriodicBoundaryConditions::EnforceBoundaryConditions(double *position) const
{
    for(int coord=0;coord<DIMENSION;coord++)
    {
        if(wrap[coord])
        {
            if(position[coord] >= upperEdge[coord])
            {
                position[coord] -= length[coord];			// Do this if fmod is more expensive than two "if"s.
                if (position[coord] >= upperEdge[coord])
                {
                    position[coord] = lowerEdge[coord]+fmod(position[coord]-lowerEdge[coord], length[coord]);
                }
            }
            if(position[coord] < lowerEdge[coord])
            {
                position[coord] += length[coord];			// Do this if fmod is more expensive than two "if"s.
                if(position[coord] < lowerEdge[coord])
                {
                    position[coord] = lowerEdge[coord]+fmod(position[coord]-lowerEdge[coord],length[coord])+length[coord];
                }
            }
        }
    }
}


void SmartPeriodicBoundaryConditions::DifferenceBoundaryConditions(
                                 DMArray<double, DIMENSION> *displacements) const
{
    DMArrayIterator<double, DIMENSION> dispIter(displacements);	
    for(dispIter.Init(); !dispIter.IsDone(); dispIter.Next())
    {
        double *tmp = dispIter.Current();
        for(int coord=0;coord<DIMENSION;coord++)
            if(wrap[coord])
            {
                if(tmp[coord] >= 0.5*length[coord])
                    tmp[coord] -= length[coord];
                if(tmp[coord] < -0.5*length[coord])
                    tmp[coord] += length[coord]; 
            }
    }
}


void SmartPeriodicBoundaryConditions::DifferenceBoundaryConditions(
                                double (*displacements)[DIMENSION],
                                int nAtoms) const
{
    
    for(int i=0;i<nAtoms;i++)
        for(int coord=0;coord<DIMENSION;coord++)
            if(wrap[coord])
            {
                if(displacements[i][coord] >= 0.5*length[coord])
                    displacements[i][coord] -= length[coord];
                if(displacements[i][coord] < -0.5*length[coord])
                    displacements[i][coord] += length[coord]; 
            }
}


BoundaryConditions *SmartPeriodicBoundaryConditions::Copy() const
{
    SmartPeriodicBoundaryConditions *newPBC = 
            new SmartPeriodicBoundaryConditions(wrap);
    newPBC->SetLength(length);
    newPBC->SetBoxCenter(center);
    return newPBC;
}


void SmartPeriodicBoundaryConditions::Save(DMWriter *writer) const
{
    int intWrap[DIMENSION];
    for(int coord=0; coord<DIMENSION; ++coord) intWrap[coord] = (wrap[coord] ? 1: 0);
    std::vector<int> dim;
    dim.push_back(DIMENSION);
    writer->PutIntArray(intWrap, dim, "wrap");
    
    PeriodicBoundaryConditions::Save(writer); // get length
    
    writer->PutDoubleArray(center, dim, "center");
}

void SmartPeriodicBoundaryConditions::Load(DMReader *reader)
{
    std::vector<int> dim;
    int *intWrap = reader->GetIntArray(dim, "wrap");

    if((dim.size() != 1) || (dim[0] != DIMENSION)) 
        throw DMException("SmartPeriodicBoundaryConditions", "Load",
            "Problem while reading wrap");

    for(int coord=0; coord<DIMENSION; ++coord)
        wrap[coord] = (intWrap[coord] == 1);
  
    delete intWrap;

    PeriodicBoundaryConditions::Load(reader);

    double *data = reader->GetDoubleArray(dim, "center");
    if((dim.size() != 1) || (dim[0] != DIMENSION)) 
        throw DMException("SmartPeriodicBoundaryConditions", "Load",
            "Problem while reading center");
    SetBoxCenter(data);
    delete data;
}
