// BravaisLatticeWithBasis.cpp: implementation of the BravaisLatticeWithBasis class.
//
//////////////////////////////////////////////////////////////////////
#include "DMFirstHeader.h"

#include "DigitalMaterial.h"

#include "BravaisLatticeWithBasis.h"
#include "DMException.h"
#include "DMWriter.h"
#include "DMReader.h"
#include "SerialBuilder.h"
#include "SprintString.h"
#include "DMLastHeader.h"

using namespace std;

static SerialBuilder<GeneralLattice> *_GeneralLatticecreator = 
    new SerialBuilder<GeneralLattice>("GeneralLattice");

const string BravaisLatticeWithBasis::className = string("BravaisLatticeWithBasis");

const string GeneralLattice::className = string("GeneralLattice");


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

BravaisLatticeWithBasis::BravaisLatticeWithBasis()
{
    for(int coord=0; coord<DIMENSION; coord++) origin[coord] = 0.;
	// we temporarily set a cubic lattice
    for(int vec=0; vec<DIMENSION; vec++)
        for(int coord=0; coord<DIMENSION; coord++)
            latticeVectors[vec][coord] = (vec==coord? 1. :0.);

    vector<TinyVector<double,DIMENSION> > atomsCoord;
    atomsCoord.push_back(TinyVector<double, DIMENSION> (0.));
    SetAtomsCoordinatesInCell(atomsCoord);
}

BravaisLatticeWithBasis::~BravaisLatticeWithBasis()
{

}

void BravaisLatticeWithBasis::CartesianCoordinatesFromLatticeCoordinates(
                              double cartesianCoordinates[DIMENSION], 
                              const double latticeCoordinates[DIMENSION]) const
{
    for (int i=0; i<DIMENSION; i++)
    {
        cartesianCoordinates[i] = origin[i];
        for (int j=0; j<DIMENSION; j++)
        {
            cartesianCoordinates[i] += latticeVectors[j][i] * latticeCoordinates[j];
        }
    }
}

void BravaisLatticeWithBasis::LatticeCoordinatesFromCartesianCoordinates(
                              double latticeCoordinates[DIMENSION],
                              const double cartesianCoordinates[DIMENSION]) const
{
    for (int i=0; i<DIMENSION; i++)
    {
        latticeCoordinates[i]=0;
        for (int j=0; j<DIMENSION; j++)
        {
            latticeCoordinates[i] += inverseOfLatticeVectors[j][i] * 
                                    (cartesianCoordinates[j]-origin[j]);
        }
    }
}

void BravaisLatticeWithBasis::ClosestLatticeSiteIndices(
                              int latticeIndices[DIMENSION],
                              const double cartesianCoordinates[DIMENSION]) const
{
    double latticeCoordinates[DIMENSION] ;
    
    this->LatticeCoordinatesFromCartesianCoordinates(latticeCoordinates, cartesianCoordinates) ;
    
    int i ;
    for (i=0; i<DIMENSION; ++i)
    { 
        latticeIndices[i] = (int)floor(latticeCoordinates[i]+0.5) ; // truncate the double
    }
}

void BravaisLatticeWithBasis::ClosestAtomCartesianCoordinates(
                              double atomCartesianCoordinates[DIMENSION],
                              const double position[DIMENSION]) const
{
    int coord;
    double distSq,closestDistanceSq=1.0e99,atomIndices[DIMENSION],atomPosition[DIMENSION];
    // find closest lattice site
    int latticeIndices[DIMENSION];
    this->ClosestLatticeSiteIndices(latticeIndices,position);
    
    const int nAtomsPerCell = GetNAtomsPerCell();

    if(nAtomsPerCell == 1)
    {
        for(coord=0;coord<DIMENSION;++coord)
            atomIndices[coord] = latticeIndices[coord];
        CartesianCoordinatesFromLatticeCoordinates(atomCartesianCoordinates,atomIndices);
        return;
    }
    
    //loop through all neighboring unit cells
    int neighborIndices[DIMENSION],nCells = 1;
    for(coord=0;coord<DIMENSION;++coord)
    { 
        nCells *= 3;
        neighborIndices[coord] = - 1;
    }
    
    for(int cell=0;cell<nCells;++cell)
    {
        // loop through all atoms within each unit cell
        for(int basisAtom=0;basisAtom<nAtomsPerCell;++basisAtom)
        {
            // lattice site is latticeIndices plus neighborIndices
            // atom site is lattice site plus atomsCoordinatesInCell
            for(coord=0;coord<DIMENSION;++coord)
                atomIndices[coord] = latticeIndices[coord] + neighborIndices[coord] 
                                   + atomsCoordinatesInCell[basisAtom][coord];
            CartesianCoordinatesFromLatticeCoordinates(atomPosition, atomIndices);
            // find separation and see if smaller than closestDistanceSq
            distSq = 0.;
            for(coord=0;coord<DIMENSION;++coord)
            {
                double difference = position[coord] - atomPosition[coord]; 
                distSq += difference*difference;
            }
            if(distSq < closestDistanceSq)
            {
                closestDistanceSq = distSq;
                for(coord=0;coord<DIMENSION;++coord)
                    atomCartesianCoordinates[coord] = atomPosition[coord];
            }
        }
        
        ++neighborIndices[0];
        for(coord=0;coord<DIMENSION-1;++coord)
        {		
            if(neighborIndices[coord] > 1) 
            {
                neighborIndices[coord] -= 3;
                ++neighborIndices[coord+1];
            }

        }
    }
    ASSERT(neighborIndices[DIMENSION-1] == 2);
}


void BravaisLatticeWithBasis::RotateLatticeVectors(double rotation[DIMENSION][DIMENSION])
{
    double newLatticeVectors[DIMENSION][DIMENSION];
    if(fabs(Determinant(rotation,DIMENSION)-1.) > 1.0e-11)
        throw DMException("BravaisLatticeWithBasis", "RotateLatticeVectors", "this is not a rotation matrix");
    double test1, test2;   // check that rotation is a rotation matrix (ie det R = +1 , and the norm of its colums and row is 1)
    for(int vec=0; vec<DIMENSION; vec++)
    {
        test1 = test2 = 0.;
        for(int coord=0; coord<DIMENSION; coord++) 
        {
            test1 += rotation[coord][vec]*rotation[coord][vec];
            test2 += rotation[vec][coord]*rotation[vec][coord];
            newLatticeVectors[vec][coord] = 0.;
            for(int k=0; k<DIMENSION;k++) newLatticeVectors[vec][coord] += rotation[coord][k]*latticeVectors[vec][k];
        }
        if((fabs(test1-1.) > 1.0e-10) || (fabs(test2-1.)> 1.0e-10))
            throw DMException("BravaisLatticeWithBasis", "RotateLatticeVectors", "this is not a rotation matrix");
    }
    SetLatticeVectors(newLatticeVectors);
}

void BravaisLatticeWithBasis::ResetOrientation()
{
    assert((DIMENSION == 2) || (DIMENSION == 3)); // currently set only for 2 or 3 D
    int coord;
    double newVectors[DIMENSION][DIMENSION];
    double norms[DIMENSION];
    for(int vec=0; vec<DIMENSION; ++vec)
    {
        TinyVector<double, DIMENSION> tmp(latticeVectors[vec]);
        norms[vec] = sqrt(tmp.norm2());
    }
    newVectors[0][0] = norms[0];
    for(coord=1; coord<DIMENSION; coord++) newVectors[0][coord] = 0.;
    double cosTheta = 0., sinTheta;  // Theta is the angle between v1 and v2
    for(coord=0; coord<DIMENSION; coord++)
        cosTheta += latticeVectors[0][coord] * latticeVectors[1][coord];
    cosTheta /= norms[0]*norms[1];
    sinTheta = sqrt(1.-cosTheta*cosTheta);  // Theta > 0
    newVectors[1][0] = norms[1]*cosTheta;
    newVectors[1][1] = norms[1]*sinTheta;
    if(DIMENSION > 2)
    {
        for(coord=2; coord<DIMENSION; coord++) newVectors[1][coord] = 0.;
        double cosPsi, tmp = 0.; // Psi is the angle between v3 and v1 X v2.
        // inverseOfLatticeVectors[*][2] is orthogonal to v1 and v2 and its norm is 1/|v3|cos Psi
        for(coord=0; coord<DIMENSION; coord++)
            tmp += inverseOfLatticeVectors[coord][2]*inverseOfLatticeVectors[coord][2];
        cosPsi = 1./sqrt(tmp)/norms[2];
        newVectors[2][2] = 1./sqrt(tmp);   // = cosPsi * norms[2]
        double proj[DIMENSION];
        for(coord=0; coord<DIMENSION; coord++) proj[coord] = latticeVectors[2][coord]
            - inverseOfLatticeVectors[coord][2]/tmp;
        newVectors[2][0] = DotProduct(latticeVectors[0], proj, DIMENSION) / norms[0];
        
        tmp = DotProduct(latticeVectors[1], latticeVectors[2], DIMENSION);
        newVectors[2][1] = (tmp - newVectors[1][0]*newVectors[2][0])/newVectors[1][1];
    }
    double oldvolume = volumeOfCell;
    SetLatticeVectors(newVectors);
    ASSERT(fabs(volumeOfCell/oldvolume - 1.) < 1.0e-8);
}


void BravaisLatticeWithBasis::RescaleLatticeVectors(double a)
{
    int vec, coord;
    for(vec=0; vec<DIMENSION; vec++)
    {
        for(coord=0; coord<DIMENSION; coord++) latticeVectors[vec][coord] *= a;
    }
    UpdateMemberVariables();
}


void BravaisLatticeWithBasis::UpdateMemberVariables()
{  // called when we change the latticeVectors. Update Lengths, inverseOfLatticeVectors, volumeOfCell
    volumeOfCell = Determinant(latticeVectors,DIMENSION); 
    ASSERT(volumeOfCell > 0.) ; // we impose a positively oriented basis
    
    // we look for the smallest length of any vector of the form \sum a_i v_i, 
    //			where v_i is a lattice vector and a_i = -1, 0, 1 
    double maxi = 0.,mini=1.E99, miniinside = 1.E99, vect[DIMENSION], norm;
    int i, k, index[DIMENSION];

    const int nAtomsPerCell = GetNAtomsPerCell();
    ASSERT(nAtomsPerCell > 0);
	int threeToTheDIM = 3;
	for(k = 1; k < DIMENSION; ++k) threeToTheDIM *= 3;

    for(i=0;i<threeToTheDIM;i++)
    {
        int tmp = i;
        for(k=0;k<DIMENSION;k++)
        {
            index[k] = tmp%3;
            tmp /= 3;
        }
        norm = 0.;
        for(k=0; k<DIMENSION; k++)
        {
            vect[k] = 0.;
            for(int l=0; l<DIMENSION; l++)
                vect[k] += (index[l]-1)*latticeVectors[l][k];
            norm += vect[k]*vect[k];
        }
        norm = sqrt(norm);
        if(norm>maxi) maxi = norm;
        if(norm<mini && norm != 0.) mini = norm;
        for(int l=0; l<nAtomsPerCell; l++)
            for(int m=0; m<nAtomsPerCell; m++)
            {
                norm = 0;
                for(k=0; k<DIMENSION; k++)
                {
                    vect[k] = 0.;
                    for(int p=0; p<DIMENSION; p++) vect[k] += 
                        (index[p]+atomsCoordinatesInCell[l][p]-atomsCoordinatesInCell[m][p])*latticeVectors[p][k];
                    norm += vect[k]*vect[k];
                }
                norm = sqrt(norm);
                if(norm<miniinside && norm != 0.) miniinside = norm;
            }	
    }
    latticeSpacing = mini;
    maxIntraCellDistance = maxi;
    minInteratomicSpacing = miniinside;
    
    // Inverse. this routine is very basic, but we'll use it a few times and for small matrices
    double tempMat[DIMENSION][DIMENSION];
    for(i=0;i<DIMENSION;i++)
    {
        for(int j=0;j<DIMENSION;j++)
        {
            for(k=0;k<i;k++)
            {
                int l;
                for(l=0;l<j;l++) tempMat[k][l] = latticeVectors[k][l];
                for(l=j+1;l<DIMENSION;l++) tempMat[k][l-1] = latticeVectors[k][l];
            }
            for(k=i+1;k<DIMENSION;k++)
            {
                int l;
                for(l=0;l<j;l++) tempMat[k-1][l] = latticeVectors[k][l];
                for(l=j+1;l<DIMENSION;l++) tempMat[k-1][l-1] = latticeVectors[k][l];
            }
            inverseOfLatticeVectors[j][i] = 
                Determinant(tempMat,DIMENSION-1)/volumeOfCell*( (i+j)%2 == 0? 1.:-1.);
        }
    }
} 


double BravaisLatticeWithBasis::Determinant(const double M[DIMENSION][DIMENSION],const int dim)
{
    // this routine is very basic, but we'll use it a few times and for small matrices (TC)
    if(dim==1)
    {
        return M[0][0];
    }
    else if(dim==2)
    {
        return M[0][0]*M[1][1] - M[1][0]*M[0][1];
    }
    else 
    {
        double temp = 0., matrixtemp[DIMENSION][DIMENSION]; 
        int i, j, k;
        {
            for(i=0; i<dim; i++)
            {
                for(j=1;j<dim;j++)
                {
                    for(k=0;k<i;k++) matrixtemp[j-1][k] = M[j][k];
                    for(k=i+1;k<dim;k++) matrixtemp[j-1][k-1] = M[j][k];
                }
                temp += M[0][i]*Determinant(matrixtemp,dim-1)* (i%2 == 0 ? 1. : -1.); 
            }
            return temp;
        }
    }
}

double BravaisLatticeWithBasis::DotProduct(double const *v1, double const *v2, int n)
{
    double tmp=0.;
    for(int coord=0; coord<n; coord++) tmp += v1[coord] * v2[coord];
    return tmp;
}

void BravaisLatticeWithBasis::SetLatticeVectors(const double latticeVectors[DIMENSION][DIMENSION])
{
    for (int vec = 0; vec < DIMENSION; vec++)
    {
        for (int coord = 0; coord < DIMENSION; coord++)
        {
            this->latticeVectors[vec][coord] = latticeVectors[vec][coord];
        }
    }
    UpdateMemberVariables();
}

void BravaisLatticeWithBasis::SetAtomsCoordinatesInCell(
                              vector<TinyVector<double, DIMENSION> > atomsCoord,
                              vector<std::string> *pAtomTypes)
{
    atomTypes.clear();
    atomsCoordinatesInCell.clear();
    const int nAtomsPerCell = atomsCoord.size();
    for(int at=0; at<nAtomsPerCell; ++at)
    {
        string name = (pAtomTypes ? (*pAtomTypes)[at] : "");
        AddAtomInCell(atomsCoord[at], name);
    }
}

void BravaisLatticeWithBasis::AddAtomInCell(TinyVector<double, DIMENSION> atomCoord,
                                            string name)
{
    // make sure the atoms coordinates are between 0 and 1.
    for(int coord=0; coord<DIMENSION; ++coord)
    {
        atomCoord[coord] = fmod(atomCoord[coord], 1.);
        if(atomCoord[coord] < 0.) atomCoord[coord] += 1.;
    }
    atomsCoordinatesInCell.push_back(atomCoord);
    atomTypes.push_back(name);
    UpdateMemberVariables();
}


void BravaisLatticeWithBasis::Save(DMWriter *writer) const
{
    vector<int> dims(2,DIMENSION);
    writer->PutDoubleArray((double *)latticeVectors, dims, "latVectors");
    dims.pop_back();
    writer->PutDoubleArray(origin, dims, "origin");
    
    const int nAtoms = atomsCoordinatesInCell.size();
    int atom;
    double (*atomCoord)[DIMENSION] = new double[nAtoms][DIMENSION];
    for(atom=0; atom<nAtoms; ++atom) memcpy(atomCoord[atom], 
                atomsCoordinatesInCell[atom], DIMENSION*sizeof(double));
    dims.clear();
    dims.push_back(nAtoms);
    int dimension = DIMENSION;
    dims.push_back(dimension);
    writer->PutDoubleArray((double *)atomCoord, dims, "atomCoord");

    for(atom=0; atom<nAtoms; ++atom)
    {
        SprintString name("atomType%i",atom);
        string tmpStr = atomTypes[atom];
        if(tmpStr == "") tmpStr = " "; // do not save 0 length string.
        writer->PutString(tmpStr, name);
    }
}

void BravaisLatticeWithBasis::Load(DMReader *reader)
{
    vector<int> dims;
    double *tmp = reader->GetDoubleArray(dims, "latVectors");
    if(dims.size() != 2 || dims[1] != DIMENSION || dims[0] != DIMENSION)
        throw DMException("BravaisLatticeWithBasis", "Load", "corrupted lattice vectors");
    memcpy(this->latticeVectors, tmp, DIMENSION*DIMENSION*sizeof(double));
    delete [] tmp;

    tmp = reader->GetDoubleArray(dims, "origin");
    memcpy(this->origin, tmp, DIMENSION*sizeof(double));
    delete [] tmp;

    tmp = reader->GetDoubleArray(dims, "atomCoord");
    if(dims.size() != 2 || dims[1] != DIMENSION)
        throw DMException("BravaisLatticeWithBasis", "Load", "corrupted atomCoordinates");
    const int nAtoms = dims[0];
    atomsCoordinatesInCell.clear();
    atomTypes.clear();
    double (*tmp2)[DIMENSION] = reinterpret_cast<double(*)[DIMENSION]> (tmp);
    for(int atom=0; atom<nAtoms; ++atom)
    {
        atomsCoordinatesInCell.push_back(tmp2[atom]);
        SprintString name("atomType%i", atom);
        string tmpStr = reader->GetString(name);
        if(tmpStr == " ") tmpStr = "";
        atomTypes.push_back(tmpStr);
    }
    delete [] tmp2;
    UpdateMemberVariables();
}


GeneralLattice::GeneralLattice(double a)
{
    double tmpMat[DIMENSION][DIMENSION];
    for(int vec=0; vec<DIMENSION; vec++)
        for(int coord=0; coord<DIMENSION; coord++) tmpMat[vec][coord] = (vec==coord? a :0.);
    SetLatticeVectors(tmpMat);
}

GeneralLattice::GeneralLattice(const double latticeVectors[DIMENSION][DIMENSION])
{
    SetLatticeVectors(latticeVectors);
}


