// BravaisLatticeWithBasis.h: interface for the BravaisLatticeWithBasis class.
//
//////////////////////////////////////////////////////////////////////


#ifndef BRAVAISLATTICEWITHBASIS_H
#define BRAVAISLATTICEWITHBASIS_H

// Unfortunately, the word "basis" is used both for the basis vectors for the crystal lattice
// and for the coordinates of the atoms within the unit cell.
// We're calling the former "LatticeVector" and the latter "AtomCoordinatesInCell"

#include "DMFirstHeader.h"

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string>
#include <vector>

#include "DigitalMaterial.h"
#include "TinyVector.h"
#include "Serializable.h"

class BravaisLatticeWithBasis : public Serializable
{
protected:    // we do not want to instantiate this class. 
    BravaisLatticeWithBasis();
public:
    virtual ~BravaisLatticeWithBasis();
    
    void SetOrigin(const double origin[DIMENSION])
    {for (int coord=0; coord<DIMENSION; coord++) this->origin[coord] = origin[coord];}
#ifndef SWIG

    const double (&GetOrigin() const)[DIMENSION] {return origin;}

    const double (&GetLatticeVectors() const)[DIMENSION][DIMENSION]
    { return latticeVectors; }

    void CartesianCoordinatesFromLatticeCoordinates(
                                   double cartesianCoordinates[DIMENSION],
                                   const double latticeCoordinates[DIMENSION]) const;
 
    void LatticeCoordinatesFromCartesianCoordinates(
                                   double latticeCoordinates[DIMENSION],
                                   const double cartesianCoordinates[DIMENSION]) const;

    void ClosestLatticeSiteIndices(int latticeIndices[DIMENSION],
                                   const double cartesianCoordinates[DIMENSION]) const;
    
    void ClosestAtomCartesianCoordinates(double atomCartesianCoordinates[DIMENSION],
                                         const double position[DIMENSION]) const;
#endif //SWIG

    //! Rotate the current lattice vectors using \a rotation as a rotation matrix.
    /*! Note that we check that \a rotation is really a rotation matrix, namely
    that \a det \a rotation = +1 , and the norm of its colums and rows is 1.
    */
    void RotateLatticeVectors(double rotation[DIMENSION][DIMENSION]);

    //! aligns the 1st lattice vector along the 1st coordinate, the 2nd vector in the 1-2 plane etc...
    virtual void ResetOrientation();     
    
    void RescaleLatticeVectors(double a);
    
    double GetVolumeOfCell() const{ return volumeOfCell;} 
    
    int GetNAtomsPerCell() const { return atomsCoordinatesInCell.size(); }
    
    double GetLatticeSpacing() const { return latticeSpacing;}
    
    double GetMaxIntraCellDistance() const { return maxIntraCellDistance;}
    
    //! the minimal distance separating equivalent atoms
    double GetMinInteratomicSpacing() const { return minInteratomicSpacing;}

#ifndef SWIG

    //! This function is swigged in BravaisLatticeWithBasis.i
    void SetAtomsCoordinatesInCell(
                    std::vector<TinyVector<double, DIMENSION> > atomsCoord,
                    std::vector<std::string> *atomTypes=0);

    virtual void AddAtomInCell(TinyVector<double, DIMENSION> atomCoord,
                               std::string name="");

    const TinyVector<double, DIMENSION> &GetAtomsCoordinatesInCell(int atom) const
    {return atomsCoordinatesInCell[atom];}

    const std::vector<std::string> &GetAtomTypes() const
    {return atomTypes;}

#endif // SWIG
    //! Save the class.
    void Save(DMWriter *writer) const ;

    //! Load the class.
    void Load(DMReader *reader) ;
	
	const std::string &GetType() {return className;}
protected:
#ifndef SWIG
    double origin[DIMENSION];
    
    //! lattice vectors: first index is which vector, second is component
    /*! First index is which vector, second is component
        This is natural in that the basisVector[0] is the first basis vector, but
        is backwards from the point of view of thinking of the array of basis vectors
        as a transformation matrix.00
    */
    double latticeVectors[DIMENSION][DIMENSION];

    // The transpose of the transformation from real space to the lattice coordinate system  
    double inverseOfLatticeVectors[DIMENSION][DIMENSION]; 
    
    double latticeSpacing;  // this is the minimal distance separating two equivallent atoms.
    double minInteratomicSpacing;
    double maxIntraCellDistance;
    
    double volumeOfCell;
//! The coordinates of the atoms in units of the lattice vectors (coordinates must be between 0 and 1)
    std::vector<TinyVector<double, DIMENSION> > atomsCoordinatesInCell;
//! The type name of the atoms in the basis
    std::vector<std::string> atomTypes;
    
    void UpdateMemberVariables();  //called when we change the latticeVectors
    
    double Determinant(const double M[DIMENSION][DIMENSION],const int dim);
    double DotProduct(double const *v1 , double const *v2, int n);
    
    virtual void SetLatticeVectors(const double latticeVectors[DIMENSION][DIMENSION]);
private:
	static const std::string className;


#endif //SWIG
};

class GeneralLattice:public BravaisLatticeWithBasis
{
public:
    void SetLatticeVectors(const double latticeVectors[DIMENSION][DIMENSION])
    {BravaisLatticeWithBasis::SetLatticeVectors(latticeVectors);};
#ifndef SWIG
    GeneralLattice(double a = 1.0);  // by default, a simple cubic lattice of lattice constant a.
    GeneralLattice(const double latticeVectors[DIMENSION][DIMENSION]);
    ~GeneralLattice() {};
#endif //SWIG
	const std::string &GetType() const {return className;}
private:
    static const std::string className;
};

#endif // !defined(BRAVAISLATTICEWITHBASIS_H)

