// AndersonDynamics.h: interface for the AndersonDynamics class.
//
//////////////////////////////////////////////////////////////////////

#ifndef ANDERSONDYNAMICS_H
#define ANDERSONDYNAMICS_H

#include "DigitalMaterial.h"
#include "Constraint.h"
#include "Rand.h"

//! The following constraint can be used to thermalize atoms. The algorithm is described in D. Frenkel, "Understanding Molecular Simulations"
/*!
	\ingroup moleculardynamics
	The idea behind the Andersen dynamics is that some of the atoms undergo random collisions with fictious 
	particles that lead to their thermalization. The occurence of these collisions is given by a Poisson 
	process: the probability that a collision occurs in some interval of time is proportionnal to the length
	of the interval. The proportionnality coefficient (called here \a frequency because it is the frequency of 
	the collisions) is the only parameter. It seems that its actual value is not crucial to get a canonical distribution
	but we shouldn't use this method to measure any dynamical quantity. 

	We can implement the method in the form of a constraint where we adjust the velocities at each time step. So in 
	principle, we can use the constraint only with a Verlet-like algorithm. In fact the important parameter is 
	frequency * dt, so this is the parameter which is passed in the constructor. If we change the time step, we
	should change frequencyDt accordingly.
*/
class AndersenDynamics : public Constraint
{
public: 
	AndersenDynamics(double frequencyDt, double sigma, int seed=1);
	~AndersenDynamics(){}

	void AdjustVelocities(DMArray<double, DIMENSION> *velocities);

	void SetFrequencyDt(double frequencyDt) {this->frequencyDt = frequencyDt;}
	double GetFrequencyDt() const {return frequencyDt;}

	void SetSigma(double sigma) {this->sigma = sigma;}
	double GetSigma() const {return sigma;}


	const std::string &GetType() const
	{ 
		return className; 
	}

private:
	double frequencyDt;
	double sigma;
	Rand rand;
	static const std::string className;
};



#endif // !defined(ANDERSONDYNAMICS_H)
