// NonRotatingRigidBody.h: interface for the NonRotatingRigidBody class.
//
//////////////////////////////////////////////////////////////////////

#ifndef NONROTATINGRIGIDBODY_H
#define NONROTATINGRIGIDBODY_H

#include "DigitalMaterial.h"
#include "ExternalForceConstraint.h"


//! Constrains a ListOfAtoms to move rigidly without rotation

/*!

  The independent degrees of freedom here are the DIMENSION components of the 
  center of mass. No rotation is allowed (a true rigid body would have 6 DOFs
  in 3D, 3 in 2D etc). Using the Adjust functions, this is achieved by
  averaging forces and velocities. Using the GeneralizedCoordinates functions
  you control the center of mass directly and the functions translate changes 
  in the COM into changes in the positions.

  A weight factor may be specified, such that the degrees fo freedom are the
  components of the center of mass divided by the weight---an increment in a
  generalized coordinate moves the center of mass by the increment times the
  weight factor. This is useful when mass-weighted coordinates are needed (eg
  calculating normal mode frequencies)

  NonRotatingRigidBody inherits from ExternalForceConstraint, so you can add 
  an external force when you adjust the forces.
  
  \ingroup MolecularDynamics
  \author Nicholas Bailey
  \date March 1, 2001
  
 */ 


class NonRotatingRigidBody : public ExternalForceConstraint
{
public:
	NonRotatingRigidBody(double weight = 1.0); 
	// eg weight would be 1/sqrt(atoms->GetNumber()) for mass weighted coords
	virtual ~NonRotatingRigidBody() ;
	Constraint *Copy() const;
	void AdjustForces(DMArray<double, DIMENSION> *forces);
	void AdjustForceIncrements(DMArray<double, DIMENSION> *forces);
	void AdjustVelocities(DMArray<double, DIMENSION> *velocities);

	int NumberGeneralizedCoordinates();
	void SetGeneralizedCoordinates(double *generalizedCoordinates) ;
    void GetGeneralizedCoordinates(double *generalizedCoordinates); 
	void IncrementGeneralizedCoordinate(int n,double inc);
#ifndef SWIG
    void CalculateGeneralizedForces(DMArray<double, DIMENSION> *forces, double *generalizedForces);
    void DerivativesWRTGeneralizedCoordinates(int atom, double (*gradient)[DIMENSION]);
#endif // SWIG
	const std::string &GetType() const
	{
		return className;
	}


private:
	void InitializeCOMAndDisplacements();
	void SetCartesianPositionsFromCOM();
	double centerOfMass[DIMENSION];
	bool centerOfMassIsDefined;
	DMArray<double, DIMENSION> *displacements;
	
	double weightFactor;
	static const std::string className;
};


#endif // !defined(NONROTATINGRIGIDBODY_H)
