// TimeDependentConstraint.h: interface for the TimeDependentConstraint class.
//
//////////////////////////////////////////////////////////////////////

#ifndef TIMEDEPENDENTCONSTRAINT_H
#define TIMEDEPENDENTCONSTRAINT_H

#include "DigitalMaterial.h"
#include "Constraint.h"
#include "Functor.h"

//!  zero DOFs, positions relative to an origin whose position is determined by a given function of time

/*
 
  This is like FixedBody. There are no independent degrees of
  freedom. In FixedBody the positions are always set to be those stored in the
  local array which was set up when the constraint was created (usually the
  "initial"  positions of the atoms). Here the local array stores the intial
  positions, but during each call to AdjustPositions an "origin" is added whose
  position is  calculated using a Functor object. A Functor has a function 
  called Evaluate which takes a time and computes a position. Thus the local
  array really stores displacements with respect to the moving origin.

  The weakness of this approach is that it is up to the user to ensure that the
  Constraint is accumulating time correctly---it adds an amount timeStep to 
  its own time variable each time, which the user can set. However it is easy 
  to imagine cases where the time gets incremented more than it should, etc

  \ingroup MolecularDynamics
  \author Nicholas Bailey
  \date March 1, 2001
  
 */ 

class TimeDependentConstraint : public Constraint
{
public:
	TimeDependentConstraint(double initialTime, double timeStep, const DMArray<double, DIMENSION> * positions, Functor *func);
  ~TimeDependentConstraint();
	void Reset(double initialTime, double timeStep,const DMArray<double, DIMENSION> *positions);
  void AdjustPositions(DMArray<double, DIMENSION> *positions);
	Constraint *Copy() const;


	int NumberGeneralizedCoordinates();
	void SetGeneralizedCoordinates(double *generalizedCoordinates) ;
    void GetGeneralizedCoordinates(double *generalizedCoordinates); 
	void IncrementGeneralizedCoordinate(int n,double inc);
#ifndef SWIG
    void CalculateGeneralizedForces(DMArray<double, DIMENSION> *forces, double *generalizedForces);
    void DerivativesWRTGeneralizedCoordinates(int atom, double (*gradient)[DIMENSION]);
#endif // SWIG
	const std::string &GetType() const
	{
		return className;
	}


private:
	double time;
	double timeStep;
	DMArray<double, DIMENSION> *positionsZero;
	Functor *func;
	static const std::string className;
};


#endif // !defined(TIMEDEPENDENTCONSTRAINT_H)
