#ifndef _PROGRAMMABLEDMITERATOR
#define _PROGRAMMABLEDMITERATOR

#include "DMArrayIterator.h"

// ######################################################################
// ProgrammableDMIterator Declaration
// ######################################################################


/*! 
  This is a new kind of iterator designed to iterate over a chosen subset of a
  DMArray. The subset is specified as a list of leaf numbers. The iterator 
  acts like the normal DMArrayIterator---in fact, it derives from it---except
  that it only includes those leaves that have been specified when it was
  created. The leaves to be iterated over are stored in a vector called
  leavesToIterate. An iterator over this keeps track of which leaf we are going
  through.

  At this time it is possible to have the leavesToIterate vector contain a leaf
  index more than once, so a given leaf can be interated over twice. This may
  turn out to be a bug, but conceivably it is a feature--though I can't think 
  of an obvious application. To stop this, we would have to use a set instead 
  of a vector or check in the constructor that no leaf number appears more than
  once. Using a set is a bit risky---I want to be able to guarantee that 
  ProgrammableIterators acting on different arrays of the same shape will 
  access the leaves (and the individual elements) in the same order, even
  though I don't care what that order is. set doesn't guarantee anything about
  order of traversal
  
  \ingroup moleculardynamics
  \author nbailey
  \date November 2001
*/

#ifdef _WIN32
template <class Type>
struct ProgrammableDMIterator : public DMArrayIterator<Type,DIMENSION>
{
public:
	ProgrammableDMIterator(const DMArray<Type, DIMENSION> *dmArray, std::vector<int> leavesToIterate);
#else
template <class Type, int N>
class ProgrammableDMIterator : public DMArrayIterator<Type,N>
{
public:
	ProgrammableDMIterator(const DMArray<Type, N> *dmArray, std::vector<int> leavesToIterate);
#endif
	~ProgrammableDMIterator() {}
	void Init() ;
	void Next();

protected:
	const std::vector<int> leavesToIterate;
	std::vector<int>::const_iterator leavesIter;
};


// #############################################################################
// ITERATOR Definitions
// #############################################################################

	
#ifdef _WIN32
	template<class Type>
	ProgrammableDMIterator<Type>::ProgrammableDMIterator(const DMArray<Type, DIMENSION> *dmArray, std::vector<int> leavesToIterate) : DMArrayIterator<Type,DIMENSION>(dmArray), leavesToIterate(leavesToIterate)
#else
		template<class Type, int N>
	ProgrammableDMIterator<Type, N>::ProgrammableDMIterator(const DMArray<Type, N> *dmArray, std::vector<int> leavesToIterate) : DMArrayIterator<Type,N>(dmArray), leavesToIterate(leavesToIterate)
#endif
{
	numberOfLists = leavesToIterate.size();
	size = 0;
	for(int list = 0; list < numberOfLists; ++list)
		size += dmArray->GetLeaf(leavesToIterate[list])->GetShape().GetSize();

	done = (size == 0 ? true : false);
}

#ifdef _WIN32
	template<class Type>
	void ProgrammableDMIterator<Type>::Init()
#else
	template<class Type, int N>
	void ProgrammableDMIterator<Type, N>::Init()
#endif
{
	leavesIter = leavesToIterate.begin();
	current = ptrs[*leavesIter];
	index.leaf = *leavesIter;
	index.inLeaf = 0;
	currentLimit = sizes[index.leaf];
	if(currentLimit == 0) Next();
	done = (size == 0 ? true : false);
}


#ifdef _WIN32
	template<class Type>
	void ProgrammableDMIterator<Type>::Next()
#else
	template<class Type, int N>
	void ProgrammableDMIterator<Type, N>::Next()
#endif
{
	if(++index.inLeaf < currentLimit)
		++current;
	else
	{
		index.inLeaf = 0;
		do
		{
			++leavesIter;
			if(leavesIter == leavesToIterate.end())
			{
				done = true;
				currentLimit = 1; // to make sure to get out of the while loop
			}
			else 
			{
				index.leaf = *leavesIter;
				current = ptrs[index.leaf];
				currentLimit = sizes[index.leaf];
			}
		}
		while(currentLimit == 0);
	}

}



#endif // ! def PROGRAMMABLEDMITERATOR
