// Shape.cpp: implementation of the Shape class.
//
//////////////////////////////////////////////////////////////////////

#include "DMFirstHeader.h"

#include <assert.h>

#include "DigitalMaterial.h"

#include "Shape.h"
#include "DMException.h"
#include "IndexError.h"
#include "DMWriter.h"
#include "DMReader.h"
#include "SerialBuilder.h"

#include "DMLastHeader.h"

using namespace std;

const string Shape::className = string("Shape");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

const bool Shape::Up   = true;
const bool Shape::Down = false;

static SerialBuilder<Shape> *_Shapecreator = new SerialBuilder<Shape>("Shape");


Shape::Shape(int n) : total(n), leavesSize(1,n)
{
}


Shape::Shape(const Shape &shape):total(shape.total)
{
    treeStruct = shape.treeStruct;
    leavesSize = shape.leavesSize;
}

Shape::Shape(const TreeStructure &treeStruct, const vector<int> &leavesSize)
{
    this->treeStruct = treeStruct;
    this->leavesSize = leavesSize;
    int size = leavesSize.size();
    total = 0;
    for(int i=0; i<size; ++i) total += leavesSize[i];
}

Shape &Shape::operator=(const Shape &shape)
{
    total = shape.total;
    treeStruct = shape.treeStruct;
    leavesSize = shape.leavesSize;
#ifndef NDEBUG
    int test = 0;
    for(int i=0; i<leavesSize.size(); ++i) test += leavesSize[i];
    ASSERT(test == total);  // this is supposed to be true in the passed shape
#endif
    return *this;
}


Shape::~Shape()
{
}

Shape Shape::GetSubShape(int i) const
{
    vector<int>::const_iterator leafSizeIt = leavesSize.begin();
    TreeStructure::const_iterator start = treeStruct.begin();
    int sub = 0;
    while(sub < i)
    {
        int level = 0;
        do
        {
            if(*start == Down) level--;
            else
            {
                level++;
                if(*(start-1) == Down) leafSizeIt++;
            }
            ++start;
        }
        while (level != 0);
        sub++;
        if(start == treeStruct.end())
            throw DMException("Shape", "GetSubShape", "index too large");
    }
    TreeStructure subTree;
    vector<int> subLeafSize;
    start++; // go down one step in order to extract the subShape
    if(*start == Up)
        subLeafSize.push_back(*leafSizeIt); // means the subShape is just a leaf
    else
    {
        int level = 0;
        do
        {
            subTree.push_back(*start);
            if(*start == Down) level--;
            else
            {
                level++;
                if(*(start-1) == Down) subLeafSize.push_back(*leafSizeIt++);
            }
            ++start;
        }
        while (level != 0 || *start == Down);
    }
    return Shape(subTree, subLeafSize);
}

int Shape::GetNSubShapes() const
{
    int nSub = 0;
    TreeStructure::const_iterator it=treeStruct.begin();
    while(it != treeStruct.end())
    {
        int level = 0;
        do level += (*it++ == Down ? -1: 1) ;
        while (level != 0);
        nSub++;
    }
    return nSub;
}	


void Shape::Clear()
{
    treeStruct.clear();
    leavesSize.clear();
    total = 0;
}

void Shape::Merge(const Shape &shape)
{
    if(!IsSimilar(shape)) throw DMException("Shape","Merge","Shapes are not similar.");
    for(unsigned int i=0; i<leavesSize.size(); ++i)
        leavesSize[i] += shape.leavesSize[i];
    total += shape.total;
}

void Shape::Attach(const Shape &shape)
{
    if(treeStruct.size() == 0)
    {
        if(total == 0) leavesSize.clear(); // make this shape a composite
        else throw DMException("Shape","Attach", 
            "Cannot attach a shape because I'm not a composite shape");
    }
    leavesSize.insert(leavesSize.end(), shape.leavesSize.begin(), 
        shape.leavesSize.end());
    treeStruct.push_back(Down);
    treeStruct.insert(treeStruct.end(), shape.treeStruct.begin(), 
        shape.treeStruct.end());
    treeStruct.push_back(Up);	
    total += shape.total;
}


void Shape::Save(DMWriter *writer) const
{
    vector<int> v(1, leavesSize.size());
    writer->PutIntArray(&(leavesSize[0]),v,"LS");
    v[0] = treeStruct.size();
    int *tmp = new int[v[0]];
    for(int i=0; i<v[0]; ++i) tmp[i] = treeStruct[i];
    writer->PutIntArray(tmp, v, "TS");
    delete [] tmp;
}

void Shape::Load(DMReader *reader)
{
    vector<int> dim;
    int *ls = reader->GetIntArray(dim, "LS");
    ASSERT(dim.size() == 1);
    this->leavesSize.clear();
    this->leavesSize.insert(leavesSize.begin(), ls, ls+dim[0]);
    int size = dim[0];
    total = 0;
    for(int i=0; i<size; ++i) total += leavesSize[i];
    
    int *ts = reader->GetIntArray(dim, "TS");
    ASSERT(dim.size() == 1);
    this->treeStruct.clear();
    for(int j=0; j<dim[0]; ++j) treeStruct.push_back((ts[j] == 0? false: true));
    delete [] ls;
    delete [] ts;
}


/*const Shape Shape::Global() const
{
#ifdef DMPARALLEL
    int nLeaves = leavesSize.size();
    int *oldLeavesSize = new int[nLeaves];  // in case vector not contiguous.
    int *newLeavesSize = new int[nLeaves];
    copy(leavesSize.begin(), leavesSize.end(), oldLeavesSize);

    Protocol *protocol = ProtocolFactory::Instance()->New();
    protocol->AllReduce(oldLeavesSize, newLeavesSize, nLeaves, DMProtocol::SUM);
    delete protocol;

    vector<int> vec(newLeavesSize, newLeavesSize + nLeaves);
    delete [] oldLeavesSize;
    delete [] newLeavesSize;
    return Shape(treeStruct, vec);
#else
    return *this;
#endif
}

*/

bool operator==(const Shape &shape1, const Shape &shape2)
{
    return (shape1.treeStruct == shape2.treeStruct) &&
        (shape1.leavesSize == shape2.leavesSize);
}


std::ostream &operator<<(ostream &out, const Shape &shape)
{
    Shape::TreeStructure::const_iterator treeIt = shape.treeStruct.begin();
    if(shape.treeStruct.empty())
    {
        out << shape.total << " " ;
        return out;
    }
    Shape::UpDown previous = Shape::Down;
    vector<int>::const_iterator leafIt = shape.leavesSize.begin();
    while(treeIt != shape.treeStruct.end())
    {
        if(*treeIt == Shape::Up && previous == Shape::Down)
            out << *leafIt++;
        if(*treeIt == Shape::Down && previous == Shape::Up)
            out << ", ";
        if(*treeIt == Shape::Up && previous == Shape::Up)
            out << "] ";
        if(*treeIt == Shape::Down && previous == Shape::Down)
            out << "[ ";
        previous = *treeIt++;
    }
    return out << " ] " ;
}
