// ListOfAtoms.h: interface for the ListOfAtoms class.
//
//////////////////////////////////////////////////////////////////////

#ifndef LISTOFATOMS_H
#define LISTOFATOMS_H

#include "DMFirstHeader.h"
#include <stdio.h>
#include <vector>
#include <string>
#include <set>

#include "DigitalMaterial.h"

#include "Shape.h"
#include "RWLock.h"
#include "Serializable.h"

#include "DMArray.h"
// in order to decouple header files, it's better to declare the classes instead of including them.
class Constraint;
class NeighborLocator;
class NoNeighborLocator;
class BoundaryConditions;
class FreeBoundaryConditions;

class ListOfAtomsObserver;
#include "ObserverList.h"

/*!
 A ListOfAtoms is the container that agregates the data related to the atoms. Very often
 we want to compose ListsOfAtoms in a tree structure in order to represent the various 
 types of atoms and the constraints associated to each type (or group of types). 
 ListOfAtoms defines the generic interface and serves as the base class for the composition: 
 ListOfAtoms can be Merged or a ListOfAtoms can be attached to another (forming a branch 
 of the tree). A ListOfAtoms has a \a positions array (more precisely a 
 DMArray<double,DIMENSION>) and an \a index array the stores the global index of the atoms
 (useful for parallel simulation when one wants to follow a particular atom).

 The \a shape of a ListOfAtoms (and of a DMArray) characterizes its tree structure: 
 it's a tree itself whose leaves are integer numbers representing the number of atoms 
 on each leaf.
  
 \ingroup moleculardynamics
*/

#ifndef SWIG
class ListOfAtoms: public RWLock, public Serializable
{
public:
    //! Create a basic list of atoms with a given shape
    /*! You can pass an integer or nothing in place of \a shape. Passing an integer \a length means
    we want a list of atoms of one type and of size \a length, while passing nothing means 
    the list of atoms can be either a tree or a leaf 
    */	
    ListOfAtoms(const Shape shape = Shape(0),
        const BoundaryConditions *bc=0,
        NeighborLocator *neighborLocator=0);
#else
class ListOfAtoms: public Serializable
{
public:
#endif
    //! Destructor
    virtual ~ListOfAtoms();
    //! Returns the type of the atoms in the list of atoms.
    const std::string &GetTypeName() const {return typeName;}
    //! Set the type name of the atoms in the list.    
    void SetTypeName(const std::string &newName) {typeName = newName;}
    
    //! Set the mass of all the atoms in the list of atoms.
    virtual void SetMass(double mass) {
        throw DMException("ListOfAtoms","SetMass","No mass to set in ListOfAtoms");
    }
    //! Get the mass of all the atoms.
    virtual double GetMass() {
        throw DMException("ListOfAtoms","GetMass","No mass to set in ListOfAtoms");
    }   
    
#ifndef SWIG
    //! Return the positions array
    /*! The Get routines and Set routines that take arrays are much more efficient */
    const DMArray<double, DIMENSION> *GetCartesianPositions() const {return &positions;}
    //! Set the positions array
    void SetCartesianPositions(const DMArray<double, DIMENSION> *newPositions) ;

#endif

    //! Get individual positions. Shouldn't use these in MD! Too slow.
    double GetCartesianPosition(int atomNumber, int coord) const {return positions[atomNumber][coord];}
    //! Set individual positions. Shouldn't use these in MD! Too slow.
    void   SetCartesianPosition(int atomNumber, int coord, double newPosition);
    
    //! equivalent to : positions += scaleFactor * shift
    void IncrementCartesianPositions(const DMArray<double, DIMENSION> *shift, double scaleFactor=1.0);
#ifndef SWIG
    //! Same as previous, but \a scaleFactor depends on the coordinate.
    void IncrementCartesianPositions(const DMArray<double, DIMENSION> *shift, double scaleFactor[DIMENSION]);
#endif
    //! Increment individual position.  Shouldn't use these in MD! Too slow.
    void IncrementCartesianPosition (int atomNumber, int coord, double deltaPosition);
    

    //!  Functions to access and modifiy velocities. They all throw error in ListOfAtoms
    //@{
    virtual void   SetCartesianVelocity(int atomNumber, int coord, double newVelocity);
    virtual double GetCartesianVelocity(int atomNumber, int coord) const;
    virtual double GetKineticEnergy() const;
    
#ifndef SWIG  // We need to reimplement this in the .i file !!!
    virtual const DMArray<double, DIMENSION> *GetCartesianVelocities() const;
    virtual void SetCartesianVelocities(const DMArray<double, DIMENSION> *newVelocities);

    virtual void IncrementCartesianVelocities(const DMArray<double, DIMENSION> *shift,
                                              double scaleFactor=1.0);
    virtual void IncrementCartesianVelocities(const DMArray<double, DIMENSION> *shift,
                                              double scaleFactor[DIMENSION]);
    virtual void IncrementCartesianMomenta   (const DMArray<double, DIMENSION> *forces,
                                              double dt)
    {IncrementCartesianVelocities(forces,dt);} // if mass == 1
#endif //SWIG
    
    virtual void IncrementCartesianVelocity(int atomNumber, int coord, double deltaVelocity);
    
    virtual void ScaleCartesianVelocities(double scaleFactor=1.0);
#ifndef SWIG
    virtual void ScaleCartesianVelocities(double scaleFactor[DIMENSION]);
#endif //SWIG    
    //@}
    
    //! Set the boundary conditions for the current list of atoms and its branches.
    virtual void SetBoundaryConditions(const BoundaryConditions *bc);
    //! Get the boundary conditions.
    const BoundaryConditions *GetBoundaryConditions() const {return this->bc;};
    //! Set the neighbor locator for this list of atoms.
    virtual void SetNeighborLocator(NeighborLocator *neighborLocator);
    //! Return the neighbor locator
    NeighborLocator *GetNeighborLocator() const {return this->neighborLocator;};
    
    //!  Methods related to the application of constraints
    //@{
    virtual void AdjustPositions();
    void AdjustForces(DMArray<double, DIMENSION> *forces);
    void AdjustForceIncrements(DMArray<double, DIMENSION> *forces);
    virtual void AdjustVelocities(){}
    
    int  NumberGeneralizedCoordinates();
    void SetGeneralizedCoordinates(double *generalizedCoordinates);
    void GetGeneralizedCoordinates(double *generalizedCoordinates);
    void IncrementGeneralizedCoordinate(int n,double inc);
    void CalculateGeneralizedForces(DMArray<double, DIMENSION> *forces, double *generalizedForces);
#ifndef SWIG
    void DerivativesWRTGeneralizedCoordinates(int atom, double (*gradient)[DIMENSION]);
#endif
    
    // the constraint is on all atoms of the list
    virtual void SetConstraint(Constraint *constraint);
    Constraint *GetConstraint() {return constraint;}
    //@}    

#ifndef SWIG
    virtual void RemoveAtoms(const std::set<int> &atomNumbers);
#endif //SWIG
    
    //!copy function: will eventually include this as part of serialization
    virtual ListOfAtoms *Copy() const;
    
#ifndef SWIG
    //! Functions associated to the serialization
    //@{
    //! Save this list of atoms to a writer
    virtual void Save(DMWriter *writer) const ;
    //! Read a list of atoms from a reader. (In parallel, only node 0 gets the data).
    virtual void Load(DMReader *reader) ;
#endif
    //! Returns "ListOfAtoms"
    const std::string &GetType() const;
    //@}

    //! Methods related to composition
    //@{
    //! returns the total number of atoms in this list	
    int GetNumber() const { return positions.Size();}
    
    //! returns the shape of the list of atoms.
    /*! This is the shape of the positions array */
    const Shape &GetShape() const {return positions.GetShape(); }
    
    //! Set the shape of the list of atoms (equivalent to Allocate)
    void SetShape(const Shape &shape) {Allocate(shape);}
    
    //! Merge \a atoms into \this. They must have similar shapes.
    /*! Calls MergeData on each leaf. */
    virtual void Merge(const ListOfAtoms *atoms);
    
    //! Attach \a atoms as a new branch of the current tree.
    /*! (\a this must not be a non-empty leaf)  */
    virtual void Attach(ListOfAtoms *atoms);
    
    //! Remove one of our branches
    virtual void RemoveBranch(ListOfAtoms *atomsToRemove);
    
    //! Delete all the branches. \a this becomes an empty leaf.
    virtual void Clear();
    
    //! Returns true if the list of atoms has no branch
    bool IsLeaf() const {return (0 == branches);}
    
    //! Returns the number of branches
    int NumberOfBranches( ) const { return (0 == branches ? 0: branches->size());}
    
    //! Get the i th branch. The validity of the index \a i is checked.
    ListOfAtoms *GetBranch      (int i);
    
#ifndef SWIG
    //! Same as before
    const ListOfAtoms *GetBranch(int i) const;
#endif //SWIG    
    //! Returns the number of leaves
    int NumberOfLeaves() const             {return flatList.size();}
    
    //! Get the i th leaf. The validity of the index \a i is not checked.
    ListOfAtoms *GetLeaf      (int i)      {return flatList[i];}
#ifndef SWIG
    //! Same as before
    const ListOfAtoms *GetLeaf(int i)const {return flatList[i];}
#endif // SWIG    
    
    //! Set the parent list of atoms.
    void SetUpperNode(ListOfAtoms *upperNode) {this->upperNode = upperNode;}
    //@}

    
    //! Functions associated to the Observers.
    //@{
    //!	Add a new observer on this list of atoms
    bool AddObserver(ListOfAtomsObserver *observer);
    //! Remove a particular observer. Returns true if successful
    bool RemoveObserver(ListOfAtomsObserver *observer);
    //! Notifies all the observers that an update is needed.
    void Notify(int notifyLevel = 0);
    //@}
    
protected:
#ifndef SWIG
    //! The global ID of the atoms (such that atoms can be traced among processors).
    DMArray<int> identities;
    //! The positions of the atoms
    DMArray<double, DIMENSION> positions;
    

    //! Member data and methods related to composition
    //@{
    //! The branches, stored in a vector. Its address is 0 for a leaf. 
    std::vector<ListOfAtoms *> *branches;
    //! The list of all the leaves.
    std::vector<ListOfAtoms *>  flatList;
    //! The parent list of atoms
    ListOfAtoms                *upperNode;
    //! Make sure the tree structure is sane all along the tree.
    void UpdateTree();
    //! Set the shape of the current list of atoms 
    virtual void Allocate(const Shape &newShape);
    //! Returns a new ListOfAtoms. Must be overloaded in each derived class.
    virtual ListOfAtoms *LOAAllocator();
    //@}

    //! The type of the atoms in this particular list.
    std::string typeName;
    
    //! The constraint acting on the atoms
    Constraint *constraint;
    
    //! The current boundary conditions
    const BoundaryConditions *bc;
    //! The default boundary conditions
    FreeBoundaryConditions *freeBC;

    //! The current neighbor locator for these atoms
    NeighborLocator *neighborLocator;
    //! The default neighbor locator. Throws error if a routine calls it.
    NoNeighborLocator *noNeighborLocator;  

    //! The list of the observers of this list of atoms    
    ObserverList<ListOfAtomsObserver> observers;

	//! The name of the class, used in Serialization via the GetType function
	static const std::string className;
#endif // SWIG
};


#endif // !defined(LISTOFATOMS_H)
