// BondBreakerNeighborLocator.cpp: implementation of the BondBreakerNeighborLocator class.
//
//////////////////////////////////////////////////////////////////////
#include "DMFirstHeader.h"

#include <math.h>
#include "DigitalMaterial.h"

#include "BondBreakerNeighborLocator.h"
#include "ListOfAtoms.h"
#include "DMArray.h"

#include "DMLastHeader.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

BondBreakerNeighborLocator::BondBreakerNeighborLocator(NeighborLocator *neighborLocator,
		const double tip1Location[DIMENSION], const double tip2Location[DIMENSION], 
		const double tipOrientation[DIMENSION], bool stripe): NeighborLocator(0.)	
{
	this->stripe = stripe;
	this->neighborLocator = neighborLocator;
	memcpy(this->tip1Location, tip1Location, DIMENSION*sizeof(double));
	memcpy(this->tip2Location, tip2Location, DIMENSION*sizeof(double));
	memcpy(this->tipOrientation, tipOrientation, DIMENSION*sizeof(double));
	int coord;
	double tip1tip2[DIMENSION];
	for(coord=0; coord<DIMENSION; ++coord) tip1tip2[coord] = tip2Location[coord] - tip1Location[coord];
// first determine the plane orientation:
	if(DIMENSION==2)
	{
		planeOrientation[0] = -tip1tip2[1];
		planeOrientation[1] =  tip1tip2[0];
		cutOrientation[0] = tip1tip2[0];
		cutOrientation[1] = tip1tip2[1];
	}
	else
	{
		planeOrientation[0] = tip1tip2[1]*tipOrientation[2] - tip1tip2[2]*tipOrientation[1];
		planeOrientation[1] = tip1tip2[2]*tipOrientation[0] - tip1tip2[0]*tipOrientation[2];
		planeOrientation[2] = tip1tip2[0]*tipOrientation[1] - tip1tip2[1]*tipOrientation[0];
		cutOrientation[0] = planeOrientation[2]*tipOrientation[1] - planeOrientation[1]*tipOrientation[2];
		cutOrientation[1] = planeOrientation[0]*tipOrientation[2] - planeOrientation[2]*tipOrientation[0];
		cutOrientation[2] = planeOrientation[1]*tipOrientation[0] - planeOrientation[0]*tipOrientation[1];
	}
	
	// normalize cutOrientation and planeOrientation

	double normCutOrient = 0, normPlaneOrient = 0;

	for(coord=0; coord<DIMENSION; ++coord)
	{
		normCutOrient   += cutOrientation[coord] * cutOrientation[coord];
		normPlaneOrient += planeOrientation[coord] * planeOrientation[coord];
	}
	normCutOrient = sqrt(normCutOrient);
	normPlaneOrient = sqrt(normPlaneOrient);
	double scalProd = 0;
	for(coord=0; coord<DIMENSION; ++coord)
		this->cutOrientation[coord] = cutOrientation[coord]/normCutOrient;
		
	if(DIMENSION==2)
	{
		this->planeOrientation[0] = - this->cutOrientation[1];
		this->planeOrientation[1] =   this->cutOrientation[0];
	}
	else 
	{
		for(coord=0; coord<DIMENSION; ++coord)
		{
			this->planeOrientation[coord] = (DIMENSION==2 ? 0. : planeOrientation[coord]/normPlaneOrient);
			scalProd += this->cutOrientation[coord] * this->planeOrientation[coord];
		}
		if(fabs(scalProd) > 1.0e-10) throw DMException("BondBreakerNeighborLocator", 
			"BondBreakerNeighborLocator", "cutOrientation and planeOrientation should be perpendicular!");
	}
}


/*BondBreakerNeighborLocator::BondBreakerNeighborLocator(NeighborLocator *neighborLocator,
		const double tipLocation[DIMENSION], const double cutOrientation[DIMENSION],
		const double planeOrientation[DIMENSION]): NeighborLocator(0.)
	
{
	stripe = false;
	this->neighborLocator = neighborLocator;
	memcpy(this->tip1Location, tipLocation, DIMENSION*sizeof(double));
// normalize cutOrientation and planeOrientation
	int coord;
	double normCutOrient = 0, normPlaneOrient = 0;
	for(coord=0; coord<DIMENSION; ++coord)
	{
		normCutOrient   += cutOrientation[coord] * cutOrientation[coord];
		normPlaneOrient += planeOrientation[coord] * planeOrientation[coord];
	}
	normCutOrient = sqrt(normCutOrient);
	normPlaneOrient = sqrt(normPlaneOrient);
	double scalProd = 0;
	for(coord=0; coord<DIMENSION; ++coord)
		this->cutOrientation[coord] = cutOrientation[coord]/normCutOrient;
		
	if(DIMENSION==2)
	{
		this->planeOrientation[0] = - this->cutOrientation[1];
		this->planeOrientation[1] =   this->cutOrientation[0];
	}
	else 
	{
		for(coord=0; coord<DIMENSION; ++coord)
		{
			this->planeOrientation[coord] = (DIMENSION==2 ? 0. : planeOrientation[coord]/normPlaneOrient);
			scalProd += this->cutOrientation[coord] * this->planeOrientation[coord];
		}
		if(fabs(scalProd) > 1.0e-10) throw DMException("BondBreakerNeighborLocator", 
			"BondBreakerNeighborLocator", "cutOrientation and planeOrientation should be perpendicular!");
	}
}
*/
BondBreakerNeighborLocator::~BondBreakerNeighborLocator()
{
	delete this->neighborLocator;
}

void BondBreakerNeighborLocator::SetPositionArray(const DMArray<double, DIMENSION> *positions)
{
  	neighborLocator->SetPositionArray(positions);
	NeighborLocator::SetPositionArray(positions);
}


bool BondBreakerNeighborLocator::Neighbors(int atomNumber, int* neighbors, int& numNeighbors, double (*displacements)[DIMENSION],
			double* displacementsSquared, int& spacesLeftInArrays)
{
	bool result = neighborLocator->Neighbors(atomNumber,neighbors,numNeighbors,displacements,
			displacementsSquared, spacesLeftInArrays);
	if(result)
	{
		FindCutBonds((*positions)[atomNumber], neighbors, numNeighbors,
			displacements, displacementsSquared, spacesLeftInArrays, neighborLocator->GetCutoffDistance());
		return true;
	}
	else return false;
}

bool BondBreakerNeighborLocator::HalfNeighbors(int atomNumber, int* neighbors, int& numNeighbors, double (*displacements)[DIMENSION],
			double* displacementsSquared, int& spacesLeftInArrays)
{
	bool result = neighborLocator->HalfNeighbors(atomNumber,neighbors,numNeighbors,displacements,
			displacementsSquared, spacesLeftInArrays);
	if(result)
	{
		FindCutBonds((*positions)[atomNumber], neighbors, numNeighbors,
			displacements, displacementsSquared, spacesLeftInArrays, neighborLocator->GetCutoffDistance());
		return true;
	}
	else return false;
}

bool BondBreakerNeighborLocator::Neighbors(const double *center, int* neighbors, int& numNeighbors, double (*displacements)[DIMENSION],	
	double* displacementsSquared, int& spacesLeftInArrays, double cutoff)
{
	bool result = neighborLocator->Neighbors(center,neighbors,numNeighbors,displacements,
			displacementsSquared, spacesLeftInArrays, cutoff);
	if(result)
	{
		FindCutBonds(center, neighbors, numNeighbors,
			displacements, displacementsSquared, spacesLeftInArrays, cutoff);
		return true;
	}
	else return false;
}

bool BondBreakerNeighborLocator::Neighbors(int atomNumber, int type1, int type2, int* neighbors, int& numNeighbors, 
			double (*displacements)[DIMENSION],	double* displacementsSquared, int& spacesLeftInArrays)
{
	bool result = neighborLocator->Neighbors(atomNumber, type1, type2, neighbors,numNeighbors,
		displacements, displacementsSquared, spacesLeftInArrays);
	if(result)
	{
		FindCutBonds((*positions)[atomNumber], neighbors, numNeighbors,
			displacements, displacementsSquared, spacesLeftInArrays, neighborLocator->GetCutoffDistance());
		return true;
	}
	else return false;

}

void BondBreakerNeighborLocator::SetBoundaryConditions(const BoundaryConditions *bc)
{
	this->bc=bc;
	neighborLocator->SetBoundaryConditions(bc);
}

NeighborLocator *BondBreakerNeighborLocator::Copy() const
{
	return new BondBreakerNeighborLocator(neighborLocator->Copy(), tip1Location, tip2Location, tipOrientation, stripe);
}


void BondBreakerNeighborLocator::FindCutBonds(const double *center, int *neighbors, int &numNeighbors, double (*displacements)[DIMENSION],
		double *displacementsSquared, int &slotsLeft, double cutoff)
{
	int coord;
// first check the distance from the center to the cutting plane is smaller than the cutoff distance.
	double distToPlane = 0, tip1ToCenter[DIMENSION];
	for(coord=0; coord<DIMENSION; ++coord)
	{
		tip1ToCenter[coord] = center[coord] - tip1Location[coord];
		distToPlane += tip1ToCenter[coord] * planeOrientation[coord];
	}
	if(fabs(distToPlane) > cutoff) return; // no need to check further
// second, check that the distance from the atom to the tip is smaller than the cutoff distance.
	double scalProd1 = 0, scalProd2 = 0;
	for(coord=0; coord<DIMENSION; ++coord) scalProd1 += tip1ToCenter[coord] * cutOrientation[coord];
	if(scalProd1 < - cutoff) return; // no need to check further
	if(stripe)
	{
		for(coord=0; coord<DIMENSION; ++coord) scalProd2 += (center[coord] - tip2Location[coord]) * cutOrientation[coord];
		if(scalProd2 > cutoff) return; // no need to check further
	}

	// now it makes sense to loop over all the bonds
	double tip1ToCenter2D[2], tip2ToCenter2D[2];
	tip1ToCenter2D[0] = scalProd1;
	tip1ToCenter2D[1] = distToPlane;
	if(stripe)
	{	
		tip2ToCenter2D[0] = scalProd2;
		tip2ToCenter2D[1] = distToPlane;
	}
	for(int nb=numNeighbors-1; nb>=0; --nb)
	{
		double bond[2] = {0.,0.};
		for(coord=0; coord<DIMENSION; ++coord)
		{
			bond[0] += displacements[nb][coord]*cutOrientation[coord];
			bond[1] += displacements[nb][coord]*planeOrientation[coord];
		}
		if( ((bond[1]+tip1ToCenter2D[1])*tip1ToCenter2D[1] < 0.0) &&
			(tip1ToCenter2D[0] > tip1ToCenter2D[1]*bond[0]/bond[1]) )   // here is a broken bond.
			if(!stripe || ( ((bond[1]+tip2ToCenter2D[1])*tip2ToCenter2D[1] < 0.0) &&
			(tip2ToCenter2D[0] < tip2ToCenter2D[1]*bond[0]/bond[1])))
// remove this bond. (ie bring the last in the list in its place).
			{
				--numNeighbors;
				++slotsLeft;
				if(nb < numNeighbors)
				{
					neighbors[nb] = neighbors[numNeighbors];
					memcpy(displacements[nb], displacements[numNeighbors], DIMENSION*sizeof(double));
					displacementsSquared[nb] = displacementsSquared[numNeighbors];
				}
			}
	}
}
