// CellNeighborList.h: interface for the CellNeighborList class.
//
//////////////////////////////////////////////////////////////////////

#ifndef CELLNEIGHBORLIST_H
#define CELLNEIGHBORLIST_H

#include "DMFirstHeader.h"
#include <vector>

#include "DigitalMaterial.h"
#include "CellNeighborLocator.h"

//! A CellNeighborList should be used if you want to store a full neighbor list 
/*!
	This is of course the most efficient way to obtain the neighbors, but it's
	also in general the most memory consuming class.
	
	This class uses 2 features of \a CellNeighborLocator, from which it derives:
	1. To build the neighbor list, it first builds a cell list.
	2. If we ask for the atoms near a given center, it uses CellNeighborLocator's
	version.

  /sa CellNeighborLocator, NeighborLocator
*/

class CellNeighborList : public CellNeighborLocator  
{
public:
	//! Constructor
	CellNeighborList(double cutoffDistance = 1., double drift = 1.);
	//! Destructor
	virtual ~CellNeighborList();
	// Returns the number of time the full neighbor list have been rebuilt.
	int GetTimesRebuilt() {return timesRebuilt;}
#ifndef SWIG
	virtual bool HalfNeighbors(int atomNumber,
                               int *neighbors,
                               int& numNeighbors,
                               double (*displacements)[DIMENSION],
                               double *displacementsSquared,
                               int& spacesLeftInArrays); 

	virtual bool Neighbors(int atomNumber, 
                           int *neighbors,
                           int& numNeighbors,
                           double (*displacements)[DIMENSION],
                           double *displacementsSquared,
                           int& spacesLeftInArrays);

	virtual bool Neighbors(int atomNumber,
                           int type1,
                           int type2,
                           int* neighbors,
                           int& numNeighbors,
                           double (*displacements)[DIMENSION],
                           double* displacementsSquared,
                           int& spacesLeftInArrays);

	virtual bool HalfNeighbors(int atomNumber,
                               int type1,
                               int type2,
                               int* neighbors,
                               int& numNeighbors,
                               double (*displacements)[DIMENSION],
                               double* displacementsSquared,
                               int& spacesLeftInArrays);


	NeighborLocator *Copy() const;

	const std::string &GetType() const
	{
		return className;
	}

	//! A struct used to translate between the full neighbor list and the typed neighbor list
	/*! We do not want to store both the regular neighbor list and the typed neighbor list. 
	One solution could be to store only the typed one and we may want to do that if we decide
	to reduce the interface of NeighborLocators to deal only with Index (instead of int).
	For now, we store the regular neighbor list and the typed neighbor list is just a set 
	of pointers to the elements of the regular list (which is sorted).
	*/
	struct TypedNBListElement
	{
		int *begin;
		int size;
	};
protected:
	//! UpdateMyData calls CellNeighborLocator's UpdateMyData and rebuilds the neighborList
	virtual void UpdateMyData();

private:
	//! The number of times the full neighbor list has been rebuilt
	int timesRebuilt;
	//! Guess what it does.
	void MakeNeighborList();
	//! The neighbor list is stored as an array (one for each atom) of vectors 
	std::vector<int> *neighborList;
	//! We make it only if somebody calls Neigbors(atomsNumber, type1, type2, ...)
	void MakeTypedNeighborList();


	int nTypes;
	std::vector<int> partialSum;
	//! Each atom has \a numberOfTypes different TypedNBListElements.
	DMArray<TypedNBListElement *> *typedNeighborList;
	//! Tells whether or not we have built the typedNeighborList
	bool typedNBListBuilt;

	int *numberOfHalfNeighbors;
//! An estimation of the number of neighbors (for efficiency, it's better to allocate STL vectors of the right size) 
	const int maxNeighbors;
	static const std::string className;
#endif //SWIG
};

#endif // CELLNEIGHBORLIST_H
