// CompositeNeighborLocator.cpp: implementation of the CompositeNeighborLocator class.
//
//////////////////////////////////////////////////////////////////////
#include "DigitalMaterial.h"

#include "CompositeNeighborLocator.h"
#include "ListOfAtoms.h"
#include "DMException.h"

#include "DMLastHeader.h"

using namespace std;

const string CompositeNeighborLocator::className = string("CompositeNeighborLocator");

const string SimpleCompositeNeighborLocator::className = string("SimpleCompositeNeighborLocator");

const string CompositeNeighborList::className = string("CompositeNeighborList");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CompositeNeighborLocator::CompositeNeighborLocator(double cutoffDistance, const ListOfAtoms *atoms)
	: NeighborLocator(cutoffDistance)
{
	this->atoms = atoms;
	positions = atoms->GetCartesianPositions();
	SetChildren();
}

CompositeNeighborLocator::~CompositeNeighborLocator()
{

}

void CompositeNeighborLocator::SetPositionArray(const DMArray<double, DIMENSION> *positions)
{
	if(positions != this->positions) 
		throw DMException("CompositeNeighborLocator","SetPositionArray",
			"you cannot change the position array in a compositeNeighborLocators");
}

void CompositeNeighborLocator::SetChildren()
{
	children.clear();
	for(int subList=0; subList<atoms->NumberOfBranches(); ++subList)
	{
		const ListOfAtoms *subListOfAtoms = atoms->GetBranch(subList);
		if(atoms != subListOfAtoms) children.push_back(subListOfAtoms->GetNeighborLocator());
	}
}

double CompositeNeighborLocator::GetDrift() const
{
    const int nChildren = children.size();
    if(nChildren == 0) return 0.;
    double drift = children[0]->GetDrift();
    for(int i=0; i<nChildren; ++i)
    {
        double tmp = children[i]->GetDrift();
        if(tmp < drift) drift = tmp;
    }
    return drift;
}

// **************************************************************

SimpleCompositeNeighborLocator::SimpleCompositeNeighborLocator(double cutoffDistance, const ListOfAtoms *atoms):
	CompositeNeighborLocator(cutoffDistance, atoms)
{	
	UpdatePosition();
}

SimpleCompositeNeighborLocator::~SimpleCompositeNeighborLocator()
{
}

void SimpleCompositeNeighborLocator::UpdatePosition()
{
	if(children.size() != atoms->NumberOfBranches())
	{
		SetChildren();
	}
	for(std::vector<NeighborLocator *>::iterator it=children.begin(); it != children.end(); ++it)
		(*it)->UpdatePosition();
	bc = atoms->GetBoundaryConditions();
}

bool SimpleCompositeNeighborLocator::HalfNeighbors(int atomNumber,	int* neighbors, 
			int& numNeighbors, double (*displacements)[DIMENSION], double* displacementsSquared, 
		       int& spacesLeftInArrays)
{
	int type=0;
	int i, index2;
	bool append=true;
	numNeighbors = 0;
	double *center = (*(atoms->GetCartesianPositions()))[atomNumber];
	// 1. find which type is atom atomNumber
	while(atomNumber >= atoms->GetBranch(type)->GetNumber()) 
	{
		atomNumber -= atoms->GetBranch(type)->GetNumber();
		type++;
	}
	// 2. For all types with smaller index, call children[type1]->neighbors(center)
	int numNbrsTmp;
	index2 = 0;
	for(int type2=0; append && (type2<type); ++type2)
	{
		append &= children[type2]->Neighbors(center, neighbors+numNeighbors, numNbrsTmp,
			displacements+numNeighbors, displacementsSquared+numNeighbors, spacesLeftInArrays, cutoffDistance);
		for(i=numNeighbors; i<numNeighbors+numNbrsTmp;++i)neighbors[i] += index2;
		index2 += atoms->GetBranch(type2)->GetNumber();
		numNeighbors += numNbrsTmp;
	}
	if(!append) return false;
	// 3. call children[type]->HalfNeighbors
	append &= children[type]->HalfNeighbors(atomNumber, neighbors+numNeighbors, numNbrsTmp,
			displacements+numNeighbors, displacementsSquared+numNeighbors, spacesLeftInArrays);
	for(i=numNeighbors; i<numNeighbors+numNbrsTmp;++i)neighbors[i] += index2;
	numNeighbors += numNbrsTmp;
	return append;
}


bool SimpleCompositeNeighborLocator::Neighbors(int atomNumber, int* neighbors, int& numNeighbors,
		       double (*displacements)[DIMENSION], double* displacementsSquared, 
		       int& spacesLeftInArrays)
{
	int type=0;
	bool append=true;
	numNeighbors = 0;
	double *center = (*(atoms->GetCartesianPositions()))[atomNumber];
	// 1. find which type is atom atomNumber
	while(atomNumber >= atoms->GetBranch(type)->GetNumber()) 
	{
		atomNumber -= atoms->GetBranch(type)->GetNumber();
		type++;
	}
	int i, numNbrsTmp;
	int index2 = 0;
	for(int type2=0; append && (type2<atoms->NumberOfBranches()); ++type2)
	{
		if(type2 != type)
			append &= children[type2]->Neighbors(center, neighbors+numNeighbors, numNbrsTmp,
			displacements+numNeighbors, displacementsSquared+numNeighbors, spacesLeftInArrays, cutoffDistance);
		else 
			append &= children[type]->Neighbors(atomNumber, neighbors+numNeighbors, numNbrsTmp,
			displacements+numNeighbors, displacementsSquared+numNeighbors, spacesLeftInArrays);
		for(i=numNeighbors; i<numNeighbors+numNbrsTmp;++i)neighbors[i] += index2;
		index2 += atoms->GetBranch(type2)->GetNumber();
		numNeighbors += numNbrsTmp;
	}
	return append;
}

bool SimpleCompositeNeighborLocator::Neighbors(const double *center, int* neighbors, int& numNeighbors,
				double (*displacements)[DIMENSION],	double* displacementsSquared,
				int& spacesLeftInArrays, double cutoff)
{
	if(cutoff<0.) cutoff = cutoffDistance;
	bool append=true;
	numNeighbors = 0;
	int i, numNbrsTmp;
	int index2 = 0;
	for(int type=0; append && (type<atoms->NumberOfBranches()); ++type)
	{
		append &= children[type]->Neighbors(center, neighbors+numNeighbors, numNbrsTmp,
			displacements+numNeighbors, displacementsSquared+numNeighbors, spacesLeftInArrays, cutoff);
		for(i=numNeighbors; i<numNeighbors+numNbrsTmp;++i)neighbors[i] += index2;
		index2 += atoms->GetBranch(type)->GetNumber();
		numNeighbors += numNbrsTmp;
	}
	return append;
}

bool SimpleCompositeNeighborLocator::Neighbors(int atomNumber, int type1, int type2,
				int* neighbors, int& numNeighbors, double (*displacements)[DIMENSION],
				double* displacementsSquared, int& spacesLeftInArrays)
{
	bool append=true;
	numNeighbors = 0;
	if(type1 == type2)
		append &= children[type1]->Neighbors(atomNumber, neighbors, numNeighbors,
			displacements, displacementsSquared, spacesLeftInArrays);
	else 
	{
		double *center = (*(atoms->GetBranch(type1)->GetCartesianPositions()))[atomNumber];
		append &= children[type2]->Neighbors(center, neighbors, numNeighbors,
			displacements, displacementsSquared, spacesLeftInArrays, cutoffDistance);
	}
	int index2 = 0;
	for(int type=0; type<type2; ++type) index2 += atoms->GetBranch(type)->GetNumber();
	for(int i=0; i<numNeighbors;++i)neighbors[i] += index2;

	return append;
}

NeighborLocator *SimpleCompositeNeighborLocator::Copy() const
{
  SimpleCompositeNeighborLocator *newNeighborLocator = 
	  new SimpleCompositeNeighborLocator(cutoffDistance);
  return newNeighborLocator;
}


// **************************************************************

CompositeNeighborList::CompositeNeighborList(double cutoffDistance, const ListOfAtoms *atoms):
	CompositeNeighborLocator(cutoffDistance, atoms)
{
	neighborList = 0;
	atomsShape = 0;
	UpdatePosition();
}

CompositeNeighborList::~CompositeNeighborList()
{
	if(neighborList)  // delete the old neighborlist
	{
		for(int atom=0; atom<atomsShape.GetSize(); ++atom) delete [] neighborList[atom];
		delete [] neighborList;
	}
}

bool CompositeNeighborList::HalfNeighbors(int atomNumber,	int* neighbors, 
			int& numNeighbors, double (*displacements)[DIMENSION], double* displacementsSquared, 
		       int& spacesLeftInArrays)
{
	int type=0, atomNumberInSubList = atomNumber;
	// find which type is atom atomNumber
	while(atomNumberInSubList >= atoms->GetBranch(type)->GetNumber()) 
	{
		atomNumberInSubList -= atoms->GetBranch(type)->GetNumber();
		type++;
	}
	numNeighbors = 0;
	int numNbrsTmp, shiftIndex = 0;
	for(int type2=type+1; type2<atoms->NumberOfBranches(); ++type2)
	{
		int *neighborCandidates = &(neighborList[atomNumber][type2][0]);
		int maxNumberOfNeighbors = neighborList[atomNumber][type2].size();
		if(maxNumberOfNeighbors <= spacesLeftInArrays)
			memcpy(neighbors+numNeighbors, neighborCandidates, maxNumberOfNeighbors*sizeof(int));
		else return false;
		double *center = (*(atoms->GetCartesianPositions()))[atomNumber];
		PruneNeighborCandidates(center,maxNumberOfNeighbors,neighbors+numNeighbors,atoms->GetCartesianPositions(),
			displacements+numNeighbors,displacementsSquared+numNeighbors, cutoffDistance2, numNbrsTmp);
		spacesLeftInArrays -= numNbrsTmp;
		numNeighbors += numNbrsTmp;
	}
	if(!children[type]->HalfNeighbors(atomNumberInSubList, neighbors+numNeighbors, numNbrsTmp,
		displacements+numNeighbors, displacementsSquared+numNeighbors, spacesLeftInArrays))
			return false;
	shiftIndex = 0;
	for(int ty=0; ty<type; ++ty) shiftIndex += atoms->GetBranch(ty)->GetNumber();
	for(int i=numNeighbors; i<numNeighbors+numNbrsTmp;++i)neighbors[i] += shiftIndex;
	numNeighbors += numNbrsTmp;
	return true;
}

bool CompositeNeighborList::Neighbors(int atomNumber, int* neighbors, int& numNeighbors,
		       double (*displacements)[DIMENSION], double* displacementsSquared, 
		       int& spacesLeftInArrays)
{
	int type=0;
	bool append=true;
	int atomNumberInSubList = atomNumber;
// 1. find which type is atom atomNumber
	while(atomNumberInSubList >= atoms->GetBranch(type)->GetNumber()) 
	{
		atomNumberInSubList -= atoms->GetBranch(type)->GetNumber();
		type++;
	}
	numNeighbors = 0;
	int numNbrsTmp, shiftIndex = 0;
	for(int type2=0; type2<atoms->NumberOfBranches(); ++type2)
	{
		if(type == type2)
		{
			if(!children[type]->Neighbors(atomNumberInSubList, neighbors+numNeighbors, numNbrsTmp,
				displacements+numNeighbors, displacementsSquared+numNeighbors, spacesLeftInArrays))
					return false;
			shiftIndex = 0;
			for(int ty=0; ty<type; ++ty) shiftIndex += atoms->GetBranch(ty)->GetNumber();
			for(int i=numNeighbors; i<numNeighbors+numNbrsTmp;++i)neighbors[i] += shiftIndex;
			numNeighbors += numNbrsTmp;
		}
		else 
		{
			int *neighborCandidates = &(neighborList[atomNumber][type2][0]);
			int maxNumberOfNeighbors = neighborList[atomNumber][type2].size();
			if(maxNumberOfNeighbors <= spacesLeftInArrays)
				memcpy(neighbors+numNeighbors, neighborCandidates, maxNumberOfNeighbors*sizeof(int));
			else return false;
			double *center = (*(atoms->GetCartesianPositions()))[atomNumber];
			PruneNeighborCandidates(center,maxNumberOfNeighbors,neighbors+numNeighbors,atoms->GetCartesianPositions(),
				displacements+numNeighbors,displacementsSquared+numNeighbors, cutoffDistance2, numNbrsTmp);
			spacesLeftInArrays -= numNbrsTmp;
			numNeighbors += numNbrsTmp;
		}
	}
	return true;
}

bool CompositeNeighborList::Neighbors(const double *center, int* neighbors, int& numNeighbors,
				double (*displacements)[DIMENSION],	double* displacementsSquared,
				int& spacesLeftInArrays, double cutoff)
{
	bool append=true;
	numNeighbors = 0;
	int numNbrsTmp;
	int index2 = 0;
	for(int type=0; append && (type<atoms->NumberOfBranches()); ++type)
	{
		append &= children[type]->Neighbors(center, neighbors+numNeighbors, numNbrsTmp,
			displacements+numNeighbors, displacementsSquared+numNeighbors, spacesLeftInArrays, cutoff);
		for(int i=numNeighbors; i<numNeighbors+numNbrsTmp;++i)neighbors[i] += index2;
		index2 += atoms->GetBranch(type)->GetNumber();
		numNeighbors += numNbrsTmp;
	}
	return append;
}

bool CompositeNeighborList::Neighbors(int atomNumber, int type1, int type2,
				int* neighbors, int& numNeighbors, double (*displacements)[DIMENSION],
				double* displacementsSquared, int& spacesLeftInArrays)
{
	numNeighbors = 0;
	if(type1 == type2)
	{
		bool append = children[type1]->Neighbors(atomNumber, neighbors, numNeighbors,
			displacements, displacementsSquared, spacesLeftInArrays);
		int shiftIndex = 0;
		for(int ty=0; ty<type1; ++ty) shiftIndex += atoms->GetBranch(ty)->GetNumber();
		for(int i=0; i<numNeighbors;++i)neighbors[i] += shiftIndex;
		return append;
	}
	else 
	{
		for(int ty=0; ty<type1; ++ty) atomNumber += atoms->GetBranch(ty)->GetNumber();
		int *neighborCandidates = &(neighborList[atomNumber][type2][0]);
		int maxNumberOfNeighbors = neighborList[atomNumber][type2].size();
		if(maxNumberOfNeighbors <= spacesLeftInArrays)
			memcpy(neighbors, neighborCandidates, maxNumberOfNeighbors*sizeof(int));
		else return false;
		double *center = (*(atoms->GetCartesianPositions()))[atomNumber];
		PruneNeighborCandidates(center,maxNumberOfNeighbors,neighbors,atoms->GetCartesianPositions(),
			displacements,displacementsSquared, cutoffDistance2, numNeighbors);
		spacesLeftInArrays -= numNeighbors; 
	}
	return true;
}

void CompositeNeighborList::UpdatePosition()
{
	if(children.size() != atoms->NumberOfBranches())
	{
		SetChildren();
	}
	if(!(atomsShape == atoms->GetShape()) || (bc != atoms->GetBoundaryConditions()))
	{
		bc = atoms->GetBoundaryConditions();
		partialSum.clear();
		int nTypes = atoms->NumberOfBranches();
		if(nTypes > 0) partialSum.push_back(atoms->GetBranch(0)->GetNumber());
		for(int type=1; type<nTypes; ++type) 
		  partialSum.push_back(partialSum.back()+atoms->GetBranch(type)->GetNumber());
		UpdateMyData();
   		atomsShape = atoms->GetShape();
		return;
	}
	if(!DataIsValid()) UpdateMyData();
}

void CompositeNeighborList::SetCutoffDistance(double cutoff)
{
	double oldCutoff = cutoffDistance;
	cutoffDistance = cutoff;
	cutoffDistance2 = cutoff*cutoff;
	if((oldCutoff < cutoffDistance) && (atomsShape.GetSize() > 0)) MakeNeighborList();
}


void CompositeNeighborList::MakeNeighborList()
{
//	std::cout << "Making NL for "<<atoms->GetTypeName() << std::endl;
	double cutList;
	int atom;
	const int nBranches = atoms->NumberOfBranches();
	const int nAtoms = atoms->GetNumber();
// delete the old neighborlist
	if(neighborList)
	{
		for(int atom=0; atom<atomsShape.GetSize(); ++atom) delete [] neighborList[atom];
		delete [] neighborList;
	}
	neighborList = new vector<int> * [nAtoms];  // allocate the new neighborlist.
	for(atom=0; atom<nAtoms; ++atom) neighborList[atom] = new vector<int> [nBranches];

// allocate arrays for the use of children's Neighbors functions
	const int maxNeighbors = 1000;
    int neighbors[maxNeighbors];
	double displacements[maxNeighbors][DIMENSION];
	double displacementsSquared[maxNeighbors];

// build the neighbor list
	DMArrayIterator<double, DIMENSION> pos(positions);
	pos.Init();
	atom = 0;
	for(int branch = 0; branch < nBranches; ++branch)
	{
		const double cutList0 = cutoffDistance + children[branch]->GetDrift();
		for(int i = 0; i < atoms->GetBranch(branch)->GetNumber(); ++i)
		{
			int indexShift = 0;
			for(int branch2 = 0; branch2 < branch; ++branch2)
			{
                int numNeighbors;
				cutList = cutList0 + children[branch2]->GetDrift();
				int spacesLeftInArrays = maxNeighbors;
				if(!children[branch2]->Neighbors(pos.Current(), neighbors, numNeighbors, displacements,
					displacementsSquared, spacesLeftInArrays, cutList))
					throw DMException("CompositeNeighborList","MakeNeighborList",
					"need more maxNeighbors");
				for(int j = 0; j < numNeighbors; ++j)
				{
					int neigh = neighbors[j]+indexShift;   
// we symmetrize the neighborlist at the same time
					neighborList[atom][branch2].push_back(neigh);
					neighborList[neigh][branch].push_back(atom);
				}
				indexShift += atoms->GetBranch(branch2)->GetNumber();
			}
			// we do not need pairs of neighbors of the same type, so we are done with atom.
			++atom;
			pos.Next();
		}
	}
}


bool CompositeNeighborList::DataIsValid()
{
	bool ok = true;
	for(std::vector<NeighborLocator *>::iterator it=children.begin(); it != children.end(); ++it)
		ok &= (*it)->DataIsValid();
	return ok;
}


void CompositeNeighborList::UpdateMyData()
{
	for(std::vector<NeighborLocator *>::iterator it=children.begin(); it != children.end(); ++it)
		(*it)->UpdateMyData();
	MakeNeighborList();
}

NeighborLocator *CompositeNeighborList::Copy() const
{
  CompositeNeighborList *newNeighborLocator =
	  new CompositeNeighborList(cutoffDistance, atoms);
  
  return newNeighborLocator;
}
