// ExtendedNeighborLocator.h
//
//////////////////////////////////////////////////////////////////////

#ifndef EXTENDEDNEIGHBORLOCATOR_H
#define EXTENDEDNEIGHBORLOCATOR_H

#include "DigitalMaterial.h"
#include "NeighborLocator.h"
#include "DMArray.h"

class ExtendedNeighborLocator : public NeighborLocator
{
public:
	ExtendedNeighborLocator(NeighborLocator *neighborLocator); 

	virtual ~ExtendedNeighborLocator();

	void SetPositionArray(const DMArray<double, DIMENSION> *positions);

#ifndef SWIG
//! Neighbors determines the index of the points located within the cutoff distance to point # atomNumber
/*! 
	On exit (if the function returns true) neighbors contains the numNeighbors neighbors of point atomNumber, 
	there are  p
	\param atomNumber the index of the point whose neighbors are needed.
	\param neighbors on exit contains the index of the neighbors
	\param numNeighbors on exit contains the number of neighbors
	\param displacements on exit contains the vector binding point \a atomNumber and its neighbors
	\param displacementsSquared on exit contains the squared distances
	\param spacesLeftInArrays on entry tells the avaliable space in the arrays. If it's too small the function return false, otherwise
	\a spacesLeftInArrays is decreased by \a numNeighbors.
*/
	virtual bool Neighbors(int atomNumber,
							int * neighbors, 
							int& numNeighbors,
							double (*displacements)[DIMENSION],		// Displacement returns place of other guy wrt main guy
							double* displacementsSquared,
							int& spacesLeftInArrays);

//! HalfNeighbors, when looped over all atoms, is guaranteed to return each pair within the cutoff exactly once.
/*!
  This function is called HalfNeighbors because on exit neighbors contains only part of the neighbors, but if it
  is called for all the points then we are sure we get all the pairs. Warning: the neighbors
  which are returned by HalfNeighbors aren't necessarily the neighbors with smaller index than 
  \a atomNumber.
*/
	bool HalfNeighbors(int atomNumber,
							int * neighbors, 
							int& numNeighbors,
							double (*displacements)[DIMENSION],	// Displacement returns place of other guy wrt main guy
							double* displacementsSquared,
							int& spacesLeftInArrays);


//! The following version of Neighbors returns all atoms within the cutoff distance to center (cutoff=-1 means this->cutoffDistance)
	bool Neighbors(const double *center,
							int* neighbors, 
							int& numNeighbors,
							double (*displacements)[DIMENSION],		// Displacement returns place of other guy wrt main guy
							double* displacementsSquared,
							int& spacesLeftInArrays, double cutoff=-1.);

//! For Lists of Atoms made of several leaves, we need to know the neighbors in other leaves.
/*! Here atomNumber refers to the index in leaf \a type1. On exit, \a neighbors returns the index of the neighbors
with respect to the leaf \a type2. 
*/
	bool Neighbors(int atomNumber, int type1, int type2,
							Index* neighbors, int& numNeighbors,
							double (*displacements)[DIMENSION],		// Displacement returns place of other guy wrt main guy
							double* displacementsSquared,
							int& spacesLeftInArrays) ;
#endif //SWIG
	void UpdatePosition();
	
	void SetBoundaryConditions(const BoundaryConditions *bc) {neighborLocator->SetBoundaryConditions(bc);}
	void SetCutoffDistance(double cutoff) {neighborLocator->SetCutoffDistance(cutoff);}
	double GetCutoffDistance() const {return neighborLocator->GetCutoffDistance();} 
	void SetDrift(double drift) { neighborLocator->SetDrift(drift);}
	double GetDrift() { return neighborLocator->GetDrift();}


	NeighborLocator *Copy() const {return new ExtendedNeighborLocator(neighborLocator->Copy());}

protected:
	NeighborLocator *neighborLocator;

//! Test whether the internal data has to be updated.
	bool DataIsValid();
//! Recalculate all the data (ie. update CellList, NeighborList, etc...)
	void UpdateMyData() ;

private:
};

#endif // EXTENDEDNEIGHBORLOCATOR_H
