#ifndef NEIGHBORLIST_H
#define NEIGHBORLIST_H

#include "DigitalMaterial.h"

#include "NeighborLocator.h"
#include "DMArray.h"
#include "AttributeError.h"

// C-ish definitions employed by Jakob Schiotz

#define MAXTMP 12					// Max number of temporary arrays
#define MAXNBR 200				// Max number of neighbors

/** 
 * This neighbor locator is based on Jakob Schiotz's code. It works in 3D only
 * and doesn't support boundary conditions.
 **/

class NeighborList : public NeighborLocator
{
public:
	NeighborList(double cutdist, double drift);
	~NeighborList();
    virtual void UpdatePosition();

	//	These two functions return true if appending to the stacks is successful,
	// and false if appending to the stacks is not successful

#ifndef SWIG
	virtual bool HalfNeighbors(	int atomNumber,
								int* neighbors, 
								int& numNeighbors,
								double (*displacements)[DIMENSION], 
								double* displacementsSquared,
								int& spacesLeftInArrays);
	virtual bool Neighbors(int atomNumber, int* neighbors, int& numNeighbors, double (*displacements)[DIMENSION], double* displacementsSquared, int& spacesLeftInArrays);
	virtual bool Neighbors(const double *center, int *neighbors, int& numNeighbors, double (*displacements)[DIMENSION], double *displacementsSquared,
		int& spacesLeftInArray, double cutoff=-1){
		throw AttributeError("NeighborList","Neighbors");
	}
#endif
	
	int MaxNumberOfNeighbors(int atomNumber);
	int MaxNumberOfHalfNeighbors(int atomNumber);
	virtual void SetCutoffDistance(double cutoff) {cutdist = cutoff;}
	virtual void SetDrift(double drift) {cutlist = cutlist+2*drift;}
	NeighborLocator *Copy() const;

	const std::string &GetType() { return className; }

#ifndef SWIG
protected:

	// protected methods

	virtual void Allocate(int nAtoms, const DMArray<double, DIMENSION> *positions);
	virtual void MakeCellList(const DMArray<double, DIMENSION> *positions);
	virtual void CheckCriteria(const DMArray<double, DIMENSION> *positions);
	virtual void MakeNeighborListWithoutCellList();
	virtual void MakeNeighborList(const DMArray<double, DIMENSION> *positions);
	virtual double max_abs_vec_diff(const DMArray<double, DIMENSION> *positions, const DMArray<double, DIMENSION> *oldPositions);
	void SetMaximumMinimumPosition(const DMArray<double, DIMENSION> *positions);

	// protected data

	int nAtoms;
	double cutdist, cutlist;
	double min[DIMENSION], max[DIMENSION];
	bool initialUpdate;
	// This should all be declared as pointers!
	int (*nblist)[MAXNBR];		// nblist[i][inbr] = atom index of
										// neighbor inbr of atom i
										// This is an uneven stack!
	int *nnblo;						// nnblo[i] = number of neighbors of
										// atom i with atom index less than i
	int *nnbhi;						// nnbhi[i] = number of neighbors of
										// atom i with atom index greater than i

	// data array to be maintained
	// In the event that an acceleration scheme other than neighbor lists and
	// cell lists is to be implemented, the checking of the criteria to rebuild
	// a neighbor/cell list cannot be implemented in the force computation
	// loop, because criterion vary from acceleration scheme to acceleration 
	// scheme.

	// Store old positions to facilitate checking

	DMArray<double, DIMENSION> *oldPositions;

	int ncell[DIMENSION];		//	Number of cells in the three directions
	int ntotcell;					//	Total number of cells
	int **cell;						//	Array of pointers to the cells
	int *cellmax, *cellcnt;		//	Used for counting the atoms in each cell
	int (*ix)[DIMENSION];		//	Cell coordinates of the atoms, used to be
										// *ix, *iy, *iz
	double cellwidth;				//	Min. size of the cells
	double icw[DIMENSION];		//	double version of ncell

	// Temporary memory space (something that Jakob Schiotz uses frequently)

	int ndouble;
	int nfloat;
	double *dTmp[MAXTMP];	// Temporary double precision storage
	float *fTmp[2*MAXTMP];	// do. single precision.   *** dTmp, fTmp, and iTmp
	int *iTmp[2*MAXTMP];		// do. integer.            *** overlayed in memory
private:
	static const std::string className;
#endif // not SWIG
};

#endif	// NEIGHBORLIST_H
