#ifndef NEIGHBORLOCATOR_H
#define NEIGHBORLOCATOR_H


#include "DigitalMaterial.h"
#include "Serializable.h"

template<class Type, int N> class DMArray ;
#include "DMException.h"

// #define PRUNESTATS

#ifndef NDEBUG
#define PRUNESTATS
#endif

//#include "BoundaryConditions.h"
// Avoid cyclic include files and decouple base classes
class BoundaryConditions;
/*!
 The duty of a neighbor locator is to determine in a set of points the pairs which 
 are separated by a distance smaller than a given cutoff distance. In general it has
 to take into account the boundary conditions (note: as of May 18th 2000, NeighborList 
 doesn't treat Boundary Conditions correctly). Since a neighbor locator certainly has
 to calculate distances beween points, it makes sense that it returns these distances 
 because they are usually needed by a potential to calculate energy and forces. This is 
 why the standard interface of a neighbor locator contains quite a lot of arguments:
 in addition to the index of the neighbors of a point, it also returns the vector of
 displacements and the distances (displacementsSquared). We have decided to use standard 
 pointers instead of the STL in order to make use of the nice trick developped by 
 J. Schiotz where all the pairs (and displacements) are accumulated in a big array and
 all the calculations on the distances are processed together (it's a kind of vectorization).
 This is why the neighbor locator not only returns the number of neighbors it found but
 also keeps track of the number of spaces avaliable in the allocated arrays (this
 number is "spacesLeftInArrays"). The idea is that the potential computes the forces once
 the array is full (ie when neighbor locator returns false). 
 
 As seen below, there are basically 6 kinds of requests a neighbor locator can handle. 
*/

class NeighborLocator : public Serializable
{
public:
	NeighborLocator(double cutoffDistance = 1.);

	virtual ~NeighborLocator();
	virtual void SetPositionArray(const DMArray<double, DIMENSION> *positions);

//! Neighbors determines the index of the points located within the cutoff distance to point # atomNumber
/*! 
	On exit (if the function returns true) neighbors contains the numNeighbors neighbors of point atomNumber, 
	there are  p
	\param atomNumber the index of the point whose neighbors are needed.
	\param neighbors on exit contains the index of the neighbors
	\param numNeighbors on exit contains the number of neighbors
	\param displacements on exit contains the vector binding point \a atomNumber and its neighbors
	\param displacementsSquared on exit contains the squared distances
	\param spacesLeftInArrays on entry tells the avaliable space in the arrays. If it's too small the function return false, otherwise
	\a spacesLeftInArrays is decreased by \a numNeighbors.
*/
    virtual bool Neighbors(int atomNumber,
                           int * neighbors,
                           int& numNeighbors,
                           double (*displacements)[DIMENSION],		// Displacement returns place of other guy wrt main guy
                           double* displacementsSquared,
                           int& spacesLeftInArrays) = 0;

//! HalfNeighbors, when looped over all atoms, is guaranteed to return each pair within the cutoff exactly once.
/*!
  This function is called HalfNeighbors because on exit neighbors contains only part of the neighbors, but if it
  is called for all the points then we are sure we get all the pairs. Warning: the neighbors
  which are returned by HalfNeighbors aren't necessarily the neighbors with smaller index than 
  \a atomNumber.
*/
	virtual bool HalfNeighbors(int atomNumber,
							int * neighbors, 
							int& numNeighbors,
							double (*displacements)[DIMENSION],	// Displacement returns place of other guy wrt main guy
							double* displacementsSquared,
							int& spacesLeftInArrays) = 0;

    //! Returns all atoms within the cutoff distance to \a center
    /*! The default value cutoff=-1 means this->cutoffDistance should be taken
    */
    virtual bool Neighbors(const double *center,
                           int* neighbors, 
                           int& numNeighbors,
                           double (*displacements)[DIMENSION],
                           double* displacementsSquared,
                           int& spacesLeftInArrays, double cutoff=-1.) = 0;

    //! Return the neighbors of type \a type2 of atom \a atomNumber of type \type1.
    /*! \a atomNumber refers to the index within leaf \a type1. On exit, \a neighbors
    returns the index of the neighbors within leaf \a type2. 
    */
    virtual bool Neighbors(int atomNumber, int type1, int type2,
							int* neighbors, int& numNeighbors,
							double (*displacements)[DIMENSION],		// Displacement returns place of other guy wrt main guy
							double* displacementsSquared,
							int& spacesLeftInArrays) 
	{
		throw DMException("NeibhorLocator",
						  "Neighbors",
						  "Types not dealt by this NeighborLocator ");
	}

    virtual bool HalfNeighbors(int atomNumber, int type1, int type2,
                               int* neighbors, int& numNeighbors,
                               double (*displacements)[DIMENSION],		// Displacement returns place of other guy wrt main guy
                               double* displacementsSquared,
                               int& spacesLeftInArrays) 
	{
		throw DMException("NeibhorLocator",
						  "Neighbors",
						  "Types not dealt by this NeighborLocator ");
	}

    //! Call to update state when positions of atoms change
	virtual void UpdatePosition()=0;
	
    virtual void SetBoundaryConditions(const BoundaryConditions *bc) {this->bc = bc;}
	virtual void SetCutoffDistance(double cutoff);
	virtual double GetCutoffDistance() const {return cutoffDistance;} 
	virtual void SetDrift(double drift)
	{
		throw DMException("NeighborLocator","SetDrift", "This NeighborLocator doesn't use drifts");
	}
	virtual double GetDrift() const
	{
		throw DMException("NeighborLocator","GetDrift", "This NeighborLocator doesn't use drifts");
	}


	virtual NeighborLocator *Copy() const = 0;


	//! Test whether the internal data has to be updated.
	virtual bool DataIsValid() {return true;}
	//! Recalculate all the data (ie. update CellList, NeighborList, etc...)
	virtual void UpdateMyData() {}

	virtual void DisplayStatistics() {
#ifdef PRUNESTATS
		std::cerr << "PruneNeighborCandidates: " << okay << " accepted (" << 100*okay/(okay+rejected) <<
			"%), " << rejected << " rejected, " << okay+rejected << "total." << std::endl;
#else
		std::cerr << "No statistics from PruneNeighborCandidates" << std::endl;
#endif // PRUNESTATS
	}
	

	//! Functions associated with the serialization
    //@{
    //! Save this neighbor locator to a writer
    void Save(DMWriter *writer) const ;
    //! Read a neighbor locator from a reader. (In parallel, only node 0 gets the data).
	void Load(DMReader *reader) ;
    //! Returns "NeighborLocator"
    const std::string &GetType() const;
    //@}


protected:
	const DMArray<double, DIMENSION> *positions;
	double cutoffDistance;
	double cutoffDistance2;
#ifdef PRUNESTATS 
	long okay, rejected;  // we use this to measure the ratio of true neighbors among the candidates that are submitted to PruneNeighborCandidates
#endif //PRUNESTATS
/*!
 This is a quite general operation in Neighbor Locators to fill an array with neighbor
 candidates, then apply the boundary conditions, then prune the list to keep only 
 the neighbors within some cutoff distance CutDist (we pass its square). 
 So here is the function: (make sure the boundary conditions are set correctly (by the UpdatePositions)
*/
	void PruneNeighborCandidates(const double *center, int numberOfCandidates,
			int *indexOfCandidates, const DMArray<double, DIMENSION> *positionArray,  // this is the array wrt which the index of the candidates in given
			double (*displacements)[DIMENSION],
			double *displacementsSquared, double sqCutDist, int &numberOfNeighbors);
	const BoundaryConditions *bc;
private:
	static const std::string className;
};


/*!
  This is a kind of NULL object, that doesn't do anything, but that can be plugged in where 
  we need a neighbor locator and do not use it. It's also useful if we set the cutoff distance
  to 0, to specify that the particles do not interact.  
*/
class NoNeighborLocator: public NeighborLocator
{
public:
	NoNeighborLocator(double cutoffDistance=1.);
	~NoNeighborLocator() {;}
	bool HalfNeighbors(int atomNumber, int* neighbors, int& numNeighbors,
		       double (*displacements)[DIMENSION], double* displacementsSquared, 
		       int& spacesLeftInArrays){return ReturnTrue(numNeighbors);}

	bool Neighbors(int atomNumber, int* neighbors, int& numNeighbors,
		       double (*displacements)[DIMENSION], double* displacementsSquared, 
		       int& spacesLeftInArrays){return ReturnTrue(numNeighbors);}

	bool Neighbors(const double *center, int* neighbors, int& numNeighbors,
				double (*displacements)[DIMENSION],	double* displacementsSquared,
				int& spacesLeftInArrays, double cutoff=-1.);

	bool Neighbors(int atomNumber, int type1, int type2,
										int* neighbors, int& numNeighbors,
										double (*displacements)[DIMENSION],
										double* displacementsSquared,
										int& spacesLeftInArrays) {return ReturnTrue(numNeighbors);}

	bool HalfNeighbors(int atomNumber, int type1, int type2,
										int* neighbors, int& numNeighbors,
										double (*displacements)[DIMENSION],
										double* displacementsSquared,
										int& spacesLeftInArrays) {return ReturnTrue(numNeighbors);}

    virtual void UpdatePosition() {}

	NeighborLocator *Copy() const;
	const std::string &GetType() const { return className; }

private:
	bool ReturnTrue(int &numNeighbors);
	static const std::string className;
};

#endif
