// CrackPlaneObserver.cpp: implementation of the CrackPlaneObserver class.
//
//////////////////////////////////////////////////////////////////////

#include "DigitalMaterial.h"
#include "CrackPlaneObserver.h"
#include "ListOfAtoms.h"
#include "DMLastHeader.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CrackPlaneObserver::CrackPlaneObserver(double planePoint[DIMENSION], double planeNormal[DIMENSION], double cutoffDistance,  int notifyLevel) : ListOfAtomsObserver(notifyLevel)
{
	SetPlanePoint(planePoint);
	SetPlaneNormal(planeNormal);
	SetCutoffDistance(cutoffDistance);
	
	maxIncrease = 0.;
	meanIncrease = 0.;
}

CrackPlaneObserver::~CrackPlaneObserver()
{

}

void CrackPlaneObserver::SetPlanePoint(double planePoint[DIMENSION])
{ 
	for( int coord = 0; coord < DIMENSION; ++coord )
		this->planePoint[coord] = planePoint[coord];
}


void CrackPlaneObserver::SetPlaneNormal(double planeNormal[DIMENSION]) 
{ 
	double normNorm = 0.;

	for( int coord = 0; coord < DIMENSION; ++coord )
		normNorm +=  planeNormal[coord] * planeNormal[coord];

	// insure normal is a unit vector
	normNorm = sqrt(normNorm);
	ASSERT(normNorm > 0.);
	for( int coord = 0; coord < DIMENSION; ++coord )
		this->planeNormal[coord] = planeNormal[coord] / normNorm;
} 


double CrackPlaneObserver::GetDistanceFromPlane(double position[DIMENSION]) const
{
	// need to figure out how to deal with periodic boundary conditions!
	// (maybe apply DifferenceBC to disp)
	double disp[DIMENSION];
	double dotProd = 0.;
	for( int coord = 0; coord < DIMENSION; ++coord )
	{
		disp[coord] = position[coord] - planePoint[coord];
		dotProd += disp[coord] * planeNormal[coord];
	}
	return dotProd;
}


void CrackPlaneObserver::Update(ListOfAtoms *atoms)
{
	const DMArray<double, DIMENSION> *positions = atoms->GetCartesianPositions();
	maxIncrease = -1.e99;
	meanIncrease = 0.;

	vector<pair<int, int> >::iterator listIter;
	vector<double>::iterator sepIter, initialSepIter;
	for( listIter = pairList.begin(), sepIter = pairSeparations.begin(), initialSepIter = initialSeparations.begin(); listIter != pairList.end(); listIter++, sepIter++, initialSepIter++ )
	{
		int atom1 = listIter->first;
		int atom2 = listIter->second;

		double separationSq = 0.;
		for( int coord = 0; coord < DIMENSION; ++coord )
		{
			double diff = (*positions)[atom2][coord] - (*positions)[atom1][coord];
			separationSq += diff * diff;
		}		

		*sepIter = sqrt(separationSq);
		
		double increase = (*sepIter) - (*initialSepIter);

		if( increase > maxIncrease)
			maxIncrease = increase;

		meanIncrease += increase;
	} // end for loop over the iterators
	
	meanIncrease /= pairSeparations.size();
}
void CrackPlaneObserver::MakePairList(ListOfAtoms *atoms)
{
	pairList.clear();
	initialSeparations.clear();
	pairSeparations.clear();

	const DMArray<double, DIMENSION> *positions = atoms->GetCartesianPositions();
	DMArrayIterator<double, DIMENSION> positionIter(positions);
	vector<int> negNearAtoms;
	vector<int> posNearAtoms;

	// first make two lists of int, one for atoms near the plane
	// on the negative side of it and one for atoms near the plane on the
	// positive side of it (ie same side as the normal vector)
	int atom;
	for( positionIter.Init(), atom = 0; !positionIter.IsDone(); positionIter.Next(), ++atom )
	{
		double *position = positionIter.Current();
		double distFromPlane = GetDistanceFromPlane(position);
		
		if(fabs(distFromPlane) < cutoffDistance)
		{
			if(distFromPlane < 0.)
				negNearAtoms.push_back(atom);
			else
				posNearAtoms.push_back(atom);
		}
	}


	double cutoffDistanceSq = cutoffDistance * cutoffDistance;
	// loop through neg and pos atoms
	vector<int>::iterator negIter, posIter;
	for( negIter = negNearAtoms.begin(); negIter != negNearAtoms.end(); negIter++ )
	{
		int negAtom = *negIter;
		for( posIter = posNearAtoms.begin(); posIter != posNearAtoms.end(); posIter++ )
		{
			int posAtom = *posIter;
			
			double separationSq = 0.;
			for( int coord = 0; coord < DIMENSION; ++coord )
			{
				double diff = (*positions)[negAtom][coord] - (*positions)[posAtom][coord];
				separationSq += diff * diff;
			}
				
			if(separationSq <= cutoffDistanceSq)
			{
				pairList.push_back(make_pair(negAtom,posAtom));
				initialSeparations.push_back(sqrt(separationSq));
				pairSeparations.push_back(sqrt(separationSq));			
			}
		}
	}
}
