// LocalForceObserver.cpp: implementation of the LocalForceObserver class.
//
//////////////////////////////////////////////////////////////////////
#include "DigitalMaterial.h"
#include "LocalForceObserver.h"
#include "Potential.h"
#include "ListOfAtoms.h"
#include "DMArray.h"
#include "DMLastHeader.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

LocalForceObserver::LocalForceObserver(Potential *potential, double forceCutoff, std::string fileName, int notifyLevel) : ListOfAtomsObserver(notifyLevel)
{
	this->potential = potential;
	this->forceCutoff = forceCutoff;
	fileObject = new ofstream(fileName.c_str());
	nAtoms = 0;
	forces = 0;
}

LocalForceObserver::~LocalForceObserver()
{
	delete fileObject;
	if(forces)
		delete forces;
}

void LocalForceObserver::Allocate(ListOfAtoms *atoms)
{
	if(forces)
		delete forces;
	forces = new DMArray<double, DIMENSION>(atoms->GetShape());
}

void LocalForceObserver::Update(ListOfAtoms *atoms)
{
	if (!forces || nAtoms != atoms->GetNumber())
		Allocate(atoms);
	
	potential->CalculateForces(atoms,forces);
	
	int coord, maxAtom[DIMENSION], atomIdx;
	double maxForce[DIMENSION];
	for( coord = 0; coord < DIMENSION; ++coord )
	{
		maxForce[coord] = 0.;
		maxAtom[coord] = 0;
	}
	int nLargeForces = 0;
	double fCOSq = forceCutoff * forceCutoff;

	DMArrayIterator<double, DIMENSION> forceIter(forces);
	for( forceIter.Init(),atomIdx = 0; !forceIter.IsDone(); forceIter.Next(),++atomIdx )
	{
		double *currentForce = forceIter.Current();
		double forceMagSq = 0.;
		for( coord = 0; coord < DIMENSION; ++coord )
		{
			if(fabs(currentForce[coord]) > maxForce[coord]) 
			{
				maxForce[coord] = currentForce[coord];
				maxAtom[coord] = atomIdx;
			}
			forceMagSq += currentForce[coord] * currentForce[coord];
			if(forceMagSq > fCOSq)
				++nLargeForces;
		}

	}
	const DMArray<double, DIMENSION> *positions = atoms->GetCartesianPositions();
	cout << "For each coordinate, atom with maximum that component of force, with position and full force:" << endl;
	int oldPrec = cout.precision();
	cout.precision(5);
	for( coord = 0; coord < DIMENSION; ++coord )
	{		
		int coord2;
		cout << maxAtom[coord] << " (";
		for ( coord2 = 0; coord2 < DIMENSION; ++coord2 )
			cout << (*positions)[maxAtom[coord]][coord2] << " ";
		cout << "), (";
		for ( coord2 = 0; coord2 < DIMENSION; ++coord2 )
			cout << (*forces)[maxAtom[coord]][coord2] << " ";
		cout << ")" << endl;
	}
	cout << "Number of large forces " << nLargeForces << endl << endl;
	cout.precision(oldPrec);
}

