// ParallelNeighborLocator.cpp: implementation of the ParallelNeighborLocator class.
//
//////////////////////////////////////////////////////////////////////
#include "DigitalMaterial.h"

#include "ParallelNeighborLocator.h"
#include "DMException.h"

#include "DMLastHeader.h"

#ifdef DMPARALLEL

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ParallelNeighborLocator::ParallelNeighborLocator(Migrator *m,
                                                 NeighborLocator *neighborLocator):
NeighborLocator(0.) 
{
    if (m == 0)
        throw DMException("ParallelCellNeighborList", "ParallelCellNeighborList",
						  "No migrator");
    migrator = m;
    this->neighborLocator = neighborLocator;
}

ParallelNeighborLocator::~ParallelNeighborLocator()
{
}

void ParallelNeighborLocator::SetPositionArray(const DMArray<double, DIMENSION> *positions) 
{
    this->positions = positions;
    positions->GetSynchronizationKit()->Init();	
    migrator->Migrate();                   /* Changes atoms */
    positions->GetSynchronizationKit()->Synchronize();
    neighborLocator->SetPositionArray(positions->GetSynchronizationKit()->GetExtended());
}


void ParallelNeighborLocator::UpdatePosition()
{
    if(!DataIsValid()) UpdateMyData();
    else positions->GetSynchronizationKit()->Synchronize();
}

void ParallelNeighborLocator::UpdateMyData()
{
    migrator->Migrate();                   /* Changes atoms */
    positions->GetSynchronizationKit()->Synchronize();
    // Do NOT call atoms->SetCartesianPositions() as it would cause a loop.
    neighborLocator->UpdateMyData();
}

bool ParallelNeighborLocator::DataIsValid()
{
    int needUpdateHere = (neighborLocator->DataIsValid() ? 0: 1);
    int needUpdateSomewhere;
    
    dmProtocol->AllReduce(&needUpdateHere, &needUpdateSomewhere, 1, DMProtocol::MAX);
    if (needUpdateSomewhere) return false;
    else return true;
}


bool ParallelNeighborLocator::Neighbors(int atomNumber,
                                        int * neighbors, 
                                        int& numNeighbors,
                                        double (*displacements)[DIMENSION],
                                        double* displacementsSquared,
                                        int& spacesLeftInArrays)
{
    return neighborLocator->Neighbors(atomNumber, neighbors, numNeighbors, 
        displacements, displacementsSquared, spacesLeftInArrays);
}

bool ParallelNeighborLocator::HalfNeighbors(int atomNumber,
                                            int * neighbors, 
                                            int& numNeighbors,
                                            double (*displacements)[DIMENSION],
                                            double* displacementsSquared,
                                            int& spacesLeftInArrays)
{
    return neighborLocator->HalfNeighbors(atomNumber, neighbors, numNeighbors, 
        displacements, displacementsSquared, spacesLeftInArrays);
}


/*bool ParallelNeighborLocator::HalfNeighbors(Index atomNumber,
                                            Index* neighbors, 
                                            int& numNeighbors,
                                            double (*displacements)[DIMENSION],
                                            double* displacementsSquared,
                                            int& spacesLeftInArrays)
{
    return neighborLocator->HalfNeighbors(atomNumber, neighbors, numNeighbors, 
        displacements, displacementsSquared, spacesLeftInArrays);
}
*/


bool ParallelNeighborLocator::Neighbors(const double *center,
                                        int* neighbors, 
                                        int& numNeighbors,
                                        double (*displacements)[DIMENSION],
                                        double* displacementsSquared,
                                        int& spacesLeftInArrays, double cutoff)
{
    return neighborLocator->Neighbors(center, neighbors, numNeighbors, 
        displacements, displacementsSquared, spacesLeftInArrays, cutoff);
}


bool ParallelNeighborLocator::Neighbors(int atomNumber,
                                        int type1,
                                        int type2,
                                        int* neighbors,
                                        int& numNeighbors,
                                        double (*displacements)[DIMENSION],
                                        double* displacementsSquared,
                                        int& spacesLeftInArrays) 
{
    const int nTypes = positions->NumberOfLeaves();
    ASSERT(type1 < nTypes);
    ASSERT(type2 < nTypes);
    bool append = neighborLocator->Neighbors(atomNumber, type1, type2,
        neighbors, numNeighbors, displacements, displacementsSquared,
        spacesLeftInArrays);
    if(!append) return false;
    int numGhosts;
    int ghostType2 = type2 + nTypes; // ghosts come after reals
    append = neighborLocator->Neighbors(atomNumber, type1, ghostType2,
        neighbors+numNeighbors, numGhosts, displacements+numNeighbors,
        displacementsSquared+numNeighbors, spacesLeftInArrays);
    const int shift = positions->GetShape().GetLeavesSize()[type2];
    for(int gh=0; gh<numGhosts; ++gh)
        neighbors[numNeighbors+gh] += shift;
    numNeighbors += numGhosts;
    return append;
}

bool ParallelNeighborLocator::HalfNeighbors(int atomNumber,
                                            int type1,
                                            int type2,
                                            int* neighbors,
                                            int& numNeighbors,
                                            double (*displacements)[DIMENSION],
                                            double* displacementsSquared,
                                            int& spacesLeftInArrays) 
{
    const int nTypes = positions->NumberOfLeaves();
    ASSERT(type1 < nTypes);
    ASSERT(type2 < nTypes);
    if(type2 < type1) numNeighbors = 0;
    else
    {
        bool append = neighborLocator->HalfNeighbors(atomNumber, type1, type2,
            neighbors, numNeighbors, displacements, displacementsSquared,
            spacesLeftInArrays);
        if(!append) return false;
    }
    int numGhosts;
    int ghostType2 = type2 + nTypes; // ghosts come after reals
    bool append = neighborLocator->HalfNeighbors(atomNumber, type1, ghostType2,
        neighbors+numNeighbors, numGhosts, displacements+numNeighbors,
        displacementsSquared+numNeighbors, spacesLeftInArrays);
    const int shift = positions->GetShape().GetLeavesSize()[type2];
    for(int gh=0; gh<numGhosts; ++gh)
        neighbors[numNeighbors+gh] += shift;
    numNeighbors += numGhosts;
    return append;
}

#endif // DMPARALLEL
