// CutLennardJonesPotential.h: interface for the CutLennardJonesPotential class.
//
//////////////////////////////////////////////////////////////////////

#ifndef CUTLENNARDJONESPOTENTIAL_H
#define CUTLENNARDJONESPOTENTIAL_H

#include "DMFirstHeader.h"

#include "DigitalMaterial.h"
#include "Potential.h"

//! Lennard Jones potential 4*(r12inv - r6inv); smooth cut off, using parabola in r^2 (fourth order) between cut1 and cut2


/*
  It seems we should put the elastic constants here too.
  For this potential in 3D we have C11=96.6031,C44 = 55.01, C12=54.9209


 */

class CutLennardJonesPotential : public Potential
{
public:
	CutLennardJonesPotential();
	virtual ~CutLennardJonesPotential();
	void CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);
	void CalculateForcesBetweenTypes(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces, int type1, int type2) ; // the force array isn't initialized in this function

	//! Computes energy of a single atom, and the forces arising due to just this term
	double AtomicForcesEnergy(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces, int i);
	double CalculatePairForceEnergy(double *displacement, double displacementSq, double *force);


	double CalculateEnergy(ListOfAtoms *atoms);
	double CalculateEnergyAndForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);
	double CalculateAtomicEnergy(ListOfAtoms *atoms, int i);
	double GetInnerCutoffDistance() const {return cut1;}

	void CalculateStressTensor(ListOfAtoms *atoms, int i, double stress[DIMENSION*(DIMENSION+1)/2]);
	void CalculateStressTensor(ListOfAtoms *atoms, DMArray<double, DIMENSION*(DIMENSION+1)/2> *stress);
	

	// Roughly returns the diameter of the atom
	// For fcc, should return 1/sqrt(2) of the lattice constant, for example
	// Currently used to set up the lattice structure: a good initial number is important
	double GetLengthScale() const
	{
	  if(DIMENSION == 2) return 1.1132013 ;   //   the energy per atom in 2D is : -3.20572570331
	  else if(DIMENSION==3) return 1.095693678895;  //   the energy per atom in 3D is : -7.6092447344
	  else	return pow(2.0, 1.0/6.0);	// Minimum of pair potential: a bit bigger than the lattice constant!
		// It would be fine to put the real answers here, as a function of dimension
	}

	// For each potential, we need some kind of energy scale to set time constants
	// for movers and such
	// It doesn't have to be precisely defined! A crude measure is OK.
	// This should return, for a pair potential V(x) with LengthScale ell, V''(ell)
	double GetEnergyCurvatureScale() const
	{
		return 72.0/pow(2.0, 1.0/3.0);
	}

	
	void Save(DMWriter *writer) const;
	void Load(DMReader *reader);
	//! Get the string name of the class for saving or loading.
	const std::string &GetType() const
	{
		return className;
	}

 protected:
#ifndef SWIG
	//! Do the main work for calculating atomic stress. This is called by the two versions of CalculateStressTensor
	virtual void StressTensorKernel(ListOfAtoms *atoms, int i, double stress[(DIMENSION+1)*DIMENSION/2]);
	//! Allocate the \a neighbors, \a displacements, and \a displacementsSq arrays
	void AllocateForNeighborLocator(int nPairs);
	//! a rough (over-)estimation of the number of neighbors
	int         nPairs;
	double      *displacementsSq;
	double      (*displacements)[DIMENSION];
	int         *neighbors;

private:
	// parameters of the potential: cutoffs
	const double cut1;
	const double cut2;
	const double r2cut1;

	// parameters of the potential (relating the polynomial part)
	const double CLJ_alpha;
	const double CLJ_beta;
	const double CLJ_gamma;
	const double CLJ_delta;

	static const std::string className;
#endif // SWIG 
};

#endif // !defined(CUTLENNARDJONESPOTENTIAL_H
