// MultiCutLennardJonesPotential.h: interface for the MultiCutLennardJonesPotential class.
//
//////////////////////////////////////////////////////////////////////

#ifndef MULTICUTLENNARDJONESPOTENTIAL_H
#define MULTICUTLENNARDJONESPOTENTIAL_H

#include "DMFirstHeader.h"

#include "DigitalMaterial.h"
#include "Potential.h"
#include <map>

//! Lennard-Jones potential with variable parameters and multiple species (parameters differ for each pair of types). Same as CutLennardJones

/*!

  For now I am identifying leaves with chemical types (species)---ie types to which are associated different values of the LJ interaction parameters. Leaf 0 is species 0 etc. This choice is somewhat arbitrary, except NeighborLocator has a function that takes leaf indices to get the Neighbors of an atom on oe given leaf which are members of another given leaf.

  Another choice would be to do it by first level branches Or you could use typeNames of leaves to identify species, although if more than one leaf was of the same species it might be inefficient.

  


*/

class MultiCutLennardJonesPotential : public Potential
{
public:
	MultiCutLennardJonesPotential();
	virtual ~MultiCutLennardJonesPotential();
	void Allocate(int nAtoms);
	void CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);
	void CalculateForcesBetweenTypes(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces, int type1, int type2) ; // the force array isn't initialized in this function

	double CalculateEnergy(ListOfAtoms *atoms);
	double CalculateEnergyAndForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);
	double CalculateAtomicEnergy(ListOfAtoms *atoms, int i);
	double GetInnerCutoffDistance() const {return cut1;}

	// Set the parameters

	void SetNumberOfTypes(int numTypes){AllocateParameters(numTypes);}
	int GetNumberOfTypes() const {return numberOfTypes;}
	void SetSigmas(double **sigma);
	double GetSigma(int a, int b) const { return sigmas[a][b]; }
	void SetEpsilons(double **epsilon);
	double GetEpsilon(int a, int b) const { return epsilons[a][b]; }

	int GetNumberElementNames() { return elementNames.size(); }
	//const std::string &GetElementName(int n) { return elementNames[n].first; }
	void AddElementName(std::string name) { elementNames.insert(std::make_pair(name,elementNames.size())); }

	// Roughly returns the diameter of the atom
	// Here I will assume that the 00 components of the epsilons and sigmas
	// are typical and use them for the length scale and energy curvature scale
	double GetLengthScale() const
	{
		double lengthScale;

		if(DIMENSION == 2) lengthScale = 1.1132013 ;   //   the energy per atom in 2D is : -3.20572570331
		else if(DIMENSION==3) lengthScale =  1.09569360401;  //   the energy per atom in 2D is : -7.6092447344
		else	lengthScale =  pow(2.0, 1.0/6.0);	// Minimum of pair potential: a bit bigger than the lattice constant!
		
		if(sigmas)
			lengthScale *= sigmas[0][0];
		return lengthScale;

	}

	// This should return, for a pair potential V(x) with LengthScale ell, V''(ell)
	double GetEnergyCurvatureScale() const
	{
	    double energyCurvatureScale = 72.0/pow(2.0, 1.0/3.0);
		if(epsilons && sigmas)
			energyCurvatureScale *= epsilons[0][0]/(sigmas[0][0]*sigmas[0][0]);
		return energyCurvatureScale;
	}

	void Save(DMWriter *writer) const;
	void Load(DMReader *reader);
	//! Get the string name of the class for saving or loading.
	const std::string &GetType() const
	{
		return className;
	}


protected:
#ifndef SWIG	
	int         nAtoms;
	double      *displacementsSq;
	double      (*displacements)[DIMENSION];
	int         *neighbors;

	double		cut1;
	double		cut2;
	double		r2cut1;
	double		r2cut2;

	// main parameters
	int numberOfTypes;
	double **epsilons;
	double **sigmas;

	std::map<std::string, int> elementNames;

	void AllocateParameters(int numTypes);

	// other parameters of the potential (relating the polynomial part)
	const double CLJ_alpha;
	const double CLJ_beta;
	const double CLJ_gamma;
	const double CLJ_delta;
private:
	static const std::string className;

#endif // SWIG 
};

#endif // !defined(MULTICUTLENNARDJONESPOTENTIAL_H
