// Potential.h: interface for the Potential class.
//
//////////////////////////////////////////////////////////////////////

#ifndef POTENTIAL_H
#define POTENTIAL_H

// Think of this as an abstract class, except we don't insist that you 
// define all three member functions. 
// We imagine not everyone will need all three functions.
#include "DMFirstHeader.h"
#include "Serializable.h"
#include "DMWriter.h"
#include "DMReader.h"
#include "SprintString.h"

#include <math.h>
#include "DigitalMaterial.h"
#include "AttributeError.h"

#ifndef SWIG
template<class Type, int N> class DMArray;
class ListOfAtoms;
#endif


class Potential : public Serializable
{
public:
	virtual double CalculateEnergy(ListOfAtoms *atoms) 
	{ 
		throw AttributeError("Potential","CalculateEnergy");
		return 0;
	}

	// XXX I'm confused about whether the array should be passed in 
	// from the outside and copied to, or whether passing a const pointer 
	// with the array is better.
	// So far, we have two conventions: the forces have an array passed in 
	// that gets filled, and the Get routines in ListOfAtoms send back 
	// the local copy. This is very bad.

	virtual void CalculateForces (ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces) 
		{ throw AttributeError("Potential","CalculateForces");}

	virtual double CalculatePairForceEnergy(double displacement[DIMENSION], double displacementSq, double *force)
		{
			throw AttributeError("Potential","CalculatePairForce");
		}

	virtual double AtomicForcesEnergy(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces, int i)
		{
			throw AttributeError("Potential","AtomicForcesEnergy");
		}

	virtual void CalculateForcesBetweenTypes(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces, int type1, int type2) 
	{ // **** the force array isn't initialized in this function
		throw AttributeError("Potential","CalculateForcesBetweenTypes");}

	virtual double CalculateEnergyAndForces (ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces) 
	{ 
		throw AttributeError("Potential","CalculateEnergyAndForces");
		return 0;
	}

	virtual double CalculateAtomicEnergy(ListOfAtoms *atoms, int i)
	{ 
		throw AttributeError("Potential","CalculateAtomicEnergy");
		return 0;
	}
	
#ifndef SWIG
	virtual void CalculateStressTensor(ListOfAtoms *atoms, int i, double stress[(DIMENSION+1)*DIMENSION/2])
		{
			throw AttributeError("Potential","CalculateStressTensor");
		}
	
	virtual void CalculateStressTensor(ListOfAtoms *atoms, DMArray<double, DIMENSION*(DIMENSION+1)/2> *stress)
		{
			throw AttributeError("Potential","CalculateStressTensor");
		}
#endif // SWIG
	virtual void CalculateHessianMatrix(ListOfAtoms *atoms, double *matrix)
		{
			throw AttributeError("Potential","CalculateHessianMatrix");
		}


	// For each potential, we need some kind of energy scale to set time constants
	// for movers and such
	// It doesn't have to be precisely defined! A crude measure is OK.
	// This should return, for a pair potential V(x) with LengthScale ell, V''(ell)
	virtual double GetEnergyCurvatureScale() const
	{
		throw AttributeError("Potential","GetEnergyCurvatureScale");
		return 0;
	}

	// Roughly returns the diameter of the atom
	// For fcc, should return 1/sqrt(2) of the lattice constant, for example
	// Currently used to set up the lattice structure: a good initial number is important
	virtual double GetLengthScale() const
	{
		throw AttributeError("Potential","GetLengthScale");
		return 0;
	}

	// XXX Shouldn't be virtual: performance hit
	// EMTPotential stores cutoff in separate class, though
	// We should fix this when we go to multiple atom types

	virtual double GetCutoffDistance2() const {return cutoffDistance2;}
	virtual double GetCutoffDistance() const {return sqrt(cutoffDistance2);}

	Potential()
	{
#if DIMENSION==2
		stresscomp[0][0] = 0; stresscomp[1][1] = 1; stresscomp[0][1] = 2; stresscomp[1][0] = 2;
#else // if DIMENSION==3
		stresscomp[0][0] = 0; stresscomp[1][1] = 1; stresscomp[2][2] = 2; 
		stresscomp[2][1] = stresscomp[1][2] = 3; stresscomp[2][0] = stresscomp[0][2] = 4; 
		stresscomp[1][0] = stresscomp[0][1] = 5;
#endif
	}
	virtual ~Potential() { /* cout << "Potential deleted" << endl; */ }

#ifndef SWIG
    void Save(DMWriter *writer) const 
    {
        writer->PutDouble(cutoffDistance2, "cutoffDistance2");
    }

	void Load(DMReader *reader)
    {
        this->cutoffDistance2 = reader->GetDouble("cutoffDistance2");
    }
#endif //SWIG
	const std::string &GetType() const
	{
		throw AttributeError("Potential","GetType");
	}

protected:
	void SetCutoffDistance(double cutoffDistance){cutoffDistance2 = cutoffDistance*cutoffDistance;}
	//! The standard mapping of the independent parts of the stress tensor to vector notation 
	int stresscomp[DIMENSION][DIMENSION];
	
	
	void CheckPotentialParameter(DMReader *reader, double parameter, const std::string parameterName)
		{
			double tmp;
			tmp = reader->GetDouble(parameterName);
			if (tmp != parameter)
			{
				std::string errorString("Value of ");
				errorString += parameterName;
				errorString += SprintString(" has changed from %f to %f", tmp, parameter);
				throw DMException(this->GetType(),"Load -> Potential::CheckPotentialParameter", errorString);
			}
		}
	

private:
	double cutoffDistance2;
};

#endif // !defined(POTENTIAL_H)
