// StillingerWeberPotential.h: interface for the StillingerWeberPotential class.
//
//////////////////////////////////////////////////////////////////////

#ifndef STILLINGERWEBERPOTENTIAL_H
#define STILLINGERWEBERPOTENTIAL_H

#include "DMFirstHeader.h"

#include "DigitalMaterial.h"
#include "Potential.h"

//! Silicon potential with three body terms developed by Stillinger and Weber.

/*!
  This potential was developed to include the directionality of bonds in
  Silicon. The three-body terms explicitly favour the angles to have 
  cosine -1/3. See Stillinger and Weber, Phys Rev B, volume 31, pp 5262-5271
  (1985).
  
  I haven't written CalculateForcesBetweenTypes yet it's now the (commented)
  one from Lennard-Jones
  
  WARNING: when using SWIG, you need to be aware of the following---the default
  string constructor argument "natural" is not handled properly by SWIG. You 
  get seg. faults. To fix it you need to interrupt the SWIG build process after
  the wrapper file has been generated, do a search on "new_Still" to take you
  to the wrapper for the constructor and 
  
  REPLACE
  std::string *arg1 = (std::string *) &"natural" ;
  WITH
  std::string *arg1 = new std::string("natural");

  Alternatively, to save you this trouble, remove the argument if you don't
  need it.

  \date January 2001
  \author Nicholas Bailey

 */

class StillingerWeberPotential : public Potential
{
public:
	StillingerWeberPotential(double BRITTLE_FACTOR = 1., std::string unitString = "natural");
	virtual ~StillingerWeberPotential();
	void Allocate(int nAtoms);
	void CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);
	void CalculateForcesBetweenTypes(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces,
									 int type1, int type2) ; 
	// the force array isn't initialized in this function

	double CalculateEnergy(ListOfAtoms *atoms);
	double CalculateEnergyAndForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);
	double CalculateAtomicEnergy(ListOfAtoms *atoms, int i);

	void CalculateStressTensor(ListOfAtoms *atoms, int i, double stress[DIMENSION*(DIMENSION+1)/2]);
	void CalculateStressTensor(ListOfAtoms *atoms, DMArray<double, DIMENSION*(DIMENSION+1)/2> *stress);

	void CalculateHessianMatrix(ListOfAtoms *atoms, double *matrix);

	double GetLengthScale() const
	{
		return 1.122462 * lengthMultiplier; // interatomic spacing in ground state---in this case
		// sqrt(3)/4 times the lattice constant (diamond lattice), which is 
		// 2^(1/6) in natural units
	}

	double GetEnergyCurvatureScale() const
	{
		return 0.; // needs to be determined.
	}

	void Save(DMWriter *writer) const;
	void Load(DMReader *reader);
	//! Get the string name of the class for saving or loading.
	const std::string &GetType() const
	{
		return className;
	}

private:
#ifndef SWIG	
	void StressTensorKernel(ListOfAtoms *atoms, int i, double stress[(DIMENSION+1)*DIMENSION/2]);
	int         nAtoms;
	double      *displacementsSq;
	double      (*displacements)[DIMENSION];
	int         *neighbors;

	// Units---an enum to specify which units arebeing used, and conversion 
	// factors between natural and other units
	enum Units{NATURAL, EV_ANGSTROMS1, EV_ANGSTROMS2} units;
	double lengthMultiplier;
	double energyMultiplier;

	// parameters for the SW potential
	const double SW_A;
	const double SW_B;
	const double SW_lambda;
	const double SW_gamma;
	const double SW_a; // same as cutoffDistance
private:
	static const std::string className;


#endif // SWIG 
};

#endif // !defined(STILLINGERWEBERPOTENTIAL_H)
