// ConfigurationShifter.cpp: implementation of the ConfigurationShifter class.
//
//////////////////////////////////////////////////////////////////////

#include "DMFirstHeader.h"
#include <math.h>
#include <vector>

#include "DigitalMaterial.h"
#include "ConfigurationShifter.h"
#include "BoundaryConditions.h"
#include "CellNeighborLocator.h"
#include "SimpleNeighborLocator.h"
#include "ListOfAtoms.h"

#include "DMLastHeader.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ConfigurationShifter::ConfigurationShifter(double shift[DIMENSION],double transWeightFactor)
{
    cellNL = new CellNeighborLocator(1.0,0.0);
    simpleNL = new SimpleNeighborLocator(1.0);
    SetShift(shift);	
    this -> nAtoms = 0;
    displacementsSq = NULL;
    neighbors = NULL;
    displacements = NULL;
    this->transWeightFactor = transWeightFactor;
}

ConfigurationShifter::~ConfigurationShifter()
{
    if(displacementsSq) delete [] displacementsSq;
    if(displacements) delete [] displacements;
    if(neighbors) delete [] neighbors;
    delete cellNL;
    delete simpleNL;
}

void ConfigurationShifter::Allocate(int nAtoms, const DMArray<double, DIMENSION> *array)
{
    this->nAtoms = nAtoms;	
    if(displacementsSq)delete [] displacementsSq;
    displacementsSq = new double[nAtoms];
    
    if(displacements) delete [] displacements;
    displacements = new double[nAtoms][DIMENSION];
    
    if(neighbors) delete [] neighbors;
    neighbors = new int[nAtoms];
}


void ConfigurationShifter::SetShift(const double shift[DIMENSION])
{
    shiftMag = 0.;
    for(int coord=0;coord<DIMENSION;++coord)
    { 
        this->shift[coord] = shift[coord];
        shiftMag += shift[coord]*shift[coord];
    }
    shiftMag = sqrt(shiftMag);
    // Had trouble with CellNeighborLocator when the cutoff was very small!
    double cutoff = (shiftMag > 1.0 ? shiftMag : 1.0);
    
    cellNL->SetCutoffDistance(cutoff);
    cellNL->SetDrift(0);
    simpleNL->SetCutoffDistance(cutoff);
}

void ConfigurationShifter::Transform(ListOfAtoms *atoms)
{
    if(atoms->GetNumber() != nAtoms) Allocate(atoms->GetNumber(),atoms->GetCartesianPositions());
    
    int i,j,coord, numNeighbors,nearest,spacesLeftInArrays;
    bool arrayNotTooFull;
    double minDistSq,parallelComp,atomDisplacement[DIMENSION],netDisplacement[DIMENSION];
    std::vector<int> atomsAlreadyUsed;
    for(coord=0;coord<DIMENSION;++coord)
        netDisplacement[coord] = 0.;
    
    DMArray<double, DIMENSION> *newPositions =
        new DMArray<double, DIMENSION>(atoms->GetCartesianPositions()->GetShape()); // temporary vector. No need to declare it sibling
    
    NeighborLocator *oldNeighborLocator = atoms->GetNeighborLocator();
    
    atoms->SetNeighborLocator(cellNL);
    //atoms->SetNeighborLocator(simpleNL);
    const BoundaryConditions *bc = atoms->GetBoundaryConditions();
    
    for(i=0;i<nAtoms;++i)
    {
        double shiftedPos[DIMENSION];
        for(coord=0;coord<DIMENSION;++coord)
            shiftedPos[coord] = atoms->GetCartesianPosition(i,coord) + shift[coord];
        bc->EnforceBoundaryConditions(shiftedPos);
        spacesLeftInArrays = nAtoms;     
        arrayNotTooFull = cellNL->Neighbors(shiftedPos,neighbors,numNeighbors, displacements,displacementsSq,spacesLeftInArrays);
        //arrayNotTooFull = simpleNL->Neighbors(shiftedPos,neighbors,numNeighbors, displacements,displacementsSq,spacesLeftInArrays);
        assert(arrayNotTooFull && numNeighbors > 0);
        // Recalculated the distances, giving a different weight to the transverse direction
        for(j=0;j<numNeighbors;++j)
        {
            parallelComp = 0.;
            for(coord=0;coord<DIMENSION;++coord)
            {
                parallelComp += displacements[j][coord] * shift[coord]/shiftMag;
            }
            double perpCompSq= displacementsSq[j] - parallelComp * parallelComp;
            displacementsSq[j] = parallelComp*parallelComp + transWeightFactor*perpCompSq;
        }
        if(numNeighbors == 1)
            nearest = neighbors[0];
        else
        {
            minDistSq = 1.0e99;  
            for(j=0;j<numNeighbors;++j)
            {
                if(displacementsSq[j] < minDistSq)
                {
                    minDistSq = displacementsSq[j];
                    nearest = neighbors[j];
                }
            }
        }
        
        for(unsigned int a=0;a<atomsAlreadyUsed.size();++a)
        {
            if(atomsAlreadyUsed[a] == nearest)
            {
                Map_error error(nearest,a,i);
                for(int coord=0;coord<DIMENSION;++coord)
                {
                    error.nearest_pos[coord] = atoms->GetCartesianPosition(nearest,coord);
                    error.a_pos[coord] = atoms->GetCartesianPosition(a,coord);
                    error.i_pos[coord] = atoms->GetCartesianPosition(i,coord);
                }
                throw error;
            }
        }
        
        atomsAlreadyUsed.push_back(nearest);
        for(coord=0;coord<DIMENSION;++coord)
        {
            atomDisplacement[coord] = shiftedPos[coord] - atoms->GetCartesianPosition(nearest,coord);
            (*newPositions)[nearest][coord] = shiftedPos[coord];
        }
        bc->DifferenceBoundaryConditions(&atomDisplacement,1);
        for(coord=0;coord<DIMENSION;++coord)
            netDisplacement[coord] += atomDisplacement[coord];
        
    }
    
    
    
    
    for(coord=0;coord<DIMENSION;++coord)
        COMShift[coord] = netDisplacement[coord]/nAtoms;
    
    atoms->SetNeighborLocator(oldNeighborLocator);
    atoms->SetCartesianPositions(newPositions);
    delete newPositions;
    
}


