// EulerCoordinateTransformer.cpp: implementation of the EulerCoordinateTransformer class.
//
//////////////////////////////////////////////////////////////////////


#include "DMFirstHeader.h"
#include "EulerCoordinateTransformer.h"
#include "DigitalMaterial.h"
#include <fstream>
#include "ElasticFieldTransformer.h"
#include "ListOfAtoms.h"
#include "BravaisLatticeWithBasis.h"
#include "DMArray.h"
#include "DMException.h"
#include "lapack.h"


#include "DMLastHeader.h"

//using namespace std;


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

EulerCoordinateTransformer::EulerCoordinateTransformer(double tolerance, BoundaryConditions *bc)
{  
	this->tolerance = tolerance;
	this->bc = bc;
	error = 0.;
	undeformedLattice = 0;
	undeformedPositions = 0;
	maxIterations = 20;
}

EulerCoordinateTransformer::~EulerCoordinateTransformer()
{

}

void EulerCoordinateTransformer::SetUndeformedPositions(const DMArray<double, DIMENSION> *currentPositions)
{
	if(currentPositions == 0)
	{
		if(undeformedPositions != 0) delete undeformedPositions;
		undeformedPositions = 0;
		return;
	}

	// get here if currentPositions is not null

	if( undeformedPositions && !(undeformedPositions->GetShape() == currentPositions->GetShape()) )
	{
		delete undeformedPositions;
		undeformedPositions = new DMArray<double, DIMENSION>(currentPositions);
	}		
	if(!undeformedPositions)
		undeformedPositions = new DMArray<double, DIMENSION>(currentPositions);



    double eulerDisplacement[DIMENSION], latticePosition[DIMENSION],guessLatticePosition[DIMENSION];
	double *position, *undefPos;
	int coord;

	DMArrayIterator<double,DIMENSION> posIter(currentPositions);
	DMArrayIterator<double,DIMENSION> undefIter(undeformedPositions);

	for(posIter.Init(),undefIter.Init(); !posIter.IsDone(); posIter.Next(),undefIter.Next())
	{
		position = posIter.Current();
		undefPos = undefIter.Current();

		for(coord=0;coord<DIMENSION;++coord)
			eulerDisplacement[coord] = 0.;
		
		std::vector<ElasticFieldTransformer*>::iterator iter;
		for(iter=strainTerms.begin();iter!=strainTerms.end();++iter)
			(*iter)->ElasticDisplacement(position, eulerDisplacement);
		// need to find the position in the undeformed lattice closest to positions[atom] - eulerDisplacement
		for(coord=0;coord<DIMENSION;++coord)
			guessLatticePosition[coord] = position[coord] - eulerDisplacement[coord];

		
		undeformedLattice->ClosestAtomCartesianCoordinates(latticePosition,guessLatticePosition);
		for( coord = 0 ; coord < DIMENSION; ++coord )
			undefPos[coord] = latticePosition[coord];

	} // end of loop over atoms positions

}	


void EulerCoordinateTransformer::Transform(ListOfAtoms *atoms)
{
	if(atoms->GetNumber() == 0) return;
	int coord, nAtoms = atoms->GetNumber();
	const int tmpInt = nAtoms;
	dmProtocol->AllReduce(&tmpInt, &nAtoms, 1, DMProtocol::SUM);
    nIterations = 0;
    double eulerDisplacement[DIMENSION],actualDisplacement[DIMENSION],difference, latticePosition[DIMENSION],guessLatticePosition[DIMENSION];
	DMArrayIterator<double,DIMENSION> posIter(atoms->GetCartesianPositions());
	double *position, *posIncrement,*undefPos;

	DMArray<double, DIMENSION> increments(atoms->GetShape());
	DMArrayIterator<double,DIMENSION> incIter(&increments);
	DMArrayIterator<double,DIMENSION> *undefIter = 0;

	if(this->undeformedPositions) undefIter = new DMArrayIterator<double,DIMENSION> (this->undeformedPositions);

	
	do
    {
		error = 0.;
		if(undefIter) undefIter->Init();

		for(posIter.Init(),incIter.Init(); !posIter.IsDone(); posIter.Next(),incIter.Next())
		{
			position = posIter.Current();
			posIncrement = incIter.Current();
			

			for(coord=0;coord<DIMENSION;++coord)
				eulerDisplacement[coord] = 0.;

			std::vector<ElasticFieldTransformer*>::iterator iter;
			for(iter=strainTerms.begin();iter!=strainTerms.end();++iter)
				(*iter)->ElasticDisplacement(position, eulerDisplacement);
			
			// need the undeformed position of this atom
	
				// figure out the most likely undeformed position---the 
				// position in the undeformed lattice closest to 
				// positions[atom] - eulerDisplacement
			for(coord=0;coord<DIMENSION;++coord)
				guessLatticePosition[coord] = position[coord] - eulerDisplacement[coord];
			
			undeformedLattice->ClosestAtomCartesianCoordinates(latticePosition,guessLatticePosition);
			
			// use stored undeformed positions if available, and check that
			// they are not too different
			if(undefIter)
			{
				undefPos = undefIter->Current();
				for(coord=0;coord<DIMENSION;++coord)
				{
					if (fabs(latticePosition[coord] - undefPos[coord]) > 1.0)
						{
							throw DMException("EulerCoordinateTransformer","Transform","bad undeformedPosition");
						}
					latticePosition[coord] = undefPos[coord];
				}
				undefIter->Next();
			}


			// need to this as an array operation. Should also have boundary conditions applied to actualDisplacement and difference. And latticePosition
			for(coord=0;coord<DIMENSION;++coord)
			{
				actualDisplacement[coord] = position[coord] - latticePosition[coord];
				difference = eulerDisplacement[coord]-actualDisplacement[coord];
				posIncrement[coord] = difference;
				error += difference*difference;
			} // end of coord loop which calculates the increments
		} // end of loop over atoms positions

		++nIterations;
		const double tmp = error;
		dmProtocol->AllReduce(&tmp, &error, 1, DMProtocol::SUM);
		error /= nAtoms;

		if(error > tolerance)
			atoms->IncrementCartesianPositions(&increments);

		if(undefIter) undefIter->Next();
    }
    while(error>tolerance && nIterations < maxIterations);


	if(undefIter) delete undefIter;
}


/*!
  This computes the matrix (1 - dF/dr)^(-1) where F is the function which gives
  a displacement given a position---this function is independent of whether
  we interpret the position as Eulerian or Lagrangian. But the matrix
  we now calculate multiplies its derivative to give the gradient wrt to 
  position when using an Eulerian interpretation. Although we apply to
  individual terms in a displacement field, this matrix must be constructed
  from their sum.

 */

void EulerCoordinateTransformer::CalculateDerivativeMatrix(const double position[DIMENSION],double derivativeMatrix[DIMENSION][DIMENSION])
{
	int coord1,coord2;
	for(coord1=0;coord1<DIMENSION;++coord1)
		for(coord2=0;coord2<DIMENSION;++coord2)
			derivativeMatrix[coord1][coord2] = 0.;
			
	// need the derivative with repect to field variable
	std::vector<ElasticFieldTransformer*>::iterator iter;
	for(iter=strainTerms.begin();iter!=strainTerms.end();++iter)
		(*iter)->GradientWRTPosition(position, derivativeMatrix);
	
	// multiply by minus 1
	for(coord1=0;coord1<DIMENSION;++coord1)
		for(coord2=0;coord2<DIMENSION;++coord2)
			derivativeMatrix[coord1][coord2] *= -1.;
	
	// add the identity matrix
	for(coord1=0;coord1<DIMENSION;++coord1) derivativeMatrix[coord1][coord1] += 1.0;
	
	// call lapack to invert the matrix
#if DIMENSION == 2
		double determinantInv = 1./(derivativeMatrix[0][0]*derivativeMatrix[1][1] - derivativeMatrix[0][1]*derivativeMatrix[1][0]);
		derivativeMatrix[0][1] *= -1.;
		derivativeMatrix[1][0] *= -1.;
		double temp = derivativeMatrix[0][0];
		derivativeMatrix[0][0] = derivativeMatrix[1][1];
		derivativeMatrix[1][1] = temp;

		for(int a=0;a<DIMENSION;++a)
			for(int b=0;b<DIMENSION;++b)
				derivativeMatrix[a][b] *= determinantInv;
#else 
		int N=DIMENSION,lwork = DIMENSION*DIMENSION,ipiv[DIMENSION],info;;
		double M[DIMENSION*DIMENSION], work[DIMENSION*DIMENSION];
		for(coord1=0;coord1<DIMENSION;++coord1)
			for(coord2=0;coord2<DIMENSION;++coord2)
				M[coord2*DIMENSION+coord1] = derivativeMatrix[coord1][coord2];
#ifdef _WIN32
		DGETRF(&N,&N,M,&N,ipiv,&info);   // LU decompose the matrix
		DGETRI(&N,M,&N,ipiv,work,&lwork,&info); // compute the inverse of the matrix from the factorization
#else
		dgetrf_(&N,&N,M,&N,ipiv,&info);   // LU decompose the matrix
		dgetri_(&N,M,&N,ipiv,work,&lwork,&info); // compute the inverse of the matrix from the factorization
#endif

		for(coord1=0;coord1<DIMENSION;++coord1)
			for(coord2=0;coord2<DIMENSION;++coord2)
				derivativeMatrix[coord1][coord2] = M[coord2*DIMENSION+coord1];
#endif

}

/*!

  The ElasticDisplacement function calculates the displacement that an atom at
  the given position would receive from the Transform function; that is, the
  iterations are performed.

 */

void EulerCoordinateTransformer::ElasticDisplacement(const double position[DIMENSION], double displacement[DIMENSION])
{
	int coord;
	double actualDisplacement[DIMENSION], difference, latticePosition[DIMENSION], guessLatticePosition[DIMENSION], localError, tempPos[DIMENSION];
	for( coord = 0; coord < DIMENSION; ++coord )
		tempPos[coord] = position[coord];
	nIterations = 0;
	
	do
	{		
		for(coord=0;coord<DIMENSION;++coord)
			displacement[coord] = 0.;
		
		std::vector<ElasticFieldTransformer*>::iterator iter;
		for(iter=strainTerms.begin();iter!=strainTerms.end();++iter)
			(*iter)->ElasticDisplacement(tempPos, displacement);
		// need to find the position in the undeformed lattice closest to positions[atom] - eulerDisplacement
		for(coord=0;coord<DIMENSION;++coord)
			guessLatticePosition[coord] = tempPos[coord] - displacement[coord];
		
		undeformedLattice->ClosestAtomCartesianCoordinates(latticePosition,guessLatticePosition);
		
		//  Should also have boundary conditions applied to actualDisplacement and difference. And latticePosition
		localError = 0.;
		for(coord=0;coord<DIMENSION;++coord)
		{
			actualDisplacement[coord] = tempPos[coord] - latticePosition[coord];
			difference = displacement[coord]-actualDisplacement[coord];
			tempPos[coord] += difference;
			localError += difference*difference;
		} // end of coord loop which calculates the increment
		
		++nIterations;
    }
    while(localError>tolerance && nIterations < maxIterations);
}

/*!
  The following funciton simply evaluates the sum of elastic fields from
  different terms at the given point.

 */

void EulerCoordinateTransformer::CurrentDisplacementField(const double position[DIMENSION], double displacement[DIMENSION])
{
	for( int coord=0;coord<DIMENSION;++coord)
		displacement[coord] = 0.;
	
	std::vector<ElasticFieldTransformer*>::iterator iter;
	for(iter=strainTerms.begin();iter!=strainTerms.end();++iter)
		(*iter)->ElasticDisplacement(position, displacement);
	
}

void EulerCoordinateTransformer::InitialDisplace(ListOfAtoms *atoms)
{
	if(atoms->GetNumber() == 0) return;
	int coord;
	double *position, *posIncrement;
    double displacement[DIMENSION];
	
	DMArrayIterator<double,DIMENSION> posIter(atoms->GetCartesianPositions());
	DMArray<double, DIMENSION> increments(atoms->GetShape());
	DMArrayIterator<double,DIMENSION> incIter(&increments);

	int i=0;
	for(posIter.Init(),incIter.Init(); !posIter.IsDone(); posIter.Next(),incIter.Next())	
	{
		position = posIter.Current();
		posIncrement = incIter.Current();
		
		ElasticDisplacement(position, displacement);

		for(coord=0;coord<DIMENSION;++coord)
		{
			posIncrement[coord] = displacement[coord];
		}
		++i;
	}
	atoms->IncrementCartesianPositions(&increments);

}

void EulerCoordinateTransformer::GradientWRTPosition(const double position[DIMENSION],  double gradientWRTPosition[DIMENSION][DIMENSION])
{
	for(int a=0;a<DIMENSION;++a)
		for(int b=0;b<DIMENSION;++b)
			gradientWRTPosition[a][b] = 0;
	
	std::vector<ElasticFieldTransformer*>::iterator iter;
	for(iter=strainTerms.begin();iter!=strainTerms.end();++iter)
		(*iter)->GradientWRTPosition(position, gradientWRTPosition);
}
