// IncrementalMinimizer.cpp: implementation of the IncrementalMinimizer class.
//
//////////////////////////////////////////////////////////////////////
#include "DigitalMaterial.h"
#include "IncrementalMinimizer.h"
#include "ListOfAtoms.h"
#include "DMArray.h"
#include "Potential.h"
#include "lapack.h"
#include "AttributeError.h"
#include "ElasticFieldTransformer.h"
#include <fstream>

#include "DMLastHeader.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

/*!

  We have two possible modes (so far) for using this class. Mode 0 is when 
  there is a fixed force increment on a subset of the atoms. Mode 1 is when one
  branch of the ListOfAtoms have constrained positions, and thos epositions are
  changed (externally) and we wish to find the new relaxed positions of the
  unconstrained atoms.

  The array forceIncrements is only used for mode 0 and the array
  positionIncrements is only used for mode 1---otherwise s null pointer
  should be passed.

  epsilon is needed if there is no calculate hessian function for the
  potential--it's the finite difference for numerically calculating it.
 */

IncrementalMinimizer::IncrementalMinimizer(int mode, Potential *potential, double epsilon)
{
	if( mode < 0 || mode > 1)
		throw DMException("IncrementalMinimizer","IncrementalMinimizer","Mode value not allowed");
	this->mode = mode;
	this->epsilon = epsilon;
	this->potential = potential;

	forceIncrements = 0;
	positionIncrements = 0;
	constrainedBranch = 0;

	forceDerivatives = 0;
	positionCorrections = 0;
	nAtoms = 0;

	verbose = false;
}

IncrementalMinimizer::~IncrementalMinimizer()
{
	if(forceDerivatives) delete forceDerivatives;
	if(positionCorrections) delete positionCorrections;
}



void IncrementalMinimizer::SetForceIncrements(const DMArray<double, DIMENSION> *forceIncrements)
{
	if (0 == mode)
	{
		this->forceIncrements = forceIncrements;
	}
	else
	{		
		this->forceIncrements = 0;
		if(verbose) cout << "Warning: SetForceIncrements attempted in mode " << mode << endl;
	}
}

void IncrementalMinimizer::SetPositionIncrements(int constrainedBranch, const DMArray<double, DIMENSION> *positionIncrements)
{
	if(1 == mode)
	{
		this->constrainedBranch = constrainedBranch;
		this->positionIncrements = positionIncrements;
	}
	else
	{
		this->constrainedBranch = 0;
		this->positionIncrements = 0;
		if(verbose) cout << "Warning: SetPositionIncrements attempted in mode " << mode << endl;
	}
}

void IncrementalMinimizer::Allocate(int nAtoms)
{
	if(verbose)
		cout << "In IncrementalMinimizer::Allocate" << endl;
	// allocate space for the matrix and the column vector
	this->nAtoms = nAtoms;

	if(forceDerivatives) delete forceDerivatives;
	if(positionCorrections) delete positionCorrections;
	// in future need to allow for constraints; will operate on independent DOFs
	forceDerivatives = new double[nAtoms * nAtoms * DIMENSION * DIMENSION];
	positionCorrections = new double[nAtoms * DIMENSION];
	
}

void IncrementalMinimizer::Transform(ListOfAtoms *atoms)
{
	if(atoms->GetNumber() != nAtoms) Allocate(atoms->GetNumber());

	int i, j, coord, index = 0;
	int matrixSize = 0; // set below
	double *arrayForLinearSolver = 0; // will say which array to pass to Lapack
	double *subHessian = 0; // needed for mode 1--will be the 'unconstrained' submatrix of the hessian
	int beginConstrained = 0, endConstrained = 0; // also for mode 1

	// first calculate the Hessian
	try
	{
		// works if CalculateHessian has been implemented for that potential.
		potential->CalculateHessianMatrix(atoms, forceDerivatives);
	}
	catch (AttributeError)
	{
		// Otherwise do numerical differentiation to calculate the hessian
		if(verbose) cout << "No analytic hessian available; performing numerical differentiation" << endl;
		CalculateNumericalHessian(atoms);
		if(verbose) cout << "Numerical Hessian done" << endl;
	}


	if(0 == mode) // BEGIN MODE 0 SPECIFIC SECTION
	{
		ASSERT(forceIncrements);
		if(!forceIncrements)
			throw DMException("IncrementalMinimizer","Transform","forceIncrements has not been set");
			
		ASSERT(forceIncrements->GetShape() == atoms->GetShape());
		if(!(forceIncrements->GetShape() == atoms->GetShape()))
			throw DMException("IncrementalMinimizer","Transform","forceIncrements has different shape to ListOfAtoms");
		matrixSize = nAtoms * DIMENSION;
		DMArrayIterator<double, DIMENSION> forceIncIter(forceIncrements);
		for( forceIncIter.Init(); !forceIncIter.IsDone(); forceIncIter.Next()) 
		{
			double *current = forceIncIter.Current();
			for( coord = 0; coord < DIMENSION; ++coord )
				positionCorrections[index++] = current[coord];
		}

		arrayForLinearSolver = forceDerivatives;
	} // END MODE 0 SPECIFIC SECTION


	else if (1 == mode) // BEGIN MODE 1 SPECIFIC SECTION
	{
		ASSERT(positionIncrements);
		if(!positionIncrements)
			throw DMException("IncrementalMinimizer","Transform","positionIncrements has not been set");
		ASSERT(positionIncrements->GetShape() == atoms->GetBranch(constrainedBranch)->GetShape());
		if(!(positionIncrements->GetShape() == atoms->GetBranch(constrainedBranch)->GetShape()))
			throw DMException("IncrementalMinimizer","Transform","positionIncrements has different shape to ListOfAtoms");
		matrixSize = DIMENSION * (nAtoms - atoms->GetBranch(constrainedBranch)->GetNumber());
		// allocate for submatrix of hessian		
		int totalArraySize = matrixSize * matrixSize;
		subHessian = new double[totalArraySize];
		
		// zero the right hand side array
		for( index = 0; index < matrixSize; ++index )
			positionCorrections[index] = 0.;
		
		// identify limits corresponding to constrained atoms
		int br = 0;
		while(br != constrainedBranch)
			beginConstrained += atoms->GetBranch(br++)->GetNumber();
		endConstrained = beginConstrained + atoms->GetBranch(constrainedBranch)->GetNumber();
		// copy elements of hessian into submatrix that aren't part of the constrained branch
		int subIndex = 0, mainIndex = 0, unConstrainedI_Index = 0, iCoord, jCoord;
		DMArrayIterator<double, DIMENSION> posIncIter(positionIncrements);
		for( i = 0; i < nAtoms; ++i )
		{
			bool iConstrained = (i >= beginConstrained && i < endConstrained);
			if(!iConstrained)
			{
				for( iCoord = 0; iCoord < DIMENSION; ++iCoord)
				{
					// loop over all atoms j. At the same time start the 
					// iterator looping over positionIncrements so that for
					// constrained  values of j we can access the correct 
					// items of that array
					for(j = 0, posIncIter.Init(); j < nAtoms; ++j )
					{
						bool jConstrained = (j >= beginConstrained && j < endConstrained);
						
						if( !jConstrained ) // if in the main submatrix
						{
							for( jCoord = 0; jCoord < DIMENSION; ++jCoord)
							{
								subHessian[subIndex] = forceDerivatives[mainIndex];
								++subIndex;
								++mainIndex;
							}
						}
						// the following case increments mainIndex but not subIndex
						else  //  i not constrained and j is constrained
						{
							// in this case we can add to the right hand
							// side, which  uses elements of hessian 
							// connecting unconstrained (i) with constrained
							// (j) DOFs. Corresponding to the sum over 
							// constrained j is a loop over the elements of
							// the  positionIncrements array, because these
							// are to multiply the corresponding elements of
							// the hessian.
							double *current = posIncIter.Current();
							for( jCoord = 0; jCoord < DIMENSION; ++jCoord )
								positionCorrections[unConstrainedI_Index] += forceDerivatives[mainIndex++] * current[jCoord];
							posIncIter.Next();
						} // end of if( !jConstrained ) ... else ... 
					} // end loop over atoms j
					++unConstrainedI_Index;
				} // end loop over iCoord
			} 
			else // ie iConstrained is true. These values of i are not useful.
				mainIndex += nAtoms * DIMENSION * DIMENSION; 
		} // end loop over of atoms i
		ASSERT(subIndex == totalArraySize);
		ASSERT(mainIndex == nAtoms * nAtoms * DIMENSION * DIMENSION);
		
		arrayForLinearSolver = subHessian; // avoid copying back this way (we set this pointer to whichever array and pass it to the Lapack routine)
	}  // END MODE 1 SPECIFIC SECTION
	
	// Call the Lapack routine that solves for the corrections
	SolveLinearSystem(arrayForLinearSolver, positionCorrections, matrixSize);

	
	if(1 == mode)
		delete [] subHessian; // I may want to put the allocation and freeing of this array somewhere else

	// Now add the position corrections to the appropriate subset of atoms
/*	ListOfAtoms *toBeAdjusted;
	if (0 == mode) toBeAdjusted = atoms;
	else toBeAdjusted = atoms->GetBranch(0);
*/

	DMArray<double, DIMENSION> *incrementsDMArray = new DMArray<double, DIMENSION>(atoms->GetShape());
	DMArrayIterator<double, DIMENSION> corrIter(incrementsDMArray);
	
	index = 0;
	int atomIndex =0;
	for( corrIter.Init(); !corrIter.IsDone(); corrIter.Next() )
	{
		double *current = corrIter.Current();
		if(1 == mode && atomIndex >= beginConstrained && atomIndex < endConstrained)
		{
			for( coord = 0; coord < DIMENSION; ++coord)
				// index doesn't increase!
				current[coord] = 0.;
		}
		else{
			for( coord = 0; coord < DIMENSION; ++coord, ++index )
				current[coord] = - positionCorrections[index];
			// the minus sign is part of the equation that we are solving--we could have put it elsewhere
		}
		
		++atomIndex;
	}
	ASSERT(index == matrixSize);
	
	atoms->IncrementCartesianPositions(incrementsDMArray);	
	delete incrementsDMArray;
}

void IncrementalMinimizer::SolveLinearSystem(double *arrayForLinearSolver, double *positionCorrections, int matrixSize)
{
	// make Lapack call to solve for positionCorrections
	int nrhs = 1;
	int info;
	int rank;
	double rCond = 1.e-12;
	double *singularValues = new double[matrixSize];
	//int lwork = 10 * size; // lapack documentation for dgelss has a formula for minimum work length which works out as 5*size in this case
	int lwork = 200 * matrixSize; //  documentation that I found for dgelsd has a formula for minimum work length which works out as 650*size in this case, as far as I could tell without knowing SMLSIZ which is returned by ILAENV
	int liwork = 40 * matrixSize; // documentation seems to indicate around 30*size is enough
	double *work = new double[lwork];
	int *iwork = new int[liwork];

	fstream matrixFile("matrix.dat", ios::out);
	int t;
	for (t = 0; t< matrixSize*matrixSize; ++t )
		matrixFile << arrayForLinearSolver[t] << endl;

	fstream rhsFile("rhs.dat", ios::out);
	for (t = 0; t < matrixSize; ++t )
		rhsFile << positionCorrections[t] << endl;
	cout << "calling dgelsd" << endl;
	// maybe try using dgelsd, which is apparently significantly faster for large matrices--make sure a newer version of lapack is available!
	dgelsd_(&matrixSize, &matrixSize, &nrhs, arrayForLinearSolver, &matrixSize, positionCorrections, &matrixSize, singularValues, &rCond, &rank, work, &lwork, iwork, &info);
	cout << "end dgelsd" << endl;
	

	// the following is an older, slower lapack function. dgelsd is not 
	// available in all lapack releases so I'm leaving htis here commented out
	// dgelss_(&matrixSize, &matrixSize, &nrhs, arrayForLinearSolver, &matrixSize, positionCorrections, &matrixSize, singularValues, &rCond, &rank, work, &lwork, &info);
	
	
	ASSERT(info == 0);
	if(info < 0) throw DMException("IncrementalMinimizer","Transform",SprintString("Lapack function dgelss has bad argument %d",info));
	if(info > 0) throw DMException("IncrementalMinimizer","Transform","Lapack dgelss algorithm failed to converge");

	if(verbose)
	{
		cout << "Maximum eigenvalue " << singularValues[0] << endl;
		cout << "Minimum eigenvalue " << singularValues[rank-1] << endl;
		cout << "Rank: " << rank << endl;
		cout << "Optimal value of lwork: " << work[0] << " (used " << lwork << ")" << endl;
	}

	delete [] singularValues;
	delete [] work;
	delete [] iwork;	
}


void IncrementalMinimizer::CalculateNumericalHessian(ListOfAtoms *atoms)
{
	int i, j, coord, coord2, index;
	
	DMArray<double, DIMENSION> *forcesPlus = new DMArray<double, DIMENSION> (atoms->GetShape());
	DMArray<double, DIMENSION> *forcesMinus = new DMArray<double, DIMENSION> (atoms->GetShape());

	// fill forceDerivatives by looping over positions, moving them, computing numerical derivative etc.

	index = 0;
	for( i = 0; i < nAtoms; ++i )
		for( coord = 0; coord < DIMENSION; ++coord )
		{
			// 'i'th derivative of the forces
			atoms->IncrementCartesianPosition(i,coord,epsilon);
			
			potential->CalculateForces(atoms,forcesPlus);
			
			// now subtract 2*epsilon to get the original value minus epsilon
			atoms->IncrementCartesianPosition(i,coord,-2*epsilon); 
			
			potential->CalculateForces(atoms,forcesMinus);
			
			for( j = 0; j < nAtoms; ++j )
				for( coord2 = 0; coord2 < DIMENSION; ++coord2 )
				{
					forceDerivatives[index++] = 0.5*((*forcesPlus)[j][coord2] - (*forcesMinus)[j][coord2])/epsilon;
				}
			
			// Now add epsilon to get back the original value
			atoms->IncrementCartesianPosition(i,coord,epsilon);	
		}
	delete forcesPlus;
	delete forcesMinus;
}
