// OverlapPruner.cpp: implementation of the OverlapPruner class.
//
//////////////////////////////////////////////////////////////////////

#include "DMFirstHeader.h"
#include <vector>
#include <set>

#include "DigitalMaterial.h"

#include "OverlapPruner.h"
#include "CellNeighborList.h"
#include "ListOfAtoms.h"
#include "BoundaryConditions.h"
#include "DMException.h"

#include "DMLastHeader.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

OverlapPruner::OverlapPruner(double minSeparation, const DMArray<double, DIMENSION> *fixed, double *metric):
    minSeparation(minSeparation), fixed(fixed)
{
	cellSize = 0.;
	this->metric = new double[DIMENSION * DIMENSION];
	memset(this->metric,0,DIMENSION*DIMENSION*sizeof(double));
	if(metric)
	{
		for(int coord1 = 0; coord1 < DIMENSION; ++coord1)
			for(int coord2 = 0; coord2 < DIMENSION; ++coord2)
			{
				this->metric[coord1 + DIMENSION*coord2] = metric[coord1 + DIMENSION*coord2];
				ASSERT(metric[coord1 + DIMENSION*coord2] == metric[coord2 + DIMENSION*coord1]);
				if(metric[coord1 + DIMENSION*coord2] != metric[coord2 + DIMENSION*coord1])
					throw DMException("OverlapPruner","OverlapPruner","non-symmetric metric passed");}
	}
	else
		for(int coord = 0; coord < DIMENSION; ++coord)
			this->metric[coord + DIMENSION*coord] = 1.0;
}

OverlapPruner::~OverlapPruner()
{
}

void OverlapPruner::SetMinSeparation(double minSeparation)
{
	this->minSeparation = minSeparation;
}

void OverlapPruner::SetCellSize(double cellSize)
{
	this->cellSize = cellSize;
}

void OverlapPruner::Transform(ListOfAtoms *atoms)
{
	if (fixed) RemoveToCloseAtoms(atoms);
	else SelfPrune(atoms);
}

void OverlapPruner::RemoveToCloseAtoms(ListOfAtoms *atoms)
{
	double drift = (cellSize > 2*minSeparation ? cellSize - minSeparation : 4.*minSeparation);
	NeighborLocator *tempNList = new CellNeighborLocator(minSeparation*2., drift);
	tempNList->SetBoundaryConditions(atoms->GetBoundaryConditions());
	tempNList->SetPositionArray(atoms->GetCartesianPositions());
	DMArrayIterator<double, DIMENSION> iter(fixed);
	int nAtoms = atoms->GetNumber();
	int spacesLeftInArrays, numNeighbors;
	int *neighbors = new int[nAtoms];
	double (*displacements)[DIMENSION] = new double [nAtoms][DIMENSION];
	double *displacementsSquared = new double[nAtoms];
	atomsRemoved.clear();
	for(iter.Init(); !iter.IsDone(); iter.Next())
	{
		spacesLeftInArrays = nAtoms;
		tempNList->Neighbors(iter.Current(), neighbors, numNeighbors, displacements, displacementsSquared, spacesLeftInArrays);
		atomsRemoved.insert(neighbors, neighbors+numNeighbors);
	}
	std::cout << "Remove "<< atomsRemoved.size() << " atoms" << std::endl;
	atoms->RemoveAtoms(atomsRemoved);
	delete [] neighbors;
	delete [] displacements;
	delete [] displacementsSquared;
	delete tempNList;
}

void OverlapPruner::SelfPrune(ListOfAtoms *atoms)
{	
	const BoundaryConditions *bc = atoms->GetBoundaryConditions();
	int i,j, coord, coord2, pairsTooClose, pairmin, atom1, atom2;
	double distsq;
	double drift = (cellSize > minSeparation ? cellSize - minSeparation : 0.5*minSeparation);
	const DMArray<double, DIMENSION> *positions = atoms->GetCartesianPositions();
	int nAtoms = atoms->GetNumber();
	NeighborLocator *tempNList = new CellNeighborList(minSeparation, drift);
	// The first has to be bigger than the distance criterion, while the second will determine the size of the cells. It should not be too small because we'll need to allocate many cells in the neighbor locator
	tempNList->SetBoundaryConditions(atoms->GetBoundaryConditions());
	tempNList->SetPositionArray(atoms->GetCartesianPositions());
	int spaceLeftInArray, numberOfHalfNeighbors;
	int *neighbors = new int[nAtoms];
	double (*displacement)[DIMENSION] = new double [nAtoms][DIMENSION];
	double *distSQ = new double[nAtoms];
	bool alldone = false, *touched = new bool[nAtoms];
//	bool newAtom;
	std::vector<int> *ListOfMerged = new std::vector<int> [nAtoms];  // stores the indices of the atoms being considered the same.
     // Only the atom with the smallest index stores the list. The others point to it via belongsTo.
	int *belongsTo = new int[nAtoms];
	int maxi, mini;
	struct tooClose   // stores the distance and the indices of the atoms forming a too close pair
	{
		double distsq; 
		int at1, at2;
	};

	int maxPairs = 5*nAtoms;
	tooClose *toBeMoved = new tooClose[maxPairs];
	for(i=0;i<nAtoms;i++)
	{
		belongsTo[i] = i;
		ListOfMerged[i].push_back(i);
	}

	while (!alldone)
	{
		tempNList->UpdatePosition();
		pairsTooClose = 0;
	    for(i=0;i<nAtoms;i++)
		{
			spaceLeftInArray = nAtoms;

			bool ok = tempNList->HalfNeighbors(i,neighbors,numberOfHalfNeighbors,displacement,distSQ, spaceLeftInArray);
			assert(ok);
//						printf("%d has %d neighbors\n",i,numberOfHalfNeighbors);
//						for(j=0; j<numberOfHalfNeighbors; j++) std::cout << distSQ[j] << " " ;
//						std::cout << std::endl;
			for(j=0; j<numberOfHalfNeighbors; j++)
			{
				distsq = 0.;
				for( coord = 0; coord < DIMENSION; ++coord )
					for( coord2 = 0; coord2 < DIMENSION; ++coord2 )
						distsq += displacement[j][coord] * metric[coord + DIMENSION * coord2] * displacement[j][coord2];
				//distsq = distSQ[j];
				if((distsq <= minSeparation*minSeparation) && (belongsTo[i] != belongsTo[neighbors[j]]))
				{
					toBeMoved[pairsTooClose].distsq = distsq;
					toBeMoved[pairsTooClose].at1 = i;toBeMoved[pairsTooClose].at2 = neighbors[j];
					pairsTooClose++;
					if(pairsTooClose == maxPairs)  //  in the unlikely event that we haven't allocated enough slots
					{
						tooClose *tmp = new tooClose[2*maxPairs];
						memcpy(tmp, toBeMoved, maxPairs*sizeof(tooClose));
						delete [] toBeMoved;
						toBeMoved = tmp;
						maxPairs *= 2;
					}
				}
			}
		}
//	    	        		printf("%d pairs too close\n", pairsTooClose);
		if (pairsTooClose==0) alldone = true;

		for(i=0;i<nAtoms;i++)  touched[i] = false;
		for(i=0;i<pairsTooClose;i++)
		{
			distsq = toBeMoved[i].distsq;
			pairmin = i;   // look for the smallest distance (and put element i at its place)
			for(j=i+1;j<pairsTooClose;j++) 
			if (toBeMoved[j].distsq < distsq) 
			{
				distsq = toBeMoved[j].distsq;
                pairmin = j;
			}
	        atom1 = toBeMoved[pairmin].at1; atom2 = toBeMoved[pairmin].at2;
			toBeMoved[pairmin].distsq = toBeMoved[i].distsq;
			toBeMoved[pairmin].at1 = toBeMoved[i].at1;
			toBeMoved[pairmin].at2 = toBeMoved[i].at2;

			if (!touched[belongsTo[atom1]] && !touched[belongsTo[atom2]])
			{		 
				mini = belongsTo[atom1];
				maxi = belongsTo[atom2];
				if(mini > maxi)
				{
					int tmp = maxi;
					maxi = mini; mini = tmp;
				}
				for(unsigned int it=0;it<ListOfMerged[maxi].size();it++){
					ListOfMerged[mini].push_back(ListOfMerged[maxi][it]);
					belongsTo[ListOfMerged[maxi][it]] = mini;
				}
				ListOfMerged[maxi].clear();
				touched[mini] = true;
				double dr[DIMENSION];

				for(coord = 0;coord<DIMENSION;coord++)
				      dr[coord] = (*positions)[atom1][coord] - (*positions)[atom2][coord];
				bc->DifferenceBoundaryConditions(&dr,1);
				for(coord = 0;coord<DIMENSION;coord++)
				{       
                  	double meanx;
//					meanx = 0.5*(positions[atom1][coord] + positions[atom2][coord]); do not use it like that because of the boundary conditions
					meanx = (*positions)[atom1][coord] - 0.5*dr[coord];
					for(unsigned int it=0;it<ListOfMerged[mini].size();it++)
						atoms->SetCartesianPosition(ListOfMerged[mini][it],coord, meanx);
				}				 
			}
		}
	}

	atomsRemoved.clear();
    for(i=0;i<nAtoms;i++) if(ListOfMerged[i].size() == 0) atomsRemoved.insert(i);
	atoms->RemoveAtoms(atomsRemoved);

	delete [] touched;
	delete [] neighbors;
	delete [] displacement;
	delete [] toBeMoved;
	delete [] belongsTo;
	delete [] ListOfMerged;
	delete tempNList;
	delete [] distSQ;
}
