// OverlapPruner.h: interface for the OverlapPruner class.
//
//////////////////////////////////////////////////////////////////////

#ifndef OVERLAPPRUNER_H
#define OVERLAPPRUNER_H

#include "DMFirstHeader.h"
#include <vector>

#include "Transformer.h"

/**
The duty of this class is to remove overlapping atoms when Transform(atoms) is called.
There are two possibilities:
1. If no fixed atoms (from a different list of atoms) are specified then we take all pairs 
of atoms closer than minSeparation and replaces the atoms by one at the mean position.
2. If we specify some fixed atoms (again a different list of atoms) then we remove all atoms that are
closer than minSeparation to any fixed position.

Note that in order to determine the pairs too close, we use a CellNeighborLocator. If the cutoff distance 
is very small, then the cells will we small too and we'll have problems of allocating too many cells.
This is why we can set cellSize by hand. 
\author Thierry Cretegny and Nick Bailey
\date May 2000
\ingroup moleculardynamics
**/

class ListOfAtoms;
#ifndef SWIG
template<class Type, int N> class DMArray ;
#endif

class OverlapPruner : public Transformer  
{
public:
	OverlapPruner(double minSeparation, const DMArray<double, DIMENSION> *fixed=NULL, double *metric = 0);
	virtual ~OverlapPruner();
	void SetMinSeparation(double minSeparation);
	double GetMinSeparation() {return minSeparation;}
	void SetCellSize(double cellSize);
	double GetCellSize() {return cellSize;}
	void Transform(ListOfAtoms *atoms);
#ifndef SWIG
	const std::set<int> &GetAtomsRemoved() {return atomsRemoved;}
private:
/**The principle of the SelfPruner is to treat the pairs by increasing separation. 
All pairs are treated once before an averaged atome is treated again.
Atoms are really removed only at the end, such that the number of atoms in the list doesn't change 
during the calculation of the list of neighbors.
**/
	void SelfPrune(ListOfAtoms *atoms);
	void RemoveToCloseAtoms(ListOfAtoms *atoms);
	double minSeparation;
	double cellSize;
	const DMArray<double, DIMENSION> *fixed;
	std::set<int> atomsRemoved;
	double *metric;
#endif //SWIG
};

#endif // !defined(OVERLAPPRUNER_H)
