// RectangleAtomsLocator.cpp: implementation of the RectangleAtomsLocator class.
//
//////////////////////////////////////////////////////////////////////

#include "DigitalMaterial.h"
#include "RectangleAtomsLocator.h"
#include "ListOfAtoms.h"
#include "NeighborLocator.h"

#include "DMLastHeader.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

RectangleAtomsLocator::RectangleAtomsLocator(double lower[DIMENSION], double upper[DIMENSION], bool bounded[DIMENSION])
{


	for( int coord = 0; coord < DIMENSION; ++coord )
	{
		this->lower[coord] = lower[coord];
		this->upper[coord] = upper[coord];
		this->bounded[coord] = bounded[coord];	
	}
	
	CalculateCenterRadius();

	neighbors       = 0;
	displacements   = 0;
	displacementsSq = 0;
}

RectangleAtomsLocator::~RectangleAtomsLocator()
{

}


void RectangleAtomsLocator::CalculateCenterRadius()
{
	radius = 0.;

	for( int coord = 0; coord < DIMENSION; ++coord )
	{
		this->center[coord] = 0.5*(lower[coord] + upper[coord]);
		double diff = 0.5 * (upper[coord] - lower[coord]);
		radius += diff * diff;		
	}

	radius = sqrt(radius);
}

void RectangleAtomsLocator::AllocateForNeighborLocator(int nSpaces)
{
	// nSpaces is playing the role of EMTPotential's arrayLim here
	this->nSpaces = nSpaces;
	
	if(neighbors) 
	{	
		delete [] neighbors;
		delete [] displacements;
		delete [] displacementsSq;
	}

	if(nSpaces > 0)
	{
		neighbors       = new (std::nothrow) int[nSpaces];
		displacements   = new (std::nothrow) double[nSpaces][DIMENSION];
		displacementsSq = new (std::nothrow) double[nSpaces];
		
		ASSERT(neighbors);
		ASSERT(displacements);
		ASSERT(displacementsSq);
	}
	else
	{
		neighbors       = 0;
		displacements   = 0;
		displacementsSq = 0;
	}
}



const std::vector<int> &RectangleAtomsLocator::GetAtomNumbers(ListOfAtoms *atoms)
{
	listOfAtomNumbers.clear();
	
	int  numNeighbors, spacesLeftInArrays;
    bool arrayNotTooFull;
	int nAtoms = atoms->GetNumber();
	AllocateForNeighborLocator(nAtoms);	
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();


	spacesLeftInArrays = nSpaces;     
	arrayNotTooFull = neighborLocator->Neighbors(center, neighbors, numNeighbors, displacements, displacementsSq, spacesLeftInArrays, radius);
	
	while(!arrayNotTooFull)
	{
		AllocateForNeighborLocator(2*nSpaces);
		spacesLeftInArrays = nSpaces; // nSpaces has been doubled
		arrayNotTooFull = neighborLocator->Neighbors(center,neighbors, numNeighbors, displacements, displacementsSq, spacesLeftInArrays, radius);
	}

	for(int i = 0; i < numNeighbors; ++i)
	{
		double position[DIMENSION];
		for( int coord = 0; coord < DIMENSION; ++coord )
			position[coord] = atoms->GetCartesianPosition(neighbors[i],coord);
		
		if (Inside(position))
			listOfAtomNumbers.push_back(neighbors[i]);
	}
	return listOfAtomNumbers;
}

bool RectangleAtomsLocator::Inside(const double *position)
{
	bool inside = true;

	for( int coord = 0; coord < DIMENSION; ++coord )
	{
		if(bounded[coord] && (position[coord] < lower[coord] || position[coord] > upper[coord]))
			inside = false;
	}
	return inside;
}
