// SubsetElasticTransformer.cpp: implementation of the SubsetElasticTransformer class.
//
//////////////////////////////////////////////////////////////////////

#include "DigitalMaterial.h"
#include "SubsetElasticTransformer.h"
#include "AtomsLocator.h"

#include "DMLastHeader.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

SubsetElasticTransformer::SubsetElasticTransformer(ElasticFieldTransformer *transformerToApply, AtomsLocator *locator)
{
	this->transformerToApply = transformerToApply;
	this->locator = locator;
}

SubsetElasticTransformer::~SubsetElasticTransformer()
{

}


/*! The Transform function makes a copy of the positions, iterates over the 
  atom numbers in listOfAtomNumbers, gets the position of that atom in the
  copied array, calculates the displacement, and increments
  the position (in the copy). Then we use SetCartesianPositions to change
  the positions in the LOA. I'm assuming this is faster than using
  IncrementCartesianPosition for each atom to be moved--the neighborLocator is 
  called just once, but all atoms outside the subset get copied over 
  identically

  
  It would be great if we could set up a mechanism for
  IncrementCartesianPositions to work on a subset of the LOA, perhaps by 
  creating an LOA which is essentially a reference into the given one.


 */

void SubsetElasticTransformer::Transform(ListOfAtoms *atoms)
{
	std::vector<int>::const_iterator atomIter;
	const std::vector<int> listOfAtomNumbers = locator->GetAtomNumbers(atoms);
	
	DMArray<double, DIMENSION> positions(atoms->GetShape());
	positions.Copy(atoms->GetCartesianPositions());

	int coord;
	for (atomIter = listOfAtomNumbers.begin(); atomIter != listOfAtomNumbers.end(); atomIter++ )
	{
		const int atom  = *atomIter;
		
		double *position = positions[atom], displacement[DIMENSION];
		
        for( coord=0;coord<DIMENSION;++coord )
            displacement[coord] = 0.;
		
        transformerToApply->ElasticDisplacement(position, displacement);
		
		for(coord=0;coord<DIMENSION;++coord)
			position[coord] += displacement[coord];
		
	}
	atoms->SetCartesianPositions(&positions);
}


/*!
  The elastic displacement function checks whether the position is within the 
  region defined by the AtomsLocator and if so passes the request onto the
  ElasticDisplacement function of the real Transformer

 */

void SubsetElasticTransformer::ElasticDisplacement(const double position[DIMENSION], double displacement[DIMENSION])
{
	if(locator->Inside(position))
		transformerToApply->ElasticDisplacement(position, displacement);
}

/*!
  The GradientWRTPosition function checks whether the position is within the 
  region defined by the AtomsLocator and if so passes the request onto the
  GradientWRTPosition function of the real Transformer

 */

void SubsetElasticTransformer::GradientWRTPosition(const double position[DIMENSION],  double gradientWRTPosition[DIMENSION][DIMENSION])
{
	if(locator->Inside(position))
		transformerToApply->GradientWRTPosition(position, gradientWRTPosition);
}
