// RasMolObserverCPP.cpp: implementation of the RasMolObserverCPP class.
//
//////////////////////////////////////////////////////////////////////

#include "DMFirstHeader.h"
#include <string>
#include <fstream>

#include "DigitalMaterial.h"

#include "RasMolObserver.h"
#include "ListOfAtoms.h"
#include "SortAtoms.h"
#include "DMArray.h"
#include "LeafToLeafIterator.h"



#include "DMLastHeader.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

/*!	Opens a pipe to the RasMol program and sets initial parameters.

	\param radius Radius of each atom.
	\param colorMethod class which provides a coloring function

	*/


RasMolObserverCPP::RasMolObserverCPP(int radius,ColorMethod *colorMethod) 
{

	char command[256];

	sprintf(command, "rasmol");
	pipe_file = popen(command,"w");
	
	this->radius = radius;
	this->colorMethod = colorMethod;
	this->colorScale = 1.;
	this->lengthScale = 1.;
}



/*!	Closes the pipe to the RasMol program.
	*/


RasMolObserverCPP::~RasMolObserverCPP()
{
	pclose(pipe_file);
}

/*!	Sends ListOfAtoms positions to the RasMol program.

	\param atoms Must be a pointer to a ListOfAtoms.
	*/

void RasMolObserverCPP::Update(ListOfAtoms *atoms)
{

    DMArrayIterator<double, DIMENSION> positionIter(atoms->GetCartesianPositions());

	fstream rasmolFile("atomConfig.rm",ios::out);
	rasmolFile.setf(ios::fixed,ios::floatfield);
	int i;
	double aveColorValue = 0.,maxColorValue = 0.;
	if(colorMethod)
		colorMethod->Update(atoms);

	for(positionIter.Init(),i=0; !positionIter.IsDone(); positionIter.Next(),++i)
	{
		double *position = positionIter.Current();

		double x =  position[0] / lengthScale;
		double y =  position[1] / lengthScale;
		double z;
		
		if(DIMENSION > 2)
			z = position[2] / lengthScale;
		else z = 0.;

		double color;
		if(colorMethod)
		{
			double value = fabs((*colorMethod)(atoms,i));
			aveColorValue += value;
			if(value > maxColorValue)
				maxColorValue = value;
			color = 1000 * value/colorScale;
		}
		else
		{
			int leaf = positionIter.GetIndex().leaf;
			std::map<int, int >::const_iterator colorToUse = colorMap.find(leaf);
			if(colorToUse == colorMap.end())
				color = 0;
			else 
				color = colorToUse->second;
		}
		
		rasmolFile << "ATOM  ";
		rasmolFile.width(5);
		rasmolFile << i%100000 << "  NN               ";
		rasmolFile.precision(3);
		rasmolFile.width(8);
		rasmolFile << x;
		rasmolFile.width(8);
		rasmolFile << y;
		rasmolFile.width(8);
		rasmolFile << z;
		rasmolFile << "  0.00  "; // "occupancy factor"
		rasmolFile.width(8);
		rasmolFile << color << endl; 
	}	
	
	fprintf(pipe_file, "zap\n");
	fflush( pipe_file );	
	
	fprintf(pipe_file, "load atomConfig.rm\n");	
	fflush( pipe_file );	
	
	

	fprintf( pipe_file, "color temperature\n");	
	fprintf( pipe_file, "spacefill %d\n",radius);	
	fflush( pipe_file );	
	
	colorScale = maxColorValue;	

}


void RasMolObserverCPP::Plot(ListOfAtoms *atoms)
{
    Update(atoms);
}

