"""GeoFmt.py provides a Python implementation of the GeoFmt geometry 
format defined by the CPTC project.  I/O for GeoFmt objects is provided
by the GeoFmtIO module."""
import Numeric

PATCHTYPE_TRIANGLE = 0
PATCHTYPE_QUAD = 1

def TRI_INDX(i,j,k):      return (((i)+(k))*((i)+(k)+1)/2+(i))
def QUAD_INDX(i,j,ord_u): return ((j)*((ord_u)+1)+(i))

# ------------------------------------------------------------------------

class Cptc_GPoint:
    
    def __init__(self, coord=Numeric.array([0.,0.,0.])):
	self.coord = Numeric.array(coord)
	self.dim   = len(coord)
	return None

    def __repr__(self):
	return '\tCptc_GPoint(%s)\n' % (self.coord)

    def __add__(self, other):
	return Cptc_GPoint(self.coord + other.coord)

    def __sub__(self, other):
	return Cptc_GPoint(self.coord - other.coord)

    def __getattr__(self, attr):
	return apply(getattr(self.coord, attr), (self, other))


Point = Cptc_GPoint

# ------------------------------------------------------------------------

class Cptc_GCurve:

    def __init__(self, order=0, control_pnt_indx=Numeric.array([], 'Float')):
	self.order = order
	self.control_pnt_indx = control_pnt_indx
	return None

    def __repr__(self):
	s = 'Cptc_GCurve(order = %d, ...)\n' % (self.order)
        s = s + '\t{\n'
        s = s + '\t   control_pnt_indx = %s\n' % (self.control_pnt_indx)
        s = s + '\t}\n\t'
        return s

Curve = Cptc_GCurve

# ------------------------------------------------------------------------

class Cptc_GPatch:

    def __init__(self, type=0, order_u=0, order_v=0, \
		 control_pnt_indx=Numeric.array([], 'Float')):
        self.type = type
        self.order_u = order_u
        self.order_v = order_v
        self.control_pnt_indx = control_pnt_indx
	return None

    def __repr__(self):
        s = 'Patch(type = %d; order_u = %d, order_v = %d, ...)\n' % \
            (self.type, self.order_u, self.order_v)
        s = s + '\t{\n'
        s = s + '\t   control_pnt_indx = %s\n' % (self.control_pnt_indx)
        s = s + '\t}\n\t'
        return s

Patch = Cptc_GPatch

# ------------------------------------------------------------------------

class TopoElement:
    
    def update_properties(self):
	self.properties = {}
	for i in range(len(self.props)):
	    self.properties[self.props[i]] = self.values[i]

class Cptc_TVertex(TopoElement):
    
    def __init__(self, id='', control_pnt_indx=0, num_properties=0,
                 props=[], values=[]):
	self.id = id
	self.control_pnt_indx = control_pnt_indx
	self.num_properties = num_properties
	self.props  = props
	self.values = values
	self.update_properties()
	return None
	
    def __repr__(self):
	s = 'Cptc_TVertex(id = %s, control_pnt_indx = %d, num_properties = %d, ...)\n' % (self.id, self.control_pnt_indx, self.num_properties)
	s = s + '{\n'
	s = s + '   properties = %s\n' % (self.properties)
	s = s + '}\n'
	return s

Vertex = Cptc_TVertex

# ------------------------------------------------------------------------

class Cptc_TEdge(TopoElement):
    
    def __init__(self, id='', num_curves=0, curves=[], num_bound_vertices=0, 
		 bound_vertices=[], num_properties=0, props=[], values=[]):
	self.id = id
	self.num_curves = num_curves
        self.curves = curves
        self.num_bound_vertices = num_bound_vertices
        self.num_properties = num_properties
        self.bound_vertices = bound_vertices
        self.props = props
        self.values = values
	self.update_properties()
	return None

    def __repr__(self):
	s = 'Edge(id = %s, num_curves = %d, num_bound_vertices = %d, num_properties = %d, ...)\n' % \
            (self.id, self.num_curves, self.num_bound_vertices, \
             self.num_properties)
        s = s + '{\n'
        s = s + '   curves = %s\n' % (self.curves)
        s = s + '   bound_vertices = %s\n' % (self.bound_vertices)
	s = s + '   properties = %s\n' % (self.properties)
        s = s + '}\n'
        return s
	

Edge = Cptc_TEdge

# ------------------------------------------------------------------------

class Cptc_TSurface(TopoElement):
	
    def __init__(self, id='', num_patches=0, patches=[], num_bound_edges=0, 
		 bound_edges=[], num_bound_vertices=0, bound_vertices=[],
		 num_properties=0, props=[], values=[]):
        
        self.id = id
        self.num_patches = num_patches
        self.patches = patches
        self.num_bound_edges = num_bound_edges
        self.bound_edges = bound_edges
        self.num_bound_vertices = num_bound_vertices
        self.bound_vertices = bound_vertices
        self.num_properties = num_properties
        self.props = props
	self.values = values
	self.update_properties()
	return None

    def __repr__(self):
        s = 'Cptc_TSurface(id = %s, num_patches = %d, num_bound_edges = %d, num_bound_vertices = %d, num_properties = %d, ...)\n' % \
	    (self.id, self.num_patches, self.num_bound_edges, \
	     self.num_bound_vertices, self.num_properties)
        s = s + '{\n'
        s = s + '   patches = %s\n' % (self.patches)
        s = s + '   bound_edges = %s\n' % (self.bound_edges)
        s = s + '   bound_vertices = %s\n' % (self.bound_vertices)
	s = s + '   properties = %s\n' % (self.properties)
        s = s + '}\n'
        return s


Surface = Cptc_TSurface

# ------------------------------------------------------------------------

class Cptc_TRegion(TopoElement):

    def __init__(self, id='', num_bound_surfaces=0, bound_surfaces=[], \
		 num_bound_edges=0, bound_edges=[], \
		 num_bound_vertices=0, bound_vertices=[], \
		 surface_orientation=[], num_properties=0, \
		 props=[], values=[]):

        self.id                 = id
        self.num_bound_surfaces = num_bound_surfaces
        self.bound_surfaces     = bound_surfaces
	self.num_bound_edges    = num_bound_edges
	self.bound_edges        = bound_edges
	self.num_bound_vertices = num_bound_vertices
	self.bound_vertices     = bound_vertices
        self.surface_orientation= surface_orientation
	self.num_properties     = num_properties
        self.props              = props
        self.values             = values
	self.update_properties()
	return None

    def __repr__(self):
        s = 'Cptc_TRegion(id = %s, num_bound_surfaces = %d, num_bound_edges = %d, \
	num_bound_vertices = %d, num_properties = %d)\n' % \
            (self.id, self.num_bound_surfaces, self.num_bound_edges,
	     self.num_bound_vertices, self.num_properties)
        s = s + '{\n'
        s = s + '   bound_surfaces = %s\n' % (self.bound_surfaces)
        s = s + '   bound_edges = %s\n' % (self.bound_edges)
        s = s + '   bound_vertices = %s\n' % (self.bound_vertices)
        s = s + '   surface_orientation = %s\n' % (self.surface_orientation)
	s = s + '   properties = %s\n' % (self.properties)
        s = s + '}\n'
        return s


Region = Cptc_TRegion

# ------------------------------------------------------------------------

class Cptc_GeomDesc(TopoElement):

    def __init__(self, intrinsic_dimension=3, embedded_dimension=3,
		 num_properties=0, props=[], values=[], 
		 num_control_pts=0, num_top_vertices=0, 
		 num_top_edges=0, num_top_surfaces=0, num_top_regions=0, 
		 control_points=[], top_vertices=[], 
		 top_edges=[], top_surfaces=[], top_regions=[]):

        self.intrinsic_dimension = intrinsic_dimension
        self.embedded_dimension  = embedded_dimension
	self.num_properties      = num_properties
	self.props               = props
	self.values              = values
        self.num_control_pts     = num_control_pts
        self.num_top_vertices    = num_top_vertices
        self.num_top_edges       = num_top_edges
        self.num_top_surfaces    = num_top_surfaces
        self.num_top_regions     = num_top_regions
        self.control_points      = control_points
        self.top_vertices        = top_vertices
        self.top_edges           = top_edges
        self.top_surfaces        = top_surfaces
        self.top_regions         = top_regions
        self.header              = ''
        
    def __repr__(self):
        s = '{\n'
        s = s + '  intrinsic_dimension \t = %d\n' % (self.intrinsic_dimension)
        s = s + '  embedded_dimension \t = %d\n' % (self.embedded_dimension)
	s = s + '  num_properties \t = %d\n' % (self.num_properties)
	s = s + '  props \t\t = %s\n' % (self.props)
	s = s + '  values \t\t = %s\n' % (self.values)
        s = s + '  num_control_pts \t = %d\n' % (self.num_control_pts)
        s = s + '  num_top_vertices \t = %d\n' % (self.num_top_vertices)
        s = s + '  num_top_edges \t = %d\n' % (self.num_top_edges)
        s = s + '  num_top_surfaces \t = %d\n' % (self.num_top_surfaces)
        s = s + '  num_top_regions \t = %d\n' % (self.num_top_regions)
        s = s + '  control_points \t = [...]\n' 
        s = s + '  top_vertices \t\t = [...]\n' 
        s = s + '  top_edges \t\t = [...]\n' 
        s = s + '  top_surfaces \t\t = [...]\n' 
        s = s + '  top_regions \t\t = [...]\n' 
        s = s + '}\n'
        return s

    def update_properties(self):
	self.properties = {}
	for i in range(self.num_properties):
	    self.properties[self.props[i]] = self.values[i]
	for i in range(self.num_top_vertices):
	    self.top_vertices[i].update_properties()
	for i in range(self.num_top_edges):
	    self.top_edges[i].update_properties()
	for i in range(self.num_top_surfaces):
	    self.top_surfaces[i].update_properties()
	for i in range(self.num_top_regions):
	    self.top_regions[i].update_properties()
	return None


	    
