#include <stdio.h>
/*  Global variables and symbols                */

#include    <math.h>
#include    "geoPoly3D.h"

static float GeoTolerance = 0.000001 ;

#define RAD(x) ((x)*(acos(-1.0)/180.0))

#define dsin(x) sin(RAD(x))
#define dcos(x) cos(RAD(x))
#define dtan(x) tan(RAD(x))


/*
** -----------------------------------------------------------------
** C/SWIG/Python Interface
*/

GeoPoint* new_GeoPoints(int num_pts) {
    GeoPoint* pts ;
    pts = (GeoPoint *)malloc( num_pts*sizeof(GeoPoint) ) ;
    return pts ;
}
    	
void set_GeoPoint(GeoPoint* pts, int ptnum, float x, float y, float z) {
    pts[ptnum].x = x ;
    pts[ptnum].y = y ;
    pts[ptnum].z = z ;
}
    
void delete_GeoPoints(GeoPoint* pts) {
    free(pts) ;
}

GeoLine* new_GeoLines(int num_lines) {
    GeoLine* lines ;
    lines = (GeoLine *)malloc( num_lines*sizeof(GeoLine) ) ;
    return lines ;
}

void delete_GeoLines(GeoLine* lines) {
    free(lines) ;
}

GeoPoly* new_GeoPolys(int num_polys) {
    GeoPoly* polys ;
    polys = (GeoPoly *)malloc( num_polys*sizeof(GeoPoly) ) ;
    return polys ;
}

void delete_GeoPolys(GeoPoly* polys) {
    free(polys) ;
}

/*      --------------------------------------------------------
**      Private entry points
*/

    /*  --------------------------------------------------------
    **  GeoIntscPolyLine - returns a flag (GeoTrue of GeoFalse) 
    **      indicating whether a horizontal scan line in 2D space 
    **      passing at the given 2D point intersects the given line 
    **      in 2D space from minus infinite to the given point.
    */

GeoPredicate    GeoIntscPolyLine( line, spar )

    GeoSurfPar  line[] ;        /* given 2D line                    */
    GeoSurfPar  *spar ;         /* 2D coordinates of pt location    */
{
    float   u0,v0,u1,v1 ;   /* line endpoint 2D coords                      */
    float   ua,va,ub,vb ;   /* corrected line endpoint coordinates          */
    float   max_u,min_u ;   /* maximum and minimum u coordinates            */
    float   max_v,min_v ;   /* maximum and minimum v coordinates            */

    u0 = line[0].u ;
    v0 = line[0].v ;
    u1 = line[1].u ;
    v1 = line[1].v ;

    /*  find the max and min coordinates                        */

    max_u = (u0 > u1) ? u0 : u1 ;
    min_u = (u0 < u1) ? u0 : u1 ;
    max_v = (v0 > v1) ? v0 : v1 ;
    min_v = (v0 < v1) ? v0 : v1 ;

/*
    the horizontal line does not intersect the line to the 
    left of location pt if:
*/
    /*  (1) if line is horizontal                                   */

    if( v0 == v1 ) return( GeoFalse ) ;

    /*  (2) if the v coordinate of point is outside the range 
            max_v and min_v (but not including min_v)               */

    if( (spar->v < min_v) || (spar->v >= max_v) ) return( GeoFalse ) ;

    /*  (3) if given line is vertical and u coordinate of point is 
            smaller than that of line                               */

    if( (u0 == u1) && (spar->u < u0) ) return( GeoFalse ) ;

    /*  (4) if inclined line is located to the right of given point     */
    /*     (first reduce the problem to a case where vb >va, always)*/

    if( u0 != u1 ) {
        if( ((u1 > u0) && (v1 > v0)) || ((u1 < u0) && (v1 > v0)) ) {
            ua = u0 ;    va = v0 ;
            ub = u1 ;    vb = v1 ;
        }
        else {
            ua = u1 ;    va = v1 ;
            ub = u0 ;    vb = v0 ;
        }

        if( ((spar->v - va) * (ub - ua)) > ((spar->u - ua) * (vb - va)) ) {
            return( GeoFalse ) ;
        }
    }

/*
    if we get here that is because the horizontal line passing 
    at the location pt intersects the given line to the left of the pt
*/

return( GeoTrue ) ;
}

/*  --------------------------------------------------------------------
**   The following two routines allow the user to get the current value
**   of the tolerance being used in this module, and to change its value.
**   -sss (27-Jan-90)
*/

void GeoSetTolerance( tol )
  float  tol ;
{
    GeoTolerance = tol ;
}

void GeoGetTolerance( tol )
  float  *tol ;
{
    *tol = GeoTolerance ;
}


    /*  --------------------------------------------------------
    **  GeoIsInsidePoly - returns a flag (GeoTrue of GeoFalse) 
    **      indicating whether a point lying on the plane of the 
    **      given polygon is inside it.
    **      A scan line algorithm in the parametric space of the 
    **      polygon plane is used to check whether the given point 
    **      is inside the polygon.  The algorith works as follows:
    **      It gets the projection of each polygon boundary edge onto 
    **      its plane and counts the number of intersections from 
    **      minus infinite in the u direction to the given point 
    **      of a horizontal scan line in the plane parametric space 
    **      with the boundary of the polygon.  If the number of 
    **      intersections is odd the point is inside, otherwise it 
    **      is outside.
    */

GeoPredicate    GeoIsInsidePoly( poly, spar )

    GeoPoly     *poly ;         /* given polygon parametric rep.    */
    GeoSurfPar  *spar ;         /* parametric values of pt location */
{
    GeoSurfPar  edge[2] ;       /* projection of a polygon edge on plane    */
    GeoPoint    local_pt ;      /* vector from origin of local plane system 
                                    to current vertex on polygon            */
    int         n_cross = 0 ;   /* counts the number of intersections       */
    int         i ;

    for( i = 0; i < poly->n_verts; i++ ) {

        local_pt.x = poly->verts[i].x - poly->a.x ;
        local_pt.y = poly->verts[i].y - poly->a.y ;
        local_pt.z = poly->verts[i].z - poly->a.z ;
        edge[0].u = GeoDotProd( &local_pt, &poly->u ) ;
        edge[0].v = GeoDotProd( &local_pt, &poly->v ) ;

        local_pt.x = poly->verts[(i+1) % poly->n_verts].x - poly->a.x ;
        local_pt.y = poly->verts[(i+1) % poly->n_verts].y - poly->a.y ;
        local_pt.z = poly->verts[(i+1) % poly->n_verts].z - poly->a.z ;
        edge[1].u = GeoDotProd( &local_pt, &poly->u ) ;
        edge[1].v = GeoDotProd( &local_pt, &poly->v ) ;

        if( GeoIntscPolyLine( edge, spar ) == GeoTrue ) n_cross++ ;
    }

    if( (n_cross % 2) == 0 ) return( GeoFalse ) ;
    else                     return( GeoTrue ) ;
}


    /*  --------------------------------------------------------
    **  GeoIsOnPolyBdry - returns a flag (GeoTrue of GeoFalse) 
    **      indicating whether a point in 3D space is close to 
    **      one of the edges of a given polygon.
    */

GeoPredicate    GeoIsOnPolyBdry( poly, pt )

    GeoPoly     *poly ;         /* given polygon parametric rep.    */
    GeoPoint    *pt ;           /* given pt location                */
{
    GeoLine     edge ;          /* current polygon edge             */
    GeoPoint    clst ;          /* closest point on edge            */
    GeoPointPar par ;           /* parametric value of point on edge*/
    int         is_on_bdry = 0 ;
    int         i ;

    for( i = 0; i < poly->n_verts; i++ ) {
        GeoGenLine( &poly->verts[i], &poly->verts[(i+1) % poly->n_verts], 
                    &edge ) ;
        GeoClstPtLine( &edge, pt, &clst, &par ) ;
        if( ( (float)par > - GeoTolerance ) 
                    && ( (float)par < 1.0 + GeoTolerance ) 
                    && ( GeoDist( pt, &clst ) < GeoTolerance ) ) {
            is_on_bdry = 1 ;
            break ;
        }
    }

    if( is_on_bdry == 0 ) return( GeoFalse ) ;
    else                  return( GeoTrue ) ;
}



/*      --------------------------------------------------------
**      Public entry points
*/

/*      --------------------------------------------------------
**      Generation Routines:

        void GeoGenLine( pt0, pt1, line ) 
            GeoPoint    *pt0, *pt1 ;
            GeoLine     *line ;

        pt0     first input point on the line.
        pt1     second input point on the line.
        line    output parametric description of the line.

        given two input points, this function creates the parametric 
        description of the line  p = D + t*E where D = pt0 and
        E = pt1 - pt0.


        void GeoGenPlane( pt0, pt1, pt2, plane ) 
            GeoPoint    *pt0, *pt1, *pt2 ; 
            GeoPlane    *plane ; 

        pt0     first input point on the plane.
        pt1     second input point on the plane.
        pt2     third input point on the plane.
        plane   output parametric description of the plane.

        given three input point, this function creates the parametric
        description of the plane  p = A + u*B + v*C where A = pt0,
        B = pt1 - pt0, and C = pt2 - pt0.


        void GeoGenPlanEqn( pt0, pt1, pt2, plan_eqn ) 
            GeoPoint    *pt0, *pt1, *pt2 ; 
            GeoPlanEqn  *plan_eqn ; 

        pt0     first input point on the plane.
        pt1     second input point on the plane.
        pt2     third input point on the plane.
        plane   output general (minimal) description of the plane

        given three input points, this function creates the general
        description of the plane such that the following eqn. is satisfied
        by any point (X,Y,Z) lying on the plane:  a*X + b*Y + c*Z + d = 0.
        The unit normal to the plane is given by (a,b,c) and points in
        direction defined by  {vec(0,1) cross vec(0,2)} (see below).


                     normal |  + pt2
                            | /
                            |/ 
                            + ------+
                         pt0        pt1



        GeoStatus GeoGenPoly( n_verts, verts, plane )
            int          *n_verts ;
            GeoPoint     *verts ; 
            GeoNormPlane *plane ;

        n_verts number of polygon vertices.
        verts   a list of the polygon vertices.
        plane   output parametric description of the normalized 
                plane of the polygon.

        given a list of polygon vertex coordinates, this function creates
        the normalized parametric description of polygon plane, which 
        means that the two plane local directions are unit perpendicular 
        vectors and the unit normal vector is also contained in the 
        descriptor.
        The vertices are assumed given in clockwise direction around 
        the polygon.  If the computed normal vector is null, a GeoInvalid 
        status is returned.  Otherwise a GeoValid status is returned.
*/

void GeoGenLine( pt0, pt1, line ) 
    GeoPoint    *pt0, *pt1 ;
    GeoLine     *line ;
{                          
    line->d = *pt0 ;
    line->e.x = pt1->x - pt0->x ;
    line->e.y = pt1->y - pt0->y ;
    line->e.z = pt1->z - pt0->z ;
}

void GeoGenPlane( pt0, pt1, pt2, plane ) 
    GeoPoint    *pt0, *pt1, *pt2 ; 
    GeoPlane    *plane ; 
{
    plane->a = *pt0 ;
    plane->b.x = pt1->x - pt0->x ;
    plane->b.y = pt1->y - pt0->y ;
    plane->b.z = pt1->z - pt0->z ;
    plane->c.x = pt2->x - pt0->x ;
    plane->c.y = pt2->y - pt0->y ;
    plane->c.z = pt2->z - pt0->z ;
}


void GeoGenPlanEqn( pt0, pt1, pt2, plan_eqn ) 
    GeoPoint    *pt0, *pt1, *pt2 ; 
    GeoPlanEqn  *plan_eqn ; 
{
    GeoPoint    vec1, vec2, normal ;
    float       length ;

    vec1.x =  pt1->x - pt0->x ;
    vec1.y =  pt1->y - pt0->y ;
    vec1.z =  pt1->z - pt0->z ;
    vec2.x =  pt2->x - pt0->x ;
    vec2.y =  pt2->y - pt0->y ;
    vec2.z =  pt2->z - pt0->z ;
 
    normal.x = (vec1.y * vec2.z) - (vec1.z * vec2.y) ; 
    normal.y = (vec1.z * vec2.x) - (vec1.x * vec2.z) ; 
    normal.z = (vec1.x * vec2.y) - (vec1.y * vec2.x) ; 

    length = sqrt(     normal.x*normal.x
                    +  normal.y*normal.y
                    +  normal.z*normal.z  ) ;

    plan_eqn->a = normal.x / length ;
    plan_eqn->b = normal.y / length ;
    plan_eqn->c = normal.z / length ;
    plan_eqn->d = (  - plan_eqn->a * pt0->x
                     - plan_eqn->b * pt0->y
                     - plan_eqn->c * pt0->z  ) ;

return;
}


GeoStatus GeoGenNormPlane( n_verts, verts, plane )
    int          *n_verts ;
    GeoPoint     *verts ; 
    GeoNormPlane *plane ;
{
    float  e_size ;                 /* length of last polygon edge  */

    if ( GeoPolyNormal( n_verts, verts, &plane->n ) != GeoValid )
        return( GeoInvalid ) ;

    e_size = GeoDist( &verts[0], &verts[*n_verts-1] ) ;
    plane->a = verts[0] ;

    plane->u.x = (verts[*n_verts-1].x - verts[0].x) / e_size ;
    plane->u.y = (verts[*n_verts-1].y - verts[0].y) / e_size ;
    plane->u.z = (verts[*n_verts-1].z - verts[0].z) / e_size ;

    GeoCrossProd( &plane->n, &plane->u, &plane->v ) ;
    return( GeoValid );
}

int GeoGenPoly( n_verts, verts, poly )
    int          n_verts ;
    GeoPoint     *verts ; 
    GeoPoly   *poly ;
{
    GeoNormPlane plane ;
    if ( GeoGenNormPlane( &n_verts, verts, &plane ) != GeoValid) 
        return( 0 ) ;
    
    poly->a.x = plane.a.x ; poly->a.y = plane.a.y ; poly->a.z = plane.a.z ; 
    poly->u.x = plane.u.x ; poly->u.y = plane.u.y ; poly->u.z = plane.u.z ;
    poly->v.x = plane.v.x ; poly->v.y = plane.v.y ; poly->v.z = plane.v.z ;
    poly->n.x = plane.n.x ; poly->n.y = plane.n.y ; poly->n.z = plane.n.z ;
    
    poly->n_verts = n_verts ;
    poly->verts = verts ;
    
    return( 1 );
}


/*      --------------------------------------------------------
**      Evaluation Routines:

        void GeoEvalLine( par, line, pt ) 
            GeoPointPar     *par ;
            GeoLine         *line ;
            GeoPoint        *pt ;

        par     input parametric value along the line.
        line    input parametric description of a line.
        pt      output evaluated cartesian point.

        given a parametric description of a line and a parameter
        value along the line, this routine generates the cartesian
        point coresponding to the given parametric value.


        void GeoEvalPlane( par, plane, pt )  
            GeoSurfPar  *par ;
            GeoPlane    *plane ;
            GeoPoint    *pt ;

        par     input parametric surface values.
        plane   input parametric representation of a plane.
        pt      output evaluated cartesian point.

        given a parametric description of a plane and surface
        parameter values (u and v), this routine generates the
        cartesian point coresponding to the given surface parameter.


*/

void GeoEvalLine( par, line, pt ) 
    GeoPointPar     *par ;
    GeoLine         *line ;
    GeoPoint        *pt ;
{
    pt->x = line->d.x + (*par * line->e.x) ;   
    pt->y = line->d.y + (*par * line->e.y) ;   
    pt->z = line->d.z + (*par * line->e.z) ;   
    return ;
}


void GeoEvalPlane( par, plane, pt )  
    GeoSurfPar  *par ;
    GeoPlane    *plane ;
    GeoPoint    *pt ;
{
    pt->x = plane->a.x + (par->u * plane->b.x) + (par->v * plane->c.x) ;   
    pt->y = plane->a.y + (par->u * plane->b.y) + (par->v * plane->c.y) ;   
    pt->z = plane->a.z + (par->u * plane->b.z) + (par->v * plane->c.z) ;   
}


/*      --------------------------------------------------------
**      Vector Routines:


        float GeoDotProd( vec0, vec1 )  
            GeoPoint    *vec0, *vec1 ;

        vec0    first input vector.
        vec1    second input vector.

        The function value is the value of the dot product of
        the two three element input vectors.


        void GeoCrossProd( in0, in1, out )
            GeoPoint    *in0, *in1, *out ;

        in0     first input vector.
        in1     second input vector.
        out     output cross product vector.

        This function computes the cross product of the two
        three element input vectors.  The cross product components
        are returned as the third argument.


        float GeoTripleProd( vec0, vec1, vec2 ) 
            GeoPoint    *vec0, *vec1, *vec2 ;

        vec0    first input vector.
        vec1    second input vector.
        vec2    third input vector.
        
        The returned function value is the value of the triple
        product of the three input vectors.


        GeoPredicate GeoNullVec( vec, tolerance )
            GeoPoint    *vec ;
            float       *tolerance ;

        vec             input vector.
        tolerance       check tolerance.

        This function return a value of GeoTrue if all absolute values 
        of the three elements of vec are less the the value of tolerance.  
        Otherwise a value of GeoFalse is returned.


        float GeoVecLen( vec )
            GeoPoint    *vec ;

        vec     input vector.

        The value of this function is the length of the input
        vector.


        void GeoVecNormalize( vec, n )
            GeoPoint    *vec ;
            GeoPoint    *n ;

        vec     input vector.
        n       normalized vector.

        This function normalizes a vector (i.e. finds a vector
        with the same direction but a magnitude of one).


        float   GeoDist( in0, in1 )
            GeoPoint    *in0 ;
            GeoPoint    *in1 ;

        in0     input point0
        in1     input point1

        The value of this function is the distance between the
        two given points.


        void    GeoDiffVec( in0, in1, out)
            GeoPoint    *in0, *in1, *out ;

        in0, in1 : input vectors
        out : output vector = in0 - in1 ;

        computes the difference between two vectors 


        void    GeoSumVec( in0, in1, out)
            GeoPoint    *in0, *in1, *out ;

        in0, in1 : input vectors
        out : output vector = in0 + in1 ;

        computes the sum of two vectors

 
        void GeoProdScalVec ( in, scalar, out )
            GeoPoint    *in, *out ;
            float       *csi ;

        in:     input vector
        out :   output vector = csi * in ;

        computes the product of a vector by a scalar

        void    GeoLinCombVec( in0, csi0, in1, csi1, out )
            GeoPoint    *in1, *in2, *out ; 
            float       *csi0, *csi1 ;
        in0, in1, :     input vectors ;
        csi0, csi1 :    coefficients corresponding to in0 and in1
        out :   output vector = csi0 * in0 + csi1 * in1 ;

        computes the linear combination of two vectors

        void    GeoVecLinComb( csi0, in0, csi1, in1, out )
                float     csi0 ;
                GeoPoint  *in0 ;
                float     csi1 ;
                GeoPoint  *in1 ;
                GeoPoint  *out ;  out =  csi0 * in0 + csi1 * in1 

        computes the linear combination of two vectors

*/

float GeoDotProd( vec0, vec1 )  
    GeoPoint    *vec0, *vec1 ;
{
    return ( (vec0->x * vec1->x) + 
             (vec0->y * vec1->y) +
             (vec0->z * vec1->z) ) ;
}    

void GeoCrossProd( in0, in1, out )
    GeoPoint    *in0, *in1, *out ;
{
    out->x = (in0->y * in1->z) - (in0->z * in1->y) ; 
    out->y = (in0->z * in1->x) - (in0->x * in1->z) ; 
    out->z = (in0->x * in1->y) - (in0->y * in1->x) ; 
}    

float GeoTripleProd( vec0, vec1, vec2 ) 
    GeoPoint    *vec0, *vec1, *vec2 ;
{
    GeoPoint    tmp ;

    GeoCrossProd( vec0, vec1, &tmp ) ;
    return( GeoDotProd( &tmp, vec2 ) ) ;
}

GeoPredicate GeoNullVec( vec, tolerance )
    GeoPoint    *vec ;
    float       *tolerance ;
{
    if (( fabs( vec->x ) < *tolerance ) &&  
        ( fabs( vec->y ) < *tolerance ) &&  
        ( fabs( vec->z ) < *tolerance ))
        return( GeoTrue ) ;
    else
        return( GeoFalse );
}

float GeoVecLen( vec )
    GeoPoint    *vec ;
{
    return ( sqrt( (vec->x * vec->x) +
                   (vec->y * vec->y) +
                   (vec->z * vec->z) ) ) ;
}


void  GeoVecNormalize( vec, n )
    GeoPoint    *vec ;
    GeoPoint    *n ;
{
    float       len ;

    len = GeoVecLen( vec ) ;

    n->x = vec->x / len ;
    n->y = vec->y / len ;
    n->z = vec->z / len ;
}


float   GeoDist( in0,in1 )
    GeoPoint    *in0 ;
    GeoPoint    *in1 ;
{
    GeoPoint    tmp ;

    tmp.x = in1->x - in0->x ;
    tmp.y = in1->y - in0->y ;
    tmp.z = in1->z - in0->z ;

    return( GeoVecLen( &tmp ) ) ;
}

void    GeoDiffVec( in0, in1, out)
    GeoPoint    *in0, *in1, *out ;
    
{
    out->x = in0->x - in1->x ;
    out->y = in0->y - in1->y ;
    out->z = in0->z - in1->z ;
    return ;
} 

void    GeoSumVec( in0, in1, out)
    GeoPoint    *in0, *in1, *out ;
    
{
    out->x = in0->x + in1->x ;
    out->y = in0->y + in1->y ;
    out->z = in0->z + in1->z ;
    return ;
} 


void GeoProdScalVec ( in, csi, out )
    GeoPoint    *in ;
    float       *csi ;
    GeoPoint    *out ;
{
    out->x = in->x * (*csi) ;
    out->y = in->y * (*csi) ;
    out->z = in->z * (*csi) ;
    return ;
}


void    GeoLinCombVec( in0, csi0, in1, csi1, out )
    GeoPoint    *in0, *in1, *out ; 
    float       *csi0, *csi1 ;
{
    out->x = in0->x * (*csi0) + in1->x * (*csi1) ;
    out->y = in0->y * (*csi0) + in1->y * (*csi1) ;
    out->z = in0->z * (*csi0) + in1->z * (*csi1) ;
}


void    GeoVecLinComb( csi0, in0, csi1, in1, out )
    float     csi0 ;
    GeoPoint  *in0 ;
    float     csi1 ;
    GeoPoint  *in1 ;
    GeoPoint  *out ;
{
    out->x = in0->x * csi0 + in1->x * csi1 ;
    out->y = in0->y * csi0 + in1->y * csi1 ;
    out->z = in0->z * csi0 + in1->z * csi1 ;
}

/*      --------------------------------------------------------
**      Rotation Routines:


        void GeoRotateX( angle, in, out ) 
            float       *angle ;
            GeoPoint    *in, *out ;

        angle   rotation angle (in degrees).
        in      input cartesian point.
        out     rotated cartesian point.

        This function takes an input and rotates it about the X
        axis by the number of degrees specified by angle.  The
        new point position is returned as the third argument.


        void GeoRotateY( angle, in, out ) 
            float       *angle ;
            GeoPoint    *in, *out ;

        angle   rotation angle (in degrees).
        in      input cartesian point.
        out     rotated cartesian point.

        This function takes an input and rotates it about the Y
        axis by the number of degrees specified by angle.  The
        new point position is returned as the third argument.


        void GeoRotateZ( angle, in, out ) 
            float       *angle ;
            GeoPoint    *in, *out ;

        angle   rotation angle (in degrees).
        in      input cartesian point.
        out     rotated cartesian point.

        This function takes an input and rotates it about the Z
        axis by the number of degrees specified by angle.  The
        new point position is returned as the third argument.


        void GeoRotateAxis( axis, angle, in, out ) 
            GeoPoint    *axis ;
            float       *angle ;
            GeoPoint    *in, *out ;

        axis    asix about which to rotate.
        angle   rotation angle (in degrees).
        in      input cartesian point.
        out     rotated cartesian point.

        This function takes an input vector and rotates it about
        a given axis (passing through the origin) by the number 
        of degrees specified by angle.  The new point position
        is returned as the fourth argument.

*/

void GeoRotateX( angle, in, out ) 
    float       *angle ;
    GeoPoint    *in, *out ;
{
    float       s, c ;

    s = dsin( *angle) ;   c = dcos( *angle ) ;
    out->x = in->x ;
    out->y = (in->y * c) - (in->z * s) ;
    out->z = (in->y * s) + (in->z * c) ;
}


void GeoRotateY( angle, in, out ) 
    float       *angle ;
    GeoPoint    *in, *out ;
{         
    float       s, c ;

    s = dsin( *angle) ;   c = dcos( *angle ) ;
    out->x = (in->x * c) - (in->z * s) ;
    out->y = in->y ;
    out->z = (in->x * s) + (in->z * c) ;
}


void GeoRotateZ( angle, in, out ) 
    float       *angle ;
    GeoPoint    *in, *out ;
{
    float       s, c ;

    s = dsin( *angle ) ;    c = dcos( *angle ) ;
    out->x = (in->x * c) - (in->y * s) ;
    out->y = (in->x * s) + (in->y * c) ;
    out->z = in->z ;
}


void GeoRotateAxis( axis, angle, in, out ) 
    GeoPoint    *axis ;
    float       *angle ;
    GeoPoint    *in, *out ;
{
    GeoPoint    u ;
    float       s, c ;
    float       r[3][3] ;

    s = dsin( *angle ) ;    c = dcos( *angle ) ;

    GeoVecNormalize( axis, &u ) ;

    r[0][0] = u.x*u.x + c * (1.0 - u.x*u.x) ;
    r[0][1] = u.x*u.y*(1.0 - c) - u.z*s ;
    r[0][2] = u.z*u.x*(1.0 - c) + u.y*s ;

    r[1][0] = u.x*u.y*(1.0 - c) + u.z*s ;
    r[1][1] = u.y*u.y + c * (1.0 - u.y*u.y) ;
    r[1][2] = u.y*u.z*(1.0 - c) - u.x*s ;

    r[2][0] = u.z*u.x*(1.0 - c) - u.y*s ;
    r[2][1] = u.y*u.z*(1.0 - c) + u.x*s ;
    r[2][2] = u.z*u.z + c * (1.0 - u.z*u.z) ;

    out->x = r[0][0]*in->x + r[0][1]*in->y + r[0][2]*in->z ;
    out->y = r[1][0]*in->x + r[1][1]*in->y + r[1][2]*in->z ;
    out->z = r[2][0]*in->x + r[2][1]*in->y + r[2][2]*in->z ;
}


/*      --------------------------------------------------------
**      Intersection Routines:


        GeoStatus GeoIntscLinePlane( line, plane, pt, ptpar, spar ) 
            GeoLine     *line ;
            GeoPlane    *plane ;
            GeoPoint    *pt ;
            GeoPointPar *ptpar ;
            GeoSurfPar  *spar ;

        line    input parametric description of a line.
        plane   input parametric description of a plane.
        pt      output intersection point.
        ptpar   output parametric value along the line.
        spar    output surface parameters for the plane.

        This function computes the intersection point of a line and a
        plane both given as parametric representations.  The parametric
        value of the intersection along the line as well as the 
        parametric values of the intersection point on the plane are
        also returned.


        GeoStatus GeoIntscLinePoly( line, poly, pt, ptpar, spar ) 
            GeoLine     *line ;
            GeoPoly     *poly ;
            GeoPoint    *pt ;
            GeoPointPar *ptpar ;
            GeoSurfPar  *spar ;

        line    input parametric description of a line.
        poly    input parametric description of a polygon.
        pt      output intersection point.
        ptpar   output parametric value along the line.
        spar    output surface parameters for the plane.

        This function computes the intersection point of a line and a
        planar polygon both given as parametric representations.  The 
        parametric value of the intersection along the line as well as the 
        parametric values of the intersection point on the poly are
        also returned.
        First it is found whether the line intersects the plane which 
        contains the polygon, then, using a scan line algorithm in the 
        parametric space of that plane, it is checked whether the 
        intersection point is inside the polygon.
*/

GeoStatus GeoIntscLinePlane( line, plane, pt, ptpar, spar ) 
    GeoLine     *line ;
    GeoPlane    *plane ;
    GeoPoint    *pt ;
    GeoPointPar *ptpar ;
    GeoSurfPar  *spar ;
{
    float       size ;
    GeoPoint    b_cross_c, c_cross_e, b_cross_e ;
    float       b_x_c_dot_e, c_x_e_dot_b, b_x_e_dot_c ;

    size = GeoVecLen( &plane->b ) ;

    GeoCrossProd( &plane->b, &plane->c, &b_cross_c ) ;
    GeoCrossProd( &plane->c, &line->e,  &c_cross_e ) ;
    GeoCrossProd( &plane->b, &line->e,  &b_cross_e ) ;
         
    b_x_c_dot_e = GeoDotProd( &b_cross_c, &line->e ) ;
    c_x_e_dot_b = GeoDotProd( &c_cross_e, &plane->b ) ;
    b_x_e_dot_c = GeoDotProd( &b_cross_e, &plane->c ) ;

    if (( fabs(b_x_c_dot_e) < GeoTolerance*size ) || 
        ( fabs(c_x_e_dot_b) < GeoTolerance*size ) || 
        ( fabs(b_x_e_dot_c) < GeoTolerance*size )) 
        return( GeoInvalid ) ;
    else {
        *ptpar = ( GeoDotProd( &b_cross_c, &plane->a ) -
                   GeoDotProd( &b_cross_c, &line->d ) ) /  
                   b_x_c_dot_e ;
        GeoEvalLine( ptpar, line, pt ) ;
        spar->u = ( GeoDotProd( &c_cross_e, &line->d ) -
                    GeoDotProd( &c_cross_e, &plane->a ) ) /
                    c_x_e_dot_b ;
        spar->v = ( GeoDotProd( &b_cross_e, &line->d ) -
                    GeoDotProd( &b_cross_e, &plane->a ) ) /
                    b_x_e_dot_c ;
        return( GeoValid ) ;
    }
}



int GeoIntscLinePoly( line, poly, pt ) 
    GeoLine     *line ;
    GeoPoly     *poly ;
    GeoPoint    *pt ;
{
    GeoPointPar ptpar ;
    GeoSurfPar  spar ;
    /*  first find if line intersects the plane which 
        contains the polygon                                            */

    if( GeoIntscLinePlane( line, (GeoPlane *)poly, pt, &ptpar, &spar ) 
                                                            == GeoValid ) {

        /*  then check to see if intersection point is inside polygon
            or on the boundary of polygon                               */

        if( GeoIsInsidePoly( poly, &spar ) == GeoTrue ) {
            return( 1 ) ;
        }
        else {
            if( GeoIsOnPolyBdry( poly, pt ) == GeoTrue ) {
                return( GeoValid ) ;
            }
            else {
                return( 0 ) ;
            }
        }
    }
    else return( 0 ) ;
}



/*      --------------------------------------------------------
**      Find Normal Routines:

        GeoStatus GeoPolyNormal( n_verts, verts, n )
            int         *n_verts ; 
            GeoPoint    *verts ;
            GeoPoint    *n ;

        n_verts input number of polgon vertices.
        verts   input polygon vertices.
        n       ouput polygon normal.

        This function computes a unit vector normal of polygon given 
        by a list of vertex coordinates.  The vertices are assumed
        given in a clockwise order around the polygon.  If the computed
        normal vector is null, a GeoInvalid status is returned. Otherwise
        a value of GeoValid is returned.

        The algorithm for the routine is as follows:
            Traverse the loop of coordinates, assuming that it is in
            clockwise order, computing the components of the "area" of the
            enclosed polygon.  The total "area" components are computed by
            adding "area" components (cross product components) of
            triangles sides formed by the first, previous, and current
            vertices.  If the loop is not convex, some of the triangle
            areas might be negative, but those will be compensated by other
            positive triangle areas so that the final area is positive.
            (area here is actually twice the area).
            (positive here means in the direction of the face normal).


        GeoStatus  GeoLstSqrPlanEqn( num_pts, pts, plan_eqn ) 
            int         *num_pts ;
            GeoPoint    *pts ;
            GeoPlanEqn  *plan_eqn ;

        num_pts     number of points
        pts         list of points
        plan_eqn    plane equation

        Given a list of points this function computes the equation 
        of the plane which correspond to the least square fit 
        through the points.
        The equation is given in the form:
            a * x  +  b * y  +  c * z  +  d  = 0

*/


GeoStatus GeoPolyNormal( n_verts, verts, n )
    int         *n_verts ; 
    GeoPoint    *verts ;
    GeoPoint    *n ;
{
    float       e_size;         /* length of first edge         */
    float       n_size;         /* normal length                */
    float       toler ;
    int         i;              /* loop counter                 */

    n->x = 0.0 ; n->y = 0.0 ; n->z = 0.0 ;
             
    for ( i=2 ; i < *n_verts ; i++ ) {
        n->x +=  (verts[i  ].y - verts[0].y) * 
                 (verts[i-1].z - verts[0].z) -
                 (verts[i  ].z - verts[0].z) * 
                 (verts[i-1].y - verts[0].y) ;

        n->y += -(verts[i  ].x - verts[0].x) * 
                 (verts[i-1].z - verts[0].z) +
                 (verts[i  ].z - verts[0].z) * 
                 (verts[i-1].x - verts[0].x) ;

        n->z +=  (verts[i  ].x - verts[0].x) * 
                 (verts[i-1].y - verts[0].y) -
                 (verts[i  ].y - verts[0].y) * 
                 (verts[i-1].x - verts[0].x) ;
    }

/*  normalize                                                   */
/*  get the length of the first edge (to compute ratio norm_size/edge_size) */

    e_size = GeoDist( &verts[0], &verts[1] ) ;
    n_size = GeoVecLen( n ) ;

    toler = GeoTolerance ;

    if (n_size/e_size > toler) {
        n->x /= n_size ;
        n->y /= n_size ;
        n->z /= n_size ;
        return( GeoValid ) ;
    }
    else 
        return( GeoInvalid );
}


GeoStatus  GeoLstSqrPlanEqn( num_pts, pts, plan_eqn ) 
    int         *num_pts ;      /* number of points     */
    GeoPoint    *pts ;          /* list of points       */
    GeoPlanEqn  *plan_eqn ;     /* plane equation       */

{
    double      a11,a12,a13,a22,a23,a33 ;       /* A matrix         */
    double      b11,b12,b13,b22,b23,b33 ;       /* A inverse matrix */
    double      sum_x,sum_y,sum_z ;             /* summed values    */
    double      div ;
    float       norm ;                  /* auxiliar coeff.  */
    GeoPoint    seg1, seg2, cross ;
    int         i ;                             /* counter          */

    /*  loop through the points and calculate the sums and products  */

    a11 = 0.0 ;     a12 = 0.0 ;     a13 = 0.0 ;
    a22 = 0.0 ;     a23 = 0.0 ;     a33 = 0.0 ;

    sum_x = 0.0 ;   sum_y = 0.0 ;   sum_z = 0.0 ;

    
    for ( i = 0; i < *num_pts; i++ ) {
        a11 = a11 + (double)(pts[i].x * pts[i].x) ;
        a12 = a12 + (double)(pts[i].x * pts[i].y) ;
        a13 = a13 + (double)(pts[i].x * pts[i].z) ;
        a22 = a22 + (double)(pts[i].y * pts[i].y) ;
        a23 = a23 + (double)(pts[i].y * pts[i].z) ;
        a33 = a33 + (double)(pts[i].z * pts[i].z) ;

        sum_x = sum_x + (double)pts[i].x ;
        sum_y = sum_y + (double)pts[i].y ;
        sum_z = sum_z + (double)pts[i].z ;
    }

    /*  compute the inverse of A                            */

    div = a11 * ( a22*a33 - a23*a23 ) + a12 * ( a13*a23 - a12*a33 ) +
          a13 * ( a12*a23 - a13*a22 ) ;
          
    /* Check for possible trouble.  Either colinear points or
       all pts. lie near to a plane which passes through origin resulting
       in the A matrix to be singular, e.g., if only three pts. then
       third row is linear combo. of top two rows.
    */

    if ( fabs(div) < GeoTolerance ) {

        /* Find first two segments which are non-colinear.
           Use these to form the plane eqn.
        */
        for( i = 0; i < *num_pts - 2; i++ ) {
            GeoDiffVec( &pts[i+1], &pts[i], &seg1 ) ;
            GeoDiffVec( &pts[i+2], &pts[i+1], &seg2 ) ;
            GeoCrossProd( &seg1, &seg2, &cross ) ;
            if( GeoNullVec( &cross, &GeoTolerance ) == GeoFalse ) {
                GeoGenPlanEqn( &pts[i], &pts[i+1], &pts[i+2], plan_eqn ) ;
                return( GeoValid ) ;
            }
        }

        return( GeoInvalid ) ;
    }
    else {
    /*  compute the inverse of A coefficients               */

        b11 = ( a22*a33 - a23*a23 ) / div ;
        b12 = ( a13*a23 - a12*a33 ) / div ;
        b13 = ( a12*a23 - a13*a22 ) / div ;
        b22 = ( a11*a33 - a13*a13 ) / div ;
        b23 = ( a12*a13 - a11*a23 ) / div ;
        b33 = ( a11*a22 - a12*a12 ) / div ;

    /*  compute the equation coefficients                   */

        plan_eqn->a = (float)(b11*sum_x + b12*sum_y + b13*sum_z) ;
        plan_eqn->b = (float)(b12*sum_x + b22*sum_y + b23*sum_z) ;
        plan_eqn->c = (float)(b13*sum_x + b23*sum_y + b33*sum_z) ;

    /*  normalize the equation coefficients                 */

        norm = sqrt( plan_eqn->a * plan_eqn->a + plan_eqn->b * plan_eqn->b + 
                     plan_eqn->c * plan_eqn->c ) ;

        plan_eqn->a /= norm ;
        plan_eqn->b /= norm ;
        plan_eqn->c /= norm ;
        plan_eqn->d = - 1.0 / norm ;
    }

return( GeoValid ) ;
}



/*      --------------------------------------------------------
**      Closest Point Routines:

        void GeoClstPtLine( line, pt, clst, par ) 
            GeoLine      *line ;
            GeoPoint     *pt, *clst ;
            GeoPointPar  *par ;

        line    input parametric description of a line.
        pt      input cartesian point.
        clst    output closest point.
        par     parametric value of the point.

        Given the parametric description of a line, and a cartesian
        point not on the line, this function computes the cartesian
        coordinates of a point on the line which is closest to the
        input point.  The function also returns the parametric value
        of this point along the input line.


        void GeoClstPtNormPlane( plane, pt, clst, spar ) 
            GeoNormPlane *plane ;
            GeoPoint     *pt, *clst ;
            GeoSurfPar   *spar ;

        plane   input normalized parametric description of a plane.
        pt      input cartesian point.
        clst    output closest point.
        spar    parametric value of the point.

        Given the normalized parametric description of a plane, and a 
        cartesian point not necessarely on the plane, this function 
        computes the cartesian coordinates of a point on the plane 
        which is closest to the input point.  
        The function also returns the parametric values of this point on 
        the plane.  
        It is assumed that the two local vectors on the plane, 
        "u" and "v", are perpendicular-to-each-other unit vectors.


        GeoStatus GeoClstPtPoly( poly, pt, clst, spar ) 
            GeoPoly     *poly ;
            GeoPoint    *pt ;
            GeoPoint    *clst ;
            GeoSurfPar  *spar ;

        poly    input parametric description of a polygon.
        pt      input cartesian point.
        clst    output closest point.
        spar    parametric value of the point.

        Given the normalized parametric description of a polygon, and a 
        cartesian point not necessarely on its plane, this function 
        computes the cartesian coordinates of a point on the plane 
        which is closest to the input point.  
        The function also returns the parametric values of this point on 
        the plane.  
        First it is obtained the closest point on the plane which
        contains the polygon, then, using a scan line algorithm in the 
        parametric space of that plane, it is checked whether the 
        closest point is inside the polygon.  If the point is inside 
        the polygon, the function returns a GeoValid status, otherwise 
        it returns GeoInvalid.
        It is assumed that the two local vectors on the polygon plane, 
        "u" and "v", are perpendicular-to-each-other unit vectors.


        GeoPredicate GeoIsOnPlane( GeoPoint  *pt, GeoNormPlane  *pla )

        This function determines if the given point, pt,  lies on the
        given plane, pla. THE PLANE SHOULD BE GIVEN AS A POINTER 
        TO A PLANE NORMALIZED DESCRIPTION.
*/


void GeoClstPtLine( line, pt, clst, par ) 
    GeoLine     *line ;
    GeoPoint    *pt, *clst ;
    GeoPointPar *par ;
{
    GeoPoint    local_pt  ;

    local_pt.x = pt->x - line->d.x ;
    local_pt.y = pt->y - line->d.y ;
    local_pt.z = pt->z - line->d.z ;

    *par = GeoDotProd( &local_pt, &line->e ) / 
           GeoDotProd( &line->e,  &line->e ) ;

    GeoEvalLine( par, line, clst ) ;
}


void GeoClstPtNormPlane( plane, pt, clst, spar ) 
    GeoNormPlane *plane ;
    GeoPoint     *pt, *clst ;
    GeoSurfPar   *spar ;
{
    GeoPoint    local_pt ;

    local_pt.x = pt->x - plane->a.x ;
    local_pt.y = pt->y - plane->a.y ;
    local_pt.z = pt->z - plane->a.z ;

    spar->u = - GeoTripleProd( &local_pt, &plane->n, &plane->v ) ;
    spar->v =   GeoTripleProd( &local_pt, &plane->n, &plane->u ) ;

    GeoEvalPlane( spar, (GeoPlane *)plane, clst ) ;
}



GeoStatus GeoClstPtPoly( poly, pt, clst, spar ) 
    GeoPoly     *poly ;
    GeoPoint    *pt, *clst ;
    GeoSurfPar  *spar ;
{
    /*  first get the closest point on the plane which 
        contains the polygon                                    */

    GeoClstPtNormPlane( (GeoNormPlane *)poly, pt, clst, spar ) ;

    /*  then check to see if closest point is inside polygon 
        or on the boundary of polygon                           */

    if( GeoIsInsidePoly( poly, spar ) == GeoTrue ) {
        return( GeoValid ) ;
    }
    else {
        if( GeoIsOnPolyBdry( poly, pt ) == GeoTrue ) {
            return( GeoValid ) ;
        }
        else {
            return( GeoInvalid ) ;
        }
    }
}



GeoPredicate GeoIsOnPlane( pt, pla )

GeoPoint        *pt ;
GeoNormPlane    *pla ;

{
GeoLine         line ;          /* line from "pt" to origin of "pla"        */
float           proj ;          /* projection of above line on plane normal */
float           toler ;         /* tolerance for closeness of point on plane*/
float           line_len ;      /* length of line                           */

    /*  Generate a line connecting the given point to the  
        origin of the given normalized plane.                   */

    GeoGenLine( pt, &pla->a, &line ) ;

    line_len = GeoVecLen( &line.e ) ;

    /*  Get the projection of this line in the direction of the unit 
        normal of the given plane. (take the dot product of the two 
        vectors).  The division by line_len is to normalize the vector 
        line.e to unit size.                                    */

    if( fabs( line_len ) < GeoTolerance ) return( GeoTrue ) ;
    proj = fabs( GeoDotProd( &pla->n, &line.e ) / line_len ) ;

    /*  If the projection is smaller than an arbitrary value,
        the point can be assumed to be on the plane.            */

    toler = 0.01 ;      /* 0.01 ~= cos(1.56 radians) == cos(89.4 degrees) */
    if( proj < toler ) {
        return( GeoTrue ) ;
    }
    else {
        return( GeoFalse ) ;
    }
}


/*      --------------------------------------------------------
**      Misc Routines:

        GeoStatus  GeoInvert2x2( a, b ) 
            float       a[2][2] ;
            float       b[2][2] ;

        a       input 2 x 2 array of floats.
        b       output 2 x 2 inverse of a.

        This function computes the inverse of a 2 x 2 array.
        If the matrix is singular a value of GeoInvalid is
        returned.  Otherwise a value of GeoValid is returned


        GeoStatus  GeoInvert3x3( a, b ) 
            float       a[3][3] ;
            float       b[3][3] ;

        a       input 3 x 3 array of floats.
        b       output 3 x 3 inverse of a.

        This function computes the inverse of a 3 x 3 array.
        If the matrix is singular a value of GeoInvalid is
        returned.  Otherwise a value of GeoValid is returned


        GeoStatus  GeodInvert3x3( a, b ) 
            double      a[3][3] ;
            double      b[3][3] ;

        a       input 3 x 3 array of doubles.
        b       output 3 x 3 inverse of a.

        This function computes the inverse of a 3 x 3 array
        in double precision.
        If the matrix is singular a value of GeoInvalid is
        returned.  Otherwise a value of GeoValid is returned


        void    GeoCentroid( num_pts, pts, cntrd )
            int         *num_pts ;
            GeoPoint    pts[*num_pts] ;
            GeoPoint    *cntrd ;

        num_pts - number of points input
        pts     - coordinates of the points
        cntrd   - returned value of the centroid of the given list of points

        This function calculates the centroid of the given list of points
        such that cntrd->k =  sum{ pts[j]->k } / num_pts.


        GeoStatus GeoPerpLine( pl1, pl2, pl3, u, v, w ) 
            GeoLine     *pl1, *pl2, *pl3 ; 
            GeoPointPar *u, *v, *w ;

        pl1     first input parametric line.
        pl2     second input parametric line.
        pl3     output parametric line.
        u       output parametric value along pl1
        v       output parametric value along pl2
        w       output parametric value along pl3

        Given the parametric description of two input lines
        this function computes the parametric description of the
        line which is perpendicular to the two given lines.
        It also returns the parametric values for the intersection
        points along the three lines.  The w paramenter is the 
        parametric value of the intersection point of pl3 and pl2
        along pl3.  The parametric value of the intersection point
        of pl3 and pl1 measured along pl3 is always 0.0.


        void    GeoPtNormToParPlan( pt, normal, plane ) 
            GeoPoint     *pt, *normal ;
            GeoNormPlane *plane ;

        pt          point on plane
        normal      normal to plane
        plane       normalized plane parametric description

        Given a point on a plane and the plane normal direction, 
        this function finds a parametric representation of the plane.
        The representation is given as a GeoNormPlane structure, which 
        means that the two plane local directions are unit perpendicular 
        vectors and the unit normal vector is also contained in the 
        descriptor.
        The origin of the plane local system is the given point, and the 
        two unit directions on the plane are arbitrarely chosen.

        void    GeoPlanToNormPlan( plane, normplane )
            GeoPlane        *plane ;
            GeoNormPlane    *normplane ;
            
        plane       GeoPlane description of plane                (in)
        normplane   retrieved GeoNormPlane description of plane (out)

        converts plane description from GeoPlane to GeoNormPlane. 
        The vector normplane->u is plane->d normalized and normplane->v
        is the normalized difference between plane.e and its projection
        onto the direction of normplane->u.  


        void    GeoPlanEqnToParPlan( plan_eqn, plane ) 
            GeoPlanEqn   *plan_eqn ;
            GeoNormPlane *plane ;

        plan_eqn    plane equation
        plane       normalized plane parametric description

        Given a plane equation in the form:
            a * x  +  b * y  +  c * z  +  d  = 0
        this function finds a parametric representation of the plane.
        The representation is given as a GeoNormPlane structure, which 
        means that the two plane local directions are unit perpendicular 
        vectors and the unit normal vector is also contained in the 
        descriptor.
        The origin of the plane local system and the two unit directions 
        on the plane are arbitrarely chosen.


        GeoPtTest   GeoSideofPlan( plan_eqn, pt )
            GeoPlanEqn  *plan_eqn ;
            GeoPoint    *pt ;

        Given a plane equation of the form:  a*x + b*y + c*z + d = 0
        (which was created via call to GeoGenPlanEqn), whose top side
        is taken to be that side in which the normal vector points;
        this function will return one of three values.
                GeoAbove -- given pt. is above the plane
                GeoBelow -- given pt. is below the plane
                GeoOn    -- given pt. is on (within GeoTolerance) the plane


        void    GeoSurfInterp( num_pts, map_func, pts_val, interp_val ) 
            int         *num_pts ;
            float       map_func[*num_pts] ;
            GeoSurfPar  pts_val[*num_pts] ;
            GeoSurfPar  *interp_val ;

        num_pts     number of points which describe the mapping
        map_func[]  mapping function evaluated at target point
        pts_val[]   known surface values at interpolation points
        interp_val  returned surface values at target point

        This function interpolates a set of known surface values.
        The mapping function is given evaluated at the target point.
        Basically this function performs the dot product of the 
        given mapping function with the known interpolation point surface 
        values, which results in the target interpolated values.


        void    GeoSurfJacob( num_pts, map_deriv_s, map_deriv_t, 
                                        pts_val, jac ) 
            int         *num_pts ;
            float       map_deriv_s[*num_pts] ;
            float       map_deriv_t[*num_pts] ;
            GeoSurfPar  pts_val[*num_pts] ;
            float       jac[2][2] ;

        num_pts         number of points which describe the mapping
        map_deriv_s[]   mapping derivative wrt s evaluated at target point
        map_deriv_t[]   mapping derivative wrt t evaluated at target point
        pts_val[]       known surface values at interpolation points
        jac[2][2]       returned mapping jacobian evaluated at target point

        This function computes the jacobian matrix of a surface mapping at 
        a target point.
        The partial derivatives of the mapping function are given evaluated at 
        a the target point.
        Basically this function performs the dot product of the given mapping 
        function derivatives with the known interpolation point surface 
        values, which results in jacobian matrix.


        float   GeoAngVecAboutLine( l, v0, v1 ) 
            GeoLine     *l ;
            GeoPoint    *v0, *v1 ;

        l               given line
        v0, v1          reference and angle vectors normal to line

        This function determines the angle about a given line between two 
        planes that meet at this line and contain the two given vectors
        (presumed normal to line).
        The angle is measured in clockwise order going from the plane passing 
        at the second (angle) vector to the first (reference) plane.
        Clockwise is given by the right hand rule with the thumb 
        pointing in the direction of the line.


        float   GeoAngPtAboutLine( l, p0, p1 ) 
            GeoLine     *l ;
            GeoPoint    *p0, *p1 ;

        l               given line
        p0, p1          reference and angle points

        This function determines the angle about a given line between two 
        planes that meet at this line and pass at two given points 
        (presumed not on the line).
        The angle is measured in clockwise order going from the plane passing 
        at the second (angle) point to the first (reference) plane.
        Clockwise is given by the right hand rule with the thumb 
        pointing in the direction of the line.


        float   GeoAngPtAboutPt( pivot, p0, p1 ) 
            GeoPoint    *pivot ;
            GeoPoint    *p0, *p1 ;

        pivot           pivot point
        p0, p1          reference and angle points

        This function determines the angle about a given pivot point 
        between two given points.
        The angle is measured in clockwise order going from the second 
        (angle) point to the first (reference) point.


        GeoStatus GeoProjVecPlane( vec, plane, proj )
            GeoPoint    *vec ; 
            GeoPlane    *plane ;
            GeoPoint    *proj ; 

        vec     vector direction
        plane   given plane
        proj    projection of vector direction onto plane

        given a vector direction and a plane, this function finds the 
        vector direction which corresponds to the projection of the 
        vector onto the plane.
        If the given vector is normal to the plane, a GeoInvalid 
        status is returned.  Otherwise a GeoValid status is returned.


        float GeoDistPtNormPlane( plane, pt )
            GeoNormPlane    *plane ;
            GeoPoint        *pt ;

        plane   input normalized parametric description of a plane.
        pt      input cartesian point.

        returns the distance from the point to the plane


        void GeoDyadProd( vec0, vec1, matr )

            GeoPoint    *vec0, *vec1 ;
            float       matr[3][3] ;

        vec0, vec1  input vectors
        matr        output matrix, dyadic product  matr = v1 @ v2

        performs tensorial, or dyadic, product of two vectors


        void GeoProdMatrVec( matr, vec0, vec1 ) 

        float       matr[3][3] ;
        GeoPoint    *vec0 ;
        GeoPoint    *vec1 ;
        
        matr    input matrix
        vec0    input vector
        vec1    output vector   v1 = matr x v0

*/

GeoStatus  GeoInvert2x2( a, b ) 
    float       a[2][2] ;       /* matrix to invert     */
    float       b[2][2] ;       /* inverse matrix       */
{
    float       det ;       /* determinant of given matrix  */
    int         i,j ;       /* counters                     */

/*  clear inverse matrix                        */

    for( i = 0; i < 2; i++) 
        for( j = 0; j < 2; j++) b[i][j] = 0.0 ;

/*  compute matrix determinant                  */

    det = a[0][0] * a[1][1] - a[0][1] * a[1][0] ;

    if (det == 0.0) 
        return( GeoInvalid ) ;
    else {
        b[0][0] =  a[1][1] / det ;
        b[0][1] = -a[0][1] / det ;
        b[1][0] = -a[1][0] / det ;
        b[1][1] =  a[0][0] / det ;
        return( GeoValid );
    }
}


GeoStatus  GeoInvert3x3( a, b ) 
    float       a[3][3] ;       /* matrix to invert     */
    float       b[3][3] ;       /* inverse matrix       */
{
    float       det ;       /* determinant of given matrix  */
    int         i,j ;       /* counters                     */

/*  clear inverse matrix                        */

    for( i = 0; i < 3; i++) 
        for( j = 0; j < 3; j++) b[i][j] = 0.0 ;

/*  compute matrix determinant                  */

    det = a[0][0] * (a[1][1] * a[2][2] - a[1][2] * a[2][1]) - 
          a[0][1] * (a[1][0] * a[2][2] - a[1][2] * a[2][0]) + 
          a[0][2] * (a[1][0] * a[2][1] - a[1][1] * a[2][0]) ;

    if (det == 0.0) 
        return( GeoInvalid ) ;
    else {
        b[0][0] =  (a[1][1] * a[2][2] - a[1][2] * a[2][1]) / det ;
        b[0][1] = -(a[0][1] * a[2][2] - a[0][2] * a[2][1]) / det ;
        b[0][2] =  (a[0][1] * a[1][2] - a[0][2] * a[1][1]) / det ;
        b[1][0] = -(a[1][0] * a[2][2] - a[1][2] * a[2][0]) / det ;
        b[1][1] =  (a[0][0] * a[2][2] - a[0][2] * a[2][0]) / det ;
        b[1][2] = -(a[0][0] * a[1][2] - a[0][2] * a[1][0]) / det ;
        b[2][0] =  (a[1][0] * a[2][1] - a[1][1] * a[2][0]) / det ;
        b[2][1] = -(a[0][0] * a[2][1] - a[0][1] * a[2][0]) / det ;
        b[2][2] =  (a[0][0] * a[1][1] - a[0][1] * a[1][0]) / det ;
        return( GeoValid );
    }
}


GeoStatus  GeodInvert3x3( a, b ) 
    double      a[3][3] ;       /* matrix to invert     */
    double      b[3][3] ;       /* inverse matrix       */
{
    double      det ;       /* determinant of given matrix  */
    int         i,j ;       /* counters                     */

/*  clear inverse matrix                        */

    for( i = 0; i < 3; i++) 
        for( j = 0; j < 3; j++) b[i][j] = 0.0 ;

/*  compute matrix determinant                  */

    det = a[0][0] * (a[1][1] * a[2][2] - a[1][2] * a[2][1]) - 
          a[0][1] * (a[1][0] * a[2][2] - a[1][2] * a[2][0]) + 
          a[0][2] * (a[1][0] * a[2][1] - a[1][1] * a[2][0]) ;

    if (det == 0.0) 
        return( GeoInvalid ) ;
    else {
        b[0][0] =  (a[1][1] * a[2][2] - a[1][2] * a[2][1]) / det ;
        b[0][1] = -(a[0][1] * a[2][2] - a[0][2] * a[2][1]) / det ;
        b[0][2] =  (a[0][1] * a[1][2] - a[0][2] * a[1][1]) / det ;
        b[1][0] = -(a[1][0] * a[2][2] - a[1][2] * a[2][0]) / det ;
        b[1][1] =  (a[0][0] * a[2][2] - a[0][2] * a[2][0]) / det ;
        b[1][2] = -(a[0][0] * a[1][2] - a[0][2] * a[1][0]) / det ;
        b[2][0] =  (a[1][0] * a[2][1] - a[1][1] * a[2][0]) / det ;
        b[2][1] = -(a[0][0] * a[2][1] - a[0][1] * a[2][0]) / det ;
        b[2][2] =  (a[0][0] * a[1][1] - a[0][1] * a[1][0]) / det ;
        return( GeoValid );
    }
}


void    GeoCentroid( num_pts, pts, cntrd )
            int         *num_pts ;   /* number of points            */
            GeoPoint    pts[] ;     /* list of points       */
            GeoPoint    *cntrd ;    /* returned centroid    */

{
    GeoPoint    sum ;
    int         i ;

    sum.x = sum.y = sum.z = 0.0 ;
    for( i=0; i<(*num_pts); i++ ) {

        sum.x += pts[i].x ;
        sum.y += pts[i].y ;
        sum.z += pts[i].z ;
    }

    cntrd->x = sum.x / *num_pts ;
    cntrd->y = sum.y / *num_pts ;
    cntrd->z = sum.z / *num_pts ;

return ;
}


GeoStatus GeoPerpLine( pl1, pl2, pl3, u, v, w ) 
    GeoLine     *pl1, *pl2, *pl3 ; 
    GeoPointPar *u, *v, *w ;
{
    float       coef[3][3], cinv[3][3], det ;
    GeoPoint    dvec ;

    /* comput the direction of the perpendicular line   */

    GeoCrossProd( &pl1->e, &pl2->e, &pl3->e ) ;

    /* build and invert the coef matrix */

    coef[0][0] = pl1->e.x ; coef[0][1] = pl2->e.x ; coef[0][2] = pl3->e.x ;
    coef[1][0] = pl1->e.y ; coef[1][1] = pl2->e.y ; coef[1][2] = pl3->e.y ;
    coef[2][0] = pl1->e.z ; coef[2][1] = pl2->e.z ; coef[2][2] = pl3->e.z ;

    if ( GeoInvert3x3( coef, cinv ) != GeoValid ) 
        return( GeoInvalid ) ;
    else
        /* find the scale factors */

        dvec.x = pl2->d.x - pl1->d.x ;
        dvec.y = pl2->d.y - pl1->d.y ;
        dvec.z = pl2->d.z - pl1->d.z ;

        *u = cinv[0][0]*dvec.x + cinv[0][1]*dvec.y + cinv[0][2]*dvec.z ;
        *v = cinv[1][0]*dvec.x + cinv[1][1]*dvec.y + cinv[1][2]*dvec.z ;
        *w = cinv[2][0]*dvec.x + cinv[2][1]*dvec.y + cinv[2][2]*dvec.z ;
        *v = -*v ;

        GeoEvalLine( u, pl1, &pl3->d ) ;
        return(GeoValid) ;
}


void    GeoPtNormToParPlan( pt, normal, plane ) 
    GeoPoint     *pt, *normal ; /* point on plane and normal    */
    GeoNormPlane *plane ;       /* plane parametric description */

{
    GeoPoint    x_dir ;                 /* x direction              */
    GeoPoint    y_dir ;                 /* y direction              */
    GeoPoint    z_dir ;                 /* z direction              */
    float       size ;                  /* size of a vector         */
    float       tolerance ;             /* tolerance for checking   */

    x_dir.x = 1.0 ;     x_dir.y = 0.0 ;     x_dir.z = 0.0 ;
    y_dir.x = 0.0 ;     y_dir.y = 1.0 ;     y_dir.z = 0.0 ;
    z_dir.x = 0.0 ;     z_dir.y = 0.0 ;     z_dir.z = 1.0 ;

    /*  normalize given normal                          */

    size = GeoVecLen( normal ) ;

    plane->n.x = normal->x / size ;
    plane->n.y = normal->y / size ;
    plane->n.z = normal->z / size ;

    /*  choose the given point as an origin for the local system on plane */

    plane->a = *pt ;

    /*  arbitrarely choose two unit directions for the local system on plane */

    tolerance = GeoTolerance ;
    GeoCrossProd( &plane->n, &x_dir, &plane->u ) ;
    if (GeoNullVec( &plane->u, &tolerance ) == GeoTrue) {
        GeoCrossProd( &plane->n, &y_dir, &plane->u ) ;
        if (GeoNullVec( &plane->u, &tolerance ) == GeoTrue) {
            GeoCrossProd( &plane->n, &z_dir, &plane->u ) ;
        }
    }
    size = GeoVecLen( &plane->u ) ;
    plane->u.x /= size ;    plane->u.y /= size ;    plane->u.z /= size ;
    GeoCrossProd( &plane->n, &plane->u, &plane->v ) ;

return ;
}

void    GeoPlanToNormPlan( plane, normplane )
    GeoPlane        *plane ;
    GeoNormPlane    *normplane ;


{
    float       proj ;

    normplane->a = plane->a ;
    GeoVecNormalize( &plane->b, &normplane->u ) ;
    GeoCrossProd( &normplane->u, &plane->c, &normplane->n ) ;
    GeoVecNormalize( &normplane->n, &normplane->n) ;
    GeoCrossProd( &normplane->n, &normplane->u, &normplane->v ) ;

    return ;
}
        

void    GeoPlanEqnToParPlan( plan_eqn, plane ) 
    GeoPlanEqn   *plan_eqn ;    /* plane equation               */
    GeoNormPlane *plane ;       /* plane parametric description */

{
    GeoPoint    x_dir ;                 /* x direction              */
    GeoPoint    y_dir ;                 /* y direction              */
    GeoPoint    z_dir ;                 /* z direction              */
    float       tolerance ;             /* tolerance for checking   */
    float       size ;                  /* size of a vector         */

    x_dir.x = 1.0 ;     x_dir.y = 0.0 ;     x_dir.z = 0.0 ;
    y_dir.x = 0.0 ;     y_dir.y = 1.0 ;     y_dir.z = 0.0 ;
    z_dir.x = 0.0 ;     z_dir.y = 0.0 ;     z_dir.z = 1.0 ;

    /*  normal to plane is obtained directly from its equation  */

    size = sqrt( plan_eqn->a * plan_eqn->a + plan_eqn->b * plan_eqn->b + 
                 plan_eqn->c * plan_eqn->c ) ;

    plane->n.x = plan_eqn->a /size ;
    plane->n.y = plan_eqn->b /size ;
    plane->n.z = plan_eqn->c /size ;

    /*  arbitrarely choose an origin for the local system on plane */

    if (fabs(plan_eqn->a) > GeoTolerance) {
        plane->a.x = - plan_eqn->d / plan_eqn->a ;
        plane->a.y = 0.0 ;
        plane->a.z = 0.0 ;
    }
    else if (fabs(plan_eqn->b) > GeoTolerance) {
        plane->a.x = 0.0 ;
        plane->a.y = - plan_eqn->d / plan_eqn->b ;
        plane->a.z = 0.0 ;
    }
    else {
        plane->a.x = 0.0 ;
        plane->a.y = 0.0 ;
        plane->a.z = - plan_eqn->d / plan_eqn->c ;
    }

    /*  arbitrarely choose two unit directions for the local system on plane */

    if (fabs( plan_eqn->d ) == 0.0) tolerance = GeoTolerance ;
    else                        tolerance = GeoTolerance * fabs( plan_eqn->d ) ;
    GeoCrossProd( &plane->n, &x_dir, &plane->u ) ;
    if (GeoNullVec( &plane->u, &tolerance ) == GeoTrue) {
        GeoCrossProd( &plane->n, &y_dir, &plane->u ) ;
        if (GeoNullVec( &plane->u, &tolerance ) == GeoTrue) {
            GeoCrossProd( &plane->n, &z_dir, &plane->u ) ;
        }
    }
    size = GeoVecLen( &plane->u ) ;
    plane->u.x /= size ;    plane->u.y /= size ;    plane->u.z /= size ;
    GeoCrossProd( &plane->n, &plane->u, &plane->v ) ;

return ;
}


GeoPtTest  GeoSideofPlan( plan_eqn, pt )
    GeoPlanEqn  *plan_eqn ;
    GeoPoint    *pt ;

{
    float  result ;

    result =  (     plan_eqn->a * pt->x
                 +  plan_eqn->b * pt->y
                 +  plan_eqn->c * pt->z
                 +  plan_eqn->d          ) ;

    if( result > GeoTolerance )
        return( GeoAbove ) ;
    else if( result < -GeoTolerance )
        return( GeoBelow ) ;
    else
        return( GeoOn ) ;
}


void    GeoSurfInterp( num_pts, map_func, pts_val, interp_val ) 
    int         *num_pts ;      /* number of points which describe the mapping*/
    float       map_func[] ;    /* mapping function evaluated at target point */
    GeoSurfPar  pts_val[] ;     /* known surface values at interpolation pts  */
    GeoSurfPar  *interp_val ;   /* returned surface values at target point    */

{
    int     i;

    interp_val->u = 0.0 ;
    interp_val->v = 0.0 ;

    for( i = 0; i < *num_pts; i++ ) {
        interp_val->u += map_func[i] * pts_val[i].u ;
        interp_val->v += map_func[i] * pts_val[i].v ;
    }

return ;
}


void    GeoSurfJacob( num_pts, map_deriv_s, map_deriv_t, pts_val, jac ) 
    int         *num_pts ;      /* number of points which describe the mapping*/
    float       map_deriv_s[] ; /* mapping deriv. wrt s evaluated at target pt*/
    float       map_deriv_t[] ; /* mapping deriv. wrt t evaluated at target pt*/
    GeoSurfPar  pts_val[] ;     /* known surface values at interpolation pts  */
    float       jac[2][2] ;     /* mapping jacobian evaluated at target point */

{
    int     i;

    jac[0][0] = 0.0 ;       jac[0][1] = 0.0 ;
    jac[1][0] = 0.0 ;       jac[1][1] = 0.0 ;

    for( i = 0; i < *num_pts; i++ ) {
        jac[0][0] += map_deriv_s[i] * pts_val[i].u ;
        jac[0][1] += map_deriv_s[i] * pts_val[i].v ;
        jac[1][0] += map_deriv_t[i] * pts_val[i].u ;
        jac[1][1] += map_deriv_t[i] * pts_val[i].v ;
    }

return ;
}


float   GeoAngVecAboutLine( l, v0, v1 ) 
    GeoLine     *l ;            /* given line                                */
    GeoPoint    *v0, *v1 ;      /* reference and angle vectors normal to line*/
{
    float       cosang ;            /* cosine of target angle           */
    GeoPoint    cross ;             /* cross product of normal vectors  */
    float       pi = fabs( acos(-1.0) ) ;
    float       len, toler ;

    /* normalize the given vectors                          */

    len = GeoVecLen ( v0 ) ;
    if( len == 0.0 ) return( 0.0 ) ;
    v0->x /= len ;   v0->y /= len ;   v0->z /= len ;
    len = GeoVecLen ( v1 ) ;
    if( len == 0.0 ) return( 0.0 ) ;
    v1->x /= len ;   v1->y /= len ;   v1->z /= len ;

    /* cosine of angle between two vectors is obtained from dot product */

    cosang = GeoDotProd( v1, v0 ) ;
    if (cosang> 1.0) cosang =  1.0;
    if (cosang<-1.0) cosang = -1.0;
    
    /* get the cross product of these vectors.  if the cross product vector 
       has the same direction as the given line, then desired angle is 
       in the range [0 .. pi], otherwise it is in the range [pi .. 2pi] */

    GeoCrossProd( v1, v0, &cross ) ;

    toler = GeoTolerance ;
    if( GeoNullVec( &cross, &toler ) == GeoFalse ) {
        if( GeoDotProd( &l->e, &cross ) > 0.0 ) {
            return( acos( cosang ) ) ;
        }
        else {
            return( 2.0 * pi - acos( cosang ) ) ;
        }
    }
    else {
        if( cosang > 0.0 ) {
            return( 0.0 ) ;
        }
        else {
            return( pi ) ;
        }
    }
}


float   GeoAngPtAboutLine( l, p0, p1 ) 
    GeoLine     *l ;            /* given line                   */
    GeoPoint    *p0, *p1 ;      /* reference and angle points   */
{
    GeoPoint    clstp0, clstp1 ;    /* points on line which are closest 
                                       to given points respectively     */
    GeoPoint    v0, v1 ;            /* vectors normal to given line 
                                       passing on p0 and p1 respectively*/
    GeoPointPar dum_par ;

    /* get the two points on line which are closest to given points */

    GeoClstPtLine  ( l, p0, &clstp0, &dum_par ) ;
    GeoClstPtLine  ( l, p1, &clstp1, &dum_par ) ;

    /* generate two vectors normal to line passing at the given points  */

    v0.x = p0->x - clstp0.x ;
    v0.y = p0->y - clstp0.y ;
    v0.z = p0->z - clstp0.z ;
    v1.x = p1->x - clstp1.x ;
    v1.y = p1->y - clstp1.y ;
    v1.z = p1->z - clstp1.z ;

    return( GeoAngVecAboutLine( l, &v0, &v1 ) ) ;
}


float   GeoAngPtAboutPt( pivot, p0, p1 ) 
    GeoPoint    *pivot ;        /* pivot point                  */
    GeoPoint    *p0, *p1 ;      /* reference and angle points   */
{
    GeoLine     l ;              /* line normal to plane formed by 3 points*/
    GeoPoint    v0, v1 ;         /* vectors from pivot to given points  */
    float       pi = fabs( acos(-1.0) ) ;
    float       toler ;

    /* first the vectors from the pivot to the two other points.
       get the cross product of these vectors.                              */

    v0.x = p0->x - pivot->x; v0.y = p0->y - pivot->y; v0.z = p0->z - pivot->z;
    v1.x = p1->x - pivot->x; v1.y = p1->y - pivot->y; v1.z = p1->z - pivot->z;

    /* get a line which is normal to the three points.
       if the cross product vector is null, the angle is either 0 or pi.    */

    l.d = *pivot ;
    GeoCrossProd( &v1, &v0, &l.e ) ;

    toler = GeoTolerance ;
    if( GeoNullVec( &l.e, &toler ) == GeoTrue ) {
        if( GeoDotProd( &v0, &v1 ) > 0.0 ) {
            return( 0.0 ) ;
        }
        else {
            return( pi ) ;
        }
    }
    else {

    /* return angle between two vectors             */

        return( GeoAngVecAboutLine( &l, &v0, &v1 ) ) ;
    }
}


GeoStatus GeoProjVecPlane( vec, plane, proj )
    GeoPoint    *vec ; 
    GeoPlane    *plane ;
    GeoPoint    *proj ; 
{
    float       vec_dot_b, vec_dot_c ;

    vec_dot_b = GeoDotProd( vec, &plane->b ) ;
    vec_dot_c = GeoDotProd( vec, &plane->c ) ;

    proj->x = vec_dot_b * plane->b.x + vec_dot_c * plane->c.x ;
    proj->y = vec_dot_b * plane->b.y + vec_dot_c * plane->c.y ;
    proj->z = vec_dot_b * plane->b.z + vec_dot_c * plane->c.z ;

    if( (fabs(vec_dot_b) < GeoTolerance) && (fabs(vec_dot_c) < GeoTolerance) ) {
        return( GeoInvalid ) ;
    }
    else {
        return( GeoValid ) ;
    }
}

float GeoDistPtNormPlane( plane, pt )
    GeoNormPlane    *plane ;
    GeoPoint        *pt ;

{
    GeoPoint    clst ;
    GeoSurfPar  spar ;

    GeoClstPtNormPlane( plane, pt, &clst, &spar ) ;

    return( GeoDist( pt, &clst ) ) ;
}

void GeoDyadProd( vec0, vec1, matr )

    GeoPoint    *vec0, *vec1 ;
    float       matr[3][3] ;
{

    matr[0][0] = vec0->x * vec1->x ;
    matr[0][1] = vec0->x * vec1->y ;
    matr[0][2] = vec0->x * vec1->z ;
    matr[1][0] = vec0->y * vec1->x ;
    matr[1][1] = vec0->y * vec1->y ;
    matr[1][2] = vec0->y * vec1->z ;
    matr[2][0] = vec0->z * vec1->x ;
    matr[2][1] = vec0->z * vec1->y ;
    matr[2][2] = vec0->z * vec1->z ;

    return ;
}

void GeoProdMatrVec( matr, vec0, vec1 ) 
    float       matr[3][3] ;
    GeoPoint    *vec0 ;
    GeoPoint    *vec1 ;

{
    vec1->x = matr[0][0] * vec0->x + matr[0][1] * vec0->y + 
                                     matr[0][2] * vec0->z ;
    vec1->y = matr[1][0] * vec0->x + matr[1][1] * vec0->y + 
                                     matr[1][2] * vec0->z ;
    vec1->z = matr[2][0] * vec0->x + matr[2][1] * vec0->y + 
                                     matr[2][2] * vec0->z ;
    return ;
}




