# Author: C.R. Myers, 2000-2001.

import PolycrystalSource, PolycrystalIO
from DigitalMaterial.PX.Geometry.Common import PointSetSource
import Misorientations

class Polycrystal:

    """Polycrystal is the central object of interest in the DM.PX
    package.  It aggregates a tesselation geometry
    (polycrystal.geometry) and an extensible set of attributes
    (polycrystal.attributes).  A polycrystal instance is also
    identified by an ID string, which by default is a timestamp of
    creation."""

    def __init__(self, geometry=None, ID=None):

        self.ID = ID
	self.geometry = geometry
        self.attributes = {'orientations': None, 'misorientations':None,
                           'misorientationAngles':None}

    def __getattr__(self, attr):
        if hasattr(self.geometry, attr):
            return getattr(self.geometry, attr)
        elif self.attributes.has_key(attr):
            return self.attributes[attr]
        else:
            raise AttributeError, "no such attribute"
        
    def __repr__(self):
        s = 'Polycrystal( %s )\n' % self.ID
        s = s + '  {  %d grains, %d grain boundaries, %d vertices   }' % \
            (self.GetNumberOfGrains(),
             self.GetNumberOfGrainBoundaries(), \
             self.GetNumberOfVertices())
        return s

    def __getstate__(self):
        return {'geometry': self.geometry, 'ID': self.ID,
                'attributes': self.attributes}

    def __setstate__(self, dict):
        self.geometry = dict['geometry']
        self.ID = dict['ID']
        self.attributes = dict['attributes']

    def GetNumberOfGrains(self):
        return self.geometry.NumberOfRegions()

    def GetNumberOfGrainBoundaries(self):
        return self.geometry.NumberOfInterfaces()

    def GetNumberOfVertices(self):
        return self.geometry.NumberOfVertices()

    def GetGrain(self, ID):
        return Grain(self, ID)

    def GetGrainBoundary(self, ID):
        return GrainBoundary(self, ID)

    def GrainIDs(self):
	return self.geometry._regions.keys()

    def GrainBoundaryIDs(self):
	return self.geometry._interfaces.keys()

    def _build_grainBoundary_list(self):
        interfaces = {}
        for gID in self.GrainIDs():
            faces = []
            for iID in self.geometry._interfaces.keys():
                if gID in iID:
                    faces.append(iID)
            interfaces[gID] = faces
        return interfaces

    def AttachAttribute(self, name, value):
        self.attributes[name] = value

    def DetachAttribute(self, name):
        del self.attributes[name]

    def GetAttribute(self, name):
        return self.attributes[name]

    def SetOrientations(self, orientations):
        self.attributes['orientations'] = orientations

    def GetOrientations(self):
        return self.attributes['orientations']

    def GetOrientation(self, index):
        return self.attributes['orientations'][index]

    def GetMisorientations(self):
        return self.attributes['misorientations']
    
    def GetMisorientation(self, index):
        return self.attributes['misorientations'][index]
    
    def GetMisorientationAngles(self):
        return self.attributes['misorientationAngles']
    
    def GetMisorientationAngle(self, index):
        return self.attributes['misorientationAngle'][index]
    
    def GetAngularMODF(self):
        return self.attributes['AMODF']
    
    def ComputeMisorientations(self):
        self.attributes['misorientations'] = \
            Misorientations.GetAllMisorientationAxisAngles(self)
                                                
    def ComputeMisorientationAngles(self):
        self.attributes['misorientationAngles'] = \
            Misorientations.GetAllMisorientationAngles(self)
                                                
    def ComputeAngularMODF(self, nbins=0):
        if self.attributes['misorientationAngles'] is None:
            self.ComputeMisorientationAngles()
        angles = self.attributes['misorientationAngles']
        self.attributes['AMODF'] = \
            Misorientations.GetMisorientationAngleDistribution(angles, nbins)
        
    def has_Grain(self, ID):
	return self.geometry._regions.has_key(ID)

    def has_GrainBoundary(self, ID):
	return self.geometry._interfaces.has_key(ID)

    def Store(self, filename=None):
        import PolycrystalIO
        writer = PolycrystalIO.PicklePolycrystalWriter()
        if filename is None:
            filename = 'px.' + self.ID + '.dms'
        writer.SetFileName(filename)
        writer.write(self)

    def Load(self, filename):
        import PolycrystalSource
        reader = PolycrystalSource.PolycrystalReader()
        reader.SetFileName(filename)
        reader.BuildPolycrystal()
        poly = reader.GetPolycrystal()
        self.geometry = poly.geometry
        self.ID = poly.ID
        self.orientations = poly.orientations

# ------------------------------------------------------------------------

class Grain:

    def __init__(self, polycrystal, grainID):
        self.polycrystal = polycrystal
        self.ID = grainID

    def __getattr__(self, attr):
        return getattr(self.polycrystal, attr)[self.ID]

    def GrainBoundaries(self):
        return self.polycrystal._regions[self.ID].interfaces


class GrainBoundary:

    def __init__(self, polycrystal, interfaceID):
        self.polycrystal = polycrystal
        self.ID = interfaceID

    def __getattr__(self, attr):
	try:
	    return getattr(self.polycrystal, attr)[self.ID]
	except:
	    return getattr(self.polycrystal.geometry.Interface(self.ID), 
			   attr)

    def neighboringGrains(self):
        return list(self.ID)

    def area(self):
        return self.polycrystal.geometry.Interface(self.ID).area()


class GrainSet:

    def __init__(self, polycrystal=None, indexSet=[]):
        self.polycrystal = polycrystal
        self.indexSet = indexSet

    def __repr__(self):
        s = 'GrainSet: { polycrystal = (%s), indexSet = %s }' % \
            (self.polycrystal.ID, self.indexSet)
        return s

class GrainBoundarySet:

    def __init__(self, polycrystal=None, indexSet=[]):
        self.polycrystal = polycrystal
        self.indexSet = indexSet

    def __repr__(self):
        s = 'InterfaceSet: { polycrystal = (%s), indexSet = %s }' % \
            (self.polycrystal.ID, self.indexSet)
        return s

# ------------------------------------------------------------------------


def test(ngrains=10):
    import PolycrystalSource
    px = PolycrystalSource.BuildUniformGrainsWithUniformTexture(ngrains)
    return px

if __name__ == '__main__':
    px = test(10)


