"""PolycrystalSource.py defines classes and functions for constructing
polycrystals"""


import Polycrystal
from DigitalMaterial.PX.Geometry.Common import PointSetSource
from DigitalMaterial.PX.Geometry.Models import qhullModels
from DigitalMaterial.PX.Attributes import Orientations

class PolycrystalSource:

    """PolycrystalSource is a base class for items that return a
    polycrystal on a call to GetPolycrystal()"""

    def GetPolycrystal(self):

        if self.polycrystal is None:
            try:
                self.BuildPolycrystal()
            except:
                pass

        try:
            return self.polycrystal
        except:
            raise AttributeError, "source has no polycrystal"


class PolycrystalGenerator(PolycrystalSource):

    """PolycrystalGenerator is a PolycrystalSource that generates a
    polycrystal based on a specified tesselation source and a specified
    orientation source"""

    def __init__(self):
        self.polycrystal = None
        self.tesselationSource = None
        self.orientationSource = None

    def SetTesselationSource(self, source):
        self.tesselationSource = source

    def SetOrientationSource(self, source):
        self.orientationSource = source

    def _generateGrains(self):
        pass

    def _generateOrientations(self, number):
        self.orientationSource.SetNumberOfOrientations(number)
        self.orientations = self.orientationSource.GetOrientations()
        if self.polycrystal is not None:
            self.polycrystal.AttachAttribute('orientations', self.orientations)

    def BuildPolycrystal(self):
        self._generateGrains()
        numGrains = self.polycrystal.GetNumberOfGrains()
        if self.orientationSource is not None:
            self._generateOrientations(numGrains)



class VoronoiPolycrystalGenerator(PolycrystalGenerator):

    """VoronoiPolycrystalGenerator is a PolycrystalGenerator that generates a
    Voronoi polycrystal based on a specified point source and a specified
    orientation source"""

    def __init__(self):
        PolycrystalGenerator.__init__(self)

    def SetPointSource(self, source):
        self.pointSource = source
        
    def SetComponentGeometry(self, component):
        self.component = component

    def SetPointSourceAsUniform(self, ngrains):
        self.pointSource = PointSetSource.UniformPointSetGenerator()
        if self.component is not None:
            self.pointSource.SetComponentGeometry(self.component) 
        else:
            self.pointSource.SetComponentGeometry(qhullModels.UnitCube(3)) 
        self.pointSource.SetNumberOfPoints(ngrains)

    def SetPointSourceAsNonuniform(self, inputFile):
        self.pointSource = PointSetSource.NonuniformPointSetGenerator()
        self.pointSource.SetInputFile(inputFile)

    def _generateGrains(self):
        from DigitalMaterial.PX.Geometry.Tesselation import Tesselation
        import string, Numeric

        tBuilder = \
             Tesselation.VoronoiTesselationBuilder()

        if self.pointSource is not None:
            tBuilder.SetPointSource(self.pointSource)
        else:
            raise IOError, "point source for Voronoi generator not set"

        if self.component is not None:
            tBuilder.SetComponentGeometry(self.component)
        else:
            tBuilder.SetComponentGeometry(qhullModels.UnitCube(3))            

        tesselation = tBuilder.GetTesselation()

        from DigitalMaterial.PX.Tools.Utils.Timestamp import Timestamp
        ID = Timestamp().GetStamp()

        self.polycrystal = \
              Polycrystal.Polycrystal(geometry=tesselation, ID=ID)



class UniformVoronoiPolycrystalGenerator(VoronoiPolycrystalGenerator):

    def __init__(self, ngrains=None):
        VoronoiPolycrystalGenerator.__init__(self)
        self.pointSource = PointSetSource.RboxGenerator()
        if ngrains is not None: self.pointSource.SetNumberOfPoints(ngrains)

    def SetNumberOfGrains(self, ngrains):
        self.pointSource.SetNumberOfPoints(ngrains)

class NonuniformVoronoiPolycrystalGenerator(VoronoiPolycrystalGenerator):

    def __init__(self, inputFile=None):
        VoronoiPolycrystalGenerator.__init__(self)
        self.pointSource = PointSetSource.NonuniformPointSetGenerator()
        if inputFile is not None: self.pointSource.SetNumberOfPoints(ngrains)
        
    def SetInputFile(self, inputFile):
        self.pointSource.SetInputFile(inputFile)



class PolycrystalReader(PolycrystalSource):

    def __init__(self):

        self.polycrystal = None
        
    def SetFileName(self, filename):

        self.filename = filename

    def BuildPolycrystal(self):

        import cPickle
        polyFile = open(self.filename, 'rb')
        self.polycrystal = cPickle.load(polyFile)
        polyFile.close()


# ---------------------------------------------------------------------------

def BuildUniformGrainsWithUniformTexture(ngrains):
    from DigitalMaterial.PX.Geometry.Models import qhullModels
    generator = UniformVoronoiPolycrystalGenerator()
    generator.SetNumberOfGrains(ngrains)
    orientationSource = Orientations.RandomOrientationSource()
    
    #generator.SetComponentGeometry(qhullModels.ConvexPolyhedron(3,'20'))
    generator.SetComponentGeometry(qhullModels.UnitCube(3))
    #generator.SetComponentGeometry(qhullModels.UnitDiamond(3))
    generator.SetPointSourceAsUniform(ngrains)
    generator.SetOrientationSource(orientationSource)
    poly = generator.GetPolycrystal()
    return poly


def BuildNonuniformGrainsWithUniformTexture(nonuniformPointFile):
    from DigitalMaterial.PX.Geometry.Models import qhullModels

    generator = NonuniformVoronoiPolycrystalGenerator()
    generator.SetInputFile(nonuniformPointFile)
    orientationSource = Orientations.RandomOrientationSource()
    
    generator.SetComponentGeometry(qhullModels.UnitCube(3))
    generator.SetOrientationSource(orientationSource)
    poly = generator.GetPolycrystal()
    return poly


def BuildUniformGrainsWithNonuniformTexture(ngrains, texture):
    from DigitalMaterial.PX.Geometry.Models import qhullModels
    generator = UniformVoronoiPolycrystalGenerator()
    generator.SetNumberOfGrains(ngrains)
    orientationSource = Orientations.BungeOrientationReader(texture)
    
    generator.SetComponentGeometry(qhullModels.UnitCube(3))
    generator.SetOrientationSource(orientationSource)
    poly = generator.GetPolycrystal()
    return poly

def BuildUniformGrainsWithNonuniformTexture2(ngrains):
    from DigitalMaterial.PX.Geometry.Models import qhullModels
    generator = UniformVoronoiPolycrystalGenerator()
    generator.SetNumberOfGrains(ngrains)
    orientationSource = Orientations.OrientationSampler()
    
    generator.SetComponentGeometry(qhullModels.UnitCube(3))
    generator.SetOrientationSource(orientationSource)
    poly = generator.GetPolycrystal()
    return poly


def test(ngrains):
    poly = BuildUniformGrainsWithUniformTexture(ngrains)
    return poly

if __name__ == '__main__':
    poly = test(10)
    
