import VTK
from DigitalMaterial.PX.Structures.Polycrystals.Polycrystal import *
        

class OrientationColoring:

    def __init__(self, rRange=[0,1.0], gRange=[0,1.0], bRange=[0,1.0],
                 resolution=8):
        self.rRange = rRange
        self.gRange = gRange
        self.bRange = bRange
        self.resolution = resolution
        self.maxIndex = resolution**3

    def GetLookupTable(self):
        lookupTable = VTK.vtkLookupTable()
        resolution = self.resolution
        lookupTable.SetNumberOfColors(resolution**3)
        lookupTable.Build()

        index = 0
        res = self.resolution
        rMin = self.rRange[0]
        rSlope = (self.rRange[1]-rMin)/(res-1)
        gMin = self.gRange[0]
        gSlope = (self.gRange[1]-gMin)/(res-1)
        bMin = self.bRange[0]
        bSlope = (self.bRange[1]-bMin)/(res-1)
        for red in xrange( res ):
            for green in xrange( res ):
                for blue in xrange( res ):
                    rVal = rMin+rSlope*red
                    gVal = gMin+gSlope*green
                    bVal = bMin+bSlope*blue
                    lookupTable.SetTableValue(index, rVal, gVal, bVal, 1.0)
                    index = index + 1

        #lookupTable.SetTableRange(0, index-1)
        return lookupTable

    def __call__(self, orientation):
        import math
        try:
            res = self.resolution
            red = int(res*orientation[0]/(2.*math.pi))
            green = int(res*(orientation[1]+math.pi)/(2*math.pi))
            blue = int(res*orientation[2]/(2*math.pi))
            val = (red*res + green)*res + blue
            return (float(val)/self.maxIndex)
        except ValueError:
            return 0.


orientationColors = OrientationColoring()


class PolycrystalAttributeMap:

    def __init__(self, polycrystal, attribute, colorFunction):
        self.polycrystal = polycrystal
        self.attribute = attribute
        self.colorFunction = colorFunction


class GrainAttributeMap(PolycrystalAttributeMap):

    def __call__(self, index):
        attr = self.polycrystal.GetGrainAttribute(self.attribute,index)
        color = self.colorFunction(attr)
        return color


class GrainBoundaryAttributeMap(PolycrystalAttributeMap):

    def __call__(self, index):
        attr = self.polycrystal.GetGrainBoundaryAttribute(self.attribute,index)
        color = self.colorFunction(attr)
        return color


def MapOrientationsToColors(gbpolydata, polycrystal):
    """takes a GBPolyData instance, and associated polycrystal instance,
    and attaches scalar color field for each grain boundary face based on
    grain orientations"""

    attrMap = GrainAttributeMap(polycrystal, 'orientations',
                                orientationColors)
    
    GBFrontColors = VTK.vtkScalars()
    GBFrontColors.Allocate(gbpolydata.GetNumberOfPolys(), 10)

    GBBackColors = VTK.vtkScalars()
    GBBackColors.Allocate(gbpolydata.GetNumberOfPolys(), 10)

    grainIndices = \
        gbpolydata.GetCellData().GetFieldData().GetArray('grainIndices')

    for n in range(gbpolydata.GetNumberOfPolys()):
        i0 =  int(grainIndices.GetComponent(n,0))
        i1 =  int(grainIndices.GetComponent(n,1))
        try:
            color = attrMap(i0)
        except:
            color = 0
        GBFrontColors.InsertScalar(n, color)
        try:
            color = attrMap(i1)
        except:
            color = 0
        GBBackColors.InsertScalar(n, color)

    gbpolydata.GetCellData().SetScalars(GBFrontColors)
    #return gbpolydata

# --------------------------------------------------------------------------

def GetGBPolyDataFromGBIndexSet(vtkpolycrystal, GBIndexSet): 
    gbPolyData = vtkpolycrystal.GBPolyData

    vtkcells = []
    gIs = VTK.vtkIntArray()
    gIs.SetNumberOfComponents(2)

    for i in GBIndexSet:
        vtkcells.append(vtkpolycrystal._VTKGBMap[i])

    sampledPolyData = VTK.vtkPolyData()
    sampledPolyData.SetPoints(gbPolyData.GetPoints())
    sampledCells = VTK.vtkCellArray()

    for c in vtkcells:
        sampledCells.InsertNextCell(gbPolyData.GetCell(c))
        gIs.InsertNextValue(vtkpolycrystal.grainIndices.GetComponent(c,0))
        gIs.InsertNextValue(vtkpolycrystal.grainIndices.GetComponent(c,1))

    gIField = VTK.vtkFieldData()
    gIField.SetArray(0, gIs)
    gIField.SetArrayName(0, 'grainIndices')
    sampledPolyData.SetPolys(sampledCells)
    sampledPolyData.GetCellData().SetFieldData(gIField)
    sampledPolyData.Update()
    return sampledPolyData


def ExtractSurfaceGrains(vtkpolycrystal):

    from DigitalMaterial.PX.Geometry.Tesselation import Tesselation
    interfaces = []
    tess = vtkpolycrystal.polycrystal.geometry
    for i in range(6):
        interfaces.extend(Tesselation.GetInterfacesOnComponentSurface(tess,
                                                                      i))

    surfaceGBPolyData = GetGBPolyDataFromGBIndexSet(vtkpolycrystal,
                                                    interfaces)
    return surfaceGBPolyData


def ExtractGrainBoundaries(vtkpolycrystal, indexSet=None):
    if indexSet is None:
        return vtkpolycrystal.GBPolyData
    else:
        GBPolyDataSubset = GetGBPolyDataFromGBIndexSet(vtkpolycrystal,
                                                       indexSet)
        return GBPolyDataSubset

def ExtractGrains(vtkpolycrystal, indexSet=None):

    if indexSet is None:
        return ExtractSurfaceGrains(vtkpolycrystal)
    else:
        GBIndexSet = []
        for gID in indexSet:
            GBIndexSet.extend(\
                    vtkpolycrystal.polycrystal.GetGrain(gID).GrainBoundaries())
        return ExtractGrainBoundaries(vtkpolycrystal, GBIndexSet)


def ExtractPlanarCut(vtkpolycrystal, knife):

    planeNormal, planeOffset = knife.planeNormal, knife.planeOffset
    
    clipPlane = VTK.vtkPlane()
    clipPlane.SetOrigin(vtkpolycrystal.GBPolyData.GetCenter())
    clipPlane.SetNormal(planeNormal)

    clipper = VTK.vtkClipPolyData()
    clipper.SetInput(vtkpolycrystal.GBPolyData)
    clipper.SetValue(planeOffset)
    clipper.SetClipFunction(clipPlane)
    #        clipper.GenerateClippedOutputOn()
    clipData = clipper.GetOutput()
    clipData.Update()

    cutData = knife.GetSurfaceGrains()
    cutData.Update()
        
    #writer=VTK.vtkPolyDataWriter()
    #writer.SetFileName('clipData.vtk')
    #writer.SetInput(clipData)
    #writer.Write()

    concatenation = VTK.vtkAppendPolyData()
    concatenation.AddInput(clipData)
    concatenation.AddInput(cutData)
    clipAndCut = concatenation.GetOutput()
    clipAndCut.Update()
    return clipAndCut



# --------------------------------------------------------------------------

def generate(ngrains):
    from DigitalMaterial.PX.Structures.Polycrystals import VTKPolycrystal 
    vtkpoly = VTKPolycrystal.test(ngrains)
    vtkplotset = ExtractSurfaceGrains(vtkpoly)
    #vtkplotset = ExtractGrains(vtkpoly, [1,2])
    #vtkplotset = ExtractGrainBoundaries(vtkpoly)
    #import Cut
    #knife = Cut.PolycrystalCutter(vtkpolycrystal=vtkpoly)
    #knife.Cut((1.,1.,0.), 0.)
    #vtkplotset = ExtractPlanarCut(vtkpoly, knife)
    MapOrientationsToColors(vtkplotset, vtkpoly.polycrystal)
    from DigitalMaterial.PX.Tools.Viz import PolyDataAvatar
    avatar = PolyDataAvatar.Avatar(vtkplotset, \
                                   orientationColors.GetLookupTable())
    return vtkpoly, avatar

def triangulate(vtkpoly):
    from DigitalMaterial.PX.Structures.Polycrystals import VTKPolycrystal 
    filter = VTK.vtkTriangleFilter()
    filter.SetInput(vtkpoly.GBPolyData)
    vtkplotset = filter.GetOutput()
    vtkplotset.Update()
    MapOrientationsToColors(vtkplotset, vtkpoly.polycrystal)
    from DigitalMaterial.PX.Tools.Viz import PolyDataAvatar
    avatar = PolyDataAvatar.Avatar(vtkplotset, \
                                   orientationColors.GetLookupTable())
    return vtkpoly, avatar

def demo(ngrains):
    from DigitalMaterial.PX.Structures.Polycrystals import VTKPolycrystal 
    vtkpoly = VTKPolycrystal.test(ngrains)
    vtkplotset1 = ExtractSurfaceGrains(vtkpoly)
    import Cut
    knife = Cut.PolycrystalCutter(vtkpolycrystal=vtkpoly)
    knife.Cut((1.,1.,0.), 0.)
    vtkplotset2 = ExtractPlanarCut(vtkpoly, knife)
    MapOrientationsToColors(vtkplotset1, vtkpoly.polycrystal)
    MapOrientationsToColors(vtkplotset2, vtkpoly.polycrystal)
    from DigitalMaterial.PX.Tools.Viz import PolyDataAvatar
    avatar1 = PolyDataAvatar.Avatar(vtkplotset1, \
                                    orientationColors.GetLookupTable())
    avatar2 = PolyDataAvatar.Avatar(vtkplotset2, \
                                    orientationColors.GetLookupTable())
    return vtkplotset1, avatar1, avatar2


def cut(vtkpoly, planeNormal, planeOffset=0.0):
    from DigitalMaterial.PX.Structures.Polycrystals import VTKPolycrystal 
    import Cut
    knife = Cut.PolycrystalCutter(vtkpolycrystal=vtkpoly)
    knife.Cut(planeNormal, planeOffset)
    vtkplotset = ExtractPlanarCut(vtkpoly, knife)
    MapOrientationsToColors(vtkplotset, vtkpoly.polycrystal)
    from DigitalMaterial.PX.Tools.Viz import PolyDataAvatar
    avatar = PolyDataAvatar.Avatar(vtkplotset, \
                                   orientationColors.GetLookupTable())
    return avatar
    
