import VTK

def CutVTKPolyData(polydata, planeNormal, planeOffset):

    """CutVTKPolyData(polydata, planeNormal, planeOffset) takes a vtkPolyData
    instance (polydata) and a cutting plane (defined by planeNormal and 
    planeOffset) and returns the cut vtkPolyData resulting from the operation
    of a vtkCutter"""

    clipPlane = VTK.vtkPlane()
    clipPlane.SetOrigin(polydata.GetCenter())
    apply(clipPlane.SetNormal, planeNormal)

    cutter = VTK.vtkCutter()
    cutter.SetInput(polydata)
    cutter.SetValue(0, planeOffset)
    cutter.SetCutFunction(clipPlane)
    cutData = cutter.GetOutput()
    cutData.Update()
    return cutData
#    cleaner = VTK.vtkCleanPolyData()
#    cleaner.SetInput(cutData)
#    cleaner.SetTolerance(1.0e-01)
#    cleanData = cleaner.GetOutput()
#    cleanData.Update()
#    return cleanData


class VTKPolycrystalCutter:

    def __init__(self, vtkpolycrystal):

        self.vtkpolycrystal = vtkpolycrystal
        self.polydata = vtkpolycrystal.GrainBoundaries

    def CutAlongPlane(self, planeNormal, planeOffset=0.0):

	return CutVTKPolyData(self.polydata, planeNormal, planeOffset)

    def _buildPolygonsFromSurfaceCut(self, surfaceCut):

	surfacePolygons = self._extractPolygonsFromEdges(surfaceCut)
	vtkpolygons = self._buildVTKPolygons(surfaceCut, surfacePolygons)
	return vtkpolygons

    def _buildVTKPolygons(self, surfaceCut, surfacePolys):

	grains = VTK.vtkPolyData()
	polys = VTK.vtkCellArray()
        gIs = VTK.vtkIntArray()
        gIs.SetNumberOfComponents(2)
#        gIs.SetNumberOfComponents(0)        
        
	nc = 0
	for ID, poly in surfacePolys.items():
	    polys.InsertNextCell(len(poly))
	    for v in poly:
		polys.InsertCellPoint(v)
            gIs.InsertNextValue(ID)
            gIs.InsertNextValue(ID)

        grains.SetPolys(polys)
	grains.SetPoints(surfaceCut.GetPoints())
        gIField = VTK.vtkFieldData()
        gIField.SetArray(0, gIs)
        gIField.SetArrayName(0, 'grainIndices')
        grains.GetCellData().SetFieldData(gIField)
	grains.Update()

	return grains

    def _extractPolygonsFromEdges(self, surfaceCut):

        lines = surfaceCut.GetLines()
        numLines = lines.GetNumberOfCells()
        grainIndices = \
            surfaceCut.GetCellData().GetFieldData().GetArray('grainIndices')
        surfacePolygons = {}
        
        for n in range(numLines):
            line = surfaceCut.GetCell(n)
            pointIds = line.GetPointIds()
            vertices = [pointIds.GetId(0), pointIds.GetId(1)]
            grains = [int(grainIndices.GetComponent(n,0)),
                      int(grainIndices.GetComponent(n,1))]
            for g in grains:
                if g < self.vtkpolycrystal.NumberOfRegions():
                    if not surfacePolygons.has_key(g): surfacePolygons[g] = []
                    surfacePolygons[g].append(vertices)

        if 0:
            for g in surfacePolygons.keys():            
                if len(surfacePolygons[g]) < 3:
                    del surfacePolygons[g]

	surfacePolygons = self._orderPolygonVertices(surfacePolygons)
	return surfacePolygons


    def _orderPolygonVertices(self, surfacePolygons):

        import copy
        vertexList={}

        for g in surfacePolygons.keys():            
            vertexList[g] = copy.deepcopy(surfacePolygons[g][0])
            remaining = []
            for vp in surfacePolygons[g][1:]:
                remaining.append(vp)

            while len(remaining) != 1:  # don't duplicate last point
                last = vertexList[g][-1]
                nextpair = \
                    filter(lambda x, target=last: x.count(target)>0, \
                           remaining)[0]
                nextvertex = \
                    filter(lambda x, target=last: x != target, \
                           nextpair)[0]
                vertexList[g].append(nextvertex)
                remaining.remove(nextpair)

            surfacePolygons[g] = vertexList[g]

	return surfacePolygons

