import VTK

class PolycrystalViewer:

    def __init__(self, vtkfilename, colorFunc = None):

        self.vtkfilename = vtkfilename
        self.colorFunc = colorFunc

        self.reader = VTK.vtkPolyDataReader()
        self.reader.SetFileName(self.vtkfilename)
	self.polycData = self.reader.GetOutput()
        self.reader.Update()

        self.lookupTables = []
        import Numeric
        for hue in Numeric.arange(0., 1., 0.1):
            lut = VTK.vtkLookupTable()
	    lut.SetHueRange(hue, hue)
	    lut.SetValueRange(0.5,1.0)
            self.lookupTables.append(lut)

    def GetActors(self):
	return self.actors

    def DrawInterfaces(self, colorFunc=None):

        colors = {}
	interfaceColors = VTK.vtkScalars()
	interfaceColors.Allocate(self.polycData.GetNumberOfPolys(), 10)

        if colorFunc is None:
            math = VTK.vtkMath()
            for n in range(self.polycData.GetNumberOfPolys()):
                colors[n] = math.Random(0.,1.)

        for n in range(self.polycData.GetNumberOfPolys()):        
            interfaceColors.InsertScalar(n, colors[n])

	self.polycData.GetCellData().SetScalars(interfaceColors)

	interfaceMapper = VTK.vtkDataSetMapper()
	interfaceActor = VTK.vtkActor()

	if colorFunc:
	    try:
		interfaceMapper.SetLookupTable( colorFunc.GetLookupTable() )
		interfaceMapper.ScalarVisibilityOn()
	    except AttributeError:
		pass
	else:
	    interfaceMapper.SetLookupTable(self.lookupTables[1])

	interfaceMapper.SetInput(self.polycData)
	interfaceActor.SetMapper(interfaceMapper)
	interfaceActor.GetProperty().SetOpacity(1.0)

        return [interfaceActor]


    def Cut(self, planeNormal, planeOffset=0.0):

        self.clipPlane = VTK.vtkPlane()
        self.clipPlane.SetOrigin(self.polycData.GetCenter())
        apply(self.clipPlane.SetNormal, planeNormal)

        cutter = VTK.vtkCutter()
        cutter.SetInput(self.polycData)
        cutter.SetValue(0, planeOffset)
        cutter.SetCutFunction(self.clipPlane)
        self.cutData = cutter.GetOutput()

    def BuildPolygonsFromCutData(self):

        import copy

        lines = self.cutData.GetLines()  # lines is a vtkCellArray
        numLines = lines.GetNumberOfCells()
        grainIndices = \
            self.cutData.GetCellData().GetFieldData().GetArray('grainIndices')
        self.cutPolygons = {}
        
        for n in range(numLines):
            line = self.cutData.GetCell(n)
            pointIds = line.GetPointIds()
            vertices = [pointIds.GetId(0), pointIds.GetId(1)]
            grains = [int(grainIndices.GetComponent(n,0)),
                      int(grainIndices.GetComponent(n,1))]
            for g in grains:
                if not self.cutPolygons.has_key(g): self.cutPolygons[g] = []
                self.cutPolygons[g].append(vertices)

        vertexList={}

        for g in self.cutPolygons.keys():            

            if len(self.cutPolygons[g]) < 3:
                del self.cutPolygons[g]
                continue
            vertexList[g] = copy.deepcopy(self.cutPolygons[g][0])
            remaining = []
            for vp in self.cutPolygons[g][1:]:
                remaining.append(vp)

            while remaining != []:
                last = vertexList[g][-1]
                nextpair = \
                    filter(lambda x, target=last: x.count(target)>0, \
                           remaining)[0]
                nextvertex = \
                    filter(lambda x, target=last: x != target, \
                           nextpair)[0]
                vertexList[g].append(nextvertex)
                remaining.remove(nextpair)

            self.cutPolygons[g] = vertexList[g]

    def DrawClipAndCut(self, planeNormal, planeOffset=0.0, colorFunc=None):

        self.clipPlane = VTK.vtkPlane()
        self.clipPlane.SetOrigin(self.polycData.GetCenter())
        apply(self.clipPlane.SetNormal, planeNormal)

        clipper = VTK.vtkClipPolyData()
        clipper.SetInput(self.polycData)
        clipper.SetValue(planeOffset)
        clipper.SetClipFunction(self.clipPlane)
#        clipper.GenerateClippedOutputOn()

        clipMapper = VTK.vtkPolyDataMapper()
        clipMapper.SetLookupTable(self.lookupTables[2])
        clipMapper.SetInput(clipper.GetOutput())
        
        clipActor = VTK.vtkActor()
        clipActor.SetMapper(clipMapper)

        restmapper = VTK.vtkPolyDataMapper()
        restmapper.SetInput(clipper.GetClippedOutput())
        restmapper.SetLookupTable(self.lookupTables[2])
        restActor = VTK.vtkActor()
        restActor.SetMapper(restmapper)
        restActor.GetProperty().SetOpacity(0.05)

        cutter = VTK.vtkCutter()
        cutter.SetInput(self.polycData)
        cutter.SetValue(0, planeOffset)
        cutter.SetCutFunction(self.clipPlane)
        self.cutData = cutter.GetOutput()

        cutMapper = VTK.vtkPolyDataMapper()
        cutMapper.SetLookupTable(self.lookupTables[2])
        cutMapper.SetInput(self.cutData)
        cutActor = VTK.vtkActor()
        cutActor.SetMapper(cutMapper)
        cutActor.GetProperty().SetOpacity(1.0)

        return [clipActor, cutActor, restActor]


    def DrawDelaunay(self, colorFunc=None):

        delaunay = VTK.vtkDelaunay3D()
        delaunay.SetInput(self.polycData)
        self.delaunayGrid = delaunay.GetOutput()
        self.delaunayGrid.Update()
        self.grainColors = VTK.vtkScalars()
	self.grainColors.Allocate(self.delaunayGrid.GetNumberOfPoints(), 10)
        
        math = VTK.vtkMath()
        for n in range(self.delaunayGrid.GetNumberOfPoints()):
            color = math.Random(0.,1.)
            self.grainColors.InsertScalar(n, color)

        self.delaunayGrid.GetPointData().SetScalars(self.grainColors)
                
        delaunayMapper = VTK.vtkDataSetMapper()
        delaunayMapper.SetLookupTable(self.lookupTables[5])
        delaunayMapper.SetInput(self.delaunayGrid)
        delaunayActor = VTK.vtkActor()
        delaunayActor.SetMapper(delaunayMapper)

    	return [delaunayActor]
