#ifndef UNARYOPERATORS_H
#define UNARYOPERATORS_H

#include "DigitalMaterial.h" 
#include "OperatorOperand.h"
#include "Point.h"
//#include "Field.h"
//#include "SquareGridField.h"


class Laplacian : virtual public Operator {
	public:
		Operator & OperateOn(Field *f);
		Operator & OperateOn(SquareGridField *f);
};

class LaplacianSquared : virtual public Operator {
	public:
		Operator & OperateOn(Field *f);
		Operator & OperateOn(SquareGridField *f);
};

class DirDerivative : virtual public Operator {
	protected:
		Point direction;
	public:
		DirDerivative(void);
		DirDerivative(const Point & dir);
		
		Operator & operator()(const Point dir) { direction=dir; return(*this); }
		bool SetDirection(const Point & dir) { direction=dir; return(true); }
		Point GetDirection(void) const { return(direction); }

		Operator & OperateOn(Field *f);
		Operator & OperateOn(SquareGridField *f);
};

class PartialDerivative : virtual public Operator {
	protected:
		int direction;
	public:
		PartialDerivative(void);
		PartialDerivative(int n);

		Operator & operator()(int dir) {
			#ifdef _DEBUG
				assert(0<=dir && dir<DIMENSION);
			#endif
			direction=dir;
			return(*this);
		}
		bool SetDirection(int n) {
			if (0<=n && n<DIMENSION) {
				direction=n;
				return(true);
			}
			else return(false);
		}
		int GetDirection(void) const { return(direction); }

		Operator & OperateOn(Field *f);
		Operator & OperateOn(SquareGridField *f);
};

class AverageTwo : virtual public Operator {
	protected:
		int direction;
		int forwardbackward; // must be either +1 or -1
	public:
		AverageTwo(void);
		AverageTwo(int n);

		Operator & operator()(int dir) {
			#ifdef _DEBUG
				assert(0<=dir && dir<DIMENSION);
			#endif
			direction=dir;
			return(*this);
		}
		bool SetDirection(int n) {
			if (0<=n && n<DIMENSION) {
				direction=n;
				return(true);
			}
			else return(false);
		}
		bool SetForward(void) { forwardbackward=1; return(true); }
		bool SetBackward(void) { forwardbackward=-1; return(true); }
		int GetForwardBackward(void) { return(forwardbackward); }
		int GetDirection(void) const { return(direction); }

		Operator & OperateOn(Field *f);
		Operator & OperateOn(SquareGridField *f);
};

class AverageFour : virtual public Operator {
	protected:
		int forwardbackward; // must be either +1 or -1
	public:
		AverageFour(void);

		bool SetForward(void) { forwardbackward=1; return(true); }
		bool SetBackward(void) { forwardbackward=-1; return(true); }
		int GetForwardBackward(void) { return(forwardbackward); }

		Operator & OperateOn(Field *f);
		Operator & OperateOn(SquareGridField *f);
};

class AsymmetricDerivative : virtual public Operator {
	protected:
		int direction;
		int forwardbackward; // must be either +1 or -1
	public:
		AsymmetricDerivative(void);
		AsymmetricDerivative(int n);

		Operator & operator()(int dir) {
			#ifdef _DEBUG
				assert(0<=dir && dir<DIMENSION);
			#endif
			direction=dir;
			return(*this);
		}
		bool SetDirection(int n) {
			if (0<=n && n<DIMENSION) {
				direction=n;
				return(true);
			}
			else return(false);
		}
		bool SetForward(void) { forwardbackward=1; return(true); }
		bool SetBackward(void) { forwardbackward=-1; return(true); }
		int GetForwardBackward(void) { return(forwardbackward); }
		int GetDirection(void) const { return(direction); }

		Operator & OperateOn(Field *f);
		Operator & OperateOn(SquareGridField *f);
};

class DiagonalPropagator : virtual public Operator {
	protected:
		Real time;		// time to propagate
		Point delta;	// grid spacing in all directions - provides an interface for Re/ImDiagonal() 
						// to the grid spacing of the field the operator is applied to
	public:
		DiagonalPropagator(void);
		DiagonalPropagator(Real t);
		DiagonalPropagator(const DiagonalPropagator & prop);
		virtual ~DiagonalPropagator(void);

		virtual Real ReDiagonal(const Point &k)=0;	// real part of the diagonal elements of the propagator
		virtual Real ImDiagonal(const Point &k)=0;	// imaginary part ... k is the location on the Broillin zone

		Operator & operator()(Real t) { time=t; return(*this); }
		bool SetTime(Real t) { time=t; return(true); }
		Real GetTime(void) const { return(time); }

		Operator & OperateOn(Field *f);
		Operator & OperateOn(SquareGridField *f);
};

class ImplicitPropagator : virtual public Operator {
	protected:
		Real time;		// time to propagate
		Point delta;	// grid spacing in all directions - provides an interface for Re/ImDiagonal() 
						// to the grid spacing of the field the operator is applied to
	public:
		ImplicitPropagator(void);
		ImplicitPropagator(Real t);
		ImplicitPropagator(const ImplicitPropagator & prop);
		virtual ~ImplicitPropagator(void);

		virtual Real ReDiagonal(const Point &k)=0;	// real part of the diagonal elements of the propagator
		virtual Real ImDiagonal(const Point &k)=0;	// imaginary part ... k is the location on the Broillin zone

		Operator & operator()(Real t) { time=t; return(*this); }
		bool SetTime(Real t) { time=t; return(true); }
		Real GetTime(void) const { return(time); }

		Operator & OperateOn(Field *f);
		Operator & OperateOn(SquareGridField *f);
};

class AddReal : virtual public Operator {
	protected:
		Real addend;
	public:
		AddReal(void);
		AddReal(Real ad);

		Operator & operator()(Real ad) { addend=ad; return(*this); }
		bool SetAddend(Real ad) { addend=ad; return(true); }
		Real GetAddend(void) const { return(addend); }

		Operator & OperateOn(Field *f);
		Operator & OperateOn(SquareGridField *f);
};

class MultiplyReal : virtual public Operator {
	protected:
		Real factor;
	public:
		MultiplyReal(void);
		MultiplyReal(Real fac);

		Operator & operator()(Real fac) { factor=fac; return(*this); }
		bool SetFactor(Real fac) { factor=fac; return(true); }
		Real GetFactor(void) const { return(factor); }

		Operator & OperateOn(Field *f);
		Operator & OperateOn(SquareGridField *f);
};

class ApplyFunction : virtual public Operator {
	public:
		ApplyFunction(void);

		virtual Real function(Real x)=0;

		Operator & OperateOn(Field *f);
		Operator & OperateOn(SquareGridField *f);
};

class PhaseFieldPotential : virtual public Operator {
	protected:
		Real strength;
	public:
		PhaseFieldPotential(void);
		PhaseFieldPotential(Real s);

		Operator & operator()(Real s) { strength=s; return(*this); }
		
		bool SetStrength(Real s) { strength=s; return(true); }
		Real GetStrength(void) const { return(strength); }

		Operator & OperateOn(Field *f);
		Operator & OperateOn(SquareGridField *f);
};

class MultiplyFunction : virtual public Operator {
	public:	
		MultiplyFunction(void);

		virtual Real function(const Point x)=0;

		Operator & OperateOn(Field *f);
		Operator & OperateOn(SquareGridField *f);
};

class AddFunction : virtual public Operator {
	public:
		AddFunction(void);

		virtual Real function(const Point x)=0;

		Operator & OperateOn(Field *f);
		Operator & OperateOn(SquareGridField *f);
};

class L2NormalizeField : virtual public Operator {
	public:
		L2NormalizeField(void);
		Operator & OperateOn(Field *f);
		Operator & OperateOn(SquareGridField *f);
};

#endif // UNARYOPERATORS_H
