#include "Index.h"

Index::Index(const int x[DIMENSION]) 
{
	int i;
	for (i=0; i<DIMENSION; i++) coordinate[i]=x[i];
}

Index::Index(const Index & point) 
{
	int i;
	for (i=0; i<DIMENSION; i++) coordinate[i]=point(i);
}

Index::Index(const int x) 
{
	Initialize(x);
}

#if DIMENSION==2
Index::Index(const int x, const int y) 
{
	Initialize(x,y);
}
#elif DIMENSION==3
Index::Index(const int x, const int y, const int z) 
{
	Initialize(x,y,z);
}
#else
		//
#endif

Index::Index() 
{ 
	// Zero(); // For Efficiency
}

size_t Index::operator()(const Index & size) const
{
	int i;
	size_t x;
	#ifdef _DEBUG
		assert(size.IsPositive());
		assert(IsInBox(size));
	#endif
	for (i=1, x=(size_t)coordinate[0]; i<DIMENSION; i++) {
		x*=(size_t)size(i);
		x+=(size_t)coordinate[i];
	}
	return(x);
}

int * Index::operator()(int * pp) const 
{
	int i;
	for (i=0; i<DIMENSION; i++) pp[i]=coordinate[i];
	return(pp);
}

bool Index::operator==(const Index &right) const 
{
	int i;
	bool f;
	for (i=0, f=true; i<DIMENSION; i++) if (coordinate[i]!=right(i)) f=false;
	return(f);
}

bool Index::IsNonNegative() const 
{
	int i;
	bool f;
	for (i=0, f=true; i<DIMENSION; i++) if (coordinate[i]<0) f=false;
	return(f);
}

bool Index::IsPositive() const
{
	int i;
	bool f;
	for (i=0, f=true; i<DIMENSION; i++) if (coordinate[i]<=0) f=false;
	return(f);
}

bool Index::IsInBox(const Index &box) const
{
	int i;
	bool f;
	if (!IsNonNegative() || !box.IsPositive()) f=false;
	else for (i=0, f=true; i<DIMENSION; i++) if (coordinate[i]>=box(i)) f=false;
	return(f);
}

bool Index::NextInBox(const Index &box) 
{
	int i;
	#ifdef _DEBUG
		assert(IsInBox(box));
	#endif
	coordinate[DIMENSION-1]++;
	for (i=DIMENSION-1; i>0; i--) {
		coordinate[i-1]+=coordinate[i]/box(i);
		coordinate[i]%=box(i);
	}
#ifdef _DEBUG
	return(IsInBox(box));
#else 
	return(true);
#endif
}

bool Index::PrevInBox(const Index &box)
{
	int i;
	#ifdef _DEBUG
		assert(IsInBox(box));
	#endif
	coordinate[DIMENSION-1]--;
	for (i=DIMENSION-1; i>0; i--) {
	  if (coordinate[i]<0) {
		 coordinate[i]=box(i)-1;
		 coordinate[i-1]--;
	  }
	}
#ifdef _DEBUG
	return(IsInBox(box));
#else 
	return(true);
#endif
}

bool Index::WhereInBox(const Index &box, size_t n)
{
	int i;
	size_t l;
	#ifdef _DEBUG
		if (box.Volume()<=0) return(false);
	#endif

	for (i=1, l=1; i<DIMENSION; i++) l*=(size_t)box(i);
	for (i=0; i<DIMENSION-1; l/=(size_t)box(++i)) {
		coordinate[i]=(int)(n/l);
		n%=l;
	}
	coordinate[DIMENSION-1]=(int)n;
	#ifdef _DEBUG
		return(IsInBox(box));
	#else
		return(true);
	#endif
}


Index & Index::operator%=(const Index &box)
{
	int i;
	for (i=0; i<DIMENSION; i++) coordinate[i]%=box(i);
	return(*this);
}

Index & Index::operator%=(const int b)
{
	int i;
	for (i=0; i<DIMENSION; i++) coordinate[i]%=b;
	return(*this);
}

void Index::Initialize(const int x)
{
	int i;
	for (i=0; i<DIMENSION; i++) coordinate[i]=x;
}

#if DIMENSION==2
void Index::Initialize(const int x, const int y)
{
	coordinate[0]=x;
	coordinate[1]=y;
}
#elif DIMENSION==3
void Index::Initialize(const int x, const int y, const int z)
{
	coordinate[0]=x;
	coordinate[1]=y;
	coordinate[2]=z;
}
#endif

Index & Index::operator*=(const int x)
{
	int i;
	for (i=0; i<DIMENSION; i++) coordinate[i]*=x;
	return(*this);
}

Index & Index::operator/=(const int x)
{
	int i;
	#ifdef _DEBUG
		assert(x!=0);
	#endif
	for (i=0; i<DIMENSION; i++) coordinate[i]/=x;
	return(*this);
}

Index & Index::operator+=(const Index & addend)
{
	int i;
	for (i=0; i<DIMENSION; i++) coordinate[i]+=addend(i);
	return(*this);
}

Index & Index::operator-=(const Index & addend)
{
	int i;
	for (i=0; i<DIMENSION; i++) coordinate[i]-=addend(i);
	return(*this);
}

Index & Index::operator=(const Index & x)
{
	int i;
	for (i=0; i<DIMENSION; i++) coordinate[i]=x(i);
	return(*this);
}

size_t Index::Volume() const
{
	int i;
	size_t vol;
	for (i=0, vol=1; i<DIMENSION; i++)
		vol*=coordinate[i];
	return(vol);
}

//size_t Index::FFTWRealVolume() const
//{
//	int i;
//	size_t vol;
//	for (i=0, vol=1; i<DIMENSION-1; i++) vol*=(size_t)coordinate[i];
//	vol*=2*((size_t)coordinate[DIMENSION-1]/2+1);
//	if (vol==0) vol=1;
//	return(vol);
//}
//
//size_t Index::FFTWComplexVolume() const
//{
//	int i;
//	size_t vol;
//	for (i=0, vol=1; i<DIMENSION-1; i++) vol*=(size_t)coordinate[i];
//	vol*=((size_t)coordinate[DIMENSION-1]/2+1);
//	if (vol==0) vol=1;
//	return(vol);
//}

bool Index::MakeFFTWBox(const Index & box)
{
	int i;
	if (!box.IsPositive()) return(false);
	else {
		for (i=0; i<DIMENSION-1; i++) coordinate[i]=box(i);
		coordinate[DIMENSION-1]=2*((size_t)box(DIMENSION-1)/2+1);
		return(true);
	}
}

void Index::Print(const char * text) const
{
	int i;
	cout << text << "=[" << coordinate[0];
	for (i=1; i<DIMENSION; i++) cout << ", " << coordinate[i];
	cout << "]" << endl;
}



ostream & operator <<(ostream &os, const Index & point) {
	int i;
	os << "[" << point(0);
	for (i=1; i<DIMENSION; i++) os << ", " << point(i);
	os << "]";
	return(os);
}

istream & operator >>(istream &is, Index & point) {
	int i;
	for (i=0; i<DIMENSION; i++) is >> point[i];
	return(is);
}

