#include "DMFirstHeader.h"
#include <math.h>

#include "DigitalMaterial.h"

#include "Element.h"

#include "ListOfAtoms.h"
#include "BoundaryConditions.h"
#include "DMArray.h"


using namespace std;

Element::Element(const int nodes[DIMENSION+1], int nNonLocals,
                 const TinyVector<int> (&nodesCoord)[DIMENSION])
{
    int node;
    // various checks:
#ifdef _DEBUG
    for(node = 0; node < DIMENSION; ++node) ASSERT(nodes[node] < nodes[node+1]);
    for(node = 0; node < DIMENSION; ++node)
    {
        bool nonZero = false;
        for(int i=0; i<DIMENSION; ++i) nonZero |= (nodesCoord[node][i] != 0);
        ASSERT(nonZero);
        for(int node2 = 0; node2 < node; ++node2)
        {
            bool different = false;
            for(int i=0; i<DIMENSION; ++i)
                different |= (nodesCoord[node][i] != nodesCoord[node2][i]);
            ASSERT(different);
        }
    }
#endif // DEBUG
    for(int coord=0; coord<DIMENSION+1; ++coord) this->nodes[coord] = nodes[coord];
    for(node =0; node <DIMENSION; ++node )
        for(int coord=0; coord<DIMENSION; ++coord)
            this->nodesCoord[node][coord] = nodesCoord[node][coord];
    for(int i=0; i<DIMENSION; ++i) this->nodesCoord[DIMENSION][i] = 0;

    CalculateCoefficients();
    CalculateWeight(nNonLocals);
}


void Element::ComputeShapeFunctionsAtQueryLatticeIndices(
                                         const int indices[DIMENSION],
                                         double shapeValues[DIMENSION+1]) const
{
    ASSERT(determinant != 0);
    shapeValues[DIMENSION] = 1.0; // the first shapevalue is one minus the three last
    for (int node = 0; node < DIMENSION; node++)
    {
        int s = 0;
        for(int i = 0; i < DIMENSION; i++)
            s += coef[node][i] * indices[i];
        // shapeValues must be between 0 and 1:
        ASSERT(s >= 0 && s <= determinant);
        shapeValues[node] = double(s) / determinant;
        shapeValues[DIMENSION] -= shapeValues[node];
    }
}

int Element::ComputeIntegerShapeFunctions(const int indices[DIMENSION],
                                          int shape[DIMENSION + 1]) const
{
    shape[DIMENSION] = determinant; // the last shapevalue is determinant  minus
    // the DIMENSION first
    for(int node = 0; node < DIMENSION; node++)
    {
        int s = 0;
        for(int i = 0; i < DIMENSION; i++)
            s += coef[node][i] * indices[i];
        shape[node] = s;
        shape[DIMENSION] -= s;
    }
    return determinant;
}

void Element::CalculateCoefficients()
{

#if DIMENSION == 2
    coef[0][0] =  nodesCoord[1][1];
    coef[0][1] = -nodesCoord[1][0];
    coef[1][0] = -nodesCoord[0][1];
    coef[1][1] =  nodesCoord[0][0];
#endif    
#if DIMENSION == 3
    coef[0][0] =  (nodesCoord[1][1] * nodesCoord[2][2] - nodesCoord[1][2] * nodesCoord[2][1]);
    coef[0][1] = -(nodesCoord[1][0] * nodesCoord[2][2] - nodesCoord[1][2] * nodesCoord[2][0]);
    coef[0][2] =  (nodesCoord[1][0] * nodesCoord[2][1] - nodesCoord[1][1] * nodesCoord[2][0]);
    coef[1][0] = -(nodesCoord[0][1] * nodesCoord[2][2] - nodesCoord[0][2] * nodesCoord[2][1]);
    coef[1][1] =  (nodesCoord[0][0] * nodesCoord[2][2] - nodesCoord[0][2] * nodesCoord[2][0]);
    coef[1][2] = -(nodesCoord[0][0] * nodesCoord[2][1] - nodesCoord[0][1] * nodesCoord[2][0]);
    coef[2][0] =  (nodesCoord[0][1] * nodesCoord[1][2] - nodesCoord[0][2] * nodesCoord[1][1]);
    coef[2][1] = -(nodesCoord[0][0] * nodesCoord[1][2] - nodesCoord[0][2] * nodesCoord[1][0]);
    coef[2][2] =  (nodesCoord[0][0] * nodesCoord[1][1] - nodesCoord[0][1] * nodesCoord[1][0]);
#endif    
    determinant = 0;
    for(int i = 0; i < DIMENSION; i++)
    {
        determinant += nodesCoord[0][i] * coef[0][i];
        for(int j = 0; j < DIMENSION; ++j)
        {
            int test = 0;
            for(int k = 0; k < DIMENSION; ++k) test += nodesCoord[i][k] * coef[j][k];
            if(i != j) ASSERT(test == 0);
        }
    }
    if(determinant < 0)
    {
        determinant = -determinant;
        for(int i = 0; i < DIMENSION; i++)
            for(int j = 0; j < DIMENSION; j++)
                coef[i][j] = -coef[i][j];
    }
}

double Element::CalculateVolume(const ListOfAtoms* atoms) const
{
    const DMArray<double, DIMENSION>& pos = *(atoms->GetCartesianPositions());
    double R[DIMENSION][DIMENSION];
    for(int i = 0; i < DIMENSION; i++)
        for(int coord = 0; coord < DIMENSION; coord++)
            R[i][coord] = pos[nodes[i]][coord] - pos[nodes[DIMENSION]][coord];
    atoms->GetBoundaryConditions()->DifferenceBoundaryConditions(R, DIMENSION);
#if DIMENSION == 2
    return fabs((R[0][0] * R[1][1] - R[0][1] * R[1][0])/2.);
#else if DIMENSION == 3
    return fabs((R[0][0] * (R[1][1] * R[2][2] - R[1][2] * R[2][1]) -
                 R[0][1] * (R[1][0] * R[2][2] - R[1][2] * R[2][0]) +
                 R[0][2] * (R[1][0] * R[2][1] - R[1][1] * R[2][0])) /6.);
#endif
}



void Element::CalculateCenter(const ListOfAtoms* atoms,
                              double center[DIMENSION]) const
{
    const DMArray<double, DIMENSION>& pos = *(atoms->GetCartesianPositions());
    for(int coord = 0; coord < DIMENSION; coord++)
        center[coord] = pos[nodes[DIMENSION]][coord];
    double d[DIMENSION];
    for(int node = 0; node < DIMENSION; node++)
    {
        int coord;
        for(coord = 0; coord < DIMENSION; coord++)
            d[coord] = pos[nodes[node]][coord] - pos[nodes[DIMENSION]][coord];
        atoms->GetBoundaryConditions()->DifferenceBoundaryConditions(&d, 1);
        for(coord = 0; coord < DIMENSION; coord++)
            center[coord] += 1.0/(DIMENSION+1) * d[coord];
    }
    atoms->GetBoundaryConditions()->EnforceBoundaryConditions(center);
}

void Element::CalculateWeight(int nNonLocalAtoms)
{
    weight = 0.0;
    if(determinant > 0)
    {
#if DIMENSION == 2
        double angles[3];
        CalculateAngles(angles);
        for(latticePointsIterator it = this->begin(); it != this->end(); ++it)
        {
            switch(it->zeros)
            {
            case 0: // the point is totally inside the triangle
                weight += 1.0;
                break;
            case 1: // on an edge
                weight += 0.5;
                break:
            case 2:
                {
                    int one = it->nonZero[0];
                    if(nodes[one] >= nNonLocalAtoms)
                        weight += angles[one];
                    break:
                }
            default:
                throw DMException("Element", "CalculateWeight", "Problem with a lattice point");

            }
        }
#endif // DIMENSION == 2
#if DIMENSION == 3
        double angles[4][4];
        CalculateAngles(angles);
        for(latticePointsIterator it = this->begin(); it != this->end(); ++it)
        {
            switch(it->zeros)
            {
            case 0: // this lattice point is completely inside the element
                weight += 1.0;
                break;
            case 1: // face
                weight += 0.5;
                break;
            case 2: // edge
                weight += angles[it->nonZero[0]][it->nonZero[1]];
                break;
            case 3: // node
                {
                    int one = it->nonZero[0]; // "one" is which corner this lattice point is.
                    if(nodes[one] >= nNonLocalAtoms) // only if this node is to be treated locally
                    {
                        // solid angle in unit of 4*pi:  (nice formula from spherical geometry)
                        double w = (angles[one][it->zero[0]] +
                                    angles[one][it->zero[1]] +
                                    angles[one][it->zero[2]]) / 2.0 - 0.25;
                        ASSERT(w > 0.0);
                        weight += w;
                    }
                    break;
                }
            default:
                throw DMException("Element", "CalculateWeight", "Problem with a lattice point");
            }
        }
#endif // DIMENSION == 3
    }
}

#if DIMENSION == 2
void Element::CalculateAngles(double angles[3]) const
{ // not testet yet.
    for(int node = 0; node < 3; ++node)
    {
        int j = (node + 1) % 3;
        int k = (node + 2) % 3;
        TinyVector<int> v1(nodesCoord[j]);
        TinyVector<int> v2(nooesCoord[k]);
        v1 -= nodesCoord[node];
        v2 -= nodesCoord[node];
        angles[node] = acos(double(v1 * v2) / sqrt(double(v1.norm2()*v2.norm2())));
        angles[node] /= 2 * PI;
    }
}
#endif // DIMENSION == 2

#if DIMENSION == 3
void Element::CalculateAngles(double angles[4][4]) const
{
    // t represents the orientation of all the dihedra: the order of the 
    // nodes help orient the faces
    const int t[6][4] = {
        {0, 1, 2, 3},
        {0, 2, 3, 1},
        {0, 3, 1, 2},
        {1, 2, 0, 3},
        {1, 3, 2, 0},
        {2, 3, 0, 1}};
    TinyVector<int> edges[3];
    for(int e = 0; e < 6; e++)
    {
        const int (&nd)[4] = t[e];
        for(int edge = 0; edge < 3; ++edge)
            edges[edge] = TinyVector<int>(nodesCoord[nd[edge+1]]) 
                        - TinyVector<int>(nodesCoord[nd[0]]      );
        TinyVector<int> perp1 = edges[0].Cross(edges[1]); // points towards interior
        TinyVector<int> perp2 = edges[0].Cross(edges[2]); // points towards exterior
        double norms = sqrt(double(perp1.norm2() * perp2.norm2()));
        double dot = perp1 * perp2; // the minus is to compensate for perp2 in bad direction        
        angles[nd[0]][nd[1]] = acos(dot/norms) / (2*PI);
        angles[nd[1]][nd[0]] = angles[nd[0]][nd[1]];
    }
    for(int i=0; i<4; ++i) angles[i][i] = 0.; // non defined in principle.
}
#endif // DIMENSION == 3

void Element::CalculateLattice(const ListOfAtoms* atoms,
                               double latticeVectors[DIMENSION][DIMENSION]) const
{
    const DMArray<double, DIMENSION>& pos = *(atoms->GetCartesianPositions());
    for(int i = 0; i < DIMENSION; i++)
        for(int coord = 0; coord < DIMENSION; coord++)
            latticeVectors[i][coord] = 0.0;
    int node;
    double R[DIMENSION][DIMENSION];
    for(node = 0; node < DIMENSION; node++)
        for(int coord = 0; coord < DIMENSION; coord++)
            R[node][coord] = pos[nodes[node]][coord] - pos[nodes[DIMENSION]][coord];

    atoms->GetBoundaryConditions()->DifferenceBoundaryConditions(R, DIMENSION);

    for(node = 0; node < DIMENSION; ++node)
        for(int i = 0; i < DIMENSION; i++)
        {
            double c = double(coef[node][i]) / determinant;
            for(int coord = 0; coord < DIMENSION; coord++)
                latticeVectors[i][coord] += c * R[node][coord];
        }
}

Element::latticePointsIterator::latticePointsIterator(const Element& e, bool endIterator)
: element(e)
{
    for(int coord = 0; coord < DIMENSION; coord++)
    {
        minI[coord] = 0;
        maxI[coord] = 0;
    }
    for(int node = 0; node < DIMENSION; node++)
    {
        for(int coord = 0; coord < DIMENSION; coord++)
        {
            int i = element.GetNodesCoord(node)[coord];
            if(i < minI[coord])
                minI[coord] = i;
            else if(i > maxI[coord])
                maxI[coord] = i;
        }
    }
    if(endIterator)
        indices[DIMENSION - 1] = maxI[DIMENSION - 1] + 1;
    else
    {
        indices[0] = minI[0] - 1;
        for(int coord = 1; coord < DIMENSION; coord++)
            indices[coord] = minI[coord];
        ++(*this);
    }
}

Element::latticePointsIterator& Element::latticePointsIterator::operator++()
{
    const int determinant = element.GetDeterminant();
    int intShape[DIMENSION + 1];
    bool inside = false;
    while(!inside)
    {
        indices[0]++;
        for(int coord=0; coord<DIMENSION-1; ++coord)
        {
            if(indices[coord] > maxI[coord])
            {
                indices[coord] = minI[coord];
                indices[coord+1]++;
            }
            else break;
        }
        if(indices[DIMENSION-1] > maxI[DIMENSION-1]) return *this;
        // A way of checking whether the point is inside the tetrahedron is to 
        // look at the value of the shape functions.
        element.ComputeIntegerShapeFunctions(indices, intShape);
        inside = true;
        for(int node = 0; node < DIMENSION + 1; node++)
            if(intShape[node] < 0 || intShape[node] > determinant)
            {
                inside = false;
                break;
            }
    }
    zeros = 0;
    for(int node = 0; node < DIMENSION + 1; node++)
    {
        shape[node] = double(intShape[node]) / determinant;
        if(0 == intShape[node])
            zero[zeros++] = node;
        else
            nonZero[node - zeros] = node;
    }
    return *this;
}

bool Element::latticePointsIterator::operator!=(const latticePointsIterator& i)
{
    // Works as long as we compare with end(). !!
    return indices[DIMENSION - 1] != i.indices[DIMENSION - 1];
}
