#ifndef ELEMENT_H
#define ELEMENT_H

#include "DMFirstHeader.h"

#include <vector>
#include <iostream>

#include "DigitalMaterial.h"
#include "TinyVector.h"
class ListOfAtoms;

//! The element class. We use tetrahedra in 3D and triangles in 2D.
/*!
    The corners of an element are lattice points, so they are given 
    by integer numbers. In general one wants to evaluate the shape
    functions on other lattice points only. This is why we work mostly
    with ints. The shape functions are linear functions. The shape function 
    of node \a i evaluated at point x is given by

    F_i = a_ij * x_j 
*/

class Element
{
public:
#ifndef SWIG
    //! We pass the nodes index, the number of nonLocals atoms and the coordinates of the 3 first w.r.t. the last.
    /*! We must pass the number of non local atoms because it is needed for the calculation
    of the weight: whether or not a node contributes to the weight of its element depends
    whether the node is a local atom or a non local atom. This number is passed to CalculateWeight.
    */
    Element(const int nodes[4], int nNonLocals,
        const TinyVector<int> (&nodesCoord)[DIMENSION]);
#endif //SWIG
    //! Calculate the shape functions for a point of the lattice.
    /*! The indices are relative to the first node of the element */
    void ComputeShapeFunctionsAtQueryLatticeIndices(const int indices[DIMENSION],
                                                double shapeValues[DIMENSION+1]) const;
    int ComputeIntegerShapeFunctions(const int indices[DIMENSION],
                                     int shape[DIMENSION + 1]) const;
#ifndef SWIG
    //! Calculate the real lattice vectors of the element.
    /*! lat[0] contains the first lattice vector, etc... 
    */ 
    void CalculateLattice(const ListOfAtoms* atoms, double lat[DIMENSION][DIMENSION]) const;
    //! Return the real volume of a tetrahedron.
#endif //SWIG
    double CalculateVolume(const ListOfAtoms* atoms) const;
    //! Calculate the real barycenter of the element and return it in \a center
    void CalculateCenter(const ListOfAtoms* atoms, double center[DIMENSION]) const;

    double GetWeight() const {return weight;}
#ifndef SWIG
    const int (&GetNodesCoord(int node) const)[DIMENSION] {return nodesCoord[node];}
    int GetNode(int node) const {return nodes[node];}
    double GetCoef(int i, int j) const {return coef[i][j];}
    //! The determinant of an element is 6 times its volume.(2 times in 2D)
    int GetDeterminant() const {return determinant;}
    friend std::ostream& operator<<(std::ostream& out, const Element& element)
    {
        out << element.determinant / 6.0 << ' ' << element.weight << std::endl;
        for(int n = 0; n < 4; n++)
        {
            out << element.nodes[n] << '\t';
            for(int coord = 0; coord < DIMENSION; coord++)
                out << element.nodesCoord[n][coord] << ' ';
            out << std::endl;
        }
        return out;
    }
private:
    //! Calculate the \a coef matrix of coefficients for the shape functions.
    void CalculateCoefficients();
    //! \a determinant is 6 times the volume of the tetrahedron (twice in 2D).
    int determinant;
	//! the ratio coef/determinant is the coefficient of the (linear) shape functions.
    /*! So basically, the shape function of node \i for a point [a,b,c] of given by
        (coeff[i][0]*a + coeff[i][1]*b + coeff[i][2]*c)/determinant.
        (See in \a ComputeShapeFunctionsAtQueryLatticeIndices how it is done).
        Note that coef is proportional the transpose of the inverse of nodesCoord[0:DIMENSION-1][*]
    */
    int coef[DIMENSION][DIMENSION];
    //! The number of lattice sites this element containts.
    /*! It is a double because sites on the surface, edge or vertex are split
    among elements.
    */
    double weight;
	//! The index of the nodes
    int nodes[DIMENSION + 1];
	//! The coordinates of the nodes in the local lattice (the last node is the origin).
    int nodesCoord[DIMENSION+1][DIMENSION];
    //! Calculate how many atoms this element represents.
    /*! This number is equal to the number of lattice points that are 
    totally inside the element + half the number of lattice sites on 
    the surfaces (or edges in 2D) + the contribution of the lattice sites that
    are on the edges (in 3D) and on the vertices. For those on the edges (3D), 
    we compute the angle of the dihedron formed by the surfaces and divide it 
    by 2 Pi. For the vertices that are local atoms, we compute the solid angle 
    inside the element, divide it by 4 Pi (2Pi in 2D): this is its contribution
    to the element.
    */
    void CalculateWeight(int nNonLocalAtoms);

#if DIMENSION == 2
    //! Calculate the angle of the triangle (in units of 2 Pi).
    void CalculateAngles(double angles[3]) const;
#endif // DIMENSION == 2

    //! This the dihedral angles (in units of 2 Pi) between faces.
    /*! angles[i][j] = angles[j][i] is the dihedral angle at the edge between 
    node i and j.
    */
#if DIMENSION == 3
    void CalculateAngles(double angles[4][4]) const;
#endif  // DIMENSION == 3
 
#endif //SWIG

public:
    //! used to loop over all the lattice points located within the element
    /*! The nodes and lattice points located on the surfaces or on the edges are 
    taken into account
    */
    class latticePointsIterator
    {
    public:
        latticePointsIterator(const Element& e, bool endIterator);
        latticePointsIterator& operator++();
        bool operator!=(const latticePointsIterator& i);
        latticePointsIterator* operator->() {return this;}
        const Element& element;
        int indices[DIMENSION];
        int zero[DIMENSION];
        int zeros;
        int nonZero[DIMENSION + 1];
        double shape[DIMENSION + 1];
    private:
        int minI[3];
        int maxI[3];
    };
    latticePointsIterator begin() const {return latticePointsIterator(*this, false);}
    latticePointsIterator end() const {return latticePointsIterator(*this, true);}
};

#endif // ELEMENT_H



