// MeshData.h: abstract interface for the MeshData class.
//
//////////////////////////////////////////////////////////////////////

#ifndef MESHDATA_H
#define MESHDATA_H

class MeshData ;

#include "DMFirstHeader.h"
#include <vector>
#include <set>
#include "DigitalMaterial.h"
#include "Serializable.h"

class QCListOfAtoms;
class ListOfSlavedAtoms;
class BravaisLatticeWithBasis;
class NeighborLocator;

#include "ASCIIReader.h"
#include "ASCIIWriter.h"
#include "Element.h"
#include "DMArray.h"
#include "TinyVector.h"


//! MeshData class is a base class to read and wrap mesh description

class MeshData: public Serializable 
{
#ifndef SWIG
    class Face
    {
    public:
        int nodes[DIMENSION];
        mutable std::set<int> elemIds;
    public:
        bool operator<(const Face &face) const
        {
            for(int n = 0; n < DIMENSION; n++)
            {
                int d = face.nodes[n] - nodes[n];
                if(d > 0)
                    return true;
                else if(d < 0)
                    return false;
            }
            return false;
        }
        bool operator==(const Face &face) const
        {
            for(int n=0; n<DIMENSION; ++n) if (nodes[n] != face.nodes[n]) return false;
            return true;
        }
    };
    class Edge
    {
    public:
        int nodes[DIMENSION - 1];
    public:
        bool operator<(const Edge &edge) const
        {
            for(int n = 0; n < DIMENSION - 1; n++)
            {
                int d = edge.nodes[n] - nodes[n];
                if(d > 0)
                    return true;
                else if(d < 0)
                    return false;
            }
            return false;
        }
    };
#endif // SWIG
public:
    MeshData();
    ~MeshData() ; 
    //! Get element pointer
    const Element &GetElement(int elIndex) const {return elements[elIndex];}
    const BravaisLatticeWithBasis *GetLattice(int elIndex) const {return lattices[elIndex];}
    int GetNumberOfElements() const {return elements.size();}
    int GetNode(int e, int n) const {return elements[e].GetNode(n);}
    int GetDeterminant(int e) const {return elements[e].GetDeterminant();}
    void InitializeSlaves(QCListOfAtoms* masters);
    int PostProcessMesh(QCListOfAtoms* masters);

#ifndef SWIG
    //! Serializable interface.
    //@{
    //! Save the class.
    void Save(DMWriter *writer) const;
    //! Load the class.
    void Load(DMReader *reader);
    //! Get the string name of the class for saving or loading.
    const std::string GetType() const {return "MeshData";}
    //@}
    void CreateElement(const std::vector<int>& ndList, 
                       const QCListOfAtoms* masters,
                       const BravaisLatticeWithBasis* lattice);
    void DecorateElement(Element& element,
                         const QCListOfAtoms* masters);
    /*    void MeshPlot(double P0[DIMENSION],
          double n1[DIMENSION], double n2[DIMENSION],
          int N1, int N2,
          double L, double* atomicData, double* elementData);
    */
    const std::vector<int>& GetNeighbors(int number) const
        {return slaveNeighbors[number];}
private:
    //! MeshData owns the elements and stores them in a std::vector
    std::vector<Element> elements;
    //! lattices[i] contains the bravais lattice of element i.
    std::vector<const BravaisLatticeWithBasis *> lattices;
    //! The number of nonlocal atoms.
    int nAtomic;
    //! The index of the neighboring slaves. On vector per master atom.
    std::vector< std::vector<int> > slaveNeighbors;
    //! The container of the external faces of the mesh    
    std::set<Face> externalFaces;
    //! The container of the edges of the mesh
    std::set<Edge> edges;

    std::vector<int> GetElementsSharingThisFace(const Face &face) const;


    std::set<int> vertices;
    std::vector< TinyVector<int, DIMENSION> > newIndices;
    std::vector<int> numberOfSlaves;
    std::vector<Element*> decoratedElementsVector;
    double volume;
    double elementsWeight;
    int nAtomicElements;

    void BuildExternalFaces();
//    NeighborLocator* tmpnl;
#endif // SWIG
} ;

#endif // !defined(MESHDATA_H)
