// QCListOfAtoms.h: interface for the QCListOfAtoms class.
//
//////////////////////////////////////////////////////////////////////

#ifndef QCLISTOFATOMS_H
#define QCLISTOFATOMS_H

class MeshData;
class ListOfSlavedAtoms;
class BravaisLatticeWithBasis;

#include "DigitalMaterial.h"
#include "DynamicListOfAtoms.h"
#include "DMArray.h"
#include "ASCIIReader.h"
#include "ASCIIWriter.h"

//! QCListOfAtoms is the container for master atoms our quasi-continuum code
/*! QCListOfAtoms is an extension of usual lists of atoms in the sense that in addition
    to \a position and \a velocities, the masters have a \a weight (characterizing how many atoms
    they represent), a pointer to a lattice and a reference position, which isn't
    necessary a lattice site (when we start with a dislocation in the lattice, there's no
    way to have an unambiguoulsy defined lattice coordinate).

                       ----> QCListOfAtoms:masters ------> DynamicListOfAtoms: nonLocalAtoms
                       |                              |   
ListOfAtoms:allAtoms --|                              |__> DynamicListOfAtoms: localAtoms
                       |
                       |___> ListOfSlaveAtoms: slaves

The user initializes the nonLocal and local atoms, passes them to a QCListOfAtoms that
represents all the degrees of freedom. This QCListOfAtoms will create a top list of atoms,
"allAtoms", attach itself to "allAtoms" and attach a list of slave atoms.

The user sets a NeighborLocator on the masters to identify the pairs of atoms to be
treated by MD. Internally, this neighbor locator gets the positions of the nonLocalAtoms only
and the masters will set a special neighbor locator on "allAtoms" so that when ones computes the
forces in MD, we pass "allAtoms" to the potential and the special neighborLocator (a QCNeighborList)
will make sure that potential receives only pairs between nonLocal atoms and pairs between 
a nonLocal atoms and a slave.
*/

class QCListOfAtoms : public DynamicListOfAtoms  
{
public:
    QCListOfAtoms(DynamicListOfAtoms* nonLocalAtoms,
                  DynamicListOfAtoms* localAtoms,
                  const BoundaryConditions *bc=0,
                  NeighborLocator *neighborLocator=0);
    virtual ~QCListOfAtoms();
    ListOfAtoms* GetAllAtoms() const;
    ListOfSlavedAtoms *GetSlaves() const;
    DynamicListOfAtoms *GetNonLocals() const;
    DynamicListOfAtoms *GetLocals() const;
    void Save(DMWriter *writer) const;
    void Load(DMReader *reader);
    void Write(char* fileName) const {ASCIIWriter w(fileName); Save(&w);}
    void Read(char* fileName) {ASCIIReader r(fileName); Load(&r);}
    void SetInitialized(bool i) {initialized = i;}
    bool GetInitialized() {return initialized;}
    void AdjustPositions() ;
#ifndef SWIG
    const std::string GetType() const {return "QCListOfAtoms";}
#endif // SWIG

    void SetBoundaryConditions(const BoundaryConditions *bc);
    //! this neighbor locator is to be used only on the non local atoms.
    void SetNeighborLocator(NeighborLocator *neighborLocator);
#ifndef SWIG
    void SetPos(double* data);
#endif // SWIG
private:
    bool initialized;
    NeighborLocator *allAtomsNL;
    NeighborLocator *masterAtomsNL;
    NeighborLocator *localAtomsNL;
    NeighborLocator *slaveAtomsNL;

};

#endif // !defined(QCLISTOFATOMS_H)
